// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: arrays.inc
// Last updated: 10 June 2021

// Description: Array manipulation macros...sorting, resizing, etc.

//  Rand_Array_Item
//  Resize_Array      (Note v3.8, and so povr, support native growable 1d arrays)
//  Reverse_Array
//  Sort_Array

//------------------------------------------------------------------------------

#ifndef(ARRAYS_INC_TEMP)
#declare ARRAYS_INC_TEMP = version;
#version 3.8;

#ifdef(View_POV_Include_Stack)
    #debug "including arrays.inc\n"
#end

// Input:  a one dimensional array and a random stream
// Output: a random item from the input array
// Source: variate.inc by Ingo Janssen
#macro Rand_Array_Item(Array, Stream)
    #if(dimensions(Array)=1)
        Array[floor(rand(Stream)*0.9999999*dimension_size(Array,1))]
    #else
        #error "The Rand_Array_Item() macro only works for 1D arrays."
    #end
#end


#macro Resize_Array(Array, NewSize)
    #if(dimensions(Array)=1)
        #local NewArray = array[NewSize]
        #local C = 0;
        #local Max = min(NewSize, dimension_size(Array, 1));
        #while (C < Max)
            #ifdef (Array[C])
                #local NewArray[C] = Array[C];
            #end
            #local C = C + 1;
        #end
        #declare Array = NewArray
    #else
        #error "The Resize_Array() macro only works for 1D arrays."
    #end
#end


#macro Reverse_Array(Array)
    #if(dimensions(Array)=1)
        #local J = 0;
        #local N = dimension_size(Array, 1) - 1;
        #while(J < N/2)
            #local Temp = Array[J];
            #local Array[J] = Array[N-J];
            #local Array[N-J] = Temp;
            #local J = J + 1;
        #end
    #else
        #error "The Reverse_Array() macro only works for 1D arrays."
    #end
#end


// This macro will compare float values and vector lengths.  It is intended to
// be redefined by the user to suit their needs, but it should always return
// true if A is "less than" B, otherwise false.
//
// Alt: Sort_Compare(A, B) (vlength(A) < vlength(B)) #end
// Alt: Sort_Compare(Array, IdxA, IdxB) (vlength(Array[IdxA]) < vlength(Array[IdxB])) #end
#macro Sort_Compare(Array, IdxA, IdxB) (Array[IdxA] < Array[IdxB]) #end

// This macro swaps the data in two locations. The user can
// redefine it, for example to swap additional data along a
// second dimension, keeping columns of data together.
#macro Sort_Swap_Data(Array, IdxA, IdxB)
    #local Tmp = Array[IdxA];
    #declare Array[IdxA] = Array[IdxB];
    #declare Array[IdxB] = Tmp;
#end


// Sort macros slightly modified from QuickSort macros by Juha Nieminen
#macro Sort_Array(Array)
    Sort_Partial_Array(Array, 0, dimension_size(Array, 1)-1)
#end

#macro Sort_Partial_Array(Array, FirstInd, LastInd)
    ARRAYS_HybridQuickSortStep(Array, FirstInd, LastInd)
    ARRAYS_InsertionSort(Array, FirstInd, LastInd)
#end

#declare ARRAYS_QuickSortSeed = seed(0);

#macro ARRAYS_HybridQuickSortStep(Array, FirstInd, LastInd)
    #local FInd = FirstInd;
    #while(FInd < LastInd-10)
        #local RInd = FInd + rand(ARRAYS_QuickSortSeed)*(LastInd - FInd);
        Sort_Swap_Data(Array, FInd, RInd)
        #local I = FInd-1;
        #local J = LastInd+1;
        #local Mid = -1;
        #while(Mid < 0)
            #local J = J-1;
            #while(Sort_Compare(Array, FInd, J)) #local J = J-1; #end
            #local I = I+1;
            #while(Sort_Compare(Array, I, FInd)) #local I = I+1; #end

            #if(I < J)
                Sort_Swap_Data(Array, I, J)
            #else
                #local Mid = J;
            #end
        #end
        ARRAYS_HybridQuickSortStep(Array, FInd, Mid)
        #local FInd = Mid+1;
    #end
#end

#macro ARRAYS_InsertionSort(Array, FirstInd, LastInd)
    #local Ind1 = FirstInd+1;
    #while(Ind1 <= LastInd)
        #local Ind2 = Ind1;
        #while(Ind2 > FirstInd)
            #local NextInd2 = Ind2-1;
            #if(Sort_Compare(Array, Ind2, NextInd2))
                Sort_Swap_Data(Array, Ind2, NextInd2)
                #local Ind2 = NextInd2;
            #else
                #local Ind2 = 0;
            #end
        #end
        #local Ind1 = Ind1+1;
    #end
#end


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="arrays.inc")

    #fopen FID "SelfTestArraysInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
    // Run some macros to be sure parse works. Assume failure and correct later if all parse OK.
    #write (FID,"\n---\n")
    #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))

    // Use macros here...

    #write (FID,"\n")
    #fclose FID


    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestArraysInc.txt" write
    // If all macro call tests above work create output for what was tested.
    //#write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Center_Object OK\n"))
    #write (FID,"\n")
    // Rand_Array_Item
    #declare RdmStrm = seed(574647);
    #declare Ary = array[10] {0,1,2,3,4,5,6,7,8,9}
    #write (FID,concat("Rand_Array_Item +2.000000 ",str(Rand_Array_Item(Ary,RdmStrm),9,6)," \n"))
    #write (FID,concat("Rand_Array_Item +7.000000 ",str(Rand_Array_Item(Ary,RdmStrm),9,6)," \n"))
    #write (FID,concat("Rand_Array_Item +0.000000 ",str(Rand_Array_Item(Ary,RdmStrm),9,6)," \n"))
    #ifdef (RdmStrm) #undef RdmStrm #end
    #write (FID,"\n")
    // Resize_Array
    Resize_Array(Ary,20)
    #write (FID,concat("Resize_Array +20.000000 ",str(dimension_size(Ary,1),9,6)," \n"))
    #ifdef (Ary) #undef Ary #end
    #write (FID,"\n")
    // Reverse_Array
    #declare Ary = array[10] {0,1,2,3,4,5,6,7,8,9}
    Reverse_Array(Ary)
    #write (FID,concat("Reverse_Array +9.000000 ",str(Ary[0],9,6)," \n"))
    #write (FID,concat("Reverse_Array +8.000000 ",str(Ary[1],9,6)," \n"))
    #write (FID,concat("Reverse_Array +7.000000 ",str(Ary[2],9,6)," \n"))
    #write (FID,concat("Reverse_Array +6.000000 ",str(Ary[3],9,6)," \n"))
    #write (FID,concat("Reverse_Array +5.000000 ",str(Ary[4],9,6)," \n"))
    #write (FID,concat("Reverse_Array +4.000000 ",str(Ary[5],9,6)," \n"))
    #write (FID,concat("Reverse_Array +3.000000 ",str(Ary[6],9,6)," \n"))
    #write (FID,concat("Reverse_Array +2.000000 ",str(Ary[7],9,6)," \n"))
    #write (FID,concat("Reverse_Array +1.000000 ",str(Ary[8],9,6)," \n"))
    #write (FID,concat("Reverse_Array +0.000000 ",str(Ary[9],9,6)," \n"))
    #write (FID,"\n")
    // Reverse_Array
    Sort_Array(Ary)
    #write (FID,concat("Sort_Array +0.000000 ",str(Ary[0],9,6)," \n"))
    #write (FID,concat("Sort_Array +1.000000 ",str(Ary[1],9,6)," \n"))
    #write (FID,concat("Sort_Array +2.000000 ",str(Ary[2],9,6)," \n"))
    #write (FID,concat("Sort_Array +3.000000 ",str(Ary[3],9,6)," \n"))
    #write (FID,concat("Sort_Array +4.000000 ",str(Ary[4],9,6)," \n"))
    #write (FID,concat("Sort_Array +5.000000 ",str(Ary[5],9,6)," \n"))
    #write (FID,concat("Sort_Array +6.000000 ",str(Ary[6],9,6)," \n"))
    #write (FID,concat("Sort_Array +7.000000 ",str(Ary[7],9,6)," \n"))
    #write (FID,concat("Sort_Array +8.000000 ",str(Ary[8],9,6)," \n"))
    #write (FID,concat("Sort_Array +9.000000 ",str(Ary[9],9,6)," \n"))
    #write (FID,"\n")

    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"

#end // End self testing

#version ARRAYS_INC_TEMP;
#end
