// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: math.inc
// Last updated: 10 June 2021

// Description: This file contains various math macros.

// General macros:
//      Interpolate()

// Vector macros:
//      VDot(Va,Vb)   - Type aware, 2 to 5 dimension vector product.
//      VEq(Va,Vb)    - Type aware, 2 to 5 dimension vector equality test.
//      VLength(V)    - Type aware, 2 to 5 dimension vector length.
//      VNormalize(V) - Type aware, 2 to 5 dimension vector normalization.
//      VPow(V,P)     - Type aware, 2 to 5 dimension vector raised to power.
//      VSqr(V)       - Type aware, 2 to 5 dimension vector square.
//      VZero(V)      - Type aware, 2 to 5 dimension vector all zeros test.

// Notes:

// Vector macro use is always slower than coding up the component math or
// available vector math such as V*V over VSqr. However, POV-Ray's vector math
// where the vectors are not of the same size always promotes to the largest
// vector. V2d*V5d gives a 5 dimension vector result. This can lead to
// confusing, difficult to debug, issues. Lastly, most SDL vector functions
// support only 2 and 3 dimension vectors.

// Where in the macro definitions there is an overall enclosing set of
// parenthesis, it's the odd way POV-Ray SDL indicates the last calculated
// value in a block of code is to be used as the return value. For example, see
// the Interpolate macro.

// Macros NO LONGER defined herein or anywhere in povr core includes:
//      VDist()         - Use faster inbuilt SDL: vlength() / vm: f_length()
//                        Or VLength(). Pass the argument V1-V2.
//      VEq5D()         - Use VEq().
//      VLenght5D()     - Use VLength().
//      VMax() / max3() - Use much faster SDL/vm inbuilts max(V.x, V.y, V.z)
//      VMin() / min3() - Use much faster SDL/vm inbuilts min(V.x, V.y, V.z)
//      VNormalize5D()  - Use VNormalize().
//      VPow5D()        - Use VPow().
//      VZero5D()       - Use VZero().

//------------------------------------------------------------------------------

#ifndef(MATH_INC_TEMP)
#declare MATH_INC_TEMP = version;
#version 3.8;   // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including math.inc\n"
#end

#include "functions.inc"  // f_vangle,...
#include "setidtypes.inc" // Used with VEq etc.
#include "vectors.inc"    // VectorCos,...


// Interpolation
// Author: Margus Ramst
// GC - global current
// GS - global start
// GE - global end
// TS - target start
// TE - target end
// Method - interpolation method:
//          Method = 0 - cosine interpolation
//          Method > 0 - exponential (1 - linear, etc)
#macro Interpolate(GC, GS, GE, TS, TE, Method)
    (
    #if(Method!=0)
        (TS+(TE-TS)*pow((GC-GS)/(GE-GS),Method))
    #else
        #local X=(GC-GS)/(GE-GS);
        #local F=(1-cos(X*pi))*.5;
        (TS*(1-F)+TE*F)
    #end
    )
#end

// --------------------------------------------------------
// Vector macros:
// --------------------------------------------------------

// Returns dot product of two vectors
#macro VDot(V1, V2)
    (
    #if (id_type(V1) != id_type(V2))
        #error "\nVector arguments to VDot are not the same type/size.\n"
    #else
        #switch (id_type(V1))
            #case (global.dictIdTypes.v3d)
                (V1.x*V2.x + V1.y*V2.y + V1.z*V2.z)
                #break
            #case (global.dictIdTypes.v5d)
                (V1.x*V2.x + V1.y*V2.y + V1.z*V2.z +
                    V1.filter*V2.filter + V1.transmit*V2.transmit)
                #break
            #case (global.dictIdTypes.v2d)
                (V1.u*V2.u + V1.v*V2.v)
                #break
            #case (global.dictIdTypes.v4d)
                (V1.x*V2.x + V1.y*V2.y + V1.z*V2.z + V1.t*V2.t)
                #break
            #else
                #error "\nArguments to VDot not of vector type.\n"
        #end
    #end
    )
#end

// Returns true if vectors are equal, otherwise false
#macro VEq(V1, V2)
    (
    #if (id_type(V1) != id_type(V2))
        #error "\nVector arguments to VEq are not the same type/size.\n"
    #else
        #switch (id_type(V1))
            #case (global.dictIdTypes.v3d)
                (V1.x = V2.x & V1.y = V2.y & V1.z = V2.z)
                #break
            #case (global.dictIdTypes.v5d)
                (V1.x = V2.x & V1.y = V2.y & V1.z = V2.z &
                    V1.filter = V2.filter & V1.transmit = V2.transmit)
                #break
            #case (global.dictIdTypes.v2d)
                (V1.u = V2.u & V1.v = V2.v)
                #break
            #case (global.dictIdTypes.v4d)
                (V1.x = V2.x & V1.y = V2.y & V1.z = V2.z & V1.t = V2.t)
                #break
            #else
                #error "\nArguments to VEq not of vector type.\n"
        #end
    #end
    )
#end

// Returns length of a vector
#macro VLength(V)
    (
    #switch (id_type(V))
        #case (global.dictIdTypes.v3d)
            sqrt(V.x*V.x + V.y*V.y + V.z*V.z)
            #break
        #case (global.dictIdTypes.v5d)
            sqrt(V.x*V.x + V.y*V.y + V.z*V.z +
                V.filter*V.filter + V.transmit*V.transmit)
            #break
        #case (global.dictIdTypes.v2d)
            sqrt(V.u*V.u + V.v*V.v)
            #break
        #case (global.dictIdTypes.v4d)
            sqrt(V.x*V.x + V.y*V.y + V.z*V.z + V.t*V.t)
            #break
        #else
            #error "\nArgument to VLength not of vector type.\n"
    #end
    )
#end

// Returns normalization of a vector
#macro VNormalize(V)
    (
    #switch (id_type(V))
        #case (global.dictIdTypes.v3d)
            #local Len = sqrt(V.x*V.x + V.y*V.y + V.z*V.z);
            #break
        #case (global.dictIdTypes.v5d)
            #local Len = sqrt(V.x*V.x + V.y*V.y + V.z*V.z +
                V.filter*V.filter + V.transmit*V.transmit);
            #break
        #case (global.dictIdTypes.v2d)
            #local Len = sqrt(V.u*V.u + V.v*V.v);
            #break
        #case (global.dictIdTypes.v4d)
            #local Len = sqrt(V.x*V.x + V.y*V.y + V.z*V.z + V.t*V.t);
            #break
        #else
            #error "\nArgument to VNormalize not of vector type.\n"
    #end
    #if (Len)
        (V/Len)
    #else
        (V)
    #end
    )
#end

// Raises the components of a vector to a given power
#macro VPow(V, P)
    (
    #if (id_type(P) != global.dictIdTypes.flt)
        #error "\nSecond argument to VPow is not a float type.\n"
    #else
        #switch (id_type(V))
            #case (global.dictIdTypes.v3d)
                <pow(V.x,P),pow(V.y,P),pow(V.z,P)>
                #break
            #case (global.dictIdTypes.v5d)
                <pow(V.x,P),pow(V.y,P),pow(V.z,P),pow(V.filter,P),
                    pow(V.transmit,P)>
                #break
            #case (global.dictIdTypes.v2d)
                <pow(V.u,P),pow(V.v,P)>
                #break
            #case (global.dictIdTypes.v4d)
                <pow(V.x,P),pow(V.y,P),pow(V.z,P),pow(V.t,P)>
                #break
            #else
                #error "\nFirst argument to VPow not of vector type.\n"
        #end
    #end
    )
#end

// Returns square of a vector
#macro VSqr(V)
    (
    #switch (id_type(V))
        #case (global.dictIdTypes.v3d)
            #break
        #case (global.dictIdTypes.v5d)
            #break
        #case (global.dictIdTypes.v2d)
            #break
        #case (global.dictIdTypes.v4d)
            #break
        #else
            #error "\nArgument to VSqr not of vector type.\n"
    #end
    V*V
    )
#end

// Returns true if vector components are all zero, otherwise false
#macro VZero(V)
    (
    #switch (id_type(V))
        #case (global.dictIdTypes.v3d)
            (V.x = 0.0 & V.y = 0.0 & V.z = 0.0)
            #break
        #case (global.dictIdTypes.v5d)
            (V.x = 0.0 & V.y = 0.0 & V.z = 0.0 & V.filter = 0.0 &
                V.transmit = 0.0)
            #break
        #case (global.dictIdTypes.v2d)
            (V.u = 0.0 & V.v = 0.0)
            #break
        #case (global.dictIdTypes.v4d)
            (V.x = 0.0 & V.y = 0.0 & V.z = 0.0 & V.t = 0.0)
            #break
        #else
            #error "\nArgument to VZero not of vector type.\n"
    #end
    )
#end



//  Cosine of angle between V1 and V2
//  Previously: max(-1,min(1,vdot(vnormalize(V1),vnormalize(V2))))
//  Update as able to VectorCos() in vectors.inc
//  #error "VCos_Angle in math.inc called. Use vectors.inc's VectorCos()"
#macro VCos_Angle(V1,V2)
    VectorCos(V1,V2)
#end

//  Angle in radians between V1 and V2
//  Previously: acos(max(-1,min(1,vdot(vnormalize(V1),vnormalize(V2)))))
//  See also: AngleBetweenVectors, AccurateAngleBetweenVectors in vectors.inc
#macro VAngle(V1,V2) f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1) #end

//  Angle in degrees between V1 and V2
//  Previously: degrees(acos(max(-1,min(1,vdot(vnormalize(V1),vnormalize(V2))))))
//  See also: AngleBetweenVectors, AccurateAngleBetweenVectors in vectors.inc
#macro VAngleD(V1,V2) degrees(f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1)) #end

// VRotation() will find the rotation angle from V1 to V2
// around Axis. Axis should be perpendicular to both V1
// and V2. The output will be in the range between -pi and
// pi radians or between -180 degrees and 180 degrees if
// you are using the degree version. However, if Axis is
// set to <0,0,0> the output will always be positive or
// zero, the same result you will get with the VAngle() macros.
// Author: Rune S. Johansen
// Previously:
//     (acos(max(min(vdot(vnormalize(V1),vnormalize(V2)),1),-1))
//     *(vdot(Axis,vcross(V1,V2))<0?-1:1))
//
#macro VRotation(V1,V2,Axis)
    (f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1)
    *(vdot(Axis,vcross(V1,V2))<0?-1:1))
#end
#macro VRotationD(V1,V2,Axis)
    (degrees(f_vangle(V1.x,V1.y,V1.z,V2.x,V2.y,V2.z,1))
    *(vdot(Axis,vcross(V1,V2))<0?-1:1))
#end

// Returns a vector perpendicular to V
// Author: Tor Olav Kristensen
// See: OrthogonalVector() and AltOrthogonalVector() in vectors.inc
  //#error "VPerp_To_Vector in math.inc called. Use OrthogonalVector()"
#macro VPerp_To_Vector(v0)
    OrthogonalVector(v0)
#end

// Returns a vector perpendicular to V1 and V2
#macro VPerp_To_Plane(V1, V2) (vnormalize(vcross(V1, V2))) #end

// Find a vector perpendicular to Axis and in the plane of
// V and Axis. In other words, the new vector is a version
// of V adjusted to be perpendicular to Axis.
#macro VPerp_Adjust(V, Axis)
    vnormalize(vcross(vcross(Axis, V), Axis))
#end

// Projects a vector onto the plane defined by Axis.
// Based on code by Ron Parker
// Use instead VectorReject from vectors.inc
// Kept for backward compatibility.
#macro VProject_Plane(V, Axis)
    #local A = vnormalize(Axis);
    (V - vdot(V, A)*A)
#end

// Projects a vector onto an axis.
// Based on code by Ron Parker
// Use instead VectorProject from vectors.inc
// Kept for backward compatibility.
#macro VProject_Axis(V, Axis)
    (Axis*vdot(V, Axis)/vdot(Axis, Axis))
#end

// Creates a vector going in the direction of the
// given vector with the specified length
#macro VWith_Len(V, Len) (Len*vnormalize(V)) #end


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="math.inc")

 // --- Currently not 'just' calling macros herein.
 // #fopen FID "SelfTestMathInc.txt" write
 // #write (FID,"\nExpect value followed by value generated.\n")
 // // Run some macros to be sure parse works. Assume failure and correct later if all parse OK.
 // #write (FID,"\n---\n")
 // #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))
 // #fclose FID

    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestMathInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
 // // If all macro call tests above work create output for what was tested.
 // #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Center_Object OK\n"))
    #write (FID,"\n")

    // Interpolate
    #write (FID,"---\n")
    #declare TS = <0,0,0>;
    #declare TE = <1,2,3>;
    //                           now gs  ge
    #declare Vec00 = Interpolate(0.5,0.0,1.0,TS,TE,1.0);
    #write (FID,concat("Interpolate 0.500000 ",str(Vec00.x,9,6)," x lin\n"))
    #write (FID,concat("Interpolate 1.000000 ",str(Vec00.y,9,6)," y lin\n"))
    #write (FID,concat("Interpolate 1.500000 ",str(Vec00.z,9,6)," z lin\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"---\n")
    #declare Vec00 = Interpolate(0.5,0.0,1.0,TS,TE,2.0);
    #write (FID,concat("Interpolate 0.250000 ",str(Vec00.x,9,6)," x ^2\n"))
    #write (FID,concat("Interpolate 0.500000 ",str(Vec00.y,9,6)," y ^2\n"))
    #write (FID,concat("Interpolate 0.750000 ",str(Vec00.z,9,6)," z ^2\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"---\n")
    #declare Vec00 = Interpolate(0.5,0.0,1.0,TS,TE,0.0);
    #write (FID,concat("Interpolate 0.500000 ",str(Vec00.x,9,6)," x cos\n"))
    #write (FID,concat("Interpolate 1.000000 ",str(Vec00.y,9,6)," y cos\n"))
    #write (FID,concat("Interpolate 1.500000 ",str(Vec00.z,9,6)," z cos\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (TS) #undef TS #end
    #ifdef (TE) #undef TE #end
    #write (FID,"\n")
    // Really checking parser functionality in block of tests below.
    // Primarily, vector promotion on mixed input types.
    #write (FID,"---\n")
    #declare V2 = <0,0.5>;
    #declare V3 = <0,0.5,1>;
    #declare V4 = <0,0.5,1,2>;
    #declare V5 = <0,0.5,1,2,-3>;
    #declare Vec00 = V2*V2;
    #write (FID,concat("V2*V2 0.000000 ",str(Vec00.u,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V2*V2 0.250000 ",str(Vec00.v,9,6)," 0.5*0.5\n"))
  //#write (FID,concat("V3*V2 0.000000 ",str(Vec00.z,9,6)," (.z is an error)\n"))
    #write (FID,"-\n")
    #declare Vec00 = V2*V3;
    #write (FID,concat("V2*V3 0.000000 ",str(Vec00.u,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V2*V3 0.250000 ",str(Vec00.v,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V2*V3 0.000000 ",str(Vec00.z,9,6)," ?.?*1.0\n"))
    #write (FID,"-\n")
    #declare Vec00 = V3*V2;
    #write (FID,concat("V3*V2 0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V3*V2 0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V3*V2 0.000000 ",str(Vec00.z,9,6)," 1.0*?.?\n"))
  //#write (FID,concat("V3*V2 0.000000 ",str(Vec00.t,9,6)," (.t is an error)\n"))
    #write (FID,"\n")
    #write (FID,"---\n")
    #declare Vec00 = V3*V3;
    #write (FID,concat("V3*V3 0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V3*V3 0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V3*V3 1.000000 ",str(Vec00.z,9,6)," 1.0*1.0\n"))
  //#write (FID,concat("V3*V2 0.000000 ",str(Vec00.t,9,6)," (.t is an error)\n"))
    #write (FID,"-\n")
    #declare Vec00 = V3*V4;
    #write (FID,concat("V3*V4 0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V3*V4 0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V3*V4 1.000000 ",str(Vec00.z,9,6)," 1.0*1.0\n"))
    #write (FID,concat("V3*V4 0.000000 ",str(Vec00.t,9,6)," ?.?*2.0\n"))
    #write (FID,"-\n")
    #declare Vec00 = V4*V3;
    #write (FID,concat("V4*V3 0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V4*V3 0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V4*V3 1.000000 ",str(Vec00.z,9,6)," 1.0*1.0\n"))
    #write (FID,concat("V4*V3 0.000000 ",str(Vec00.t,9,6)," 2.0*?.?\n"))
  //#write (FID,concat("V3*V2 0.000000 ",str(Vec00.transmit,9,6)," (.transmit is an error)\n"))
    #write (FID,"\n")
    #write (FID,"---\n")
    #declare Vec00 = V4*V4;
    #write (FID,concat("V4*V4  0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V4*V4  0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V4*V4  1.000000 ",str(Vec00.z,9,6)," 1.0*1.0\n"))
    #write (FID,concat("V4*V4  4.000000 ",str(Vec00.t,9,6)," 2.0*2.0\n"))
  //#write (FID,concat("V3*V2 0.000000 ",str(Vec00.transmit,9,6)," (.transmit is an error)\n"))
    #write (FID,"-\n")
    #declare Vec00 = V4*V5;
    #write (FID,concat("V4*V5  0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V4*V5  0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V4*V5  1.000000 ",str(Vec00.z,9,6)," 1.0*1.0\n"))
    #write (FID,concat("V4*V5  4.000000 ",str(Vec00.t,9,6)," 2.0*2.0\n"))
    #write (FID,concat("V4*V5 -0.000000 ",str(Vec00.transmit,9,6)," ?.?*-3.0\n"))
    #write (FID,"-\n")
    #declare Vec00 = V5*V4;
    #write (FID,concat("V5*V4  0.000000 ",str(Vec00.x,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V5*V4  0.250000 ",str(Vec00.y,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V5*V4  1.000000 ",str(Vec00.z,9,6)," 1.0*1.0\n"))
    #write (FID,concat("V5*V4  4.000000 ",str(Vec00.t,9,6)," 2.0*2.0\n"))
    #write (FID,concat("V5*V4 -0.000000 ",str(Vec00.transmit,9,6)," -3.0*?.?\n"))
    #write (FID,"\n")
    #write (FID,"---\n")
    #declare Vec00 = V5*V5;
    #write (FID,concat("V5*V5 0.000000 ",str(Vec00.red,9,6)," 0.0*0.0\n"))
    #write (FID,concat("V5*V5 0.250000 ",str(Vec00.green,9,6)," 0.5*0.5\n"))
    #write (FID,concat("V5*V5 1.000000 ",str(Vec00.blue,9,6)," 1.0*1.0\n"))
    #write (FID,concat("V5*V5 4.000000 ",str(Vec00.filter,9,6)," 2.0*2.0\n"))
    #write (FID,concat("V5*V5 9.000000 ",str(Vec00.transmit,9,6)," -3.0*-3.0\n"))
    #ifdef (V2) #undef V2 #end
    #ifdef (V3) #undef V3 #end
    #ifdef (V4) #undef V4 #end
    #ifdef (V5) #undef V5 #end
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"\n")
    // VDot
    #write (FID,"---\n")
  //#write (FID,concat("VDot  1.000000 ",str(VDot(1,1),9,6)," flt - flt error\n"))
  //#write (FID,concat("VDot  1.000000 ",str(VDot(1,<1,1>),9,6)," flt - 2d error\n"))
    #write (FID,concat("VDot +3.110000 ",str(VDot(<1.1,1.2>,<+1.3,+1.4>),9,6)," 2d \n"))
    #write (FID,concat("vdot +3.110000 ",str(vdot(<1.1,1.2>,<+1.3,+1.4>),9,6)," 2d \n"))
    #write (FID,concat("VDot -3.110000 ",str(VDot(<1.1,1.2>,<-1.3,-1.4>),9,6)," 2d \n"))
    #write (FID,concat("vdot -3.110000 ",str(vdot(<1.1,1.2>,<-1.3,-1.4>),9,6)," 2d \n"))
    #write (FID,"-\n")
    #write (FID,concat("VDot +5.510000 ",str(VDot(<1.1,1.2,1.5>,<+1.3,+1.4,+1.6>),9,6)," 3d \n"))
    #write (FID,concat("vdot +5.510000 ",str(vdot(<1.1,1.2,1.5>,<+1.3,+1.4,+1.6>),9,6)," 3d \n"))
    #write (FID,concat("VDot -0.710000 ",str(VDot(<1.1,1.2,1.5>,<-1.3,-1.4,+1.6>),9,6)," 3d \n"))
    #write (FID,concat("vdot -0.710000 ",str(vdot(<1.1,1.2,1.5>,<-1.3,-1.4,+1.6>),9,6)," 3d \n"))
    #write (FID,"-\n")
    #write (FID,concat("VDot +5.300000 ",str(VDot(<1.1,1.2,1.5,0.7>,<+1.3,+1.4,+1.6,-0.3>),9,6)," 4d \n"))
  //#write (FID,concat("vdot +5.300000 ",str(vdot(<1.1,1.2,1.5,0.7>,<+1.3,+1.4,+1.6,-0.3>),9,6)," 4d not supported \n"))
    #write (FID,concat("VDot -0.920000 ",str(VDot(<1.1,1.2,1.5,0.7>,<-1.3,-1.4,+1.6,-0.3>),9,6)," 4d \n"))
  //#write (FID,concat("vdot -0.920000 ",str(vdot(<1.1,1.2,1.5,0.7>,<-1.3,-1.4,+1.6,-0.3>),9,6)," 4d not supported \n"))
    #write (FID,"-\n")
    #write (FID,concat("VDot +5.280000 ",str(VDot(<1.1,1.2,1.5,0.7,0.1>,<+1.3,+1.4,+1.6,-0.3,-0.2>),9,6)," 5d \n"))
  //#write (FID,concat("vdot +5.280000 ",str(vdot(<1.1,1.2,1.5,0.7,0.1>,<+1.3,+1.4,+1.6,-0.3,-0.2>),9,6)," 5d not supported \n"))
    #write (FID,concat("VDot -0.940000 ",str(VDot(<1.1,1.2,1.5,0.7,0.1>,<-1.3,-1.4,+1.6,-0.3,-0.2>),9,6)," 5d \n"))
  //#write (FID,concat("vdot -0.940000 ",str(vdot(<1.1,1.2,1.5,0.7,0.1>,<-1.3,-1.4,+1.6,-0.3,-0.2>),9,6)," 5d not supported \n"))
    #write (FID,"\n")
    // VEq
    #write (FID,"---\n")
  //#write (FID,concat("VEq 1.000000 ",str(VEq(1,<1,1>),9,6)," flt - 2d error\n"))
    #write (FID,concat("VEq 1.000000 ",str(VEq(<1,1>,<1,1>),9,6)," 2d - 2d pass\n"))
    #write (FID,concat("VEq 0.000000 ",str(VEq(<1,1>,<2,1>),9,6)," 2d - 2d fail\n"))
    #write (FID,"-\n")
  //#write (FID,concat("VEq 1.000000 ",str(VEq(<1,1>,<1,1,1>),9,6)," 2d - 3d error\n"))
    #write (FID,concat("VEq 1.000000 ",str(VEq(<1,1,1>,<1,1,1>),9,6)," 3d - 3d pass\n"))
    #write (FID,concat("VEq 0.000000 ",str(VEq(<1,1,1>,<1,2,1>),9,6)," 3d - 3d fail\n"))
    #write (FID,"-\n")
  //#write (FID,concat("VEq 1.000000 ",str(VEq(<1,1,1,1>,<1,1,1>),9,6)," 4d - 3d error\n"))
    #write (FID,concat("VEq 1.000000 ",str(VEq(<1,1,1,1>,<1,1,1,1>),9,6)," 4d - 4d pass\n"))
    #write (FID,concat("VEq 0.000000 ",str(VEq(<2,1,1,1>,<1,1,1,1>),9,6)," 4d - 4d fail\n"))
    #write (FID,"-\n")
  //#write (FID,concat("VEq 1.000000 ",str(VEq(<1,1,1,1,1>,<1,1,1,1>),9,6)," 5d - 4d error\n"))
    #write (FID,concat("VEq 1.000000 ",str(VEq(<1,1,1,1,1>,<1,1,1,1,1>),9,6)," 5d - 5d pass\n"))
    #write (FID,concat("VEq 0.000000 ",str(VEq(<1,2,1,1,1>,<1,2,1,1,2>),9,6)," 5d - 5d fail\n"))
    #write (FID,"\n")
    // VLength
    #write (FID,"---\n")
  //#write (FID,concat("VLength +0.999849 ",str(VLength(1),9,6)," flt error\n"))
    #write (FID,concat("VLength +0.999849 ",str(VLength(<+0.707,+0.707>),9,6)," 2d\n"))
    #write (FID,concat("vlength +0.999849 ",str(vlength(<+0.707,+0.707>),9,6)," 2d\n"))
    #write (FID,concat("VLength +0.999849 ",str(VLength(<-0.707,+0.707>),9,6)," 2d\n"))
    #write (FID,concat("vlength +0.999849 ",str(vlength(<-0.707,+0.707>),9,6)," 2d\n"))
    #write (FID,"-\n")
    #write (FID,concat("VLength +1.224560 ",str(VLength(<+0.707,+0.707,+0.707>),9,6)," 3d\n"))
    #write (FID,concat("vlength +1.224560 ",str(vlength(<+0.707,+0.707,+0.707>),9,6)," 3d\n"))
    #write (FID,concat("VLength +1.224560 ",str(VLength(<-0.707,+0.707,+0.707>),9,6)," 3d\n"))
    #write (FID,concat("vlength +1.224560 ",str(vlength(<-0.707,+0.707,+0.707>),9,6)," 3d\n"))
    #write (FID,"-\n")
    #write (FID,concat("VLength +1.414000 ",str(VLength(<+0.707,+0.707,+0.707,+0.707>),9,6)," 4d\n"))
  //#write (FID,concat("vlength +1.414000 ",str(vlength(<+0.707,+0.707,+0.707,+0.707>),9,6)," 4d not supported\n"))
    #write (FID,concat("VLength +1.414000 ",str(VLength(<-0.707,+0.707,+0.707,+0.707>),9,6)," 4d\n"))
  //#write (FID,concat("vlength +1.414000 ",str(vlength(<-0.707,+0.707,+0.707,+0.707>),9,6)," 4d not supported\n"))
    #write (FID,"-\n")
    #write (FID,concat("VLength +1.580900 ",str(VLength(<+0.707,+0.707,+0.707,+0.707,+0.707>),9,6)," 5d\n"))
  //#write (FID,concat("vlength +1.580900 ",str(vlength(<+0.707,+0.707,+0.707,+0.707,+0.707>),9,6)," 5d not supported\n"))
    #write (FID,concat("VLength +1.580900 ",str(VLength(<-0.707,+0.707,+0.707,+0.707,+0.707>),9,6)," 5d\n"))
  //#write (FID,concat("vlength +1.580900 ",str(vlength(<-0.707,+0.707,+0.707,+0.707,+0.707>),9,6)," 5d not supported\n"))
    #write (FID,"\n")
    // VSqr
    #write (FID,"---\n")
  //#write (FID,concat("VSqr +0.999849 ",str(VSqr(1),9,6)," flt error\n"))
    #declare V2 = <-0.707,+0.717>;
    #declare Vec00 = VSqr(V2);
    #declare Vec01 = V2*V2;
    #write (FID,concat("VSqr  +0.499849 ",str(Vec00.u,9,6)," 2d u\n"))
    #write (FID,concat("VSqr  +0.514089 ",str(Vec00.v,9,6)," 2d v\n"))
    #write (FID,concat("V2*V2 +0.499849 ",str(Vec01.u,9,6)," 2d u\n"))
    #write (FID,concat("V2*V2 +0.514089 ",str(Vec01.v,9,6)," 2d v\n"))
    #write (FID,"-\n")
    #declare V3 = <-0.707,+0.717,+0.727>;
    #declare Vec00 = VSqr(V3);
    #declare Vec01 = V3*V3;
    #write (FID,concat("VSqr  +0.499849 ",str(Vec00.x,9,6)," 3d x\n"))
    #write (FID,concat("VSqr  +0.514089 ",str(Vec00.y,9,6)," 3d y\n"))
    #write (FID,concat("VSqr  +0.528529 ",str(Vec00.z,9,6)," 3d z\n"))
    #write (FID,concat("V3*V3 +0.499849 ",str(Vec01.x,9,6)," 3d x\n"))
    #write (FID,concat("V3*V3 +0.514089 ",str(Vec01.y,9,6)," 3d y\n"))
    #write (FID,concat("V3*V3 +0.528529 ",str(Vec01.z,9,6)," 3d z\n"))
    #write (FID,"-\n")
    #declare V4 = <-0.707,+0.717,+0.727,+0.737>;
    #declare Vec00 = VSqr(V4);
    #declare Vec01 = V4*V4;
    #write (FID,concat("VSqr  +0.499849 ",str(Vec00.x,9,6)," 4d x\n"))
    #write (FID,concat("VSqr  +0.514089 ",str(Vec00.y,9,6)," 4d y\n"))
    #write (FID,concat("VSqr  +0.528529 ",str(Vec00.z,9,6)," 4d z\n"))
    #write (FID,concat("VSqr  +0.543169 ",str(Vec00.t,9,6)," 4d t\n"))
    #write (FID,concat("V4*V4 +0.499849 ",str(Vec01.x,9,6)," 4d x\n"))
    #write (FID,concat("V4*V4 +0.514089 ",str(Vec01.y,9,6)," 4d y\n"))
    #write (FID,concat("V4*V4 +0.528529 ",str(Vec01.z,9,6)," 4d z\n"))
    #write (FID,concat("V4*V4 +0.543169 ",str(Vec01.t,9,6)," 4d t\n"))
    #write (FID,"-\n")
    #declare V5 = <-0.707,+0.717,+0.727,+0.737,+0.747>;
    #declare Vec00 = VSqr(V5);
    #declare Vec01 = V5*V5;
    #write (FID,concat("VSqr  +0.499849 ",str(Vec00.x,9,6)," 5d x\n"))
    #write (FID,concat("VSqr  +0.514089 ",str(Vec00.y,9,6)," 5d y\n"))
    #write (FID,concat("VSqr  +0.528529 ",str(Vec00.z,9,6)," 5d z\n"))
    #write (FID,concat("VSqr  +0.543169 ",str(Vec00.t,9,6)," 5d t\n"))
    #write (FID,concat("VSqr  +0.558009 ",str(Vec00.transmit,9,6)," 5d transmit\n"))
    #write (FID,concat("V5*V5 +0.499849 ",str(Vec01.x,9,6)," 5d x\n"))
    #write (FID,concat("V5*V5 +0.514089 ",str(Vec01.y,9,6)," 5d y\n"))
    #write (FID,concat("V5*V5 +0.528529 ",str(Vec01.z,9,6)," 5d z\n"))
    #write (FID,concat("V5*V5 +0.543169 ",str(Vec01.t,9,6)," 5d t\n"))
    #write (FID,concat("V5*V5 +0.558009 ",str(Vec01.transmit,9,6)," 5d transmit\n"))
    #write (FID,"\n")
    // VPow
    #write (FID,"---\n")
  //#write (FID,concat("VPow +0.999849 ",str(VPow(1,2),9,6)," flt first error\n"))
  //#write (FID,concat("VPow +0.999849 ",str(VPow(1,"Gack"),9,6)," not flt second error\n"))
    #declare V2 = <-0.707,+0.717>;
    #declare Vec00 = VPow(V2,2);
    #write (FID,concat("VPow  +0.499849 ",str(Vec00.u,9,6)," 2d u\n"))
    #write (FID,concat("VPow  +0.514089 ",str(Vec00.v,9,6)," 2d v\n"))
    #write (FID,"-\n")
    #declare V3 = <-0.707,+0.717,+0.727>;
    #declare Vec00 = VPow(V3,2);
    #write (FID,concat("VPow  +0.499849 ",str(Vec00.x,9,6)," 3d x\n"))
    #write (FID,concat("VPow  +0.514089 ",str(Vec00.y,9,6)," 3d y\n"))
    #write (FID,concat("VPow  +0.528529 ",str(Vec00.z,9,6)," 3d z\n"))
    #write (FID,"-\n")
    #declare V4 = <-0.707,+0.717,+0.727,+0.737>;
    #declare Vec00 = VPow(V4,2);
    #write (FID,concat("VPow  +0.499849 ",str(Vec00.x,9,6)," 4d x\n"))
    #write (FID,concat("VPow  +0.514089 ",str(Vec00.y,9,6)," 4d y\n"))
    #write (FID,concat("VPow  +0.528529 ",str(Vec00.z,9,6)," 4d z\n"))
    #write (FID,concat("VPow  +0.543169 ",str(Vec00.t,9,6)," 4d t\n"))
    #write (FID,"-\n")
    #declare V5 = <-0.707,+0.717,+0.727,+0.737,+0.747>;
    #declare Vec00 = VPow(V5,2);
    #write (FID,concat("VPow  +0.499849 ",str(Vec00.x,9,6)," 5d x\n"))
    #write (FID,concat("VPow  +0.514089 ",str(Vec00.y,9,6)," 5d y\n"))
    #write (FID,concat("VPow  +0.528529 ",str(Vec00.z,9,6)," 5d z\n"))
    #write (FID,concat("VPow  +0.543169 ",str(Vec00.t,9,6)," 5d t\n"))
    #write (FID,concat("VPow  +0.558009 ",str(Vec00.transmit,9,6)," 5d transmit\n"))
    #write (FID,"\n")
    // VZero
    #write (FID,"---\n")
  //#write (FID,concat("VZero 1.000000 ",str(VZero(1),9,6)," flt error\n"))
    #write (FID,concat("VZero 1.000000 ",str(VZero(<0,0>),9,6)," 2d pass\n"))
    #write (FID,concat("VZero 0.000000 ",str(VZero(<0,1>),9,6)," 2d fail\n"))
    #write (FID,"-\n")
    #write (FID,concat("VZero 1.000000 ",str(VZero(<0,0,0>),9,6)," 3d pass\n"))
    #write (FID,concat("VZero 0.000000 ",str(VZero(<1,0,0>),9,6)," 3d fail\n"))
    #write (FID,"-\n")
    #write (FID,concat("VZero 1.000000 ",str(VZero(<0,0,0,0>),9,6)," 4d pass\n"))
    #write (FID,concat("VZero 0.000000 ",str(VZero(<0,0,1,0>),9,6)," 4d fail\n"))
    #write (FID,"-\n")
    #write (FID,concat("VZero 1.000000 ",str(VZero(<0,0,0,0,0>),9,6)," 5d pass\n"))
    #write (FID,concat("VZero 0.000000 ",str(VZero(<0,0,1,0,0>),9,6)," 5d fail\n"))
    #write (FID,"\n")
    // VNormalize
    #write (FID,"---\n")
  //#write (FID,concat("VNormalize 1.000000 ",str(VNormalize(1),9,6)," flt error\n"))
    #declare VecZ = VNormalize(<0.0,0.0>);
    #declare VecA = VNormalize(<0.1,0.2>);
    #declare vecZ = vnormalize(<0.0,0.0>);
    #declare vecA = vnormalize(<0.1,0.2>);
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.u,9,6)," 2d VecZ.u\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.v,9,6)," 2d VecZ.v\n"))
    #write (FID,concat("vnormalize 0.000000 ",str(vecZ.u,9,6)," 2d vecZ.u\n"))
    #write (FID,concat("vnormalize 0.000000 ",str(vecZ.v,9,6)," 2d vecZ.v\n"))
    #write (FID,concat("VNormalize 0.447214 ",str(VecA.u,9,6)," 2d VecA.u\n"))
    #write (FID,concat("VNormalize 0.894427 ",str(VecA.v,9,6)," 2d VecA.v\n"))
    #write (FID,concat("vnormalize 0.447214 ",str(vecA.u,9,6)," 2d vecA.u\n"))
    #write (FID,concat("vnormalize 0.894427 ",str(vecA.v,9,6)," 2d vecA.v\n"))
    #write (FID,"-\n")
    #declare VecZ = VNormalize(<0.0,0.0,0.0>);
    #declare VecA = VNormalize(<0.1,0.2,0.3>);
    #declare vecZ = vnormalize(<0.0,0.0,0.0>);
    #declare vecA = vnormalize(<0.1,0.2,0.3>);
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.x,9,6)," 3d VecZ.x\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.y,9,6)," 3d VecZ.y\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.z,9,6)," 3d VecZ.z\n"))
    #write (FID,concat("vnormalize 0.000000 ",str(vecZ.x,9,6)," 3d vecZ.x\n"))
    #write (FID,concat("vnormalize 0.000000 ",str(vecZ.y,9,6)," 3d vecZ.y\n"))
    #write (FID,concat("vnormalize 0.000000 ",str(vecZ.z,9,6)," 3d vecZ.z\n"))
    #write (FID,concat("VNormalize 0.267261 ",str(VecA.x,9,6)," 3d VecA.x\n"))
    #write (FID,concat("VNormalize 0.534522 ",str(VecA.y,9,6)," 3d VecA.y\n"))
    #write (FID,concat("VNormalize 0.801784 ",str(VecA.z,9,6)," 3d VecA.z\n"))
    #write (FID,concat("vnormalize 0.267261 ",str(vecA.x,9,6)," 3d vecA.x\n"))
    #write (FID,concat("vnormalize 0.534522 ",str(vecA.y,9,6)," 3d vecA.y\n"))
    #write (FID,concat("vnormalize 0.801784 ",str(vecA.z,9,6)," 3d vecA.z\n"))
    #write (FID,"-\n")
    #declare VecZ = VNormalize(<0.0,0.0,0.0,0.0>);
    #declare VecA = VNormalize(<0.1,0.2,0.3,0.4>);
  //#declare vecZ = vnormalize(<0.0,0.0,0.0,0.0>); // 4d vectors not supported.
  //#declare vecA = vnormalize(<0.1,0.2,0.3,0.4>); // 4d vectors not supported.
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.x,9,6)," 4d VecZ.x\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.y,9,6)," 4d VecZ.y\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.z,9,6)," 4d VecZ.z\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.t,9,6)," 4d VecZ.t\n"))
    #write (FID,concat("VNormalize 0.182574 ",str(VecA.x,9,6)," 4d VecA.x\n"))
    #write (FID,concat("VNormalize 0.365148 ",str(VecA.y,9,6)," 4d VecA.y\n"))
    #write (FID,concat("VNormalize 0.547723 ",str(VecA.z,9,6)," 4d VecA.z\n"))
    #write (FID,concat("VNormalize 0.730297 ",str(VecA.t,9,6)," 4d VecA.t\n"))
    #write (FID,"-\n")
    #declare VecZ = VNormalize(<0.0,0.0,0.0,0.0,0.0>);
    #declare VecA = VNormalize(<0.1,0.2,0.3,0.4,0.5>);
  //#declare vecZ = vnormalize(<0.0,0.0,0.0,0.0,0.0>); // 5d vectors not supported.
  //#declare vecA = vnormalize(<0.1,0.2,0.3,0.4,0.5>); // 5d vectors not supported.
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.x,9,6)," 5d VecZ.x\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.y,9,6)," 5d VecZ.y\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.z,9,6)," 5d VecZ.z\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.t,9,6)," 5d VecZ.t\n"))
    #write (FID,concat("VNormalize 0.000000 ",str(VecZ.transmit,9,6)," 5d VecZ.transmit\n"))
    #write (FID,concat("VNormalize 0.134840 ",str(VecA.x,9,6)," 5d VecA.x\n"))
    #write (FID,concat("VNormalize 0.269680 ",str(VecA.y,9,6)," 5d VecA.y\n"))
    #write (FID,concat("VNormalize 0.404520 ",str(VecA.z,9,6)," 5d VecA.z\n"))
    #write (FID,concat("VNormalize 0.539360 ",str(VecA.t,9,6)," 5d VecA.t\n"))
    #write (FID,concat("VNormalize 0.674200 ",str(VecA.transmit,9,6)," 5d VecA.transmit\n"))
    #write (FID,"\n")
    // VCos_Angle
    // Plus minor testing of new v3.8 tuple #declare style. (Parser test)
    #declare (Vec00,Vec01,Vec02,Vec03,Vec04) =
        (<+1,0,+0>,<-1,0,+0>,<+0,1,+0>,<+0,0,-1>,<+1,0,+1>);
    #declare (A00,A01,A02,A03,A04) =
        (
        VCos_Angle(Vec00,Vec00),VCos_Angle(Vec00,Vec01),
        VCos_Angle(Vec00,Vec02),VCos_Angle(Vec00,Vec03),VCos_Angle(Vec00,Vec04)
        );
    #write (FID,concat("VCos_Angle +1.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VCos_Angle -1.000000 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VCos_Angle +0.000000 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VCos_Angle +0.000000 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VCos_Angle +0.707107 ",str(A04,9,6)," A04\n"))
    #write (FID,"\n")
    // VAngle
    // Plus minor testing of new v3.8 vector components to variables style. (Parser test)
    #declare VecTmp = <VAngle(Vec00,Vec00),VAngle(Vec00,Vec01),VAngle(Vec00,Vec02),
        VAngle(Vec00,Vec03),VAngle(Vec00,Vec04)>;
    #declare <A00,A01,A02,A03,A04> = VecTmp;
    #undef VecTmp
    #write (FID,concat("VAngle +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VAngle +3.141593 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VAngle +1.570796 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VAngle +1.570796 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VAngle +0.785398 ",str(A04,9,6)," A04\n"))
    #write (FID,"---\n")
    #declare VecD = <+1.73,+1.73,+1.73>;
    #declare VecTmp = <VAngle(VecD,VecD),VAngle(VecD,-VecD)>;
    #declare <A00,A01> = VecTmp;
    #write (FID,concat("VAngle +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VAngle +3.141593 ",str(A01,9,6)," A01\n"))
    #ifdef (VecD) #undef VecD #end
    #write (FID,"\n")
    // VAngleD
    #declare A00 = VAngleD(Vec00,Vec00);
    #declare A01 = VAngleD(Vec00,Vec01);
    #declare A02 = VAngleD(Vec00,Vec02);
    #declare A03 = VAngleD(Vec00,Vec03);
    #declare A04 = VAngleD(Vec00,Vec04);
    #write (FID,concat("VAngleD +0.000000   ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VAngleD +180.000000 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VAngleD +90.000000  ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VAngleD +90.000000  ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VAngleD +45.000000  ",str(A04,9,6)," A04\n"))
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #write (FID,"\n")
    // VRotation
    // #declare (Vec00,Vec01,Vec02,Vec03,Vec04) =
    //   (<+1,0,+0>,<-1,0,+0>,<+0,1,+0>,<+0,0,-1>,<+1,0,+1>);
    #declare A00 = VRotation(Vec00,Vec00,Vec04);
    #declare A01 = VRotation(Vec00,Vec02,Vec04);
    #declare A02 = VRotation(Vec00,Vec01,Vec04);
    #declare A03 = VRotation(Vec00,Vec00,Vec03);
    #declare A04 = VRotation(Vec00,Vec02,Vec03);
    #declare A05 = VRotation(Vec00,Vec01,Vec03);
    #declare A06 = VRotation(Vec00,<-1,+1,+0>,Vec03);
    #write (FID,concat("VRotation +0.000000   ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VRotation +1.570796   ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VRotation +3.141593   ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VRotation +0.000000   ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VRotation -1.570796   ",str(A04,9,6)," A04\n"))
    #write (FID,concat("VRotation +3.141593   ",str(A05,9,6)," A05\n"))
    #write (FID,concat("VRotation -2.356194   ",str(A06,9,6)," A05\n"))
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #ifdef (A05) #undef A05 #end
    #ifdef (A06) #undef A06 #end
    #write (FID,"\n")
    // VRotationD
    #declare A00 = VRotationD(Vec00,Vec00,Vec04);
    #declare A01 = VRotationD(Vec00,Vec02,Vec04);
    #declare A02 = VRotationD(Vec00,Vec01,Vec04);
    #declare A03 = VRotationD(Vec00,Vec00,Vec03);
    #declare A04 = VRotationD(Vec00,Vec02,Vec03);
    #declare A05 = VRotationD(Vec00,Vec01,Vec03);
    #declare A06 = VRotationD(Vec00,<-1,+1,+0>,Vec03);
    #write (FID,concat("VRotationD +0.000000   ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VRotationD +90.000000  ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VRotationD +180.000000 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VRotationD +0.000000   ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VRotationD -90.000000  ",str(A04,9,6)," A04\n"))
    #write (FID,concat("VRotationD +180.000000 ",str(A05,9,6)," A05\n"))
    #write (FID,concat("VRotationD -135.000000 ",str(A06,9,6)," A05\n"))
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #ifdef (A05) #undef A05 #end
    #ifdef (A06) #undef A06 #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #ifdef (Vec03) #undef Vec03 #end
    #ifdef (Vec04) #undef Vec04 #end
    #write (FID,"\n")
    // VDist (Changed to #error recommending vlength()).
    // A few vlength tests below.
    #declare Vec00 = <+0.707,+0.707,+0.707>;
    #declare Vec01 = <-0.707,-0.707,-0.707>;
    #declare D00 = vlength(Vec00 - Vec01);
    #declare D01 = vlength(Vec01 - Vec00);
    #declare D02 = vlength(Vec00 - Vec00);
    #write (FID,concat("vlength +2.449120   ",str(D00,9,6)," D00\n"))
    #write (FID,concat("vlength +2.449120   ",str(D01,9,6)," D01\n"))
    #write (FID,concat("vlength +0.000000   ",str(D02,9,6)," D02\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (D00) #undef D00 #end
    #ifdef (D01) #undef D01 #end
    #ifdef (D02) #undef D02 #end
    #write (FID,"\n")
    // VPerp_To_Vector
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VPerp_To_Vector(Vec00);
    #declare VP01 = VPerp_To_Vector(Vec01);
    #write (FID,concat("VPerp_To_Vector +1.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VPerp_To_Vector +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VPerp_To_Vector +0.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #write (FID,concat("VPerp_To_Vector +0.707107   ",str(VP01.x,9,6)," VP01.x\n"))
    #write (FID,concat("VPerp_To_Vector -0.707107   ",str(VP01.y,9,6)," VP01.y\n"))
    #write (FID,concat("VPerp_To_Vector +0.000000   ",str(VP01.z,9,6)," VP01.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #ifdef (VP01) #undef VP01 #end
    #write (FID,"\n")
    // VPerp_To_Plane
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VPerp_To_Plane(Vec00,Vec01);
    #write (FID,concat("VPerp_To_Plane +0.707107   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VPerp_To_Plane +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VPerp_To_Plane -0.707107   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VPerp_Adjust
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VPerp_Adjust(Vec00,Vec01);
    #write (FID,concat("VPerp_Adjust -0.408248   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VPerp_Adjust +0.816497   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VPerp_Adjust -0.408248   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VProject_Plane
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VProject_Plane(Vec01,Vec00);
    #write (FID,concat("VProject_Plane +0.707000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VProject_Plane +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VProject_Plane +0.707000   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VProject_Axis
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VProject_Axis(Vec01,Vec00);
    #write (FID,concat("VProject_Axis +0.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VProject_Axis +0.707000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VProject_Axis +0.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VMin, VMax  (In performance testing pure SDL much faster these now errors if called.
//  #declare Vec00 = <+1.0,+2.0,+3.0>;
//  #declare VarMin = VMin(Vec00);
//  #declare VarMax = VMax(Vec00);
//  #write (FID,concat("Vmin +1.000000   ",str(VarMin,9,6)," Vmin\n"))
//  #write (FID,concat("Vmax +3.000000   ",str(VarMax,9,6)," Vmax\n"))
//  #ifdef (Vec00) #undef Vec00 #end
//  #ifdef (VarMin) #undef VarMin #end
//  #ifdef (VarMax) #undef VarMax #end
//  #write (FID,"\n")
    // VWith_Len
    #declare Vec00 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VWith_Len(Vec00,1.0);
    #write (FID,concat("VWith_Len +0.577350   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VWith_Len +0.577350   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VWith_Len +0.577350   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")

    #fclose FID

    #error "\n\n=== Self testing. Stop before render. ===\n"


#end // End self testing

#version MATH_INC_TEMP;
#end //math.inc

