// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File  (povr version ONLY)
// File: transforms.inc
// Last updated: 06 June 2021

// Description: Macros for dealing with transforms

// Better alternate transforms can be found in vectors.inc as indicated below.
// Traditional POV-Ray uses all lower case inbuilt looking names for the
// shipped VInv_Transform and VTransform macros. The povr branch uses inbuilt
// looking lower case names in shipped files only where functionality is truly
// inbuilt.

//  Align_Trans
//  Axial_Scale_Trans
//  Axis_Rotate_Trans    - Use AxisRotateTransform()
//  Center_Trans
//  Matrix_Trans         - Identical to TransformFromVectors()
//  Point_At_Trans
//  Reorient_Trans       - Use ReorientTransform()
//  Rotate_Around_Trans
//  Shear_Trans
//  Spline_Trans
//  VInv_Transform       - Use VectorInverseTransform()
//  VTransform           - Use VectorTransform()

//------------------------------------------------------------------------------

#ifndef(TRANSFORMS_INC_TEMP)
#declare TRANSFORMS_INC_TEMP = version;
#version 3.8; // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including transforms.inc\n"
#end

#include "math.inc"     // VPerp_To_Plane
#include "vectors.inc"  // This file now wrapping macros in vectors.inc

// -------------------------
// Transformation macros:
// -------------------------

// Reorients and deforms object so original x axis points along _a, original y
// along _b, and original z along _c.

#macro Shear_Trans(_a,_b,_c)
    transform {
        matrix <
            _a.x,_a.y,_a.z,
            _b.x,_b.y,_b.z,
            _c.x,_c.y,_c.z,
            0,     0,    0
        >
    }
#end

// An identical in function version called TransformFromVectors() exists in
// vectors.inc.
#macro Matrix_Trans(_a,_b,_c,_d)
    transform {
        matrix <
            _a.x,_a.y,_a.z,
            _b.x,_b.y,_b.z,
            _c.x,_c.y,_c.z,
            _d.x,_d.y,_d.z
        >
    }
#end


// "stretch" along a specific axis
#macro Axial_Scale_Trans(_axis,_amt)
    transform {
        transform {Point_At_Trans(_axis) inverse}
        scale <1,_amt,1>
        Point_At_Trans(_axis)
    }
#end


// rotate around a specific axis
// Author: Rune S. Johansen
// Update as able to AxisRotateTransform() in vectors.inc
#macro Axis_Rotate_Trans(_axis,_angle)
  //#error "Axis_Rotate_Trans in math.inc called. Use AxisRotateTransform()"
    AxisRotateTransform(_axis,_angle)
  //#local vX = vaxis_rotate(x,_axis,_angle);
  //#local vY = vaxis_rotate(y,_axis,_angle);
  //#local vZ = vaxis_rotate(z,_axis,_angle);
  //transform {
  //    matrix < vX.x,vX.y,vX.z, vY.x,vY.y,vY.z, vZ.x,vZ.y,vZ.z, 0,0,0 >
  //}
#end


// Rotate around a specific point
#macro Rotate_Around_Trans(_rotation,_point)
    transform {
        translate -_point
        rotate _rotation
        translate _point
    }
#end


// Aligns _axis1 with _axis2. Code based on original Reorient() macro by John
// VanSickle. Note. The ReorientTransform() version of this functionality in
// vectors.inc is recommended over Reorient_Trans().
#macro Reorient_Trans(_axis1,_axis2)
    #local vX1 = vnormalize(_axis1);
    #local vX2 = vnormalize(_axis2);
    #local Y = vcross(vX1, vX2);
    #if(vlength(Y) > 0)
        #local vY = vnormalize(Y);
        #local vZ1 = vnormalize(vcross(vX1, vY));
        #local vZ2 = vnormalize(vcross(vX2, vY));
        transform {
            matrix < vX1.x, vY.x,vZ1.x, vX1.y,vY.y,vZ1.y, vX1.z,vY.z, vZ1.z, 0,0,0 >
            matrix < vX2.x,vX2.y,vX2.z,  vY.x,vY.y, vY.z, vZ2.x,vZ2.y,vZ2.z, 0,0,0 >
        }
    #else
        #if (vlength(vX1-vX2)=0)
            transform {}
        #else
            #local vZ = OrthogonalVector(vX2);
            transform { AxisRotateTransform(vZ,180) }
        #end
    #end
#end


// Points y axis along _axis.
#macro Point_At_Trans(_axis)
    #local Y = vnormalize(_axis);
    #local X = OrthogonalVector(Y);
    #local Z = vcross(X,Y);
    Shear_Trans(X,Y,Z)
#end


// Calculates a transformation which will center the bounding box of Object
// along specified axis Axis
// Usage:
// object {MyObj
//     Center_Trans(MyObj, x) center along x axis
// You can also center along multiple axis:
//     Center_Trans(MyObj, x+y) center along x and y axis
#macro Center_Trans(_object,_axis)
    #local Mn = min_extent(_object);
    #local Mx = max_extent(_object);
    transform {translate -_axis*((Mx - Mn)/2 + Mn)}
#end

// Calculates a transformation which will align the bounding box
// of an object to a point. Negative values on _axis will align to
// the sides facing the negative ends of the coordinate system,
// positive values will align to the opposite sides, 0 means
// not to do any alignment on that axis.
// Usage:
// object {MyObj
//     Align_Trans(MyObj, x, Pt)
//     Align right side of object to be coplanar with Pt
//     Align_Trans(MyObj, -y, Pt)
//     Align bottom of object to be coplanar with Pt
#macro Align_Trans(_object,_axis,_pt)
    #local Mn = min_extent(_object);
    #local Mx = max_extent(_object);
    transform {
        #if(_axis.x < -0.5)
            translate x*(_pt - Mn.x)
        #else
            #if(_axis.x > 0.5)
                translate x*(_pt - Mx.x)
            #end
        #end

        #if(_axis.y < -0.5)
            translate y*(_pt - Mn.y)
        #else
            #if(_axis.y > 0.5)
                translate y*(_pt - Mx.y)
            #end
        #end

        #if(_axis.z < -0.5)
            translate z*(_pt - Mn.z)
        #else
            #if(_axis.z > 0.5)
                translate z*(_pt - Mx.z)
            #end
        #end
    }
#end

// Aligns an object to a spline for a given time value.
// The Z axis of the object will point in the forward direction
// of the spline and the Y axis of the object will point upwards.
//
// usage:     object {MyObj Spline_Trans(MySpline, clock, y, 0.1, 0.5)}
//
// _spline:    The spline that the object is aligned to.
//
// _time:      The time value to feed to the spline, for example clock.
//
// _sky:       The vector that is upwards in your scene, usually y.
//
// _foresight: How much in advance the object will turn and bank.
//             Values close to 0 will give precise results, while higher
//             values give smoother results. It will not affect parsing
//             speed, so just find the value that looks best.
//
// _banking:   How much the object tilts when turning. Note that the amount
//             of tilting is equally much controlled by the ForeSight value.
//
// Author: Rune S. Johansen
//
// May 27, 2021 - (povr) Now using VectorReject over original VProject_Plane
#macro Spline_Trans(_spline,_time,_sky,_foresight,_banking)
    #local Location     = <0,0,0>+_spline(_time);
    #local LocationNext = <0,0,0>+_spline(_time+_foresight);
    #local LocationPrev = <0,0,0>+_spline(_time-_foresight);
    #local Forward = vnormalize(LocationNext-Location);
    #local Right   = VPerp_To_Plane(_sky,Forward);
    #local Up      = VPerp_To_Plane(Forward,Right);
    #local Matrix  = Matrix_Trans(Right,Up,Forward,Location)
    #local BankingRotation =
    degrees(atan2(
        VRotation(
            VectorReject((LocationNext-Location),_sky),
            VectorReject((Location-LocationPrev),_sky),
            Up
        )*_banking
        ,1
    ));
    transform {
        rotate BankingRotation*z
        transform Matrix
    }
#end


// Use VectorTransform in vectors.inc
#macro VTransform(_vec, _trans)
  //#error "Use VectorTransform in vectors.inc"
    #local Fn = function { transform { _trans } }
    #local Result = (Fn(_vec.x,_vec.y,_vec.z));
    Result
#end


// Use VectorInverseTransform in vectors.inc
#macro VInv_Transform(_vec, _trans)
  //#error "Use VectorInverseTransform in vectors.inc"
    #local Fn = function { transform { _trans inverse } }
    #local Result = (Fn(_vec.x,_vec.y,_vec.z));
    Result
#end

//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="transforms.inc")

    #fopen FID "SelfTestTransformsInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
    // Run some macros to be sure parse works. Assume failure and correct later if all parse OK.
    #write (FID,"\n---\n")
    #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))

    #declare Xfrm00 = Shear_Trans(<0.1,0.1,0.0>,<-0.1,0.1,0.0>,<0.0,0.1,0.1>);
    #declare Xfrm01 = Matrix_Trans(<0.1,0.1,0.0>,<-0.1,0.1,0.0>,<0.0,0.1,0.1>,<0.1,0.1,0.1>);
    #declare Xfrm02 = Point_At_Trans(<0.707,0.707,0.707>);
    #declare Xfrm03 = Center_Trans(sphere{0.5, 0.5},y)
    #declare Xfrm04 = Align_Trans(sphere{0.5, 0.5},x,<.3,.4,.5>)
    #ifdef (Xfrm00) #undef Xfrm00 #end
    #ifdef (Xfrm01) #undef Xfrm01 #end
    #ifdef (Xfrm02) #undef Xfrm02 #end
    #ifdef (Xfrm03) #undef Xfrm03 #end
    #ifdef (Xfrm04) #undef Xfrm04 #end

    #write (FID,"\n")
    #fclose FID


    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestTransformsInc.txt" write
    // If all macro call tests above work create output for what was tested.
    #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Shear_Trans OK\n"))
    #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Matrix_Trans OK\n"))
    #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Point_At_Trans OK\n"))
    #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Center_Trans OK\n"))
    #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Align_Trans OK\n"))
    #write (FID,"\n")
    // Align_Trans
    #declare Sph00 = sphere { <0,1.5,0>, 0.5 }
    #declare Obj00 = object {Sph00 Align_Trans(Sph00,-x,<0,0,0>)}
    #declare Obj00 = object {Obj00 Align_Trans(Obj00,-y,<0,0,0>)}
    #declare MinExt = min_extent(Obj00);
    #declare MaxExt = max_extent(Obj00);
    #write (FID,concat("Align_Trans +0.000000 ",str(MinExt.x,9,6)," min x\n"))
    #write (FID,concat("Align_Trans +0.000000 ",str(MinExt.y,9,6)," min y\n"))
    #write (FID,concat("Align_Trans -0.500000 ",str(MinExt.z,9,6)," min z\n"))
    #write (FID,concat("Align_Trans +1.000000 ",str(MaxExt.x,9,6)," max x\n"))
    #write (FID,concat("Align_Trans +1.000000 ",str(MaxExt.y,9,6)," max y\n"))
    #write (FID,concat("Align_Trans +0.500000 ",str(MaxExt.z,9,6)," max z\n"))
    #ifdef (Sph00) #undef Sph00 #end
    #ifdef (Obj00) #undef Obj00 #end
    #ifdef (MinExt) #undef MinExt #end
    #ifdef (MaxExt) #undef MaxExt #end
    #write (FID,"\n")
    // Axial_Scale_Trans, Axis_Rotate_Trans <-- Now wrapper to AxisRotateTransform()
    #declare Vec00 = <0.5,0.5,0.5>;
    #declare Vec01 = VectorTransform(Vec00,Axial_Scale_Trans(Vec00,2.0));
    #write (FID,concat("Axial_Scale_Trans +1.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("Axial_Scale_Trans +1.000000 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("Axial_Scale_Trans +1.000000 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"---\n")
    #declare Vec01 = VectorTransform(Vec00,Axis_Rotate_Trans(<0,0,1>,45.0));
    #write (FID,concat("Axis_Rotate_Trans +0.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("Axis_Rotate_Trans +0.707107 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("Axis_Rotate_Trans +0.500000 ",str(Vec01.z,9,6)," z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #write (FID,"\n")
    // Center_Trans
    #declare Sph00 = sphere { <0,1.5,0>, 0.5 }
    #declare Obj00 = object {Sph00 Center_Trans(Sph00,x+y+z)}
    #declare MinExt = min_extent(Obj00);
    #declare MaxExt = max_extent(Obj00);
    #write (FID,concat("Center_Trans -0.500000 ",str(MinExt.x,9,6)," min x\n"))
    #write (FID,concat("Center_Trans -0.500000 ",str(MinExt.y,9,6)," min y\n"))
    #write (FID,concat("Center_Trans -0.500000 ",str(MinExt.z,9,6)," min z\n"))
    #write (FID,concat("Center_Trans +0.500000 ",str(MaxExt.x,9,6)," max x\n"))
    #write (FID,concat("Center_Trans +0.500000 ",str(MaxExt.y,9,6)," max y\n"))
    #write (FID,concat("Center_Trans +0.500000 ",str(MaxExt.z,9,6)," max z\n"))
    #ifdef (Sph00) #undef Sph00 #end
    #ifdef (Obj00) #undef Obj00 #end
    #ifdef (MinExt) #undef MinExt #end
    #ifdef (MaxExt) #undef MaxExt #end
    #write (FID,"\n")
    // Point_At_Trans
    #declare Vec00 = <1.5,1.5,1.5>;
    #declare Vec01 = <0.0,0.5,0>;
    #declare Vec02 = VectorTransform(Vec01,Point_At_Trans(Vec00));
    #write (FID,concat("Point_At_Trans +0.288675 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Point_At_Trans +0.288675 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Point_At_Trans +0.288675 ",str(Vec02.z,9,6)," z\n"))
    #write (FID,"---\n")
    #declare Vec00 = <0.1,0.1,0.1>;
    #declare Vec01 = <0.0,-0.5,0>;
    #declare Vec02 = VectorTransform(Vec01,Point_At_Trans(Vec00));
    #write (FID,concat("Point_At_Trans -0.288675 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Point_At_Trans -0.288675 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Point_At_Trans -0.288675 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // Reorient_Trans
    #declare Vec00 = <1.5,1.5,1.5>;
    #declare Vec01 = <0.0,0.5,0>;
    #declare Vec02 = VectorTransform(Vec01,Reorient_Trans(Vec01,Vec00));
    #write (FID,concat("Reorient_Trans +0.288675 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Reorient_Trans +0.288675 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Reorient_Trans +0.288675 ",str(Vec02.z,9,6)," z\n"))
    #write (FID,"---\n")
    #declare Vec00 = <0.1,0.1,0.1>;
    #declare Vec01 = <0.0,-0.5,0>;
    #declare Vec02 = VectorTransform(Vec01,Reorient_Trans(Vec01,Vec00));
    #write (FID,concat("Reorient_Trans +0.288675 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Reorient_Trans +0.288675 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Reorient_Trans +0.288675 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // Rotate_Around_Trans
    #declare Vec00 = <0.5,0.5,0.5>;
    #declare Vec01 = <-0.2,-0.2,-0.2>;
    #declare Vec02 = VectorTransform(Vec00,Rotate_Around_Trans(<-45,0,0>,Vec00));
    #write (FID,concat("Rotate_Around_Trans +0.500000 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Rotate_Around_Trans +0.500000 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Rotate_Around_Trans +0.500000 ",str(Vec02.z,9,6)," z\n"))
    #write (FID,"---\n")
    #declare Vec02 = VectorTransform(Vec01,Rotate_Around_Trans(<-45,0,0>,Vec00));
    #write (FID,concat("Rotate_Around_Trans -0.200000 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Rotate_Around_Trans -0.489949 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Rotate_Around_Trans +0.500000 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // Spline_Trans
    #declare Spline00 = spline {
        cubic_spline
        -2, <-5, 3, 0>, // control point
        -1, <-2, 2, 0>, // control point
        00, < 0, 2, 0>, // start
        01, < 2, 2, 0>,
        02, < 5, 3, 0>, // through yellow ring
        03, < 5, 4, 4>,
        04, < 0, 5, 5>, // around
        05, <-2, 4, 9>, // the
        06, < 2, 3, 9>, // green
        07, < 0, 2, 5>, // pole
        08, <-5, 2, 4>,
        09, <-5, 3, 0>, // through blue ring
        10, <-2, 2, 0>,
        11, < 0, 2, 0>, // stop
        12, < 2, 2, 0>, // control point
        13, < 5, 3, 0>, // control point
    }
    #declare Vec00 = <0,0.5,0>;
    #declare Vec01 = VectorTransform(Vec00,Spline_Trans(Spline00,mod((2/6),1)*11,y,0.5,0.5));
    #declare Vec02 = VectorTransform(Vec00,Spline_Trans(Spline00,mod((4/6),1)*11,y,0.5,0.5));
    #write (FID,concat("Spline_Trans +1.687805 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("Spline_Trans +5.307377 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("Spline_Trans +4.618785 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"---\n")
    #write (FID,concat("Spline_Trans -1.629590 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("Spline_Trans +2.382157 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("Spline_Trans +4.637326 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Spline00) #undef Spline00 #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // VTransform, VInv_Transform(vec, trans) See top. Use VectorTransform, VectorInverseTransform.
    #declare Vec00 = <1,0,0>;
    #declare Xfrm00 = transform { rotate z*90 }
    #declare Vec01 = VTransform(Vec00,Xfrm00);
    #declare Vec02 = VInv_Transform(Vec00,Xfrm00);
    #write (FID,concat("VTransform +0.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("VTransform +1.000000 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("VTransform +0.000000 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"---\n")
    #write (FID,concat("VInv_Transform +0.000000 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("VInv_Transform -1.000000 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("VInv_Transform +0.000000 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Xfrm00) #undef Xfrm00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")

    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"

#end // End self testing

#version TRANSFORMS_INC_TEMP;
#end
