// ===== 1 ======= 2 ======= 3 ======= 4 ======= 5 ======= 6 ======= 7 ======= 8 ======= 9 ======= 10
/*

https://github.com/t-o-k/Useful-POV-Ray-macros
Copyright (c) 2021 Tor Olav Kristensen, http://subcube.com
Use of this source code is governed by the GNU Lesser General Public License version 3,
which can be found in the LICENSE file.

Relevant Wikipedia articles:
https://en.wikipedia.org/wiki/Euclidean_vector
https://en.wikipedia.org/wiki/Multiplication_of_vectors
https://en.wikipedia.org/wiki/Scalar_projection
https://en.wikipedia.org/wiki/Vector_projection
https://en.wikipedia.org/wiki/Triple_product
https://en.wikipedia.org/wiki/Rotation_matrix
https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
https://en.wikipedia.org/wiki/Euler%E2%80%93Rodrigues_formula

6f480a197e0a35f1a27058f5bd54fb66c1aad74b  June 03, 2021

WFP - Add test cases. Add usual include file wraps.

*/
// ===== 1 ======= 2 ======= 3 ======= 4 ======= 5 ======= 6 ======= 7 ======= 8 ======= 9 ======= 10

#ifndef(VECTORS_INC_TEMP)
#declare VECTORS_INC_TEMP = version;
#version 3.7;

#ifdef(View_POV_Include_Stack)
    #debug "including vectors.inc\n"
#end

// ===== 1 ======= 2 ======= 3 ======= 4 ======= 5 ======= 6 ======= 7 ======= 8 ======= 9 ======= 10

#macro ScalarTripleProduct(vA, vB, vC)

    vdot(vA, vcross(vB, vC))

#end // macro ScalarTripleProduct


#macro VectorTripleProduct(vA, vB, vC)

    vcross(vA, vcross(vB, vC))

#end // macro VectorTripleProduct


/*
// The version above is faster

#macro VectorTripleProduct(vA, vB, vC)

    (vdot(vA, vC)*vB - vdot(vA, vB)*vC)

#end // macro VectorTripleProduct
*/


#macro VectorCos(vA, vB)

    min(max(-1, vdot(vnormalize(vA), vnormalize(vB))), +1)

#end // macro VectorCos


#macro VectorSin(vA, vB)

    min(vlength(vcross(vnormalize(vA), vnormalize(vB))), +1)

#end // macro VectorSin


#macro AngleBetweenVectors(vA, vB)

    acos(min(max(-1, vdot(vnormalize(vA), vnormalize(vB))), +1))

#end // macro AngleBetweenVectors


// Use this version if the vectors may be almost parallel or antiparallel
// The version above is faster

#macro AccurateAngleBetweenVectors(vA, vB)

    #local vAn = vnormalize(vA);
    #local vBn = vnormalize(vB);

    (2*atan2(vlength(vAn - vBn), vlength(vAn + vBn)))

#end // macro AccurateAngleBetweenVectors


#macro ScalarProject(vA, vB)

    vdot(vA, vnormalize(vB))

#end // macro ScalarProject


#macro VectorProject(vA, vB)

    (vdot(vA, vB)/vdot(vB, vB)*vB)

#end // macro VectorProject


#macro ScalarReject(vA, vB)

    vdot(vA, vnormalize(vcross(vB, vcross(vA, vB))))

#end // macro ScalarReject


#macro VectorReject(vA, vB)

    (vA - vdot(vA, vB)/vdot(vB, vB)*vB)

#end // macro VectorReject


#macro OrthogonalVector(v0)

    #if (vlength(v0) = 0)
        #local vN = <0, 0, 0>;
    #else
        #local Dm = min(abs(v0.x), abs(v0.y), abs(v0.z));
        #if (abs(v0.z) = Dm)
            #local vN = vnormalize(vcross(v0, z));
        #else
            #if (abs(v0.y) = Dm)
                #local vN = vnormalize(vcross(v0, y));
            #else
                #local vN = vnormalize(vcross(v0, x));
            #end // if
        #end // if
    #end // if

    vN

#end // macro OrthogonalVector


// Alternative to the version abvoe

#macro AltOrthogonalVector(v0)

    #switch (min(abs(v0.x), abs(v0.y), abs(v0.z)))
        #case (abs(v0.z))
            #local vE = z;
            #break
        #case (abs(v0.y))
            #local vE = y;
            #break
        #case (abs(v0.x))
            #local vE = x;
            #break
        #else
            #local vE = <0, 0, 0>;
    #end // switch
    #local vP = vdot(v0, v0)*vE - vdot(vE, v0)*v0;
    #if (vlength(vP) = 0)
        #local vN = <0, 0, 0>;
    #else
        #local vN = vnormalize(vP);
    #end // if

    vN

#end // macro AltOrthogonalVector


// Use built in vaxis_rotate() instead

#macro VectorAxisRotate(_v0,_vaxis,_angle)

    #local Phi = radians(_angle);
    #local VA = vnormalize(_vaxis);
    #local V1 = vcross(_v0,VA);
    #local V2 = vcross(V1,VA);

    (_v0 - sin(Phi)*V1 + (1 - cos(Phi))*V2)

#end // macro VectorAxisRotate


#macro VectorReorient(v0, vFrom, vTo)

    #local vF = vnormalize(vFrom);
    #local vT = vnormalize(vTo);
    #local vAxis = vcross(vF, vT);
    #local Cos = vdot(vF, vT);
    #local Sin = vlength(vAxis);
    #if (Sin = 0)
        #local vR = (Cos < 0 ? -v0 : +v0);
    #else
        #local vA = vAxis/Sin;
        #local V1 = vcross(v0, vA);
        #local V2 = vcross(V1, vA);
        #local vR = v0 - Sin*V1 + (1 - Cos)*V2;
    #end // if

    vR

#end // macro VectorReorient


#macro FunctionValue(Fn, v0)

    Fn(v0.x, v0.y, v0.z)

#end // macro FunctionValue


#macro VectorTransform(v0, Transform)

    FunctionValue(function { transform { Transform } }, v0)

#end // macro VectorTransform


#macro VectorInverseTransform(v0, Transform)

    FunctionValue(function { transform { Transform inverse } }, v0)

#end // macro VectorInverseTransform


// This identical in function to transforms.inc's Matrix_Trans. This
// so named version recommended for clarity. Having this version herein
// enables this file to be self contained with respect to dependencies.
#macro TransformFromVectors(vX, vY, vZ, pT)

    transform {
        matrix <
            vX.x, vX.y, vX.z,
            vY.x, vY.y, vY.z,
            vZ.x, vZ.y, vZ.z,
            pT.x, pT.y, pT.z
        >
    }

#end // macro TransformFromVectors


#macro TransformFromTransformFunction(TransformFn)

    #local pT = TransformFn(0, 0, 0);

    TransformFromVectors(
        TransformFn(1, 0, 0) - pT,
        TransformFn(0, 1, 0) - pT,
        TransformFn(0, 0, 1) - pT,
        pT
    )

#end // macro TransformFromTransformFunction


#macro TransformFunctionFromTransform(Transform)

    function { transform { Transform } }

#end // macro TransformFunctionFromTransform


#macro TransformFunctionFromVectors(vX, vY, vZ, pT)

    function { TransformFromVectors(vX, vY, vZ, pT) }

#end // macro TransformFunctionFromVectors


#macro VectorsFromTransformFunction(TransformFn, vX, vY, vZ, pT)

    #declare pT = TransformFn(0, 0, 0);
    #declare vX = TransformFn(1, 0, 0) - pT;
    #declare vY = TransformFn(0, 1, 0) - pT;
    #declare vZ = TransformFn(0, 0, 1) - pT;

#end // macro VectorsFromTransformFunction


#macro VectorsFromTransform(Transform, vX, vY, vZ, pT)

    VectorsFromTransformFunction(function { transform { Transform } }, vX, vY, vZ, pT)

#end // macro VectorsFromTransform


#macro AxisRotateTransform(vAxis, Angle)

    TransformFromVectors(
        vaxis_rotate(x, vAxis, Angle),
        vaxis_rotate(y, vAxis, Angle),
        vaxis_rotate(z, vAxis, Angle),
        <0, 0, 0>
    )

#end // macro AxisRotateTransform


#macro ReorientTransform(vFrom, vTo)

    #local vF = vnormalize(vFrom);
    #local vT = vnormalize(vTo);
    #local vAxis = vcross(vF, vT);
    #local Dot = min(max(-1, vdot(vF, vT)), +1);
    #local Angle = degrees(acos(Dot));

    TransformFromVectors(
        vaxis_rotate(x, vAxis, Angle),
        vaxis_rotate(y, vAxis, Angle),
        vaxis_rotate(z, vAxis, Angle),
        <0, 0, 0>
    )

#end // macro ReorientTransform

// ===== 1 ======= 2 ======= 3 ======= 4 ======= 5 ======= 6 ======= 7 ======= 8 ======= 9 ======= 10


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="vectors.inc")

    #fopen FID "SelfTestVectorsInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
    // Run some macros to be sure parse works. Assume failure and correct later if all parse OK.
    #write (FID,"\n---\n")
    #declare Xfrm00 = TransformFromVectors(<0.1,0.1,0.0>,<-0.1,0.1,0.0>,<0.0,0.1,0.1>,<0.1,0.1,0.1>);
    #fclose FID

    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestVectorsInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n\n")
    // If all macro call tests above work create output for what was tested.
    #write (FID,concat("TransformFromVectors 1.000000 ",str(1.0,9,6)," OK\n"))
    #write (FID,"\n")


    // OrthogonalVector
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = OrthogonalVector(Vec00);
    #declare VP01 = OrthogonalVector(Vec01);
    #write (FID,concat("OrthogonalVector +1.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("OrthogonalVector +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("OrthogonalVector +0.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #write (FID,concat("OrthogonalVector +0.707107   ",str(VP01.x,9,6)," VP01.x\n"))
    #write (FID,concat("OrthogonalVector -0.707107   ",str(VP01.y,9,6)," VP01.y\n"))
    #write (FID,concat("OrthogonalVector +0.000000   ",str(VP01.z,9,6)," VP01.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #ifdef (VP01) #undef VP01 #end
    #write (FID,"\n")
    // AltOrthogonalVector
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = AltOrthogonalVector(Vec00);
    #declare VP01 = AltOrthogonalVector(Vec01);
    #write (FID,concat("AltOrthogonalVector +0.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("AltOrthogonalVector +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("AltOrthogonalVector +1.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #write (FID,concat("AltOrthogonalVector -0.408248   ",str(VP01.x,9,6)," VP01.x\n"))
    #write (FID,concat("AltOrthogonalVector -0.408248   ",str(VP01.y,9,6)," VP01.y\n"))
    #write (FID,concat("AltOrthogonalVector +0.816497   ",str(VP01.z,9,6)," VP01.z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (VP00) #undef VP00 #end
    #ifdef (VP01) #undef VP01 #end
    #write (FID,"\n")
    // VectorCos, VectorSin
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = OrthogonalVector(Vec00);
    #declare VP01 = OrthogonalVector(Vec01);
    #declare Val = VectorCos(Vec00,VP00);
    #write (FID,concat("VectorCos +0.000000   ",str(Val,9,6)," Vec00,VP00\n"))
    #declare Val = VectorSin(Vec00,VP00);
    #write (FID,concat("VectorSin +1.000000   ",str(Val,9,6)," Vec00,VP00\n"))
    #declare Val = VectorCos(Vec01,VP01);
    #write (FID,concat("VectorCos +0.000000   ",str(Val,9,6)," Vec01,VP01\n"))
    #declare Val = VectorSin(Vec01,VP01);
    #write (FID,concat("VectorSin +1.000000   ",str(Val,9,6)," Vec01,VP01\n"))
    #write (FID,"-\n")
    #declare VP00 = AltOrthogonalVector(Vec00);
    #declare VP01 = AltOrthogonalVector(Vec01);
    #declare Val = VectorCos(Vec00,VP00);
    #write (FID,concat("VectorCos +0.000000   ",str(Val,9,6)," Vec00,VP00\n"))
    #declare Val = VectorSin(Vec00,VP00);
    #write (FID,concat("VectorSin +1.000000   ",str(Val,9,6)," Vec00,VP00\n"))
    #declare Val = VectorCos(Vec01,VP01);
    #write (FID,concat("VectorCos -0.000000   ",str(Val,9,6)," Vec01,VP01\n"))
    #declare Val = VectorSin(Vec01,VP01);
    #write (FID,concat("VectorSin +1.000000   ",str(Val,9,6)," Vec01,VP01\n"))
    #write (FID,"-\n")
    // More VectorCos
    // Plus minor testing of new v3.8 tuple #declare style. (Parser test)
    #declare (Vec00,Vec01,Vec02,Vec03,Vec04) =
        (<+1,0,+0>,<-1,0,+0>,<+0,1,+0>,<+0,0,-1>,<+1,0,+1>);
    #declare (A00,A01,A02,A03,A04) =
        (
        VectorCos(Vec00,Vec00),VectorCos(Vec00,Vec01),
        VectorCos(Vec00,Vec02),VectorCos(Vec00,Vec03),VectorCos(Vec00,Vec04)
        );
    #write (FID,concat("VectorCos +1.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("VectorCos -1.000000 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("VectorCos +0.000000 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("VectorCos +0.000000 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("VectorCos +0.707107 ",str(A04,9,6)," A04\n"))
    #write (FID,"\n")
    // AngleBetweenVectors
    // Plus minor testing of new v3.8 vector components to variables style. (Parser test)
    #declare VecTmp = <AngleBetweenVectors(Vec00,Vec00),AngleBetweenVectors(Vec00,Vec01),
        AngleBetweenVectors(Vec00,Vec02),AngleBetweenVectors(Vec00,Vec03),
        AngleBetweenVectors(Vec00,Vec04)>;
    #declare <A00,A01,A02,A03,A04> = VecTmp;
    #undef VecTmp
    #write (FID,concat("AngleBetweenVectors +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("AngleBetweenVectors +3.141593 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("AngleBetweenVectors +1.570796 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("AngleBetweenVectors +1.570796 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("AngleBetweenVectors +0.785398 ",str(A04,9,6)," A04\n"))
    #write (FID,"---\n")
    #declare VecD = <+1.73,+1.73,+1.73>;
    #declare VecTmp = <AngleBetweenVectors(VecD,VecD),AngleBetweenVectors(VecD,-VecD)>;
    #declare <A00,A01> = VecTmp;
    #write (FID,concat("AngleBetweenVectors +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("AngleBetweenVectors +3.141593 ",str(A01,9,6)," A01\n"))
    #ifdef (VecD) #undef VecD #end
    #write (FID,"\n")
    // AccurateAngleBetweenVectors
    // Plus minor testing of new v3.8 vector components to variables style. (Parser test)
    #declare VecTmp = <AccurateAngleBetweenVectors(Vec00,Vec00),AccurateAngleBetweenVectors(Vec00,Vec01),
        AccurateAngleBetweenVectors(Vec00,Vec02),AccurateAngleBetweenVectors(Vec00,Vec03),
        AccurateAngleBetweenVectors(Vec00,Vec04)>;
    #declare <A00,A01,A02,A03,A04> = VecTmp;
    #undef VecTmp
    #write (FID,concat("AccurateAngleBetweenVectors +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("AccurateAngleBetweenVectors +3.141593 ",str(A01,9,6)," A01\n"))
    #write (FID,concat("AccurateAngleBetweenVectors +1.570796 ",str(A02,9,6)," A02\n"))
    #write (FID,concat("AccurateAngleBetweenVectors +1.570796 ",str(A03,9,6)," A03\n"))
    #write (FID,concat("AccurateAngleBetweenVectors +0.785398 ",str(A04,9,6)," A04\n"))
    #write (FID,"---\n")
    #declare VecD = <+1.73,+1.73,+1.73>;
    #declare VecTmp = <AccurateAngleBetweenVectors(VecD,VecD),AccurateAngleBetweenVectors(VecD,-VecD)>;
    #declare <A00,A01> = VecTmp;
    #undef VecTmp
    #write (FID,concat("AccurateAngleBetweenVectors +0.000000 ",str(A00,9,6)," A00\n"))
    #write (FID,concat("AccurateAngleBetweenVectors +3.141593 ",str(A01,9,6)," A01\n"))
    #ifdef (VecD) #undef VecD #end
    #ifdef (A00) #undef A00 #end
    #ifdef (A01) #undef A01 #end
    #ifdef (A02) #undef A02 #end
    #ifdef (A03) #undef A03 #end
    #ifdef (A04) #undef A04 #end
    #write (FID,"\n")
    // AxisRotateTransform
    #declare Vec00 = <0.5,0.5,0.5>;
    #declare Vec01 = VectorTransform(Vec00,AxisRotateTransform(<0,0,1>,45.0));
    #write (FID,concat("AxisRotateTransform +0.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("AxisRotateTransform +0.707107 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("AxisRotateTransform +0.500000 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"\n")
    // VectorAxisRotate, vaxis_rotate()
    #declare Vec00 = <0.5,0.5,0.5>;
    #declare Vec01 = VectorAxisRotate(Vec00,<0,0,1>,45.0);
    #write (FID,concat("VectorAxisRotate +0.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("VectorAxisRotate +0.707107 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("VectorAxisRotate +0.500000 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"-\n")
    #declare Vec01 = vaxis_rotate(Vec00,<0,0,1>,45.0);
    #write (FID,concat("vaxis_rotate +0.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("vaxis_rotate +0.707107 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("vaxis_rotate +0.500000 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"\n")
    // ReorientTransform
    #declare Vec00 = <1.5,1.5,1.5>;
    #declare Vec01 = <0.0,0.5,0>;
    #declare Vec02 = VectorTransform(Vec01,ReorientTransform(Vec01,Vec00));
    #write (FID,concat("ReorientTransform +0.288675 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("ReorientTransform +0.288675 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("ReorientTransform +0.288675 ",str(Vec02.z,9,6)," z\n"))
    #write (FID,"---\n")
    #declare Vec00 = <0.1,0.1,0.1>;
    #declare Vec01 = <0.0,-0.5,0>;
    #declare Vec02 = VectorTransform(Vec01,ReorientTransform(Vec01,Vec00));
    #write (FID,concat("ReorientTransform +0.288675 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("ReorientTransform +0.288675 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("ReorientTransform +0.288675 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #write (FID,"\n")
    // VectorTransform, VectorInverseTransform(vec, trans)
    #declare Vec00 = <1,0,0>;
    #declare Xfrm00 = transform { rotate z*90 }
    #declare Vec01 = VectorTransform(Vec00,Xfrm00);
    #declare Vec02 = VectorInverseTransform(Vec00,Xfrm00);
    #write (FID,concat("VectorTransform +0.000000 ",str(Vec01.x,9,6)," x\n"))
    #write (FID,concat("VectorTransform +1.000000 ",str(Vec01.y,9,6)," y\n"))
    #write (FID,concat("VectorTransform +0.000000 ",str(Vec01.z,9,6)," z\n"))
    #write (FID,"---\n")
    #write (FID,concat("VectorInverseTransform +0.000000 ",str(Vec02.x,9,6)," x\n"))
    #write (FID,concat("VectorInverseTransform -1.000000 ",str(Vec02.y,9,6)," y\n"))
    #write (FID,concat("VectorInverseTransform +0.000000 ",str(Vec02.z,9,6)," z\n"))
    #ifdef (Xfrm00) #undef Xfrm00 #end
    #write (FID,"\n")
    // VectorReject    (like, no longer recommended, VProject_Plane of math.inc)
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VectorReject(Vec01,Vec00);
    #write (FID,concat("VectorReject +0.707000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VectorReject +0.000000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VectorReject +0.707000   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // VectorProject   (like, no longer recommended, VProject_Axis of math.inc)
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare VP00 = VectorProject(Vec01,Vec00);
    #write (FID,concat("VectorProject +0.000000   ",str(VP00.x,9,6)," VP00.x\n"))
    #write (FID,concat("VectorProject +0.707000   ",str(VP00.y,9,6)," VP00.y\n"))
    #write (FID,concat("VectorProject +0.000000   ",str(VP00.z,9,6)," VP00.z\n"))
    #ifdef (VP00) #undef VP00 #end
    #write (FID,"\n")
    // ScalarReject
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare Val = ScalarReject(Vec01,Vec00);
    #write (FID,concat("ScalarReject +0.999849  ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // ScalarProject
    #declare Vec00 = <+0.0,+0.707,+0.0>;
    #declare Vec01 = <+0.707,+0.707,+0.707>;
    #declare Val = ScalarProject(Vec01,Vec00);
    #write (FID,concat("ScalarProject +0.707000  ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // ScalarTripleProduct   (Signed volume...)
    #declare Vec00 = <+1.0,+0.0,+0.0>;
    #declare Vec01 = <+0.0,+1.0,+0.0>;
    #declare Vec02 = <+0.0,+0.0,+1.0>;
    #declare Val = ScalarTripleProduct(Vec00,Vec01,Vec02);
    #write (FID,concat("ScalarTripleProduct +1.000000   ",str(Val,9,6)," \n"))
    #declare Vec00 = <-1.0,+0.0,+0.0>;
    #declare Val = ScalarTripleProduct(Vec00,Vec01,Vec02);
    #write (FID,concat("ScalarTripleProduct -1.000000   ",str(Val,9,6)," \n"))
    #declare Vec00 = <+0.2,+0.3,+0.0>;
    #declare Vec01 = <+0.4,+0.1,+0.0>;
    #declare Vec02 = <-0.2,-0.5,+0.0>;
    #declare Val = ScalarTripleProduct(Vec00,Vec01,Vec02);
    #write (FID,concat("ScalarTripleProduct +0.000000   ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // VectorTripleProduct
    #declare Vec00 = <+1.0,+0.0,+0.0>;
    #declare Vec01 = <+0.0,+1.0,+0.0>;
    #declare Vec02 = <+0.0,+0.0,+1.0>;
    #declare V00 = VectorTripleProduct(Vec00,Vec01,Vec02);
    #write (FID,concat("VectorTripleProduct +0.000000   ",str(V00.x,9,6)," V00.x\n"))
    #write (FID,concat("VectorTripleProduct +0.000000   ",str(V00.y,9,6)," V00.y\n"))
    #write (FID,concat("VectorTripleProduct +0.000000   ",str(V00.z,9,6)," V00.z\n"))
    #declare Vec00 = <+1.0,+1.0,+1.0>;
    #declare Vec01 = <-1.0,+1.0,+1.0>;
    #declare Vec02 = <+1.0,+1.0,-1.0>;
    #declare V00 = VectorTripleProduct(Vec00,Vec01,Vec02);
    #write (FID,concat("VectorTripleProduct -2.000000   ",str(V00.x,9,6)," V00.x\n"))
    #write (FID,concat("VectorTripleProduct +0.000000   ",str(V00.y,9,6)," V00.y\n"))
    #write (FID,concat("VectorTripleProduct +2.000000   ",str(V00.z,9,6)," V00.z\n"))
    #ifdef (V00) #undef V00 #end
    #write (FID,"\n")
    // VectorReorient
    #declare Vec00 = <+0.0,+1.0,+0.0>;
    #declare Vec01 = <+1.0,+0.0,+0.0>;
    #declare Vec02 = <+0.0,+1.0,+0.0>; // pt
    #declare V00 = VectorReorient(Vec02,Vec00,Vec01);
    #write (FID,concat("VectorReorient +1.000000   ",str(V00.x,9,6)," V00.x\n"))
    #write (FID,concat("VectorReorient +0.000000   ",str(V00.y,9,6)," V00.y\n"))
    #write (FID,concat("VectorReorient +0.000000   ",str(V00.z,9,6)," V00.z\n"))
    #write (FID,"-\n")
    #declare Vec00 = <+0.0,-1.0,+0.0>;
    #declare Vec01 = <+1.0,+0.0,+0.0>;
    #declare Vec02 = <+0.0,-1.0,+0.0>; // pt
    #declare V00 = VectorReorient(Vec02,Vec00,Vec01);
    #write (FID,concat("VectorReorient +1.000000   ",str(V00.x,9,6)," V00.x\n"))
    #write (FID,concat("VectorReorient +0.000000   ",str(V00.y,9,6)," V00.y\n"))
    #write (FID,concat("VectorReorient +0.000000   ",str(V00.z,9,6)," V00.z\n"))
    #write (FID,"-\n")
    #declare Vec00 = <+1.0,+1.0,+1.0>;
    #declare Vec01 = <-1.0,-1.0,-1.0>;
    #declare Vec02 = <-1.0,+0.0,+1.0>; // pt
    #declare V00 = VectorReorient(Vec02,Vec00,Vec01);
    #write (FID,concat("VectorReorient +1.000000   ",str(V00.x,9,6)," V00.x\n"))
    #write (FID,concat("VectorReorient +0.000000   ",str(V00.y,9,6)," V00.y\n"))
    #write (FID,concat("VectorReorient -1.000000   ",str(V00.z,9,6)," V00.z\n"))
    #ifdef (V00) #undef V00 #end
    #write (FID,"\n")
    // FunctionValue
    #declare Vec00 = <+1.0,+2.0,+3.0>; // pt
    #declare Fn00 = function { x+y+z }
    #declare Val = FunctionValue(Fn00,Vec00);
    #write (FID,concat("FunctionValue +6.000000   ",str(Val,9,6)," \n"))
    #ifdef (Val) #undef Val #end
    #ifdef (Fn00) #undef Fn00 #end
    #write (FID,"\n")
    // VectorsFromTransform, VectorsFromTransformFunction
    #declare Trfm00 = ReorientTransform(x - z, y)
    #declare Fn00 = function { transform { Trfm00 translate <0.1,0.2,0.1>} };
    #declare (vX,vY,vZ,pT) = (<0,0,0>,<0,0,0>,<0,0,0>,<0,0,0>);
    VectorsFromTransform(Trfm00,vX,vY,vZ,pT)
    #write (FID,concat("VectorsFromTransform +0.500000   ",str(vX.x,9,6)," vX.x\n"))
    #write (FID,concat("VectorsFromTransform +0.707107   ",str(vX.y,9,6)," vX.y\n"))
    #write (FID,concat("VectorsFromTransform +0.500000   ",str(vX.z,9,6)," vX.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("VectorsFromTransform -0.707107   ",str(vY.x,9,6)," vY.x\n"))
    #write (FID,concat("VectorsFromTransform +0.000000   ",str(vY.y,9,6)," vY.y\n"))
    #write (FID,concat("VectorsFromTransform +0.707107   ",str(vY.z,9,6)," vY.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("VectorsFromTransform +0.500000   ",str(vZ.x,9,6)," vZ.x\n"))
    #write (FID,concat("VectorsFromTransform -0.707107   ",str(vZ.y,9,6)," vZ.y\n"))
    #write (FID,concat("VectorsFromTransform +0.500000   ",str(vZ.z,9,6)," vZ.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("VectorsFromTransform +0.000000   ",str(pT.x,9,6)," pT.x\n"))
    #write (FID,concat("VectorsFromTransform +0.000000   ",str(pT.y,9,6)," pT.y\n"))
    #write (FID,concat("VectorsFromTransform +0.000000   ",str(pT.z,9,6)," pT.z\n"))
    #write (FID,"---\n")
    VectorsFromTransformFunction(Fn00,vX,vY,vZ,pT)
    #write (FID,concat("VectorsFromTransformFunction +0.500000   ",str(vX.x,9,6)," vX.x\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.707107   ",str(vX.y,9,6)," vX.y\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.500000   ",str(vX.z,9,6)," vX.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("VectorsFromTransformFunction -0.707107   ",str(vY.x,9,6)," vY.x\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.000000   ",str(vY.y,9,6)," vY.y\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.707107   ",str(vY.z,9,6)," vY.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("VectorsFromTransformFunction +0.500000   ",str(vZ.x,9,6)," vZ.x\n"))
    #write (FID,concat("VectorsFromTransformFunction -0.707107   ",str(vZ.y,9,6)," vZ.y\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.500000   ",str(vZ.z,9,6)," vZ.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("VectorsFromTransformFunction +0.100000   ",str(pT.x,9,6)," pT.x\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.200000   ",str(pT.y,9,6)," pT.y\n"))
    #write (FID,concat("VectorsFromTransformFunction +0.100000   ",str(pT.z,9,6)," pT.z\n"))
    #write (FID,"\n")
    // TransformFunctionFromVectors
    #ifdef (Fn00) #undef Fn00 #end
    #declare Fn00 = TransformFunctionFromVectors(vX,vY,vZ,pT)
    VectorsFromTransformFunction(Fn00,vX,vY,vZ,pT)
    #write (FID,concat("TransformFunctionFromVectors +0.500000   ",str(vX.x,9,6)," vX.x\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.707107   ",str(vX.y,9,6)," vX.y\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.500000   ",str(vX.z,9,6)," vX.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("TransformFunctionFromVectors -0.707107   ",str(vY.x,9,6)," vY.x\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.000000   ",str(vY.y,9,6)," vY.y\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.707107   ",str(vY.z,9,6)," vY.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("TransformFunctionFromVectors +0.500000   ",str(vZ.x,9,6)," vZ.x\n"))
    #write (FID,concat("TransformFunctionFromVectors -0.707107   ",str(vZ.y,9,6)," vZ.y\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.500000   ",str(vZ.z,9,6)," vZ.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("TransformFunctionFromVectors +0.100000   ",str(pT.x,9,6)," pT.x\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.200000   ",str(pT.y,9,6)," pT.y\n"))
    #write (FID,concat("TransformFunctionFromVectors +0.100000   ",str(pT.z,9,6)," pT.z\n"))
    #write (FID,"\n")
    // TransformFunctionFromTransform
    #declare Trfm00 = ReorientTransform(x - z, y)
    #declare Trfm01 = transform { Trfm00 translate <0.1,0.2,0.1> rotate <7,14,21> }
    #ifdef (Fn00) #undef Fn00 #end
    #declare Fn00 = TransformFunctionFromTransform(Trfm01);
    VectorsFromTransformFunction(Fn00,vX,vY,vZ,pT)
    #write (FID,concat("TransformFunctionFromTransform +0.354794   ",str(vX.x,9,6)," vX.x\n"))
    #write (FID,concat("TransformFunctionFromTransform +0.822691   ",str(vX.y,9,6)," vX.y\n"))
    #write (FID,concat("TransformFunctionFromTransform +0.444186   ",str(vX.z,9,6)," vX.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("TransformFunctionFromTransform -0.451138   ",str(vY.x,9,6)," vY.x\n"))
    #write (FID,concat("TransformFunctionFromTransform -0.265481   ",str(vY.y,9,6)," vY.y\n"))
    #write (FID,concat("TransformFunctionFromTransform +0.852053   ",str(vY.z,9,6)," vY.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("TransformFunctionFromTransform +0.818899   ",str(vZ.x,9,6)," vZ.x\n"))
    #write (FID,concat("TransformFunctionFromTransform -0.502692   ",str(vZ.y,9,6)," vZ.y\n"))
    #write (FID,concat("TransformFunctionFromTransform +0.276956   ",str(vZ.z,9,6)," vZ.z\n"))
    #write (FID,"-\n")
    #write (FID,concat("TransformFunctionFromTransform +0.051735   ",str(pT.x,9,6)," pT.x\n"))
    #write (FID,concat("TransformFunctionFromTransform +0.219437   ",str(pT.y,9,6)," pT.y\n"))
    #write (FID,concat("TransformFunctionFromTransform +0.095764   ",str(pT.z,9,6)," pT.z\n"))
    #write (FID,"\n")
    // TransformFromTransformFunction
    #declare Trfm00 = TransformFromTransformFunction(Fn00)
    #declare Vec00  = <2,2,2>;
    #declare Vec01 = VectorTransform(Vec02,Trfm00);
    #write (FID,concat("TransformFunctionFromTransform +0.515840   ",str(Vec01.x,9,6)," Vec01.x\n"))
    #write (FID,concat("TransformFunctionFromTransform -1.105946   ",str(Vec01.y,9,6)," Vec01.y\n"))
    #write (FID,concat("TransformFunctionFromTransform -0.071466   ",str(Vec01.z,9,6)," Vec01.z\n"))
    #write (FID,"\n")

    #ifdef (Trfm00) #undef Trfm00 #end
    #ifdef (Trfm01) #undef Trfm01 #end
    #ifdef (Fn00) #undef Fn00 #end
    #ifdef (vX) #undef vX #end
    #ifdef (vY) #undef vY #end
    #ifdef (vZ) #undef vZ #end
    #ifdef (pT) #undef pT #end

// Singular matrix examples...  (determinant vX,vY,vZ is zero)
//  #declare Fn00 = TransformFunctionFromVectors(<1,2,3>,<4,5,6>,<7,8,9>,<0.1,0.2,0.3>)
//  #declare Trfm00 = TransformFromVectors(<1,2,3>,<4,5,6>,<7,8,9>,<0.1,0.2,0.3>)

    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #ifdef (Vec02) #undef Vec02 #end
    #ifdef (Vec03) #undef Vec03 #end
    #ifdef (Vec04) #undef Vec04 #end

    #write (FID,"\n")

    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"

#end // End self testing

#version VECTORS_INC_TEMP;
#end //vectors.inc
