///
/// @file core/render/trace.cpp
///
/// Implementations related to the @ref pov::Trace class.
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2022 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//------------------------------------------------------------------------------
// SPDX-License-Identifier: AGPL-3.0-or-later
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/render/trace.h"

// C++ variants of C standard header files
#include <cfloat>

// C++ standard header files
#include <algorithm>

// POV-Ray header files (base module)
#include "base/povassert.h"

// POV-Ray header files (core module)
#include "core/bounding/bsptree.h"
#include "core/lighting/lightsource.h"
#include "core/lighting/radiosity.h"
#include "core/lighting/subsurface.h"
#include "core/material/interior.h"
#include "core/material/noise.h"
#include "core/material/normal.h"
#include "core/material/pattern.h"
#include "core/material/pigment.h"
#include "core/material/texture.h"
#include "core/material/warp.h"
#include "core/math/matrix.h"
#include "core/render/ray.h"
#include "core/scene/atmosphere.h"
#include "core/scene/object.h"
#include "core/scene/scenedata.h"
#include "core/scene/tracethreaddata.h"
#include "core/shape/box.h"
#include "core/shape/csg.h"
#include "core/support/imageutil.h"
#include "core/support/statistics.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

using std::min;
using std::max;
using std::vector;

bool NoSomethingFlagRayObjectCondition::operator()(const Ray& ray, ConstObjectPtr object, double) const
{
    if (ray.IsImageRay() && Test_Flag(object, NO_IMAGE_FLAG))
        return false;
    if (ray.IsReflectionRay() && Test_Flag(object, NO_REFLECTION_FLAG))
        return false;
    if (ray.IsRadiosityRay() && Test_Flag(object, NO_RADIOSITY_FLAG))
        return false;
    if (ray.IsPhotonRay() && Test_Flag(object, NO_SHADOW_FLAG))
        return false;
    return true;
}

Trace::Trace(std::shared_ptr<SceneData> sd, TraceThreadData *td, const QualityFlags& qf,
             CooperateFunctor& cf, MediaFunctor& mf, RadiosityFunctor& rf) :
    threadData(td),
    sceneData(sd),
    maxFoundTraceLevel(0),
    qualityFlags(qf),
    mailbox(0),
    crandRandomNumberGenerator(0),
    randomNumbers(0.0, 1.0, 32768),
    randomNumberGenerator(&randomNumbers),
    ssltUniformDirectionGenerator(),
    ssltUniformNumberGenerator(),
    ssltCosWeightedDirectionGenerator(),
    cooperate(cf),
    media(mf),
    radiosity(rf),
    lightColorCacheIndex(-1)
{
    lightSourceLevel1ShadowCache.resize(max(1, (int) threadData->lightSources.size()));
    for (auto i(lightSourceLevel1ShadowCache.begin()); i != lightSourceLevel1ShadowCache.end(); i++)
        *i = nullptr;

    lightSourceOtherShadowCache.resize(max(1, (int) threadData->lightSources.size()));
    for (auto i(lightSourceOtherShadowCache.begin()); i != lightSourceOtherShadowCache.end(); i++)
        *i = nullptr;

    lightColorCache.resize(max(20U, sd->parsedMaxTraceLevel + 1));
    for (auto it = lightColorCache.begin(); it != lightColorCache.end(); it++)
        it->resize(max(1, (int) threadData->lightSources.size()));

    if (sceneData->boundingMethod == 2)
        mailbox = BSPTree::Mailbox(sceneData->numberOfFiniteObjects);
}

Trace::~Trace()
{
}

double Trace::TraceRay(Ray& ray, MathColour& colour, ColourChannel& transm, COLC weight, bool continuedRay,
        double maxDepth)
{
    Intersection bestisect;
    bool found;
    NoSomethingFlagRayObjectCondition precond;
    TrueRayObjectCondition postcond;

    POV_ULONG nrays = threadData->Stats()[Number_Of_Rays]++;

    if (ray.IsPrimaryRay() || (((unsigned char) nrays & 0x0f) == 0x00)) // 2nd part calls cooperate() 0 and then every 16 rays...?
    {
        cooperate(); //  TODO Are we calling cooperate too much with secondary rays?
    }

    // Check for max. trace level or ADC bailout.
    if ((ray.GetTicket().traceLevel >= ray.GetTicket().maxAllowedTraceLevel) || (weight < ray.GetTicket().adcBailout))
    {
        if (weight < ray.GetTicket().adcBailout)
        {
            threadData->Stats()[ADC_Saves]++;
        }

        colour.Clear();
        transm = 0.0;
        return HUGE_VAL;
    }

    if (maxDepth >= gkMinIsectDepthReturned)
    {
        bestisect.Depth = maxDepth;  // bestisect.Depth normally BOUND_HUGE
    }

    found = FindIntersection(bestisect, ray, precond, postcond);

    // Check if we're busy shooting too many radiosity sample rays at an unimportant object
    // NB. radiosityImportanceQueried value is set to <0 if radiosity is off so cost is only
    // conditional test and overhead for any speculative branch execution for this feature.
    if (ray.GetTicket().radiosityImportanceQueried >= 0.0)
    {
        if (found && bestisect.Object->RadiosityImportanceSet)
        {
            ray.GetTicket().radiosityImportanceFound = bestisect.Object->RadiosityImportance;
        }
        else
        {
            ray.GetTicket().radiosityImportanceFound = sceneData->radiositySettings.defaultImportance;
        }

        // The first call always reaches at least this far so 3 tests and settings above when using radiosity.
        if (ray.GetTicket().radiosityImportanceFound < ray.GetTicket().radiosityImportanceQueried)
        {
            if (found == false)
            {
                return HUGE_VAL;
            }
            else
            {
                return bestisect.Depth;
            }
        }
    }
    float oldRadiosityImportanceQueried = ray.GetTicket().radiosityImportanceQueried;
    // Setting to -1 indicates that recursive calls to TraceRay() should not check for radiosity importance
    ray.GetTicket().radiosityImportanceQueried = -1.0;

    const bool traceLevelIncremented = !continuedRay;

    if (traceLevelIncremented)
    {
        // Set highest level traced.
        ray.GetTicket().traceLevel++;
        ray.GetTicket().maxFoundTraceLevel =
            (unsigned int) max(ray.GetTicket().maxFoundTraceLevel, ray.GetTicket().traceLevel);
    }

    if (qualityFlags.media && (ray.IsPhotonRay() == true) && (ray.IsHollowRay() == true))
    {
        // Note: this version of ComputeMedia does not deposit photons. This is
        // intentional.  Even though we're processing a photon ray, we don't want
        // to deposit photons in the infinite atmosphere, only in contained
        // media, which is processed later (in ComputeLightedTexture).  [nk]
        media.ComputeMedia(sceneData->atmosphere, ray, bestisect, colour, transm);

        if (sceneData->fog != nullptr)
            ComputeFog(ray, bestisect, colour, transm);
    }

    if (found)
        ComputeTextureColour(bestisect, colour, transm, ray, weight, false);
    else
        ComputeSky(ray, colour, transm);

    if (qualityFlags.media && (ray.IsPhotonRay() == false) && (ray.IsHollowRay() == true))
    {
        if ((sceneData->rainbow != nullptr) && (ray.IsShadowTestRay() == false))
            ComputeRainbow(ray, bestisect, colour, transm);

        media.ComputeMedia(sceneData->atmosphere, ray, bestisect, colour, transm);

        if (sceneData->fog != nullptr)
            ComputeFog(ray, bestisect, colour, transm);
    }

    if (traceLevelIncremented)
        ray.GetTicket().traceLevel--;
    maxFoundTraceLevel = (unsigned int) max(maxFoundTraceLevel, ray.GetTicket().maxFoundTraceLevel);

    // Restore setting so next initial call will check for radiosity importance.
    ray.GetTicket().radiosityImportanceQueried = oldRadiosityImportanceQueried;

    if (found == false)
    {
        return HUGE_VAL;
    }
    else
    {
        return bestisect.Depth;
    }
}

bool Trace::FindIntersection(Intersection& bestisect, const Ray& ray)
{
    bool found = false;
    switch (sceneData->boundingMethod)
    {
        case 2:
        {
            BSPIntersectFunctor ifn(bestisect, ray, sceneData->objects, threadData);

            mailbox.clear();

            found = (*(sceneData->tree))(ray, ifn, mailbox, bestisect.Depth);

            // test infinite objects
            for (auto it = sceneData->objects.begin() + sceneData->numberOfFiniteObjects; it != sceneData->objects.end(); it++)
            {
                Intersection isect;

                if (FindIntersection(*it, isect, ray) && (isect.Depth < bestisect.Depth))
                {
                    bestisect = isect;
                    found = true;
                }
            }

            return found;
        }
        case 1:
        {
            if (sceneData->boundingSlabs != nullptr)
            {
                found = Intersect_BBox_Tree(priorityQueue, sceneData->boundingSlabs, ray,
                        &bestisect, threadData);
                return found;
            }
        }
        // FALLTHROUGH
        case 0:
        {
            for (auto it = sceneData->objects.begin(); it != sceneData->objects.end(); it++)
            {
                Intersection isect;

                if (FindIntersection(*it, isect, ray) && (isect.Depth < bestisect.Depth))
                {
                    bestisect = isect;
                    found = true;
                }
            }

            return found;
        }
    }

    return false;
}

bool Trace::FindIntersection(Intersection& bestisect, const Ray& ray, const RayObjectCondition& precondition,
        const RayObjectCondition& postcondition)
{
    bool found = false;
    switch (sceneData->boundingMethod)
    {
        case 2:
        {
            BSPIntersectCondFunctor ifn(bestisect, ray, sceneData->objects, threadData, precondition, postcondition);

            mailbox.clear();

            found = (*(sceneData->tree))(ray, ifn, mailbox, bestisect.Depth);

            // test infinite objects
            for (auto it = sceneData->objects.begin() + sceneData->numberOfFiniteObjects; it != sceneData->objects.end(); it++)
            {
                if (precondition(ray, *it, 0.0) == true)
                {
                    Intersection isect;

                    if (FindIntersection(*it, isect, ray, postcondition) && (isect.Depth < bestisect.Depth))
                    {
                        bestisect = isect;
                        found = true;
                    }
                }
            }

            return found;
        }
        case 1:
        {
            if (sceneData->boundingSlabs != nullptr)
            {
                found = Intersect_BBox_Tree(priorityQueue, sceneData->boundingSlabs, ray,
                        &bestisect, precondition, postcondition, threadData);
                return found;
            }
        }
        // FALLTHROUGH
        case 0:
        {
            for (auto it = sceneData->objects.begin(); it != sceneData->objects.end(); it++)
            {
                if (precondition(ray, *it, 0.0) == true)
                {
                    Intersection isect;

                    if (FindIntersection(*it, isect, ray, postcondition) && (isect.Depth < bestisect.Depth))
                    {
                        bestisect = isect;
                        found = true;
                    }
                }
            }

            return found;
        }
    }

    return found;
}


// The following Trace::FindIntersection functions are never called on
// bounding hierarchy leaves; rather only called when not traversing the
// bounding hierarchy:  Either to compute the intersections with infinite
// objects if they're not already accounted for by the bounding hierarchy
// chosen, or because bounding is disabled entirely, or because only the
// intersections with one particular object are of interest (e.g.  the
// projected_through object of a light source, but also the shadow caching
// mechanism).
//
// In the former two cases it wouldn't be necessary to perform the bounding
// box test, but in the latter case it is indispensable for good
// performance. Hence, the initial bounding test.

bool Trace::FindIntersection(ObjectPtr object, Intersection& isect, const Ray& ray, double closest)
{
    if (object != nullptr)
    {
#if (0) // Keeping example of original method in initial commit for comparison. Eliminate if well after 01/16/2022.
        BBoxVector3d origin;
        BBoxVector3d invdir;
        BBoxDirection variant;

        Vector3d tmp(1.0 / ray.GetDirection()[X], 1.0 / ray.GetDirection()[Y], 1.0 / ray.GetDirection()[Z]);
        origin = BBoxVector3d(ray.Origin);
        invdir = BBoxVector3d(tmp);
        variant = (BBoxDirection)((int(invdir[X] < 0.0) << 2) | (int(invdir[Y] < 0.0) << 1) | int(invdir[Z] < 0.0));

#else
        Vector3d rayDir = ray.GetDirection();
        // Get inverse direction, explicitly handling zero components. Avoiding divide by zero exception
        // to enable -ffast-math/-O3 use without occasional flaky issues where sometimes (invdir[*] < 0.0)
        // working and not depending on compiler options / code structure / cpu / flow. (v3.7/v3.8/v4.0)
        double tmpX = (std::abs(rayDir.x())<gkDBL_epsilon) ? std::copysign(1.0,rayDir.x())*HUGE_VAL : 1.0/rayDir.x();
        double tmpY = (std::abs(rayDir.y())<gkDBL_epsilon) ? std::copysign(1.0,rayDir.y())*HUGE_VAL : 1.0/rayDir.y();
        double tmpZ = (std::abs(rayDir.z())<gkDBL_epsilon) ? std::copysign(1.0,rayDir.z())*HUGE_VAL : 1.0/rayDir.z();

     // Vector3d tmp(tmpX,tmpY,tmpZ);
     // BBoxVector3d invdir = BBoxVector3d(tmp);
        BBoxVector3d invdir = Vector3d(tmpX,tmpY,tmpZ); // OK if BBox at doubles, otherwise compile error & use above
     // BBoxVector3d origin = BBoxVector3d(ray.Origin);
        BBoxVector3d origin = ray.Origin;  // OK while BBox at doubles
        BBoxDirection variant = (BBoxDirection)
            ((int(invdir[X] < 0.0) << 2) | (int(invdir[Y] < 0.0) << 1) | int(invdir[Z] < 0.0));
#endif

        if ((sceneData->boundingMethod != 0) && (object->Intersect_BBox(variant, origin, invdir, closest) == false))
        {
            return false;
        }

        if (object->Bound.empty() == false)
        {
            if (Ray_In_Bound(ray, object->Bound, threadData) == false)
            {
                return false;
            }
        }

        IStack depthstack(stackPool);
        POV_REFPOOL_ASSERT(depthstack->empty()); // verify that the IStack pulled from the pool is in a cleaned-up condition

        if (object->All_Intersections(ray, depthstack, threadData))
        {
            bool found = false;
            double tmpDepth = 0;

            while(depthstack->size() > 0)
            {
                tmpDepth = depthstack->top().Depth;
                POV_ASSERT(tmpDepth > 0.0);  // Shape code should never return intersections <= 0.0.
                if (tmpDepth < closest)
                {
                    isect = depthstack->top();
                    closest = tmpDepth;
                    found = true;
                }

                depthstack->pop();
            }

//          return (found == true);  // TODO FIXME Why the extra compare? (Delete lines, if direct OK well after Jan 15 2022)
            return (found);
        }

        POV_REFPOOL_ASSERT(depthstack->empty()); // verify that the IStack is in a cleaned-up condition (again)
    }

    return false;
}

bool Trace::FindIntersection(ObjectPtr object, Intersection& isect, const Ray& ray,
        const RayObjectCondition& postcondition, double closest)
{
    if (object != nullptr)
    {
        Vector3d rayDir = ray.GetDirection();
        // Get inverse direction, explicitly handling zero components. Avoiding divide by zero exception
        // to enable -ffast-math/-O3 use without occasional flaky issues where sometimes (invdir[*] < 0.0)
        // working and not depending on compiler options / code structure / cpu / flow. (v3.7/v3.8/v4.0)
        double tmpX = (std::abs(rayDir.x())<gkDBL_epsilon) ? std::copysign(1.0,rayDir.x())*HUGE_VAL : 1.0/rayDir.x();
        double tmpY = (std::abs(rayDir.y())<gkDBL_epsilon) ? std::copysign(1.0,rayDir.y())*HUGE_VAL : 1.0/rayDir.y();
        double tmpZ = (std::abs(rayDir.z())<gkDBL_epsilon) ? std::copysign(1.0,rayDir.z())*HUGE_VAL : 1.0/rayDir.z();

     // Vector3d tmp(tmpX,tmpY,tmpZ);
     // BBoxVector3d invdir = BBoxVector3d(tmp);
        BBoxVector3d invdir = Vector3d(tmpX,tmpY,tmpZ); // OK if BBox at doubles, otherwise compile error & use above
     // BBoxVector3d origin = BBoxVector3d(ray.Origin);
        BBoxVector3d origin = ray.Origin;  // OK while BBox at doubles
        BBoxDirection variant = (BBoxDirection)
            ((int(invdir[X] < 0.0) << 2) | (int(invdir[Y] < 0.0) << 1) | int(invdir[Z] < 0.0));

        if ((sceneData->boundingMethod != 0) && (object->Intersect_BBox(variant, origin, invdir, closest) == false))
        {
            return false;
        }

        if (object->Bound.empty() == false)
        {
            if (Ray_In_Bound(ray, object->Bound, threadData) == false)
            {
                return false;
            }
        }

        IStack depthstack(stackPool);
        POV_REFPOOL_ASSERT(depthstack->empty()); // verify that the IStack pulled from the pool is in a cleaned-up condition

        if (object->All_Intersections(ray, depthstack, threadData))
        {
            bool found = false;
            double tmpDepth = 0;

            while(depthstack->size() > 0)
            {
                tmpDepth = depthstack->top().Depth;
                POV_ASSERT(tmpDepth > 0.0);  // Shape code should never return intersections <= 0.0.
                if (tmpDepth < closest && postcondition(ray, object, tmpDepth))
                {
                    isect = depthstack->top();
                    closest = tmpDepth;
                    found = true;
                }

                depthstack->pop();
            }

            return (found);
//          return (found == true);  // TODO FIXME Why the extra compare? (Delete lines, if direct OK well after Jan 15 2022)
        }

        POV_REFPOOL_ASSERT(depthstack->empty()); // verify that the IStack is in a cleaned-up condition (again)
    }

    return false;
}

unsigned int Trace::GetHighestTraceLevel()
{
    return maxFoundTraceLevel;
}

void Trace::ComputeTextureColour(Intersection& isect, MathColour& colour, ColourChannel& transm,
        Ray& ray, COLC weight, bool photonPass)
{
    // NOTE: when called during the photon pass this method is used to deposit photons
    // on the surface and not, per se, to compute texture color.
    WeightedTextureVector wtextures;
    double normaldirection;
    MathColour tmpCol;
    ColourChannel tmpTransm = 0.0;
    MathColour c1;
    ColourChannel t1 = 0.0;
    Vector2d uvcoords;
    Vector3d rawnormal;
    Vector3d ipoint(isect.IPoint);

    if (++lightColorCacheIndex >= lightColorCache.size())
    {
        lightColorCache.resize(lightColorCacheIndex + 10);
        for (auto it = lightColorCache.begin() + lightColorCacheIndex; it != lightColorCache.end(); it++)
            it->resize(lightColorCache[0].size());
    }
    for (auto it = lightColorCache[lightColorCacheIndex].begin(); it != lightColorCache[lightColorCacheIndex].end(); it++)
    {
        it->tested = false;
    }

    // compute the surface normal
    isect.Object->Normal(rawnormal, &isect, threadData);

    // I added this to flip the normal if the object is inverted (for CSG).
    // However, I subsequently commented it out for speed reasons - it doesn't
    // make a difference (no pun intended). The preexisting flip code below
    // produces a similar (though more extensive) result. [NK]
    // Actually, we should keep this code to guarantee that Normal_Direction
    // is set properly. [NK]
    if (Test_Flag(isect.Object, INVERTED_FLAG))
    {
        rawnormal.invert();
    }

    // if the surface normal points away, flip its direction
    normaldirection = dot(rawnormal, ray.Direction);
    if (normaldirection > 0.0)
    {
        rawnormal.invert();
    }

    isect.INormal = rawnormal;
    isect.PNormal = rawnormal;

    // now switch to UV mapping if we need to
    if (Test_Flag(isect.Object, UV_FLAG))
    {
        // TODO FIXME
        //  I think we have a serious problem here regarding bump mapping:
        //  The UV vector doesn't contain any information about the (local) *orientation* of U and V
        //  in our XYZ co-ordinate system!
        //  This causes slopes do be applied in the wrong directions.

        // get the UV vect of the intersection
        isect.Object->UVCoord(uvcoords, &isect);
        // save the normal and UV coords into Intersection
        isect.Iuv = uvcoords;

        ipoint = Vector3d(uvcoords.u(), uvcoords.v(), 0.0);
    }

    bool isMultiTextured = Test_Flag(isect.Object, MULTITEXTURE_FLAG) ||
                           ((isect.Object->Texture == nullptr) &&
                            Test_Flag(isect.Object, CUTAWAY_TEXTURES_FLAG));

    // get textures and weights
    if (isMultiTextured == true)
    {
        isect.Object->Determine_Textures(&isect, normaldirection > 0.0, wtextures, threadData);
    }
    else if (isect.Object->Texture != nullptr)
    {
        if ((normaldirection > 0.0) && (isect.Object->Interior_Texture != nullptr))
        {
            wtextures.push_back(WeightedTexture(1.0, isect.Object->Interior_Texture)); /* Chris Huff: Interior Texture patch */
        }
        else
        {
            wtextures.push_back(WeightedTexture(1.0, isect.Object->Texture));
        }
    }
    else
    {
        // don't need to do anything as the texture list will be empty.
        // TODO: could we perform these tests earlier ? [cjc]
        lightColorCacheIndex--;
        return;
    }

    // Now, we perform the lighting calculations by stepping through
    // the list of textures and summing the weighted color.

    for (auto i(wtextures.begin()); i != wtextures.end(); i++)
    {
        TextureVector warps(texturePool);
        POV_REFPOOL_ASSERT(warps->empty()); // verify that the TextureVector pulled from the pool is in a cleaned-up condition

        // if the contribution of this texture is negligible skip ahead
        if ((i->weight < ray.GetTicket().adcBailout) || (i->texture == nullptr))
            continue;

        if (photonPass == true)
        {
            // For the photon pass, colour (and thus c1) represents the
            // light energy being transmitted by the photon.  Because of this, we
            // compute the weighted energy value, then pass it to the texture for
            // processing.
            c1 = colour * i->weight;

            // NOTE that ComputeOneTextureColor is being used for a secondary purpose, and
            // that to place photons on the surface and trigger recursive photon shooting
            ComputeOneTextureColour(c1, t1, i->texture, *warps, ipoint, rawnormal, ray, weight, isect, false, true);
        }
        else
        {
            ComputeOneTextureColour(c1, t1, i->texture, *warps, ipoint, rawnormal, ray, weight, isect, false, false);

            tmpCol    += i->weight * c1;
            tmpTransm += i->weight * t1;
        }
    }

    // [CLi] moved this here from Trace::ComputeShadowTexture() and Trace::ComputeLightedTexture(), respectively,
    // to avoid media to be computed twice when dealing with averaged textures.
    // TODO - For photon rays we're still potentially doing double work on media.
    // TODO - For shadow rays we're still potentially doing double work on distance-based attenuation.
    // Calculate participating media effects.
    if (!photonPass && qualityFlags.media && (!ray.GetInteriors().empty()) && (ray.IsHollowRay() == true))
    {
        media.ComputeMedia(ray.GetInteriors(), ray, isect, tmpCol, tmpTransm);
    }

    colour += tmpCol;
    transm += tmpTransm;

    lightColorCacheIndex--;
}

void Trace::ComputeOneTextureColour(MathColour& resultColour, ColourChannel& resultTransm, const TEXTURE *texture,
        vector<const TEXTURE *>& warps, const Vector3d& ipoint, const Vector3d& rawnormal, Ray& ray, COLC weight,
        Intersection& isect, bool shadowflag, bool photonPass)
{
    // NOTE: this method is used by the photon pass to deposit photons on the surface
    // (and not, per se, to compute texture color)
    const TextureBlendMapPtr& blendmap = texture->Blend_Map;
    const TextureBlendMapEntry *prev, *cur;
    double prevWeight, curWeight;
    double value1; // TODO FIXME - choose better name!
    Vector3d tpoint;
    Vector2d uvcoords;
    MathColour c2;
    ColourChannel t2;

    switch (texture->Type)
    {
        case NO_PATTERN:
        case PLAIN_PATTERN:
            break;
        case AVERAGE_PATTERN:
        case UV_MAP_PATTERN:
        case BITMAP_PATTERN:
        default:
            warps.push_back(texture);
            break;
    }

    // ipoint - interseciton point (and evaluation point)
    // epoint - evaluation point
    // tpoint - turbulated/transformed point

    if (texture->Type <= LAST_SPECIAL_PATTERN)
    {
        switch (texture->Type)
        {
            case NO_PATTERN:
                POV_PATTERN_ASSERT(false);  // in Create_Texture(), TEXTURE->Type is explicitly set to PLAIN_PATTERN (in deviation
                                            // from the default TPat settings), and there is no piece of code ever setting it to
                                            // NO_PATTERN (except during parsing of magnet patterns, but that code makes sure it
                                            // doesn't remain set to NO_PATTERN).
                resultColour.Clear();
                resultTransm = 1.0;
                break;
            case AVERAGE_PATTERN:
                Warp_EPoint(tpoint, ipoint, warps.back());
                ComputeAverageTextureColours(resultColour, resultTransm, texture, warps, tpoint, rawnormal,
                        ray, weight, isect, shadowflag, photonPass);
                break;
            case UV_MAP_PATTERN:
                // TODO FIXME
                //  I think we have a serious problem here regarding bump mapping:
                //  The UV vector doesn't contain any information about the (local) *orientation* of U and V
                //  in our XYZ co-ordinate system!
                //  This causes slopes do be applied in the wrong directions.

                // Don't bother warping, simply get the UV vect of the intersection
                isect.Object->UVCoord(uvcoords, &isect);
                tpoint = Vector3d(uvcoords[U], uvcoords[V], 0.0);
                cur = &(texture->Blend_Map->Blend_Map_Entries[0]);
                ComputeOneTextureColour(resultColour, resultTransm, cur->Vals, warps, tpoint, rawnormal,
                        ray, weight, isect, shadowflag, photonPass);
                break;
            case BITMAP_PATTERN:
                Warp_EPoint(tpoint, ipoint, texture);
                ComputeOneTextureColour(resultColour, resultTransm, material_map(tpoint, texture),
                        warps, tpoint, rawnormal, ray, weight, isect, shadowflag, photonPass);
                break;
            case PLAIN_PATTERN:
                if (shadowflag == true)
                {
                    ComputeShadowTexture(resultColour, texture, warps, ipoint, rawnormal, ray, isect);
                    // NB: filter and transmit components are ignored by the caller when tracing shadow rays,
                    // so no need to set them
                }
                else
                {
                    ComputeLightedTexture(resultColour, resultTransm, texture, warps, ipoint, rawnormal,
                            ray, weight, isect);
                }
                break;
            default:
                throw POV_EXCEPTION_STRING("Bad texture type in ComputeOneTextureColour");
        }
    }
    else
    {
        // NK 19 Nov 1999 added Warp_EPoint
        Warp_EPoint(tpoint, ipoint, texture);
        value1 = Evaluate_TPat(texture, tpoint, &isect, &ray, threadData);

        blendmap->Search(value1, prev, cur, prevWeight, curWeight);

        // NK phmap
        if (photonPass)
        {
            if (prev == cur)
            {
                ComputeOneTextureColour(resultColour, resultTransm, cur->Vals, warps, tpoint, rawnormal,
                        ray, weight, isect, shadowflag, photonPass);
            }
            else
            {
                c2 = resultColour * curWeight;
                ComputeOneTextureColour(c2, t2, cur->Vals, warps, tpoint, rawnormal, ray, weight,
                        isect, shadowflag, photonPass);
                c2 = resultColour * prevWeight; // modifies RGB, but leaves Filter and Transmit unchanged
                ComputeOneTextureColour(c2, t2, prev->Vals, warps, tpoint, rawnormal, ray, weight,
                        isect, shadowflag, photonPass);
            }
        }
        else
        {
            ComputeOneTextureColour(resultColour, resultTransm, cur->Vals, warps, tpoint, rawnormal,
                    ray, weight, isect, shadowflag, photonPass);

            if (prev != cur)
            {
                ComputeOneTextureColour(c2, t2, prev->Vals, warps, tpoint, rawnormal, ray, weight,
                        isect, shadowflag, photonPass);
                resultColour = curWeight * resultColour + prevWeight * c2;
                resultTransm = curWeight * resultTransm + prevWeight * t2;
            }
        }
    }
}

void Trace::ComputeAverageTextureColours(MathColour& resultColour, ColourChannel& resultTransm,
        const TEXTURE *texture, vector<const TEXTURE *>& warps, const Vector3d& ipoint,
        const Vector3d& rawnormal, Ray& ray, COLC weight, Intersection& isect, bool shadowflag,
        bool photonPass)
{
    const TextureBlendMapPtr& bmap = texture->Blend_Map;
    float total = 0.0;
    MathColour lc;
    ColourChannel lt;

    if (photonPass == false)
    {
        resultColour.Clear();
        resultTransm = 0.0;

        for (auto i = bmap->Blend_Map_Entries.cbegin(); i != bmap->Blend_Map_Entries.cend(); i++)
        {
            float val = i->value;

            ComputeOneTextureColour(lc, lt, i->Vals, warps, ipoint, rawnormal, ray, weight, isect,
                    shadowflag, photonPass);

            resultColour += lc * val;
            resultTransm += lt * val;

            total += val;
        }

        resultColour /= total;
        resultTransm /= total;
    }
    else
    {
        for (auto i = bmap->Blend_Map_Entries.cbegin(); i != bmap->Blend_Map_Entries.cend(); i++)
            total += i->value;

        for (auto i = bmap->Blend_Map_Entries.cbegin(); i != bmap->Blend_Map_Entries.cend(); i++)
        {
            lc = resultColour * (i->value / total);

            ComputeOneTextureColour(lc, lt, i->Vals, warps, ipoint, rawnormal, ray, weight, isect, shadowflag, photonPass);
        }
    }
}

void Trace::ComputeLightedTexture(MathColour& resultColour, ColourChannel& resultTransm, const TEXTURE *texture,
        vector<const TEXTURE *>& warps, const Vector3d& ipoint, const Vector3d& rawnormal, Ray& ray,
        COLC weight, Intersection& isect)
{
    Interior *interior = isect.Object->interior.get();
    POV_ASSERT(interior != nullptr);  // v3.7 onward always have interior.

    const TEXTURE *layer;
    bool radiosity_done, radiosity_back_done, radiosity_needed;
    int layer_number;
    double w1;
    double new_Weight;
    double att, trans, max_Radiosity_Contribution;
    double cos_Angle_Incidence;
    Vector3d layNormal, topNormal;
    MathColour attCol;
    TransColour layCol;
    MathColour rflCol;
    MathColour rfrCol;
    ColourChannel rfrTransm;
    MathColour filCol;
    MathColour tmpCol, tmp;
    MathColour ambCol; // Note that there is no gathering of filter or transparency
    MathColour ambBackCol;
    bool one_colour_found, colour_found;
    bool tir_occured;
    std::unique_ptr<PhotonGatherer> surfacePhotonGatherer(nullptr);

    double relativeIor;  // TODO How to mesh texture->Finish->IOR with interior->IOR...
    ComputeRelativeIOR(ray, interior, relativeIor);

    WNRXVector listWNRX(wnrxPool);         // Weight, Normal, Reflectivity, eXponent
    POV_REFPOOL_ASSERT(listWNRX->empty()); // verify that the WNRXVector pulled from the pool is in a cleaned-up condition

    // resultColour builds up the apparent visible color of the point.
    // resultTransm builds up the apparent visible color of whatever is behind the point (presuming 100% white background).
    resultColour.Clear();
    resultTransm = 0.0;

    // filCol serves two purposes. It accumulates the filter properties
    // of a multi-layer texture so that if a ray makes it all the way through
    // all layers, the color of object behind is filtered by this object.
    // It also is used to attenuate how much of an underlayer you
    // can see in a layered texture. Note that when computing the reflective
    // properties of a layered texture, the upper layers don't filter the
    // light from the lower layers -- the layer colors add together (even
    // before we added additive transparency via the "transmit" 5th
    // color channel). However when computing the transmitted rays, all layers
    // filter the light from any objects behind this object. [CY 1/95]

    // NK layers - switched transmit component to zero
    // [CLi] changed filCol to RGB, as filter and transmit were always pinned to 1.0 and 0.0 respectively anyway
    filCol = MathColour(1.0);

    trans = 1.0;

    // Add in radiosity (stochastic interreflection-based ambient light) if desired
    radiosity_done = false;
    radiosity_back_done = false;

    // This block just sets up radiosity for the code inside the loop, which is first-time-through.
    radiosity_needed = (sceneData->radiositySettings.radiosityEnabled == true) &&
                       (radiosity.CheckRadiosityTraceLevel(ray.GetTicket()) == true) &&
                       (Test_Flag(isect.Object, IGNORE_RADIOSITY_FLAG) == false);

    // Loop through the layers and compute the ambient, diffuse,
    // phong and specular for these textures.
    one_colour_found = false;

    if (sceneData->photonSettings.photonsEnabled && sceneData->surfacePhotonMap.numPhotons > 0)
    {
        surfacePhotonGatherer.reset(new PhotonGatherer(&sceneData->surfacePhotonMap, sceneData->photonSettings));
    }

    for (layer_number = 0, layer = texture; (layer != nullptr) && (trans > ray.GetTicket().adcBailout);
            layer_number++, layer = layer->Next)
    {
        // Get perturbed surface normal.
        layNormal = rawnormal;

        if (qualityFlags.normals && (layer->Tnormal != nullptr))
        {
            for (auto i(warps.begin()); i != warps.end(); i++)
                Warp_Normal(layNormal, layNormal, *i, Test_Flag(*i, DONT_SCALE_BUMPS_FLAG));

            Perturb_Normal(layNormal, layer->Tnormal, ipoint, &isect, &ray, threadData);

            if ((Test_Flag(layer->Tnormal, DONT_SCALE_BUMPS_FLAG)))
            {
                layNormal.normalize();
            }

            // TODO - Reverse iterator may be less performant than forward iterator; we might want to
            //        compare performance with using forward iterators and decrement, or using random access.
            for (auto i(warps.rbegin()); i != warps.rend(); i++)
                UnWarp_Normal(layNormal, layNormal, *i, Test_Flag(*i, DONT_SCALE_BUMPS_FLAG));
        }

        // Store top layer normal.
        if (layer_number == 0)
        {
            topNormal = layNormal;
        }

        // Get surface colour.
        new_Weight = weight * trans;
        colour_found = Compute_Pigment(layCol, layer->Pigment, ipoint, &isect, &ray, threadData);

        // If a valid color was returned set one_colour_found to true.
        // An invalid color is returned if a surface point is outside
        // an image map used just once.
        one_colour_found = (one_colour_found || colour_found);

        // This section of code used to be the routine Compute_Reflected_Colour.
        // I copied it in here to rearrange some of it more easily and to
        // see if we could eliminate passing a zillion parameters for no
        // good reason. [CY 1/95]

        if (qualityFlags.ambientOnly)
        {
            // Only use top layer and kill transparency if low quality.
            resultColour = layCol.colour();
            resultTransm = 0.0;
        }
        else
        {
            // Store vital information for later reflection.
            listWNRX->push_back(WNRX(new_Weight, layNormal, MathColour(), layer->Finish->Reflection_Exp));

            // angle-dependent reflectivity
            cos_Angle_Incidence = -dot(ray.Direction, layNormal);

            ComputeReflectivity(listWNRX->back().weight, listWNRX->back().reflec,
                                layer->Finish->Reflection_Max, layer->Finish->Reflection_Min,
                                layer->Finish->Reflection_Fresnel, layer->Finish->Reflection_Falloff,
                                cos_Angle_Incidence, relativeIor);

            ComputeMetallic(listWNRX->back().reflec, layer->Finish->Reflection_Metallic, layCol.colour(),
                    cos_Angle_Incidence);

            // We need to reduce the layer's own brightness if it is transparent.
            att = layCol.Opacity();

            if (layer->Finish->AlphaKnockout)
                listWNRX->back().reflec *= att;

            // now compute the BRDF or BSSRDF contribution
            tmpCol.Clear();

            if (sceneData->useSubsurface && layer->Finish->UseSubsurface && qualityFlags.subsurface)
            {
                // Add diffuse & single scattering contribution.
                ComputeSubsurfaceScattering(layer->Finish, layCol.colour(), isect, ray, layNormal, tmpCol, att);
                // [CLi] moved multiplication with filCol to further below

                // Radiosity-style ambient may be subject to subsurface light transport.
                // In that case, the respective computations are handled by the BSSRDF code already.
                if (sceneData->subsurfaceUseRadiosity)
                    radiosity_needed = false;
            }
            // Add radiosity ambient contribution.
            if (radiosity_needed)
            {
                double diffuse    = layer->Finish->Diffuse;
                double brilliance = 1.0;

                if (sceneData->radiositySettings.brilliance)
                {
                    diffuse    *= layer->Finish->BrillianceAdjustRad;
                    brilliance =  layer->Finish->Brilliance;
                }

                // if radiosity calculation needed, but not yet done, do it now
                // TODO FIXME - [CLi] with "normal on", shouldn't we compute radiosity for each layer separately
                //   (if it has pertubed normals)?
                // TODO FIXME - [CLi] with "brilliance on", shouldn't we compute radiosity for each layer separately
                //   (if it has non-default brilliance)?
                if (radiosity_done == false)
                {
                    // calculate max possible contribution of radiosity, to see if calculating it is worthwhile
                    // TODO FIXME - other layers may see a higher weight!
                    // Maybe we should go along and compute *first* the total contribution radiosity will make,
                    // and at the *end* apply it.
                    max_Radiosity_Contribution = (filCol * layCol.colour()).WeightGreyscale() * att * diffuse;

                    if (max_Radiosity_Contribution > ray.GetTicket().adcBailout)
                    {
                        radiosity.ComputeAmbient(isect.IPoint, rawnormal, layNormal, brilliance, ambCol,
                                weight * max_Radiosity_Contribution, ray.GetTicket());
                        radiosity_done = true;
                    }
                }

                // [CLi] moved multiplication with filCol to further below
                MathColour radiosityContribution = (layCol.colour() * ambCol) * (att * diffuse);

                diffuse = layer->Finish->DiffuseBack;
                if (sceneData->radiositySettings.brilliance)
                {
                    diffuse *= layer->Finish->BrillianceAdjustRad;
                }

                // if backside radiosity calculation needed, but not yet done, do it now
                // TODO FIXME - [CLi] with "normal on", shouldn't we compute radiosity for each layer separately
                //   (if it has pertubed normals)?
                // TODO FIXME - [CLi] with "brilliance on", shouldn't we compute radiosity for each layer separately
                //   (if it has non-default brilliance)?
                if (layer->Finish->DiffuseBack != 0.0)
                {
                    if (radiosity_back_done == false)
                    {
                        // calculate max possible contribution of radiosity, to see if calculating it is worthwhile
                        // TODO FIXME - other layers may see a higher weight!
                        // Maybe we should go along and compute *first* the total contribution radiosity will make,
                        // and at the *end* apply it.
                        max_Radiosity_Contribution = (filCol * layCol.colour()).WeightGreyscale() * att * diffuse;

                        if (max_Radiosity_Contribution > ray.GetTicket().adcBailout)
                        {
                            radiosity.ComputeAmbient(isect.IPoint, -rawnormal, -layNormal, brilliance, ambBackCol,
                                    weight * max_Radiosity_Contribution, ray.GetTicket());
                            radiosity_back_done = true;
                        }
                    }

                    // [CLi] moved multiplication with filCol to further below
                    radiosityContribution += (layCol.colour() * ambBackCol) * (att * diffuse);
                }

                if (layer->Finish->Fresnel != 0.0)
                {
                    // In diffuse reflections (which includes radiosity), the Fresnel effect is
                    // that we _lose_ the reflected component as the light enters the material
                    // (where it changes direction randomly), and then _again lose_ another
                    // reflected component as the light leaves the material. However, taking into
                    // account the former of these losses requires knowledge of the incoming light
                    // direction, which must be accounted for in radiosity sampling (TODO - not implemented yet),
                    // so we only do the latter here.
                    // NB: One might think that to properly compute the outgoing loss we would have
                    // to compute the Fresnel term for the _internal_ angle of incidence and the
                    // _inverse_ of the relative refractive index; however, the Fresnel formula
                    // is such that we can just as well plug in the external angle of incidence
                    // and straight relative refractive index.
                    double f = layer->Finish->Fresnel * FresnelR(cos_Angle_Incidence, relativeIor);
                    radiosityContribution *= (1.0 - f);
                }

                tmpCol += radiosityContribution;
            }

            // Add emissive ("classic" ambient) contribution.
            // [CLi] moved multiplication with filCol to further below
            MathColour emission = layer->Finish->Emission;
            if (!sceneData->radiositySettings.radiosityEnabled)
            {   // only use "ambient" setting when radiosity is disabled.
                // Note. In v3.8 the default ambient was changed to 0.0 so ambient less often in play.
                emission += layer->Finish->Ambient;
            }

            if (layer->Finish->Fresnel != 0.0)
            {
                // Light emitted from the material itself is subject to the Fresnel effect as it
                // leaves the material _losing_ light reflected at the interface.
                // In diffuse reflections (which includes ambient), the Fresnel effect is
                // that we _lose_ the reflected component as the light enters the material
                // (where it changes direction randomly), and then _again lose_ another
                // reflected component as the light leaves the material. The former of these
                // losses needs to be accounted for when choosing the `ambient` setting,
                // so we only do the latter here.
                // NB: One might think that to properly compute the outgoing loss we would have
                // to compute the Fresnel term for the _internal_ angle of incidence and the
                // _inverse_ of the relative refractive index; however, the Fresnel formula
                // is such that we can just as well plug in the external angle of incidence
                // and straight relative refractive index.
                double f = layer->Finish->Fresnel * FresnelR(cos_Angle_Incidence, relativeIor);
                emission *= (1.0 - f);
            }
            tmpCol += (layCol.colour() * emission * att);

            // set up the "litObjectIgnoresPhotons" flag (thread variable) so that
            // ComputeShadowColour will know whether or not this lit object is
            // ignoring photons, which affects partial-shadowing (i.e. filter and transmit)
            threadData->litObjectIgnoresPhotons = Test_Flag(isect.Object,PH_IGNORE_PHOTONS_FLAG);

            // Add diffuse, phong, specular, and iridescence contribution.
            // (We don't need to do this for (non-radiosity) rays during pretrace, as it does not affect radiosity sampling)
            if (!ray.IsPretraceRay())
            {
                if (((layer->Finish->Diffuse     != 0.0) ||
                     (layer->Finish->DiffuseBack != 0.0) ||
                     (layer->Finish->Specular    != 0.0) ||
                     (layer->Finish->Phong       != 0.0)) &&
                    ((!layer->Finish->AlphaKnockout) ||
                     (att != 0.0)))
                {
                    MathColour classicContribution;

                    ComputeDiffuseLight(layer->Finish, isect.IPoint, ray, layNormal, layCol.colour(), classicContribution,
                            att, isect.Object, relativeIor);

                    tmpCol += classicContribution;
                }
            }

            if (sceneData->photonSettings.photonsEnabled && sceneData->surfacePhotonMap.numPhotons > 0)
            {
                // NK phmap - now do the same for the photons in the area
                if (!Test_Flag(isect.Object, PH_IGNORE_PHOTONS_FLAG))
                {
                    MathColour photonsContribution;

                    ComputePhotonDiffuseLight(layer->Finish, isect.IPoint, ray, layNormal, rawnormal, layCol.colour(),
                            photonsContribution, att, isect.Object, relativeIor, *surfacePhotonGatherer);

                    tmpCol += photonsContribution;
                }
            }

            tmpCol *= filCol;
            resultColour += tmpCol;
        }

        // Get new filter color.
        if (colour_found)
        {
            filCol *= layCol.TransmittedColour();

            if (layer->Finish->Conserve_Energy != 0 && listWNRX->empty() == false)
            {
                // adjust filCol based on reflection
                // this would work so much better with r,g,b,rt,gt,bt
                filCol *= (1.0 - listWNRX->back().reflec).ClippedUpper(1.0);
            }
        }

        // Get new remaining translucency.
        // [CLi] changed filCol to RGB, as filter and transmit were always pinned to 1.0 and 0.0, respectively anyway
        // TODO CLARIFY - is this working properly if filCol.greyscale() is negative? (what would be the right thing then?)
        trans = min(1.0, (double)fabs(filCol.Greyscale()));
    }

    // Calculate transmitted component.
    //
    // If the surface is translucent a transmitted ray is traced
    // and its contribution is added to the total ResCol after
    // filtering it by filCol.
    tir_occured = false;

    if ((trans > ray.GetTicket().adcBailout) && qualityFlags.refractions)
    {
        // [CLi] changed filCol to RGB, as filter and transmit were always pinned to 1.0 and 0.0, respectively anyway
        // TODO CLARIFY - is this working properly if some filCol component is negative? (what would be the right thing then?)
        w1 = filCol.WeightMaxAbs();
        new_Weight = weight * w1;

        // Trace refracted ray.
        tir_occured = ComputeRefraction(texture->Finish, interior, isect.IPoint, ray, topNormal, rawnormal, rfrCol,
                rfrTransm, new_Weight);

        // Get distance based attenuation.
        // TODO - nearly the same code is used four times elsewhere. Search on previous line comment.
        // Note - Do NOT make a callable function/method, but maybe at some point a template. We want
        // to force the functionality inline for performance...
        attCol = interior->Amplify;

        if (ray.IsInterior(interior) == true)
        {
            if (fabs(interior->Fade_Distance) > gkMinIsectDepthReturned)
            {
                float intensity_max = interior->Intensity_Max;
                float depth = isect.Depth / interior->Fade_Distance;
                MathColour invFade_Colour = 1.0f - interior->Fade_Colour;
                if (interior->Fade_Power >= 1000)
                {
                    attCol *= Exp(-invFade_Colour * depth);
                }
                else
                {
                    float attVal = 1.0f + powf(depth, interior->Fade_Power);
                    MathColour attColTmp = (interior->Fade_Colour + invFade_Colour / attVal);
                    attColTmp = attColTmp.ClippedLower(0.0f); // If negative fade color, att moves through 0.0 -> neg
                                                              // not far beyond the fade distance.
                    attCol *= attColTmp;
                }

                // If user wishes to generate an alarm color where the resulting color is outside
                // the [0 to 1] range they can specify a negative intensity_max value.
                //
                // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
                // specify a positive positive intensity_max value.
                //
                // See similar blocks of code for diffuse, phong, specular, ...?)
                //
                if (intensity_max!=0.0)
                {
                    if (intensity_max>0.0)
                    {
                        if (attCol.Min()<0.0)
                        {
                            attCol = attCol.ClippedLower(0.0f);
                        }
                        if (attCol.Max()>intensity_max)
                        {
                            attCol = attCol/attCol.Max()*intensity_max;
                        }
                    }
                    else if ((attCol.Min()<0.0) || (attCol.Max()>std::abs(intensity_max)))
                    {
                        std::size_t hashvalue;
                        hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.y()/0.005));
                        float rndR = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                        hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.x()/0.005));
                        float rndG = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                        hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.z()/0.005));
                        float rndB = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                        attCol.Set(rndR*0.45,rndG*0.65,rndB*0.55);
                    }
                }
            }
        }

        // If total internal reflection occured the transmitted light is not filtered.
        if (tir_occured)
        {
            resultColour += attCol * rfrCol;
            // NOTE: transm() (alpha channel) stays zero
        }
        else
        {
            if (one_colour_found)
            {
                // [CLi] changed filCol to RGB, as filter and transmit were always pinned to 1.0 and 0.0, respectively anyway
                resultColour += attCol * rfrCol * filCol;
                // We need to know the transmittance value for the alpha channel. [DB]
                resultTransm = attCol.Greyscale() * rfrTransm * trans;
            }
            else
            {
                resultColour += attCol * rfrCol;
                // We need to know the transmittance value for the alpha channel. [DB]
                resultTransm = attCol.Greyscale() * rfrTransm;
            }
        }
    }

    // Calculate reflected component.
    //
    // If total internal reflection occured all reflections using
    // TopNormal are skipped.
    if (qualityFlags.reflections)
    {
        layer = texture;
        for (int i = 0; i < layer_number; i++)
        {
            if ((!tir_occured) ||
               (fabs(topNormal[X]-(*listWNRX)[i].normal[X]) > gkDBL_epsilon) ||
               (fabs(topNormal[Y]-(*listWNRX)[i].normal[Y]) > gkDBL_epsilon) ||
               (fabs(topNormal[Z]-(*listWNRX)[i].normal[Z]) > gkDBL_epsilon))
            {
                if (!(*listWNRX)[i].reflec.IsZero())
                {
                    float intensity_max = layer->Finish->Reflection_Intensity_Max;
                    rflCol.Clear();
                    ComputeReflection(layer->Finish, isect.IPoint, ray, (*listWNRX)[i].normal, rawnormal, rflCol,
                            (*listWNRX)[i].weight);

                    tmpCol.Clear();
                    if ((*listWNRX)[i].reflex != 1.0)  // TODO I think this the exponetial reflection. If povr dropped, drop this...
                    {
                        tmpCol        = (*listWNRX)[i].reflec * Pow(rflCol, (*listWNRX)[i].reflex);
                      //resultColour += (*listWNRX)[i].reflec * Pow(rflCol, (*listWNRX)[i].reflex);
                    }
                    else
                    {
                        tmpCol        = (*listWNRX)[i].reflec * rflCol;
                      //resultColour += (*listWNRX)[i].reflec * rflCol;
                    }
                    tmpCol = tmpCol + resultColour;

                    // If user wishes to generate an alarm color where the resulting color is outside
                    // the [0 to 1] range they can specify a negative intensity_max value.
                    //
                    // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
                    // specify a positive positive intensity_max value.
                    //
                    // See similar blocks of code for diffuse, phong, specular, ...?)
                    //
                    if (intensity_max!=0.0)
                    {
                        if (intensity_max>0.0)
                        {
                            if (tmpCol.Min()<0.0)
                            {
                                tmpCol = tmpCol.ClippedLower(0.0f);
                            }
                            if (tmpCol.Max()>intensity_max)
                            {
                                tmpCol = tmpCol/tmpCol.Max()*(*listWNRX)[i].reflec*intensity_max;
                            }
                        }
                        else if ((tmpCol.Min()<0.0) || (tmpCol.Max()>std::abs(intensity_max)))
                        {
                            std::size_t hashvalue;
                            hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.y()/0.005));
                            float rndR = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                            hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.x()/0.005));
                            float rndG = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                            hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.z()/0.005));
                            float rndB = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                            tmpCol.Set(rndR*0.45,rndG*0.65,rndB*0.55);
                        }
                    }

                  //resultColour += tmpCol; // Prior method.
                    resultColour = tmpCol;
                }
            }
            layer = layer->Next;
        }
    }
}

void Trace::ComputeShadowTexture(MathColour& filtercolour, const TEXTURE *texture, vector<const TEXTURE *>& warps,
        const Vector3d& ipoint, const Vector3d& rawnormal, const Ray& ray, Intersection& isect)
{
    Interior *interior = isect.Object->interior.get();
    POV_ASSERT(interior != nullptr);  // v3.7 onward always have interior.

    const TEXTURE *layer;
    double caustics, dotval, k;
    Vector3d layer_Normal;
    TransColour layer_Pigment_Colour;
    bool colour_found;

    MathColour tmpCol = MathColour(1.0);

    for (layer = texture; layer != nullptr; layer = layer->Next)
    {
        colour_found = Compute_Pigment(layer_Pigment_Colour, layer->Pigment, ipoint, &isect, &ray, threadData);

        if (colour_found)
        {
            tmpCol *= layer_Pigment_Colour.TransmittedColour();
        }

        // Get normal for faked caustics (will rewrite later to cache).
        // TODO  See no caching at all; what does comment above mean - if anything?

        // NOTE! These faked caustics also very dependent upon the position of light source
        // with respect to the object creating the refracted ray shadow.
        if ((caustics = interior->Caustics) != 0.0)
        {
            layer_Normal = rawnormal;

            if (qualityFlags.normals && (layer->Tnormal != nullptr))
            {
                for (auto i(warps.begin()); i != warps.end(); i++)
                    Warp_Normal(layer_Normal, layer_Normal, *i, Test_Flag(*i, DONT_SCALE_BUMPS_FLAG));

                Perturb_Normal(layer_Normal, layer->Tnormal, ipoint, &isect, &ray, threadData);

                if ((Test_Flag(layer->Tnormal,DONT_SCALE_BUMPS_FLAG)))
                    layer_Normal.normalize();

                // TODO - Reverse iterator may be less performant than forward iterator; we might want to
                //        compare performance with using forward iterators and decrement, or using random access.
                for (auto i(warps.rbegin()); i != warps.rend(); i++)
                    UnWarp_Normal(layer_Normal, layer_Normal, *i, Test_Flag(*i, DONT_SCALE_BUMPS_FLAG));
            }

            dotval = dot(layer_Normal, ray.Direction);

            k = (1.0 + pow(fabs(dotval), caustics));

            tmpCol *= k;
        }
    }

    // Note. On long questions about doing double work below. The short answer is not
    // too much for the reason this called during shadow ray tracing. For the initial
    // intersection point for which we start a shadow trace ray there is some duplication,
    // but doesn't look all that easy to re-use that information.

    // Get distance based attenuation.
    // TODO - nearly the same code is used four times elsewhere. Search on previous line comment.
    // Note - Do NOT make a callable function/method, but maybe at some point a template. We want
    // to force the functionality inline for performance...
    MathColour attCol = interior->Amplify;

    if (ray.IsInterior(interior) == true)
    {
        if (fabs(interior->Fade_Distance) > gkMinIsectDepthReturned)
        {
            float intensity_max = interior->Intensity_Max;
            float depth = isect.Depth / interior->Fade_Distance;
            MathColour invFade_Colour = 1.0f - interior->Fade_Colour;
            if (interior->Fade_Power >= 1000) // NK attenuate
            {
                attCol *= Exp(-invFade_Colour * depth);
            }
            else
            {
                float attVal = 1.0f + powf(depth, interior->Fade_Power);
                MathColour attColTmp = (interior->Fade_Colour + invFade_Colour / attVal);
                attColTmp = attColTmp.ClippedLower(0.0f); // If negative fade color, att moves through 0.0 -> neg
                                                          // not far beyond the fade distance.
                attCol *= attColTmp;
            }

            // If user wishes to generate an alarm color where the resulting color is outside
            // the [0 to 1] range they can specify a negative intensity_max value.
            //
            // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
            // specify a positive positive intensity_max value.
            //
            // See similar blocks of code for diffuse, phong, specular, ...?)
            //
            if (intensity_max!=0.0)
            {
                if (intensity_max>0.0)
                {
                    if (attCol.Min()<0.0)
                    {
                        attCol = attCol.ClippedLower(0.0f);
                    }
                    if (attCol.Max()>intensity_max)
                    {
                        attCol = attCol/attCol.Max()*intensity_max;
                    }
                }
                else if ((attCol.Min()<0.0) || (attCol.Max()>std::abs(intensity_max)))
                {
                    std::size_t hashvalue;
                    hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.y()/0.005));
                    float rndR = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                    hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.x()/0.005));
                    float rndG = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                    hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.z()/0.005));
                    float rndB = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                    attCol.Set(rndR*0.45,rndG*0.65,rndB*0.55);
                }
            }
        }
    }

    filtercolour = tmpCol * attCol;
}


void Trace::ComputeReflection(const FINISH* finish, const Vector3d& ipoint, Ray& ray, const Vector3d& normal,
        const Vector3d& rawnormal, MathColour& colour, COLC weight)
{
    Ray nray(ray);
    double n, n2;

    nray.SetFlags(Ray::ReflectionRay, ray);

    n = -2.0 * dot(ray.Direction, normal);
    nray.Direction = ray.Direction + n * normal;

    // The next block of code represent optional cures where we get normal
    // inversion relative to the raw surface normal or the reflected ray. Any
    // time a normal block is applied and the raw normals perturbed the
    // perturbed normal can end up pointing away from incoming or (herein)
    // reflected rays. Further, if the bump size is large a perturbed normal
    // can end up pointing against the raw surface normal. See documentation
    // for more detail.

    // The povr inv_pnormal_cure options for the reflection block are:
    //   0.0 - Long time inversion of perturbed normals turned against raw normal.
    //   1.0 - This effectively 0->(2|3) and the code run since v3.1 1998.
    //   2.0 - For reflected rays at odds with raw, but perturbed. Adjust twice toward raw ray.
    //   3.0 - For reflected rays at odds with raw, but perturbed. Creates smooth virtual reflection.
    //   4.0 - Sees 0, but where perturbed against raw, flips to follow perturbed intent.
    //   5.5 - No matter how violently perturbed flips by probability in fractional part of float.
    //  11.0 - This 1, but flipping when 3 and 2 used. 0->(3|2). More distinct at tangents.
    // Note. 0,4,5 all allow - or force - reflected rays through surface as if porous.

    int inv_normal_cure = (int)finish->Reflection_InvPNormal_Cure; // Stored as float for probability on 5.
    switch (inv_normal_cure)
    {
        case 0 : // Apply no additional fix up at all. Often OK and faster for small perturbations.
            break;
        case 1 : // This the cure always run since 1998.

            // Nathan Kopp & CEY 1998 - Reflection bugfix if the new ray is
            // going the opposite direction as raw normal, we need to fix it.
            // povr's micro with large bumps size of 9 basically treats
            // reflected rays in thirds. No fix, fix to rawnormal, double
            // reflect fix. If you have any near tangents, any normal{}
            // bump_size is causing some adjustments to the reflected ray
            // direction.

            n = dot(nray.Direction, rawnormal);
            if (n < 0.0)
            {
                n2 = dot(nray.Direction, normal); // It needs fixing. Which kind?

                if (n2 < 0.0)
                {
                    // Reflected inside rear virtual surface.
                    n = -2.0 * dot(ray.Direction, rawnormal);
                    nray.Direction = ray.Direction + n * rawnormal;
                }
                else
                {
                    n *= -2.0; // Double reflect NRay using Raw_Normal
                    nray.Direction += n * rawnormal;
                }
            }
            break;
        case 2 : // The initial test here is 1's. But we only do the double reflect fix.

            n = dot(nray.Direction, rawnormal);
            if (n < 0.0)
            {
                n *= -2.0;
                nray.Direction += n * rawnormal;
            }
            break;
        case 3 : // The initial test here is 1's. But we only do virtual inside surface rawnormal fix.

            n = dot(nray.Direction, rawnormal);
            if (n < 0.0)
            {
                n = -2.0 * dot(ray.Direction, rawnormal);
                nray.Direction = ray.Direction + n * rawnormal;
            }
            break;
        case 4 : // Attempt to allow perturbed normal into/through surface.

            // Front end/opt 0 always raw normal plane aligned.  The aim,
            // reflections with poruous surfaces. A mesh or cloth weave - lamp
            // shade say. The backside matters surface if a solid shape matters
            // here in that if it between the surface and what you want
            // 'somewhat' relfected through the surface, it will darken the
            // apparent reflection.

            n = dot(nray.Direction, rawnormal);
            if (n < 0.0)
            {
                double nit = dot(rawnormal,normal);
                double ni = -2.0 * dot(ray.Direction, normal);
                if (nit > 0.0)
                {
                    nray.Direction = ray.Direction + ni * normal;
                }
                else
                {
                    nray.Direction = ray.Direction - ni * normal; // Invert nray.Direction
                }
            }
            break;
        case 5 : // Attempt to flip nray directions based upon probabiliy.
            { // Local context for declared variables.
                // NB. The multiple of pi adders protect against somewhat common all zeros on an axis
                // which the C++ hash function returns always as zero.
                std::size_t hashvalue;
                hashvalue = std::hash<double>{}((ipoint.x()+1.987654*M_PI));
                double rndX = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())-0.5;
                hashvalue = std::hash<double>{}((ipoint.y()+2.123456*M_PI));
                double rndY = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())-0.5;
                hashvalue = std::hash<double>{}((ipoint.z()-3.456789*M_PI));
                double rndZ = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())-0.5;
                hashvalue = std::hash<double>{}(rndX+rndY+rndZ); // Final hash of all
                double value = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max());

                double ni          = -2.0 * dot(ray.Direction, normal);
                double probability = finish->Reflection_InvPNormal_Cure - (float) inv_normal_cure;
                if (value > probability)
                {
                    nray.Direction = ray.Direction + ni * normal;
                }
                else
                {
                    nray.Direction = ray.Direction - ni * normal;
                }
            }
            break;
        case 11 : // This is case (1) except internal dual fixes are flipped. Brightens edges/tangents
            n = dot(nray.Direction, rawnormal);
            if (n < 0.0)
            {
                n2 = dot(nray.Direction, normal); // It needs fixing. Which kind?

                if (n2 > 0.0)   // NOTE! This is the flip of what case 1 (1998 bugfix) does.
                {
                    n = -2.0 * dot(ray.Direction, rawnormal);
                    nray.Direction = ray.Direction + n * rawnormal; // reflected inside rear virtual surface.
                }
                else
                {
                    n *= -2.0; // Double reflect NRay using Raw_Normal
                    nray.Direction += n * rawnormal;
                }
            }
            break;
        default:
            throw POV_EXCEPTION_STRING("Invalid perturbed normal cure in ComputeReflection.");
    }

    nray.Direction.normalize();
    nray.Origin = ipoint;
    threadData->Stats()[Reflected_Rays_Traced]++;

    // Trace reflected ray.
    bool alphaBackground = ray.GetTicket().alphaBackground;
    ray.GetTicket().alphaBackground = false;
    if (!ray.IsPhotonRay() && (finish->Irid > 0.0))
    {
        MathColour tmpCol;
        ColourChannel dummyTransm;
        TraceRay(nray, tmpCol, dummyTransm, weight, false);
        ComputeIridColour(finish, nray.Direction, ray.Direction, normal, ipoint, tmpCol);
        colour += tmpCol;
    }
    else
    {
        ColourChannel dummyTransm;
        TraceRay(nray, colour, dummyTransm, weight, false);
    }

    // If user wishes to generate an alarm color where the resulting color is outside
    // the [0 to 1] range they can specify a negative intensity_max value.
    //
    // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
    // specify a positive positive intensity_max value.
    //
    // See similar blocks of code for diffuse, phong, interior(fade) ...?)
    //
    double intensity_max = finish->Intensity_Max;
    if (intensity_max!=0.0)
    {
        MathColour tmpCol = colour; // Large max values accounted for by weight. See: ComputeReflectivity.
        if (intensity_max>0.0)
        {
            if (tmpCol.Min()<0.0)
            {
                // With reflection only reach this if refected color <0.
                tmpCol += std::abs(tmpCol.Min());
            }
            if ((weight>1.0) || (colour.Max()>1.0)) // Not using if (tmpCol.Max()>1.0) due weight factor. See: ComputeReflectivity.
            {
                tmpCol = tmpCol/std::max(weight,colour.Max());
            }
            colour = tmpCol;
            weight = 1.0;
        }
        else
        {
            if ((colour.Min()<0.0) || (weight>1.0) || (colour.Max()>1.0))  // Additional weight factor.
            {
                tmpCol = tmpCol/weight;
                if (tmpCol.Min()<0.0)
                    tmpCol.Set(0.8,0.9,0.8) ; // Not ideal perhaps...
                std::size_t hashvalue;
                hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.y()/0.005));
                float rndR = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.x()/0.005));
                float rndG = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                hashvalue = std::hash<double>{}((double)(POV_INT64)(ipoint.z()/0.005));
                float rndB = ((double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                colour.Set(rndR*tmpCol.Red(),rndG*tmpCol.Green(),rndB*tmpCol.Blue());
                weight = 1.0;
            }
        }
    }

    ray.GetTicket().alphaBackground = alphaBackground;
}

bool Trace::ComputeRefraction(const FINISH* finish, Interior *interior, const Vector3d& ipoint, Ray& ray,
        const Vector3d& normal, const Vector3d& rawnormal, MathColour& colour, ColourChannel& transm, COLC weight)
{
    Ray nray(ray);
    Vector3d localnormal;
    double n, ior, dispersion;
    unsigned int dispersionelements = interior->Disp_NElems;
    POV_ASSERT(dispersionelements >= DISPERSION_ELEMENTS_MIN);
    bool totalReflection = false;

    nray.SetFlags(Ray::RefractionRay, ray);

    // Set up new ray.
    nray.Origin = ipoint;

    // Get ratio of iors depending on the interiors the ray is traversing.

    // Note:
    // For the purpose of refraction, the space occupied by "nested" objects is considered to be "outside" the containing objects,
    // i.e. when encountering (A (B B) A) we pretend that it's (A A|B B|A A).
    // (Here "(X" and "X)" denote the entering and leaving of object X, and "X|Y" denotes an interface between objects X and Y.)
    // In case of overlapping objects, the intersecting region is considered to be part of whatever object is encountered last,
    // i.e. when encountering (A (B A) B) we pretend that it's (A A|B B|B B).

    if (nray.GetInteriors().empty())
    {
        // The ray is entering from the atmosphere.
        nray.AppendInterior(interior);

        ior = sceneData->atmosphereIOR / interior->IOR;
        dispersion = sceneData->atmosphereDispersion / interior->Dispersion;
    }
    else
    {
        // The ray is currently inside an object.
        if (interior == nray.GetInteriors().back()) // The ray is leaving the "innermost" object
        {
            nray.RemoveInterior(interior);
            if (nray.GetInteriors().empty())
            {
                // The ray is leaving into the atmosphere
                ior = interior->IOR / sceneData->atmosphereIOR;
                dispersion = interior->Dispersion / sceneData->atmosphereDispersion;
            }
            else
            {
                // The ray is leaving into another object, i.e. (A (B B) ...
                // For the purpose of refraction, pretend that we weren't inside that other object,
                // i.e. pretend that we didn't encounter (A (B B) ... but (A A|B B|A ...
                ior = interior->IOR / nray.GetInteriors().back()->IOR;
                dispersion = interior->Dispersion / nray.GetInteriors().back()->Dispersion;
                dispersionelements = max(dispersionelements, (unsigned int)(nray.GetInteriors().back()->Disp_NElems));
            }
        }
        else if (nray.RemoveInterior(interior) == true) // The ray is leaving the intersection of overlapping objects, i.e. (A (B A) ...
        {
            // For the purpose of refraction, pretend that we had already left the other member of the intersection when we entered the overlap,
            // i.e. pretend that we didn't encounter (A (B A) ... but (A A|B B|B ...
            ior = 1.0;
            dispersion = 1.0;
        }
        else
        {
            // The ray is entering a new object.
            // For the purpose of refraction, pretend that we're leaving any containing objects,
            // i.e. pretend that we didn't encounter (A (B ... but (A A|B ...
            ior = nray.GetInteriors().back()->IOR / interior->IOR;
            dispersion = nray.GetInteriors().back()->Dispersion / interior->Dispersion;

            nray.AppendInterior(interior);
        }
    }

    bool haveDispersion = (fabs(dispersion - 1.0) >= gkDBL_epsilon);

    // Do the two mediums traversed have the same indices of refraction?
    if ((fabs(ior - 1.0) < gkDBL_epsilon) && !haveDispersion)
    {
        // Only transmit the ray.
        nray.Direction = ray.Direction;
        // Trace a transmitted ray.
        threadData->Stats()[Transmitted_Rays_Traced]++;

        colour.Clear();
        transm = 0.0;
        TraceRay(nray, colour, transm, weight, true);
    }
    else
    {
        // Refract the ray.
        n = dot(ray.Direction, normal);

        if (n <= 0.0)
        {
            localnormal = normal;
            n = -n;
        }
        else
        {
            localnormal = -normal;
        }


        // TODO FIXME: also for first radiosity pass ? (see line 3272 of v3.6 lighting.cpp)
        if (!haveDispersion) // TODO FIXME - radiosity: || (!isFinalTrace)
        {
            totalReflection = TraceRefractionRay(finish, ipoint, ray, nray, ior, n, normal,
                    rawnormal, localnormal, colour, transm, weight);
        }
        else if (ray.IsMonochromaticRay())
        {
            totalReflection = TraceRefractionRay(finish, ipoint, ray, nray,
                    ray.GetSpectralBand().GetDispersionIOR(ior,dispersion), n, normal,
                    rawnormal, localnormal, colour, transm, weight);
        }
        else
        {
            colour.Clear();
            transm = 0.0;

            for (unsigned int i = 0; i < dispersionelements; i++)
            {
                MathColour tempColour;
                ColourChannel tempTransm;

                // NB setting the dispersion factor also causes the MonochromaticRay flag to be set
                SpectralBand spectralBand(i, dispersionelements);
                nray.SetSpectralBand(spectralBand);

                (void)TraceRefractionRay(finish, ipoint, ray, nray, spectralBand.GetDispersionIOR(ior, dispersion), n,
                        normal, rawnormal, localnormal, tempColour, tempTransm, weight);

                colour += tempColour * spectralBand.GetHue();
                transm += tempTransm;
            }

            colour /= ColourChannel(dispersionelements);
            transm /= ColourChannel(dispersionelements);
        }
    }

    return totalReflection;
}

bool Trace::TraceRefractionRay(const FINISH* finish, const Vector3d& ipoint, Ray& ray, Ray& nray,
        double ior, double n, const Vector3d& normal, const Vector3d& rawnormal,
        const Vector3d& localnormal, MathColour& colour, ColourChannel& transm, COLC weight)
{
    // Compute refrated ray direction using Heckbert's method.
    double t = 1.0 + Sqr(ior) * (Sqr(n) - 1.0);

    if (t < 0.0)
    {
        MathColour tempcolour;

        // Total internal reflection occures.
        threadData->Stats()[Internal_Reflected_Rays_Traced]++;
        ComputeReflection(finish, ipoint, ray, normal, rawnormal, tempcolour, weight);
        colour += tempcolour;

        return true;
    }

    t = ior * n - sqrt(t);

    nray.Direction = ior * ray.Direction + t * localnormal;

    // Trace a refracted ray.
    threadData->Stats()[Refracted_Rays_Traced]++;

    colour.Clear();
    transm = 0.0;
    TraceRay(nray, colour, transm, weight, false);

    return false;
}

// see Diffuse in the v3.6 code (lighting.cpp)
void Trace::ComputeDiffuseLight(const FINISH *finish, const Vector3d& ipoint, const Ray& eye,
        const Vector3d& layer_normal, const MathColour& layer_pigment_colour, MathColour& colour,
        double attenuation, ObjectPtr object, double relativeIor)
{
    Vector3d reye(0.0);

    // TODO FIXME - [CLi] why is this computed here? Not so exciting, is it?
    if (finish->Specular != 0.0)
        reye = -eye.Direction;

    // global light sources, if not turned off for this object
    if ((object->Flags & NO_GLOBAL_LIGHTS_FLAG) != NO_GLOBAL_LIGHTS_FLAG)
    {
        for (auto i = 0ul; i < threadData->lightSources.size(); i++)
        {
            ComputeOneDiffuseLight(*threadData->lightSources[i], reye, finish, ipoint, eye, layer_normal,
                    layer_pigment_colour, colour, attenuation, object, relativeIor, i);
        }
    }

    // local light sources from a light group, if any
    if (!object->LLights.empty())
    {
        for (auto i = 0ul; i < object->LLights.size(); i++)
        {
            ComputeOneDiffuseLight(*object->LLights[i], reye, finish, ipoint, eye, layer_normal, layer_pigment_colour,
                    colour, attenuation, object, relativeIor);
        }
    }
}

void Trace::ComputePhotonDiffuseLight(const FINISH *Finish, const Vector3d& IPoint, const Ray& Eye,
        const Vector3d& Layer_Normal, const Vector3d& Raw_Normal, const MathColour& Layer_Pigment_Colour,
        MathColour& colour, double Attenuation, ConstObjectPtr Object, double relativeIor,
        PhotonGatherer& gatherer)
{
    double Cos_Shadow_Angle;
    Vector3d lightDirection(0.0);
    MathColour Light_Colour;
    MathColour tmpCol, tmpCol2;
    double r;
    int n;
    int j;
    double thisDensity=0;
    double prevDensity=gkDBL_epsilon; // avoid div-by-zero error
    int expanded = false;
    double att;  // attenuation for lambertian compensation & filters

    if (!sceneData->photonSettings.photonsEnabled || sceneData->surfacePhotonMap.numPhotons<1)
        return;

    if ((Finish->Diffuse == 0.0) && (Finish->DiffuseBack == 0.0) && (Finish->Specular == 0.0) && (Finish->Phong == 0.0))
        return;

    // statistics
    threadData->Stats()[Gather_Performed_Count]++;

    if (gatherer.gathered)
        r = gatherer.alreadyGatheredRadius;
    else
        r = gatherer.gatherPhotonsAdaptive(&IPoint, &Layer_Normal, true);

    n = gatherer.gatheredPhotons.numFound;

    tmpCol.Clear();

    // now go through these photons and add up their contribution
    for (j=0; j<n; j++)
    {
        // double theta,phi;
        int theta,phi;
        bool backside = false;

        // convert small color to normal color
        Light_Colour = ToMathColour(RGBColour(gatherer.gatheredPhotons.photonGatherList[j]->colour));

        // convert theta/phi to vector direction
        // Use a pre-computed array of sin/cos to avoid many calls to the
        // sin() and cos() functions.  These arrays were initialized in
        // InitBacktraceEverything.
        theta = gatherer.gatheredPhotons.photonGatherList[j]->theta+127;
        phi = gatherer.gatheredPhotons.photonGatherList[j]->phi+127;

        lightDirection[Y] = sinCosData.sinTheta[theta];
        lightDirection[X] = sinCosData.cosTheta[theta];

        lightDirection[Z] = lightDirection[X]*sinCosData.sinTheta[phi];
        lightDirection[X] = lightDirection[X]*sinCosData.cosTheta[phi];

        // this compensates for real lambertian (diffuse) lighting (see paper)
        // use raw normal, not layer normal
        // att = dot(Layer_Normal, Light_Source_Ray.Direction);
        att = dot(Raw_Normal, lightDirection);
        if (att>1) att=1.0;
        if (att<.1) att = 0.1; // limit to 10x - otherwise we get bright dots
        att = 1.0 / fabs(att);

        // do gaussian filter
        //att *= 0.918*(1.0-(1.0-exp((-1.953) * gatherer.photonDistances[j])) / (1.0-exp(-1.953)) );
        // do cone filter
        //att *= 1.0-(sqrt(gatherer.photonDistances[j])/(4.0 * r)) / (1.0-2.0/(3.0*4.0));

        Light_Colour *= att;

        // See if light on far side of surface from camera.
        if (!(Test_Flag(Object, DOUBLE_ILLUMINATE_FLAG)))
        {
            Cos_Shadow_Angle = dot(Layer_Normal, lightDirection);
            if (Cos_Shadow_Angle < gkDBL_epsilon)
            {
                if (Finish->DiffuseBack != 0.0)
                    backside = true;
                else
                    continue;
            }
        }

        // now add diffuse, phong, specular, irid contribution

        tmpCol2.Clear();

        if (!(sceneData->useSubsurface && Finish->UseSubsurface))
        {
            // (Diffuse contribution is not supported in combination with BSSRDF, to emphasize the fact that the BSSRDF
            // model is intended to provide for all the diffuse term by default. If users want to add some additional
            // surface-only diffuse term, they should use layered textures.
            ComputeDiffuseColour(Finish, lightDirection, Eye.Direction, Layer_Normal, tmpCol2, Light_Colour,
                    Layer_Pigment_Colour, relativeIor, Attenuation, backside);
        }

        // NK rad - don't compute highlights for radiosity gather rays, since this causes
        // problems with colors being far too bright
        // don't compute highlights for diffuse backside illumination
        if (!Eye.IsRadiosityRay() && !backside) // TODO FIXME radiosity - is this really the right way to do it (speaking of realism)?
        {
            if (Finish->Phong != 0.0)
            {
                ComputePhongColour(Finish, lightDirection, Eye.Direction, Layer_Normal, tmpCol2, Light_Colour,
                        Layer_Pigment_Colour, relativeIor);
            }
            if (Finish->Specular != 0.0)
            {
                ComputeSpecularColour(Finish, lightDirection, -Eye.Direction, Layer_Normal, tmpCol2, Light_Colour,
                        Layer_Pigment_Colour, relativeIor);
            }
        }

        if (Finish->Irid > 0.0)
        {
            ComputeIridColour(Finish, lightDirection, Eye.Direction, Layer_Normal, IPoint, tmpCol2);
        }

        tmpCol += tmpCol2;
    }

    // finish the photons equation
    tmpCol /= M_PI*r*r;

    // add photon contribution to total lighting
    colour += tmpCol;
}

// see Diffuse_One_Light in the v3.6 code (lighting.cpp)
void Trace::ComputeOneDiffuseLight(const LightSource &lightsource, const Vector3d& reye, const FINISH *finish,
        const Vector3d& ipoint, const Ray& eye, const Vector3d& layer_normal,
        const MathColour& layer_pigment_colour, MathColour& colour, double attenuation, ConstObjectPtr object,
        double relativeIor, int light_index)
{
    double lightsourcedepth, cos_shadow_angle;
    Ray lightsourceray(eye);
    MathColour lightcolour;
    bool backside = false;
    MathColour tmpCol;

    // Get a colour and a ray.
    ComputeOneLightRay(lightsource, lightsourcedepth, lightsourceray, ipoint, lightcolour);

    // Don't calculate spotlights when outside of the light's cone.
    if (lightcolour.IsNearZero(gkDBL_epsilon))
        return;

    // See if light on far side of surface from camera.
    if (!(Test_Flag(object, DOUBLE_ILLUMINATE_FLAG)) &&
        !lightsource.Use_Full_Area_Lighting) // JN2007: Easiest way of getting rid of sharp shadow lines
    {
        cos_shadow_angle = dot(layer_normal, lightsourceray.Direction);
        if (cos_shadow_angle < gkDBL_epsilon)
        {
            if (finish->DiffuseBack != 0.0)
                backside = true;
            else
                return;
        }
    }

    // If light source was not blocked by any intervening object, then
    // calculate it's contribution to the object's overall illumination.
    if (qualityFlags.shadows && ((lightsource.Projected_Through_Object != nullptr) || (lightsource.Light_Type != FILL_LIGHT_SOURCE)))
    {
        if (lightColorCacheIndex != -1 && light_index != -1)
        {
            if (lightColorCache[lightColorCacheIndex][light_index].tested == false)
            {
                // note that lightColorCache may be re-sized during trace, so we don't store a reference to it across the call
                TraceShadowRay(lightsource, lightsourcedepth, lightsourceray, ipoint, lightcolour);
                lightColorCache[lightColorCacheIndex][light_index].tested = true;
                lightColorCache[lightColorCacheIndex][light_index].colour = lightcolour;
            }
            else
                lightcolour = lightColorCache[lightColorCacheIndex][light_index].colour;
        }
        else
            TraceShadowRay(lightsource, lightsourcedepth, lightsourceray, ipoint, lightcolour);
    }

                                           // (Note there is already ONE filter like this up top (which potentially an issue too)
    if (!lightcolour.IsNearZero(gkDBL_epsilon))  // TODO The hard gkDBL_epsilon value suspect given HDR. Test not cheap - dump it or ?
    {
        if (lightsource.Use_Full_Area_Lighting && lightsource.Area_Light &&
            qualityFlags.areaLights) // JN2007: Full area lighting
        {
            ComputeFullAreaDiffuseLight(lightsource, reye, finish, ipoint, eye,
                layer_normal, layer_pigment_colour, colour, attenuation,
                lightsourcedepth, lightsourceray, lightcolour,
                object, relativeIor);
            return;
        }

        if (!(sceneData->useSubsurface && finish->UseSubsurface))
        {
            // (Diffuse contribution is not supported in combination with BSSRDF, to emphasize the fact that the BSSRDF
            // model is intended to provide for all the diffuse term by default. If users want to add some additional
            // surface-only diffuse term, they should use layered textures.
            ComputeDiffuseColour(finish, lightsourceray.Direction, eye.Direction, layer_normal, tmpCol, lightcolour,
                    layer_pigment_colour, relativeIor, attenuation, backside);
        }

        MathColour tempLightColour = (finish->AlphaKnockout ? lightcolour * attenuation : lightcolour);

        // NK rad - don't compute highlights for radiosity gather rays, since this causes
        // problems with colors being far too bright
        // don't compute highlights for diffuse backside illumination
        if ((lightsource.Light_Type != FILL_LIGHT_SOURCE) && !eye.IsRadiosityRay() && !backside) // TODO FIXME radiosity - is this really the right way to do it (speaking of realism)?
        {
            if (finish->Phong != 0.0)
            {
                ComputePhongColour(finish, lightsourceray.Direction, eye.Direction, layer_normal, tmpCol,
                        tempLightColour, layer_pigment_colour, relativeIor);
            }

            if (finish->Specular != 0.0)
            {
                ComputeSpecularColour(finish, lightsourceray.Direction, -eye.Direction, layer_normal, tmpCol,
                       tempLightColour, layer_pigment_colour, relativeIor);
            }
        }

        if (finish->Irid > 0.0)
            ComputeIridColour(finish, lightsourceray.Direction, eye.Direction, layer_normal, ipoint, tmpCol);
    }

    colour += tmpCol;
}

// JN2007: Full area lighting:
void Trace::ComputeFullAreaDiffuseLight(const LightSource &lightsource, const Vector3d& reye, const FINISH *finish,
        const Vector3d& ipoint, const Ray& eye, const Vector3d& layer_normal, const MathColour& layer_pigment_colour,
        MathColour& colour, double attenuation, double lightsourcedepth, Ray& lightsourceray,
        const MathColour& lightcolour, ConstObjectPtr object, double relativeIor)
{
    Vector3d temp;
    Vector3d axis1Temp, axis2Temp;
    double axis1_Length, cos_shadow_angle;

    axis1Temp = lightsource.Axis1;
    axis2Temp = lightsource.Axis2;

    if (lightsource.Orient == true)
    {
        // Orient the area light to face the intersection point [ENB 9/97]

        // Do Light source to get the correct lightsourceray
        ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lightsourceray, ipoint);

        // Save the lengths of the axises
        axis1_Length = axis1Temp.length();

        // Make axis 1 be perpendicular with the light-ray
        if (fabs(fabs(lightsourceray.Direction[Z]) - 1.0) < 0.01)
            // too close to vertical for comfort, so use cross product with horizon
            temp = Vector3d(0.0, 1.0, 0.0);
        else
            temp = Vector3d(0.0, 0.0, 0.1);

        axis1Temp = cross(lightsourceray.Direction, temp).normalized();

        // Make axis 2 be perpendicular with the light-ray and with Axis1.  A simple cross-product will do the trick.
        axis2Temp = cross(lightsourceray.Direction, axis1Temp).normalized();

        // make it square
        axis1Temp *= axis1_Length;
        axis2Temp *= axis1_Length;
    }

    MathColour sampleLightcolour = lightcolour / (lightsource.Area_Size1 * lightsource.Area_Size2);
    MathColour attenuatedLightcolour;

    for (int v = 0; v < lightsource.Area_Size2; ++v)
    {
        for (int u = 0; u < lightsource.Area_Size1; ++u)
        {
            Vector3d jitterAxis1, jitterAxis2;
            Ray lsr(lightsourceray);
            double jitter_u = (double)u;
            double jitter_v = (double)v;
            bool backside = false;
            MathColour tmpCol;

            if (lightsource.Jitter)
            {
                jitter_u += randomNumberGenerator() - 0.5;
                jitter_v += randomNumberGenerator() - 0.5;
            }

            // Create circular are lights [ENB 9/97]
            // First, make jitter_u and jitter_v be numbers from -1 to 1
            // Second, set scaleFactor to the abs max (jitter_u,jitter_v) (for shells)
            // Third, divide scaleFactor by the length of <jitter_u,jitter_v>
            // Fourth, scale jitter_u & jitter_v by scaleFactor
            // Finally scale Axis1 by jitter_u & Axis2 by jitter_v
            if (lightsource.Circular == true)
            {
                jitter_u = jitter_u / (lightsource.Area_Size1 - 1) - 0.5 + 0.001;
                jitter_v = jitter_v / (lightsource.Area_Size2 - 1) - 0.5 + 0.001;
                double scaleFactor = ((fabs(jitter_u) > fabs(jitter_v)) ? fabs(jitter_u) : fabs(jitter_v));
                scaleFactor /= sqrt(jitter_u * jitter_u + jitter_v * jitter_v);
                jitter_u *= scaleFactor;
                jitter_v *= scaleFactor;
                jitterAxis1 = axis1Temp * jitter_u;
                jitterAxis2 = axis2Temp * jitter_v;
            }
            else
            {
                if (lightsource.Area_Size1 > 1)
                {
                    double scaleFactor = jitter_u / (double)(lightsource.Area_Size1 - 1) - 0.5;
                    jitterAxis1 = axis1Temp * scaleFactor;
                }
                else
                    jitterAxis1 = Vector3d(0.0, 0.0, 0.0);

                if (lightsource.Area_Size2 > 1)
                {
                    double scaleFactor = jitter_v / (double)(lightsource.Area_Size2 - 1) - 0.5;
                    jitterAxis2 = axis2Temp * scaleFactor;
                }
                else
                    jitterAxis2 = Vector3d(0.0, 0.0, 0.0);
            }

            // Recalculate the light source ray but not the colour
            ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lsr, ipoint, jitterAxis1 + jitterAxis2);
            // Calculate distance- and angle-based light attenuation
            attenuatedLightcolour = sampleLightcolour * Attenuate_Light(&lightsource, lsr, lightsourcedepth);

            // If not double-illuminated, check if the normal is pointing away:
            if (!Test_Flag(object, DOUBLE_ILLUMINATE_FLAG))
            {
                cos_shadow_angle = dot(layer_normal, lsr.Direction);
                if (cos_shadow_angle < gkDBL_epsilon)
                {
                    if (finish->DiffuseBack != 0.0)
                        backside = true;
                    else
                        continue;
                }
            }

            if (!(sceneData->useSubsurface && finish->UseSubsurface))
            {
                // (Diffuse contribution is not supported in combination with BSSRDF, to emphasize the fact that the BSSRDF
                // model is intended to provide for all the diffuse term by default. If users want to add some additional
                // surface-only diffuse term, they should use layered textures.
                ComputeDiffuseColour(finish, lsr.Direction, eye.Direction, layer_normal, tmpCol, attenuatedLightcolour,
                        layer_pigment_colour, relativeIor, attenuation, backside);
            }

            // NK rad - don't compute highlights for radiosity gather rays, since this causes
            // problems with colors being far too bright
            // don't compute highlights for diffuse backside illumination
            if ((lightsource.Light_Type != FILL_LIGHT_SOURCE) && !eye.IsRadiosityRay() && !backside) // TODO FIXME radiosity - is this really the right way to do it (speaking of realism)?
            {
                if (finish->Phong != 0.0)
                {
                    ComputePhongColour(finish, lsr.Direction, eye.Direction, layer_normal, tmpCol, attenuatedLightcolour,
                            layer_pigment_colour, relativeIor);
                }

                if (finish->Specular != 0.0)
                {
                    ComputeSpecularColour(finish, lsr.Direction, -eye.Direction, layer_normal, tmpCol, attenuatedLightcolour,
                            layer_pigment_colour, relativeIor);
                }
            }

            if (finish->Irid > 0.0)
                ComputeIridColour(finish, lsr.Direction, eye.Direction, layer_normal, ipoint, tmpCol);

            colour += tmpCol;
        }
    }
}

// see do_light in v3.6's lighting.cpp
void Trace::ComputeOneLightRay(const LightSource &lightsource, double& lightsourcedepth, Ray& lightsourceray,
        const Vector3d& ipoint, MathColour& lightcolour, bool forceAttenuate)
{
    double attenuation;
    ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lightsourceray, ipoint);

    // Get the light source colour.
    lightcolour = lightsource.colour;

    // Attenuate light source color.
    if (lightsource.Area_Light && lightsource.Use_Full_Area_Lighting && qualityFlags.areaLights && !forceAttenuate)
        // for full area lighting we apply distance- and angle-based attenuation to each "lightlet" individually later
        attenuation = 1.0;
    else
        attenuation = Attenuate_Light(&lightsource, lightsourceray, lightsourcedepth);

    // Now scale the color by the attenuation
    lightcolour *= attenuation;
}

// see block_light_source in the v3.6 source
void Trace::TraceShadowRay(const LightSource &lightsource, double depth, Ray& lightsourceray, const Vector3d& point,
        MathColour& colour)
{
    // test and set highest level traced. We do it differently than TraceRay() does,
    // for compatibility with the way max_trace_level is tested and reported in v3.6
    // and earlier.
    if (lightsourceray.GetTicket().traceLevel > lightsourceray.GetTicket().maxAllowedTraceLevel)
    {
        colour.Clear();
        return;
    }

    lightsourceray.GetTicket().maxFoundTraceLevel = (unsigned int) max(lightsourceray.GetTicket().maxFoundTraceLevel,
            lightsourceray.GetTicket().traceLevel);
    lightsourceray.GetTicket().traceLevel++;

    double newdepth;
    Intersection isect;
    Ray newray(lightsourceray);

    // Store current depth and ray because they will be modified.
    newdepth = depth;

    // NOTE: shadow rays are never photon rays, so flag can be hard-coded to false
    newray.SetFlags(Ray::OtherRay, true, false);

    // Get shadows from current light source.
    if (lightsource.Area_Light && qualityFlags.areaLights)
        TraceAreaLightShadowRay(lightsource, newdepth, newray, point, colour);
    else
        TracePointLightShadowRay(lightsource, newdepth, newray, colour);

    // If there's some distance left for the ray to reach the light source
    // we have to apply atmospheric stuff to this part of the ray.

    if ((newdepth > sceneData->shadowTolerance) && (lightsource.Media_Interaction) && (lightsource.Media_Attenuation))
    {
        isect.Depth = newdepth;
        isect.Object = nullptr;
        ComputeShadowMedia(newray, isect, colour, (lightsource.Media_Interaction) && (lightsource.Media_Attenuation));
    }

    lightsourceray.GetTicket().traceLevel--;
    maxFoundTraceLevel = (unsigned int) max(maxFoundTraceLevel, lightsourceray.GetTicket().maxFoundTraceLevel);
}

// moved this here (was originally inside TracePointLightShadowRay) because
// for some reason the Intel compiler (version W_CC_PC_8.1.027) will fail
// to link the exe, complaining of an unresolved external.
//
// TODO: try moving it back in at some point in the future.
struct NoShadowFlagRayObjectCondition final : public RayObjectCondition
{
    virtual bool operator()(const Ray&, ConstObjectPtr object, double) const override { return !Test_Flag(object, NO_SHADOW_FLAG); }
};

//   Replacing with version below so I can pass in initial scene shadow tolerance... (Performance hit?)
//  struct SmallToleranceRayObjectCondition final  : public RayObjectCondition
//  {
//      virtual bool operator()(const Ray&, ConstObjectPtr, double dist) const override { return dist > 1e-3; }  // 1e-3 was SHADOW_TOLERANCE...
//  };

class SmallToleranceRayObjectCondition final : public RayObjectCondition
{
    public:
        SmallToleranceRayObjectCondition (double tolerance) : _tolerance( tolerance ) {}
        bool operator() (const Ray&, ConstObjectPtr, double dist) const override { return dist > _tolerance; }
    private:
        double _tolerance;
};


void Trace::TracePointLightShadowRay(const LightSource &lightsource, double& lightsourcedepth, Ray& lightsourceray,
        MathColour& lightcolour)
{
    Intersection boundedIntersection;
    ObjectPtr cacheObject = nullptr;
    bool foundTransparentObjects = false;
    bool foundIntersection;

    // Projected through main tests
    double projectedDepth = 0.0;

    if (lightsource.Projected_Through_Object != nullptr)
    {
        Intersection tempIntersection;

        if (FindIntersection(lightsource.Projected_Through_Object, tempIntersection, lightsourceray))
        {
            if ((tempIntersection.Depth - lightsourcedepth) < 0.0)
            {
                projectedDepth = lightsourcedepth - fabs(tempIntersection.Depth) + sceneData->shadowTolerance;
            }
            else
            {
                lightcolour.Clear();
                return;
            }
        }
        else
        {
            lightcolour.Clear();
            return;
        }

        // Make sure we don't do shadows for fill light sources.
        // (Note that if Projected_Through_Object is `nullptr`, the test for FILL_LIGHT_SOURCE happened earlier)
        if (lightsource.Light_Type == FILL_LIGHT_SOURCE)
            return;
    }

    NoShadowFlagRayObjectCondition precond;
//  SmallToleranceRayObjectCondition postcond;
    SmallToleranceRayObjectCondition postcond(sceneData->shadowTolerance);

    // check for object in the light source shadow cache (object that fully shadowed during last test) first

    if (lightsource.lightGroupLight == false) // we don't cache for light groups
    {
        if ((lightsourceray.GetTicket().traceLevel == 2) && (lightSourceLevel1ShadowCache[lightsource.index] != nullptr))
            cacheObject = lightSourceLevel1ShadowCache[lightsource.index];
        else if (lightSourceOtherShadowCache[lightsource.index] != nullptr)
            cacheObject = lightSourceOtherShadowCache[lightsource.index];

        // if there was an object in the light source shadow cache, check that first
        if (cacheObject != nullptr)
        {
            if (FindIntersection(cacheObject, boundedIntersection, lightsourceray,
                postcond, lightsourcedepth - projectedDepth) == true)
            {
                if (!Test_Flag(boundedIntersection.Object, NO_SHADOW_FLAG))
                {
                    ComputeShadowColour(lightsource, boundedIntersection, lightsourceray, lightcolour);

                    if (lightcolour.IsNearZero(gkDBL_epsilon) &&
                       (Test_Flag(boundedIntersection.Object, OPAQUE_FLAG)))
                    {
                        threadData->Stats()[Shadow_Ray_Tests]++;
                        threadData->Stats()[Shadow_Rays_Succeeded]++;
                        threadData->Stats()[Shadow_Cache_Hits]++;
                        return;
                    }
                }
                else
                    cacheObject = nullptr;
            }
            else
            {
                if (lightsourceray.GetTicket().traceLevel == 2)
                {
                    lightSourceLevel1ShadowCache[lightsource.index] = nullptr;
                }
                else
                {
                    lightSourceOtherShadowCache[lightsource.index] = nullptr;
                }
            }
        }
    }

    foundTransparentObjects = false;

    while(true)
    {
        boundedIntersection.Object = boundedIntersection.Csg = nullptr;
        boundedIntersection.Depth = lightsourcedepth - projectedDepth;

        threadData->Stats()[Shadow_Ray_Tests]++;

        foundIntersection = FindIntersection(boundedIntersection, lightsourceray, precond, postcond);

        if ((foundIntersection == true) && (boundedIntersection.Object != cacheObject) &&
           (boundedIntersection.Depth < lightsourcedepth - sceneData->shadowTolerance) &&
           (lightsourcedepth - boundedIntersection.Depth > projectedDepth) &&
           (boundedIntersection.Depth > sceneData->shadowTolerance))
        {
            threadData->Stats()[Shadow_Rays_Succeeded]++;

            ComputeShadowColour(lightsource, boundedIntersection, lightsourceray, lightcolour);

            ObjectPtr testObject(boundedIntersection.Csg != nullptr ? boundedIntersection.Csg : boundedIntersection.Object);

            if (lightcolour.IsNearZero(gkDBL_epsilon) &&
               (Test_Flag(testObject, OPAQUE_FLAG)))
            {
                // Hit a fully opaque object; cache that object, so that next time we can test for it first;
                // don't cache for light groups though (why not?)

                if ((lightsource.lightGroupLight == false) && (foundTransparentObjects == false))
                {
                    cacheObject = testObject;

                    if (lightsourceray.GetTicket().traceLevel == 2)
                        lightSourceLevel1ShadowCache[lightsource.index] = cacheObject;
                    else
                        lightSourceOtherShadowCache[lightsource.index] = cacheObject;
                }
                break;
            }

            foundTransparentObjects = true;

            // Move the ray to the point of intersection, plus some
            lightsourcedepth -= boundedIntersection.Depth;

            lightsourceray.Origin = boundedIntersection.IPoint;
        }
        else
            // No further intersections in the direction of the ray.
            break;
    }
}

void Trace::TraceAreaLightShadowRay(const LightSource &lightsource, double& lightsourcedepth,
        Ray& lightsourceray, const Vector3d& ipoint, MathColour& lightcolour)
{
    Vector3d temp;
    Vector3d axis1Temp, axis2Temp;
    double axis1_Length;

    lightGrid.resize(lightsource.Area_Size1 * lightsource.Area_Size2);

    // Flag uncalculated points with a negative value for Red
    /*
    for (i = 0; i < lightsource.Area_Size1; i++)
    {
        for (j = 0; j < lightsource.Area_Size2; j++)
            lightGrid[i * lightsource.Area_Size2 + j].Invalidate();
    }
    */
    for (size_t ind = 0; ind < lightGrid.size(); ++ind)
        lightGrid[ind].Invalidate();

    axis1Temp = lightsource.Axis1;
    axis2Temp = lightsource.Axis2;

    if (lightsource.Orient == true)
    {
        // Orient the area light to face the intersection point [ENB 9/97]

        // Do Light source to get the correct lightsourceray
        ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lightsourceray, ipoint);

        // Save the lengths of the axes
        axis1_Length = axis1Temp.length();

        // Make axis 1 be perpendicular with the light-ray
        if (fabs(fabs(lightsourceray.Direction[Z]) - 1.0) < 0.01)
            // too close to vertical for comfort, so use cross product with horizon
            temp = Vector3d(0.0, 1.0, 0.0);
        else
            temp = Vector3d(0.0, 0.0, 1.0);

        axis1Temp = cross(lightsourceray.Direction, temp).normalized();

        // Make axis 2 be perpendicular with the light-ray and with Axis1.  A simple cross-product will do the trick.
        axis2Temp = cross(lightsourceray.Direction, axis1Temp).normalized();

        // make it square
        axis1Temp *= axis1_Length;
        axis2Temp *= axis1_Length;
    }

    TraceAreaLightSubsetShadowRay(lightsource, lightsourcedepth, lightsourceray, ipoint, lightcolour, 0, 0,
            lightsource.Area_Size1 - 1, lightsource.Area_Size2 - 1, 0, axis1Temp, axis2Temp);
}

void Trace::TraceAreaLightSubsetShadowRay(const LightSource &lightsource, double& lightsourcedepth,
        Ray& lightsourceray, const Vector3d& ipoint, MathColour& lightcolour,
        int u1, int  v1, int  u2, int  v2, int level, const Vector3d& axis1, const Vector3d& axis2)
{
    MathColour sample_Colour[4];
    int i, u, v, new_u1, new_v1, new_u2, new_v2;
    double jitter_u, jitter_v, scaleFactor;

    // Sample the four corners of the region
    for (i = 0; i < 4; i++)
    {
        Vector3d center(lightsource.Center);
        Ray lsr(lightsourceray);

        switch (i)
        {
            case 0: u = u1; v = v1; break;
            case 1: u = u2; v = v1; break;
            case 2: u = u1; v = v2; break;
            case 3: u = u2; v = v2; break;
            default: u = v = 0;  // Should never happen!
        }

        if (lightGrid[u * lightsource.Area_Size2 + v].IsValid())
            // We've already calculated this point, reuse it
            sample_Colour[i] = lightGrid[u * lightsource.Area_Size2 + v];
        else
        {
            Vector3d jitterAxis1, jitterAxis2;

            jitter_u = (double)u;
            jitter_v = (double)v;

            if (lightsource.Jitter)
            {
                jitter_u += randomNumberGenerator() - 0.5;
                jitter_v += randomNumberGenerator() - 0.5;
            }

            // Create circular are lights [ENB 9/97]
            // First, make jitter_u and jitter_v be numbers from -1 to 1
            // Second, set scaleFactor to the abs max (jitter_u,jitter_v) (for shells)
            // Third, divide scaleFactor by the length of <jitter_u,jitter_v>
            // Fourth, scale jitter_u & jitter_v by scaleFactor
            // Finally scale Axis1 by jitter_u & Axis2 by jitter_v
            if (lightsource.Circular == true)
            {
                jitter_u = jitter_u / (lightsource.Area_Size1 - 1) - 0.5 + 0.001;
                jitter_v = jitter_v / (lightsource.Area_Size2 - 1) - 0.5 + 0.001;
                scaleFactor = ((fabs(jitter_u) > fabs(jitter_v)) ? fabs(jitter_u) : fabs(jitter_v));
                scaleFactor /= sqrt(jitter_u * jitter_u + jitter_v * jitter_v);
                jitter_u *= scaleFactor;
                jitter_v *= scaleFactor;
                jitterAxis1 = axis1 * jitter_u;
                jitterAxis2 = axis2 * jitter_v;
            }
            else
            {
                if (lightsource.Area_Size1 > 1)
                {
                    scaleFactor = jitter_u / (double)(lightsource.Area_Size1 - 1) - 0.5;
                    jitterAxis1 = axis1 * scaleFactor;
                }
                else
                    jitterAxis1 = Vector3d(0.0, 0.0, 0.0);

                if (lightsource.Area_Size2 > 1)
                {
                    scaleFactor = jitter_v / (double)(lightsource.Area_Size2 - 1) - 0.5;
                    jitterAxis2 = axis2 * scaleFactor;
                }
                else
                    jitterAxis2 = Vector3d(0.0, 0.0, 0.0);
            }

            // Recalculate the light source ray but not the colour
            ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lsr, ipoint, jitterAxis1 + jitterAxis2);

            sample_Colour[i] = lightcolour;

            TracePointLightShadowRay(lightsource, lightsourcedepth, lsr, sample_Colour[i]);

            lightGrid[u * lightsource.Area_Size2 + v] = sample_Colour[i];
        }
    }

    if ((u2 - u1 > 1) || (v2 - v1 > 1))
    {
        if ((level < lightsource.Adaptive_Level) ||
           (ColourDistance(sample_Colour[0], sample_Colour[1]) > 0.1) ||
           (ColourDistance(sample_Colour[1], sample_Colour[3]) > 0.1) ||
           (ColourDistance(sample_Colour[3], sample_Colour[2]) > 0.1) ||
           (ColourDistance(sample_Colour[2], sample_Colour[0]) > 0.1))
        {
            Vector3d center(lightsource.Center);

            for (i = 0; i < 4; i++)
            {
                switch (i)
                {
                    case 0:
                        new_u1 = u1;
                        new_v1 = v1;
                        new_u2 = (int)floor((u1 + u2)/2.0);
                        new_v2 = (int)floor((v1 + v2)/2.0);
                        break;
                    case 1:
                        new_u1 = (int)ceil((u1 + u2)/2.0);
                        new_v1 = v1;
                        new_u2 = u2;
                        new_v2 = (int)floor((v1 + v2)/2.0);
                        break;
                    case 2:
                        new_u1 = u1;
                        new_v1 = (int)ceil((v1 + v2)/2.0);
                        new_u2 = (int)floor((u1 + u2)/2.0);
                        new_v2 = v2;
                        break;
                    case 3:
                        new_u1 = (int)ceil((u1 + u2)/2.0);
                        new_v1 = (int)ceil((v1 + v2)/2.0);
                        new_u2 = u2;
                        new_v2 = v2;
                        break;
                    default:  // Should never happen!
                        new_u1 = new_u2 = new_v1 = new_v2 = 0;
                }

                // Recalculate the light source ray but not the colour
                ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lightsourceray, ipoint, center);

                sample_Colour[i] = lightcolour;

                TraceAreaLightSubsetShadowRay(lightsource, lightsourcedepth, lightsourceray,
                                              ipoint, sample_Colour[i], new_u1, new_v1, new_u2, new_v2, level + 1, axis1, axis2);
            }
        }
    }

    // Average up the light contributions
    lightcolour = (sample_Colour[0] + sample_Colour[1] + sample_Colour[2] + sample_Colour[3]) * 0.25;
}

// see filter_shadow_ray in v3.6's lighting.cpp
void Trace::ComputeShadowColour(const LightSource &lightsource, Intersection& isect, Ray& lightsourceray,
        MathColour& colour)
{
    WeightedTextureVector wtextures;
    Vector3d ipoint;
    Vector3d raw_Normal;
    MathColour fc1;
    ColourChannel ft1;
    MathColour temp_Colour;
    Vector2d uv_Coords;
    double normaldirection;

    // Here's the issue:
    // Imagine "LightA" shoots photons at "GlassSphereB", which refracts light and
    // hits "PlaneC".
    // When computing Diffuse/Phong/etc lighting for PlaneC, if there were no
    // photons, POV would compute a filtered shadow ray from PlaneC through
    // GlassSphereB to LightA.  If photons are used for the combination of objects,
    // this filtered shadow ray should be completely black.  The filtered shadow
    // ray should be forced to black UNLESS any of the following conditions are
    // true (which would indicate that photons were not shot from LightA through
    // GlassSphereB to PlaneC):
    // 1) PlaneC has photon collection set to "off"
    // 2) GlassSphereB is not a photon target
    // 3) GlassSphereB has photon refraction set to "off"
    // 4) LightA has photon refraction set to "off"
    // 5) Neither GlassSphereB nor LightA has photon refraction set to "on"
    if ((sceneData->photonSettings.photonsEnabled == true) &&
        (sceneData->surfacePhotonMap.numPhotons > 0) &&
        (!threadData->litObjectIgnoresPhotons) &&
        (Test_Flag(isect.Object,PH_TARGET_FLAG)) &&
        (!Test_Flag(isect.Object,PH_RFR_OFF_FLAG)) &&
        (!Test_Flag(&lightsource,PH_RFR_OFF_FLAG)) &&
        ((Test_Flag(isect.Object,PH_RFR_ON_FLAG) || Test_Flag(&lightsource,PH_RFR_ON_FLAG)))
        )
    {
        // full shadow (except for photon-based illumination)
        colour.Clear();
        return;
    }

    ipoint = isect.IPoint;

    if (!qualityFlags.shadows)
        // no shadow
        return;

    // If the object is opaque there's no need to go any further. [DB 8/94]
    if (Test_Flag(isect.Object, OPAQUE_FLAG))
    {
        // full shadow
        colour.Clear();
        return;
    }

    // Get the normal to the surface
    isect.Object->Normal(raw_Normal, &isect, threadData);

    // I added this to flip the normal if the object is inverted (for CSG).
    // However, I subsequently commented it out for speed reasons - it doesn't
    // make a difference (no pun intended). The preexisting flip code below
    // produces a similar (though more extensive) result. [NK]
    //
    // Actually, we should keep this code to guarantee that normaldirection
    // is set properly. [NK]
    if (Test_Flag(isect.Object, INVERTED_FLAG))
        raw_Normal.invert();

    // If the surface normal points away, flip its direction.
    normaldirection = dot(raw_Normal, lightsourceray.Direction);
    if (normaldirection > 0.0)
        raw_Normal.invert();

    isect.INormal = raw_Normal;
    // and save to intersection -hdf-
    isect.PNormal = raw_Normal;

    // now switch to UV mapping if we need to
    if (Test_Flag(isect.Object, UV_FLAG))
    {
        // TODO FIXME
        //  I think we have a serious problem here regarding bump mapping:
        //  The UV vector doesn't contain any information about the (local) *orientation* of U and V in our XYZ co-ordinate system!
        //  This causes slopes do be applied in the wrong directions.

        // get the UV vect of the intersection
        isect.Object->UVCoord(uv_Coords, &isect);
        // save the normal and UV coords into Intersection
        isect.Iuv = uv_Coords;

        ipoint[X] = uv_Coords[U];
        ipoint[Y] = uv_Coords[V];
        ipoint[Z] = 0;
    }

    // NB the v3.6 code doesn't set the light cache's Tested flags to false after incrementing the level.
    if (++lightColorCacheIndex >= lightColorCache.size())
    {
        lightColorCache.resize(lightColorCacheIndex + 10);
        for (auto it = lightColorCache.begin() + lightColorCacheIndex; it != lightColorCache.end(); it++)
            it->resize(lightColorCache[0].size());
    }

    bool isMultiTextured = Test_Flag(isect.Object, MULTITEXTURE_FLAG) ||
        ((isect.Object->Texture == nullptr) && Test_Flag(isect.Object, CUTAWAY_TEXTURES_FLAG));

    // get textures and weights
    if (isMultiTextured == true)
    {
        isect.Object->Determine_Textures(&isect, normaldirection > 0.0, wtextures, threadData);
    }
    else if (isect.Object->Texture != nullptr)
    {
        if ((normaldirection > 0.0) && (isect.Object->Interior_Texture != nullptr))
            wtextures.push_back(WeightedTexture(1.0, isect.Object->Interior_Texture)); /* Chris Huff: Interior Texture patch */
        else
            wtextures.push_back(WeightedTexture(1.0, isect.Object->Texture));
    }
    else
    {
        // don't need to do anything as the texture list will be empty.
        // TODO: could we perform these tests earlier ? [cjc]
        lightColorCacheIndex--;
        return;
    }

    temp_Colour.Clear();

    for (auto i(wtextures.begin()); i != wtextures.end(); i++)
    {
        TextureVector warps(texturePool);
        POV_REFPOOL_ASSERT(warps->empty()); // verify that the TextureVector pulled from the pool is in a cleaned-up condition

        // If contribution of this texture is negligible skip ahead.
        if ((i->weight < lightsourceray.GetTicket().adcBailout) || (i->texture == nullptr))
            continue;

        ComputeOneTextureColour(fc1, ft1, i->texture, *warps, ipoint, raw_Normal, lightsourceray, 0.0, isect, true, false);

        temp_Colour += i->weight * fc1;
    }

    lightColorCacheIndex--;

    if (fabs(temp_Colour.Weight()) < lightsourceray.GetTicket().adcBailout)
    {
        // close enough to full shadow - bail out to avoid media computations
        colour.Clear();
        return;
    }

    // [CLi] moved this here from Trace::ComputeShadowTexture() and Trace::ComputeLightedTexture(), respectively,
    // to avoid media to be computed twice when dealing with averaged textures.
    // TODO - For photon rays we're still potentially doing double work on media.
    // TODO - For shadow rays we're still potentially doing double work on distance-based attenuation.
    // Calculate participating media effects.
    if (qualityFlags.media && (!lightsourceray.GetInteriors().empty()) && (lightsourceray.IsHollowRay() == true))
    {
        ColourChannel dummyTransm;
        media.ComputeMedia(lightsourceray.GetInteriors(), lightsourceray, isect, temp_Colour, dummyTransm);
    }

    colour *= temp_Colour;

    // Get atmospheric attenuation.
    ComputeShadowMedia(lightsourceray, isect, colour, (lightsource.Media_Interaction) && (lightsource.Media_Attenuation));
}

void Trace::ComputeDiffuseColour(const FINISH *finish, const Vector3d& lightDirection, const Vector3d& eyeDirection,
        const Vector3d& layer_normal, MathColour& colour, const MathColour& light_colour,
        const MathColour& layer_pigment_colour, double relativeIor, double attenuation, bool backside)
{
    double diffuse = (backside ? finish->DiffuseBack : finish->Diffuse) * finish->BrillianceAdjust;
    if (diffuse <= 0.0)
        return;

    // Vector dot product here can range from -1-(3*DBL_EPSILON) to 1+(3*DBL_EPSILON)
    double cos_angle_of_incidence = dot(layer_normal, lightDirection);
    double abs_cos_angle_of_incidence = std::abs(cos_angle_of_incidence);
    double intensity_max = finish->Intensity_Max;
    double intensity;

    if (finish->Brilliance != 1.0) // Brilliance is likely to be 1.0 (default value)
    {                              // It can be anything including <1.0 and negative (~ -1) useful.
        intensity = pow(abs_cos_angle_of_incidence,(double)finish->Brilliance);
        if (intensity_max>0.0)     // Allow users to clamp to a max. Especially useful for negative brilliance
        {                          // values where a setting of <=+1 helps to keep internal values AA friendly.
            double absimax = std::abs(intensity_max);
            intensity = std::min(std::max(absimax,absimax/diffuse),intensity); // If diffuse zero, we returned above.
        }
    }
    else
    {
        intensity = abs_cos_angle_of_incidence;
    }

    intensity *= diffuse * attenuation;

    if (finish->Crand > 0.0)
        intensity -= POV_rand(crandRandomNumberGenerator) * finish->Crand; // TODO Change so thread aware.

    if (finish->Fresnel != 0.0)
    {
        // In diffuse reflections, the Fresnel effect is that we _lose_ the reflected component
        // as the light enters the material (where it changes direction randomly), and then
        // _again lose_ another reflected component as the light leaves the material.
        // ComputeFresnel() gives us the reflected component.

        double f1 = finish->Fresnel * FresnelR(cos_angle_of_incidence, relativeIor);

        // NB: One might think that to properly compute the outgoing loss we would have
        // to compute the Fresnel term for the _internal_ angle of incidence and the
        // _inverse_ of the relative refractive index; however, the Fresnel formula
        // is such that we can just as well plug in the external angle of incidence
        // and straight relative refractive index.
        cos_angle_of_incidence = -dot(layer_normal, eyeDirection);
        double f2 = finish->Fresnel * FresnelR(cos_angle_of_incidence, relativeIor);

        colour += intensity * layer_pigment_colour * light_colour * (1.0 - f1) * (1.0 - f2);
    }
    else
        colour += intensity * layer_pigment_colour * light_colour;

    // If user wishes to generate an alarm color where the resulting color is outside
    // the [0 to 1] range they can specify a negative intensity_max value.
    //
    // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
    // specify a positive positive intensity_max value.
    //
    // See similar blocks of code for phong, specular, interior(fade)...?)
    //
    // colour = ToMathColour(RGBColour(0.0, 0.0, 0.0)); // Specific alarm color vs pigment inversion?  .Set() ?
    if (intensity_max!=0.0)
    {
        if (intensity_max>0.0)
        {
            if (colour.Min()<0.0)
            {
                colour += std::abs(colour.Min());
            }
            if (colour.Max()>1.0)
            {
                colour = colour/colour.Max();
            }
        }
        else
        {
            if ((colour.Min()<0.0) || (colour.Max()>1.0))
            {
              //colour = 1.0 - layer_pigment_colour;
                std::size_t hashvalue;
                double cVal = colour.Red() + colour.Green() + colour.Blue();
                hashvalue = std::hash<double>{}((double)(POV_INT64)(cVal/0.01));
                float rndcolor = (2.0 + (double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                colour.Set(rndcolor*0.6,rndcolor*0.8,rndcolor*1.0);
            }
        }
    }
}

// TODO. What about an optional finish->Irid_Turb_Scale for ipoint given turbulence scale is fixed...
// Documentation v38 May 2020 mentions an Irid_Amount which is mapped to finish->Irid.
void Trace::ComputeIridColour(const FINISH *finish, const Vector3d& lightDirection,
        const Vector3d& eyeDirection, const Vector3d& layer_normal, const Vector3d& ipoint, MathColour& colour)
{
    double cos_angle_of_incidence_light, cos_angle_of_incidence_eye, interference;
    double film_thickness;
    double turbvalue;
    TurbulenceWarp turb;

    film_thickness = finish->Irid_Film_Thickness;

    if (finish->Irid_Turb != 0)
    {
        // Uses hardcoded octaves, lambda, omega
        turb.Omega=0.5;
        turb.Lambda=2.0;
        turb.Octaves=5;

        // Turbulence() returns a value from -1..1, centered at 0.0 (in povr, not any POV-Ray as of May 25, 2020)
        turbvalue = Turbulence(ipoint, &turb, sceneData->noiseGenerator);
        turbvalue = 1.0 + turbvalue * finish->Irid_Turb; // Was 1.0 + here...
        film_thickness *= turbvalue;
        film_thickness = std::max(0.0,film_thickness); // Adjust so thickness never <0
    }

    // NOTE: Shouldn't we compute Cos_Angle_Of_Incidence just once?
    cos_angle_of_incidence_light = abs(dot(layer_normal, lightDirection));
    cos_angle_of_incidence_eye   = abs(dot(layer_normal, eyeDirection));

    // Calculate phase offset.
    interference = 2.0 * M_PI * film_thickness * (cos_angle_of_incidence_light + cos_angle_of_incidence_eye);

    // Modify color by phase offset for each wavelength.
    colour *= 1.0 + finish->Irid * Cos(interference / sceneData->iridWavelengths);
}

void Trace::ComputePhongColour(const FINISH *finish, const Vector3d& lightDirection, const Vector3d& eyeDirection,
        const Vector3d& layer_normal, MathColour& colour, const MathColour& light_colour,
        const MathColour& layer_pigment_colour, double relativeIor)
{
    double cos_angle_of_incidence = -2.0 * dot(eyeDirection, layer_normal);
    double intensity_max = finish->Intensity_Max;
    double intensity;
    Vector3d reflect_direction = eyeDirection + cos_angle_of_incidence * layer_normal;

    cos_angle_of_incidence = dot(reflect_direction, lightDirection);

    if (cos_angle_of_incidence > 0.0)
    {
        //  cos_angle_of_incidence never > 1.0+(3*DBL_EPSILON) and in block only where it's >0.0
        intensity = pow(cos_angle_of_incidence,(double)finish->Phong_Size);
        intensity *= finish->Phong;

        MathColour colourIntoIntensity;
        if ((finish->Fresnel != 0.0) || (finish->Metallic != 0.0))
        {
            MathColour cs(1.0);
            // NB: The following dot product should technically be done with the vector halfway
            // between the in- and outgoing direction, rather than, the surface normal.
            // But since the Phong model specifically takes shortcuts to not compute that
            // vector, we'll take the surface normal as an approximation. Also, the choice to
            // compute the dot with the light rather than viewing direction is arbitrary.
            double ndotl = dot(layer_normal, lightDirection); // Values tend heavily toward 1.0 but always <=1.0
            if (finish->Fresnel != 0.0)
            {
                // In specular reflections (which includes highlights), the Fresnel effect is
                // that we only get the reflected component.
                cs *= finish->Fresnel * FresnelR(ndotl, relativeIor);
            }
            ComputeMetallic(cs, finish->Metallic, layer_pigment_colour, ndotl);
            colourIntoIntensity = light_colour * cs;
            colour += intensity * colourIntoIntensity;
        }
        else
        {
            colourIntoIntensity = light_colour;
            colour += intensity * light_colour;
        }

        // If user wishes to generate an alarm color where the resulting color is outside
        // the [0 to 1] range they can specify a negative intensity_max value.
        //
        // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
        // specify a positive positive intensity_max value.
        //
        // See similar blocks of code for diffuse, specular, interior(fade), ...?)
        //
        if (intensity_max!=0.0)
        {
            if (intensity_max>0.0)
            {
                if (colour.Min()<0.0)
                {
                    colour += std::abs(colour.Min());
                }
                if (colour.Max()>1.0)
                {
                    colour = colour/colour.Max();
                }
            }
            else
            {
                if ((colour.Min()<0.0) || (colour.Max()>1.0))
                {
                  //colour = 1.0 - colourIntoIntensity;
                    std::size_t hashvalue;
                    double cVal = colour.Red() + colour.Green() + colour.Blue();
                    hashvalue = std::hash<double>{}((double)(POV_INT64)(cVal/0.01));
                    float rndcolor = (2.0 + (double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                    colour.Set(rndcolor*1.0,rndcolor*0.6,rndcolor*0.8);
                }
            }
        }
    }
}

void Trace::ComputeSpecularColour(const FINISH *finish, const Vector3d& lightDirection, const Vector3d& eyeDirection,
        const Vector3d& layer_normal, MathColour& colour, const MathColour& light_colour,
        const MathColour& layer_pigment_colour, double relativeIor)
{
    Vector3d halfway = (eyeDirection + lightDirection) * 0.5;
    double halfway_length = halfway.length();
    double cos_angle_of_incidence, intensity;
    double intensity_max = finish->Intensity_Max;

    if (halfway_length > 0.0)
    {
        cos_angle_of_incidence = dot(halfway, layer_normal) / halfway_length;

        if (cos_angle_of_incidence > 0.0)
        {
            //  cos_angle_of_incidence never > 1.0+(3*DBL_EPSILON) and in block only where it's >0.0
            intensity = pow(cos_angle_of_incidence,(double)finish->Roughness);
            intensity *= finish->Specular;

            MathColour colourIntoIntensity;
            if ((finish->Fresnel != 0.0) || (finish->Metallic != 0.0))
            {
                MathColour cs(1.0);
                double ndotl = dot(halfway, lightDirection) / halfway_length; // Values tends heavily toward 1.0, but always <= 1.0
                if (finish->Fresnel != 0.0)
                {
                    // In specular reflections (which includes highlights), the Fresnel effect is
                    // that we only get the reflected component.
                    cs *= finish->Fresnel * FresnelR(ndotl, relativeIor);
                }
                ComputeMetallic(cs, finish->Metallic, layer_pigment_colour, ndotl);
                colourIntoIntensity = light_colour * cs;
                colour += intensity * colourIntoIntensity;
            }
            else
            {
                colourIntoIntensity = light_colour;
                colour += intensity * light_colour;
            }

            // If user wishes to generate an alarm color where the resulting color is outside
            // the [0 to 1] range they can specify a negative intensity_max value.
            //
            // If the user wishes to clamp the resultant colors inside the [0 to 1] range, they
            // specify a positive positive intensity_max value.
            //
            // See similar blocks of code for diffuse, phong, interior(fade) ...?)
            //
            if (intensity_max!=0.0)
            {
                if (intensity_max>0.0)
                {
                    if (colour.Min()<0.0)
                    {
                        colour += std::abs(colour.Min());
                    }
                    if (colour.Max()>1.0)
                    {
                        colour = colour/colour.Max();
                    }
                }
                else
                {
                    if ((colour.Min()<0.0) || (colour.Max()>1.0))
                    {
                        std::size_t hashvalue;
                        double cVal = colour.Red() + colour.Green() + colour.Blue();
                        hashvalue = std::hash<double>{}((double)(POV_INT64)(cVal/0.01));
                        float rndcolor = (2.0 + (double)hashvalue/(double)std::numeric_limits<std::size_t>::max())/3.0;
                        colour.Set(rndcolor*0.8,rndcolor*1.0,rndcolor*0.6);
                    }
                }
            }
        }
    }
}

void Trace::ComputeRelativeIOR(const Ray& ray, const Interior *interior, double& ior)
{
    // Get ratio of iors depending on the interiors the ray is traversing.

    POV_ASSERT(interior != nullptr);  // v3.7 onward always have interior.
//  if (interior == nullptr)
//  {
//      // TODO VERIFY - is this correct?
//      ior = 1.0;
//  }
//  else
//  {
        if (ray.GetInteriors().empty())
            // The ray is entering from the atmosphere.
            ior = interior->IOR / sceneData->atmosphereIOR;
        else
        {
            // The ray is currently inside an object.
            if (ray.IsInterior(interior) == true)
            {
                if (ray.GetInteriors().size() == 1)
                    // The ray is leaving into the atmosphere.
                    ior = sceneData->atmosphereIOR / interior->IOR;
                else
                    // The ray is leaving into another object.
                    ior = ray.GetInteriors().back()->IOR / interior->IOR;
            }
            else
                // The ray is entering a new object.
                ior = interior->IOR / ray.GetInteriors().back()->IOR;
        }
//  }
}

void Trace::ComputeReflectivity(double& weight, MathColour& reflectivity, const MathColour& reflection_max,
        const MathColour& reflection_min, bool fresnel, double reflection_falloff, double cos_angle,
        double relativeIor)
{
    // TODO. Where only one texture. Understand the passed weight coming in mostly at 1.0, but
    // sometimes 0.9<something>  Why?
    //  if (weight>0.9)
    //      throw POV_EXCEPTION_STRING("ComputeReflectivity incoming weight is not 1.0");
    // fprintf(stderr,"===+++ %f\n",weight);

    if (fresnel == false)
    {
        double temp_Weight_Max = reflection_max.WeightMax();
        double temp_Weight_Min = reflection_min.WeightMax();
        weight = weight * max(temp_Weight_Max, temp_Weight_Min);

        double reflection_Frac;
        if (fabs(reflection_falloff - 1.0) > gkDBL_epsilon)
        {
            reflection_Frac = pow(1.0 - cos_angle, reflection_falloff);
        }
        else
        {
            reflection_Frac = 1.0 - cos_angle;
        }

        if (fabs(reflection_Frac) < gkDBL_epsilon)
        {
            reflectivity = reflection_min;
        }
        else if (fabs(reflection_Frac - 1.0) < gkDBL_epsilon)
        {
            reflectivity = reflection_max;
        }
        else
        {
            reflectivity = reflection_Frac * reflection_max + (1.0 - reflection_Frac) * reflection_min;
        }
    }
    else
    {
        ComputeFresnel(reflectivity, reflection_max, reflection_min, cos_angle, relativeIor);
        weight = weight * reflectivity.WeightMax();
    }
}

void Trace::ComputeMetallic(MathColour& colour, double metallic, const MathColour& metallicColour,
        double cosAngle)
{
    // Calculate the reflected color by interpolating between
    // the light source color and the surface color according
    // to the (empirical) Fresnel reflectivity function. [DB 9/94]
    if (metallic != 0.0)
    {
        double x = fabs(acos(cosAngle)) / M_PI_2;
        double F = 0.014567225 / Sqr(x - 1.12) - 0.011612903;
        F = min(1.0, max(0.0, F));

        colour *= (1.0 + (metallic * (1.0 - F)) * (metallicColour - 1.0));
    }
}

void Trace::ComputeFresnel(MathColour& reflectivity, const MathColour& rMax, const MathColour& rMin,
        double cos_angle, double ior)
{
    double f = FresnelR(cos_angle, ior);
    reflectivity = f * rMax + (1.0 - f) * rMin;
}

double Trace::FresnelR(double cosTi, double n)
{
    /* NB: This is a special case of the Fresnel formula, presuming that incident light is unpolarized.

       The implemented formula is as follows:

            1     / g - cos Ti \ 2    /     / cos Ti (g + cos Ti) - 1 \ 2 \
       R = --- * ( ------------ )  * ( 1 + ( ------------------------- )   )
            2     \ g + cos Ti /      \     \ cos Ti (g - cos Ti) + 1 /   /

       where

              /---------------------------
       g = -\/ (n1/n2)^2 + (cos Ti)^2 - 1
    */

    double sqrg = Sqr(n) + Sqr(cosTi) - 1.0;

    if (sqrg <= 0.0)
    {
        // Total reflection.
        return 1.0;
    }

    double g = sqrt(sqrg);

    double quot1 = (g - cosTi) / (g + cosTi);
    double quot2 = (cosTi * (g + cosTi) - 1.0) / (cosTi * (g - cosTi) + 1.0);

    double f = 0.5 * Sqr(quot1) * (1.0 + Sqr(quot2));

    return clip(f, 0.0, 1.0);
}

void Trace::ComputeOneWhiteLightRay(const LightSource &lightsource, double& lightsourcedepth, Ray& lightsourceray,
        const Vector3d& ipoint, const Vector3d& jitter)
{
    Vector3d center = lightsource.Center + jitter;
    double a;
    Vector3d v1;

    // Get the light ray starting at the intersection point and pointing towards the light source.
    lightsourceray.Origin = ipoint;
    // NK 1998 parallel beams for cylinder source - added 'if'
    if (lightsource.Light_Type == CYLINDER_SOURCE)
    {
        double distToPointsAt;
        Vector3d toLightCtr;

        // use new code to get ray direction - use center - points_at for direction
        lightsourceray.Direction = center - lightsource.Points_At;

        // get vector pointing to center of light
        toLightCtr = center - ipoint;

        // project light_ctr-intersect_point onto light_ctr-point_at
        distToPointsAt = lightsourceray.Direction.length();
        lightsourcedepth = dot(toLightCtr, lightsourceray.Direction);

        // lenght of shadow ray is the length of the projection
        lightsourcedepth /= distToPointsAt;
        lightsourceray.Direction.normalize(); // TODO - lightsourceray.Direction /= distToPointsAt  should also work
    }
    else
    {
        // NK 1998 parallel beams for cylinder source - the stuff in this 'else'
        // block used to be all that there was... the first half of the if
        // statement (before the 'else') is new

        lightsourceray.Direction = center - ipoint;
        lightsourcedepth = lightsourceray.Direction.length();
        lightsourceray.Direction /= lightsourcedepth;
    }

    // Attenuate light source color.
    // Attenuation = Attenuate_Light(lightsource, lightsourceray, *Light_Source_Depth);
    // Recalculate for Parallel light sources
    if (lightsource.Parallel)
    {
        if (lightsource.Area_Light)
        {
            v1 = (center - lightsource.Points_At).normalized();
            a = dot(v1, lightsourceray.Direction);
            lightsourcedepth *= a;
            lightsourceray.Direction = v1;
        }
        else
        {
            a = dot(lightsource.Direction, lightsourceray.Direction);
            lightsourcedepth *= (-a);
            lightsourceray.Direction = -lightsource.Direction;
        }
    }
}

void Trace::ComputeSky(const Ray& ray, MathColour& colour, ColourChannel& transm)
{
    // This gives the same results regarding sky sphere filter as a
    // layered-texture genuine sphere

    MathColour filCol(1.0);
    double att;
    MathColour col;
    TransColour col_Temp;
    Vector3d point;

    col.Clear();

    if (sceneData->skysphere != nullptr)
    {
        // Transform point on unit sphere.
        if (sceneData->skysphere->Trans != nullptr)
            MInvTransPoint(point, ray.Direction, sceneData->skysphere->Trans);
        else
            point = ray.Direction;

        // TODO - Reverse iterator may be less performant than forward iterator; we might want to
        //        compare performance with using forward iterators and decrement, or using random access.
        //        Alternatively, reversing the vector after parsing might be another option.
        for (auto i = sceneData->skysphere->Pigments.rbegin(); i != sceneData->skysphere->Pigments.rend(); ++ i)
        {
            // Compute sky colour from colour map.
            Compute_Pigment(col_Temp, *i, point, nullptr, nullptr, threadData);

            att = col_Temp.Opacity();

            col += col_Temp.colour() * att * filCol * sceneData->skysphere->Amplify;
            filCol *= col_Temp.TransmittedColour();
        }
    }

    // apply background as if it was another sky sphere with uniform pigment
    col_Temp = sceneData->backgroundColour;

    if (!ray.GetTicket().alphaBackground)
    {
        // if rendering without alpha channel, ignore filter and transmit of background color.
        col_Temp.SetFT(0.0, 0.0); // this affects the skysphere too.
    }

    att = col_Temp.Opacity();

    col += col_Temp.colour() * att * filCol;
    filCol *= col_Temp.TransmittedColour();

    colour = col;
    transm = min(1.0f, fabs(filCol.Greyscale()));
}

void Trace::ComputeFog(const Ray& ray, const Intersection& isect, MathColour& colour, ColourChannel& transm)
{
    double att, fogDepthAtIntersection;
    MathColour col_fog;
    ColourChannel filter_fog, transm_fog;
    MathColour sum_att; // total attenuation.
    MathColour sum_col; // total color.

    // All computeFog calls being checked for nullptr at time of change to POV_ASSERT below.
    POV_ASSERT(sceneData->fog != nullptr);  // If POV_DEBUG defined, assert still true.

    // Init total attenuation and total color.
    sum_att = MathColour(1.0);
    sum_col = MathColour(0.0);

    // Loop over all fogs.
    for (FOG *fog = sceneData->fog; fog != nullptr; fog = fog->Next)
    {
        fogDepthAtIntersection = isect.Depth;
        double RangeSpan = (fog->RangeHi - fog->RangeLo);
        double RangeMid  = fog->RangeLo + RangeSpan/2.0;

        switch (fog->Type)
        {
            case gkFog:
                att = ComputeDepthFog(ray, false, fogDepthAtIntersection, fog);
                break;
            case gkFogXfrmTurb:  // Really transforming ray origin.
                att = ComputeDepthFog(ray, true, fogDepthAtIntersection, fog);
                break;
            case gkFogMist:
                att = ComputeGroundFog(ray, 0.0, fogDepthAtIntersection, fog, 0);
                break;
            case gkFogMistDist:
                att = ComputeGroundFog(ray, 0.0, fogDepthAtIntersection, fog, 1);
                break;
            case gkFogInvDepthMap:
                if ((fogDepthAtIntersection > fog->RangeLo) && (fogDepthAtIntersection < fog->RangeHi))
                    att = min(1.0,(fogDepthAtIntersection - fog->RangeLo) / RangeSpan);
                else
                    att = 1.0;
                break;
            case gkFogRange:
                att = (fabs(RangeMid - fogDepthAtIntersection) <= RangeSpan/2.0) ? 0.0 : 1.0;
                if ((fog->BlendGamma > gkDBL_epsilon) && (att == 0.0))
                {
                    double tmpVal = 1.0 - fabs(RangeMid - fogDepthAtIntersection)/(RangeSpan/2.0); // *2.71 below restores 0-1 range
                    att = exp(1.0/(pow(min(1.0-gkMinIsectDepthReturned,tmpVal),fog->BlendGamma)-1.0))*exp(1);
                }
                break;
            default:
                throw POV_EXCEPTION_STRING("Internal Error! Reached default fog type.");
                break;
        }

        if ((fog->Type != gkFogInvDepthMap) && (fog->Type != gkFogRange))
        {
            if (RangeSpan && (att < 1.0))
            {
                if ((fogDepthAtIntersection > fog->RangeLo) && (fogDepthAtIntersection < fog->RangeHi))
                {
                    double tmpVal = fabs(RangeMid - fogDepthAtIntersection)/(RangeSpan/2.0);
                    att = min(1.0,att + pow(tmpVal,2.0));
                }
                else
                    att = 1.0;
            }
            if (fog->BlendGamma > gkDBL_epsilon)
            {
                double tmpVal = 1.0 - att;                                       // *2.71 below restores 0-1 range
                att = min(1.0,exp(1.0/(pow(min(1.0-gkMinIsectDepthReturned,tmpVal),fog->BlendGamma)-1.0))*exp(1));
            }
        }

        if (fog->Inverse)
            att = 1.0 - att;

        col_fog = fog->colour.colour();
        fog->colour.GetFT(filter_fog, transm_fog);

        // Check for minimum transmittance.
        if (att < transm_fog)
            att = transm_fog;

        // Get attenuation sum due to filtered/unfiltered translucency.
        // [CLi] removed computation of sum_att.filer() and sum_att.transm(), as they were discarded anyway
        sum_att *= att * ((1.0 - filter_fog) + filter_fog * col_fog);

        if (!ray.IsShadowTestRay())  // If not a ray being traced to a light source.
            sum_col += (1.0 - att) * col_fog;
    }

    // Add light coming from background.
    colour = sum_col + sum_att * colour;
    transm *= sum_att.Greyscale();
}

double Trace::ComputeDepthFog(const Ray &ray, bool xfrmTurb, double fogDepthAtIntersection, const FOG *fog)
{
    double potentiallyTurbAdjustedDepth = fogDepthAtIntersection;

    if (fog->Turb != nullptr)
    {
        Vector3d p;
        if ((xfrmTurb) && (dot(fog->Up,Vector3d(0,1,0)) != 1.0))
        {
            Ray rayOxfrm = ray;
            Vector3d Vx, Vz;

            // Vector perpendicular to modified y up. Modeled on include VPerp_To_Vector.
            double xabs = fabs(fog->Up.x()), yabs = fabs(fog->Up.y()), zabs = fabs(fog->Up.z());
            double dimensionMax = std::max(xabs,std::max(yabs,zabs));
            if (zabs == dimensionMax)
                Vx = cross(fog->Up,Vector3d(0,0,1)).normalized();
            else if (yabs == dimensionMax)
                Vx = cross(fog->Up,Vector3d(0,1,0)).normalized();
            else
                Vx = cross(fog->Up,Vector3d(1,0,0)).normalized();
            Vz = cross(Vx,fog->Up);
            MATRIX Matrix;
            TRANSFORM Trans;
            // TODO Look to convert to Compute_Axis_Rotation_Transform. In normal.cpp work found to be
            // somewhat faster than setting up the Matrix explicitly.
            Matrix[0][0]=Vx.x();      Matrix[0][1]=Vx.y();      Matrix[0][2]=Vx.z();      Matrix[0][3]=0.0;
            Matrix[1][0]=fog->Up.x(); Matrix[1][1]=fog->Up.y(); Matrix[1][2]=fog->Up.z(); Matrix[1][3]=0.0;
            Matrix[2][0]=Vz.x();      Matrix[2][1]=Vz.y();      Matrix[2][2]=Vz.z();      Matrix[2][3]=0.0;
            Matrix[3][0]=0.0;         Matrix[3][1]=0.0;         Matrix[3][2]=0.0;         Matrix[3][3]=1.0;
            Compute_Matrix_Transform(&Trans, Matrix);
            MTransDirection(rayOxfrm.Origin,rayOxfrm.Origin,&Trans); // Transforming ray origin too and point,subjectively,
            p = rayOxfrm.Evaluate(fogDepthAtIntersection);           // looked better. Useful, but 2D turb xfrms of course hacky.
          //p = ray.Evaluate(fogDepthAtIntersection);                // Only this an last line OK too. Less compute...?
            MTransDirection(p,p,&Trans);
        }
        else
        {
            p = ray.Evaluate(fogDepthAtIntersection);
        }
        p *= fog->Turb->Turbulence;  // Here 'scaling' turbulence relative to scene x,y,z.

                                                                 // The further away the less influence turbulence has.
        double k = exp(-fogDepthAtIntersection / fog->Distance); // At fogDepthAtIntersection == fog->Distance k=0.368

        double turbValue = Turbulence(p, fog->Turb, sceneData->noiseGenerator);  // Centered at 0.0 and at omega 0.5 dist ~= -1 t 1.
        turbValue = max(0.0,(1 + turbValue) * 0.5);                              // 0 - 1+ range. '+' depending on how turbulent.

        potentiallyTurbAdjustedDepth *=
            (1.0 - (k * turbValue * fog->Turb_Depth));

          // Old min clamp below creates odd plateau, which turb depth or turbulence
          // eventually overwhelms - if large enough - in any case.
          //(1.0 - (k * min(1.0,turbValue) * fog->Turb_Depth));
    }

    double value = (exp(-potentiallyTurbAdjustedDepth / fog->Distance));

    return value;
}

/*****************************************************************************
*   Here is an ascii graph of the ground fog density, it has a maximum
*   density of 1.0 at Y <= 0, and approaches 0.0 as Y goes up:
*
*   ***********************************
*        |           |            |    ****
*        |           |            |        ***
*        |           |            |           ***
*        |           |            |            | ****
*        |           |            |            |     *****
*        |           |            |            |          *******
*   -----+-----------+------------+------------+-----------+-----
*       Y=-2        Y=-1         Y=0          Y=1         Y=2
*
*   ground fog density is 1 / (Y*Y+1) for Y >= 0 and equals 1.0 for Y <= 0.
*   (It behaves like regular fog for Y <= 0.)
*
*   The integral of the density is atan(Y) (for Y >= 0).
******************************************************************************/
// Original version by Eric Barish.
double Trace::ComputeGroundFog(const Ray& ray, double depth, double fogDepthAtIntersection,
        const FOG *fog, bool useDistanceForAttn)
{
    double fog_density, delta;
    double start, end;
    double y1, y2, k;
    double potentiallyTurbAdjustedDepth = fogDepthAtIntersection;
    Vector3d p, p1, p2;

    // Get start point.
    p1 = ray.Evaluate(depth);

    // Get end point.
    p2 = p1 + ray.Direction * fogDepthAtIntersection;

    // Could preform transfomation here to translate Start and End
    // points into ground fog space.
    y1 = dot(p1, fog->Up);
    y2 = dot(p2, fog->Up);

    start = (y1 - fog->Offset) / fog->Alt;
    end   = (y2 - fog->Offset) / fog->Alt;

    // Get integral along y-axis from start to end.
    if (start <= 0.0)
    {
        if (end <= 0.0)
            fog_density = 1.0;
        else
            fog_density = (atan(end) - start) / (end - start);
    }
    else
    {
        if (end <= 0.0)
            fog_density = (atan(start) - end) / (start - end);
        else
        {
            delta = start - end;

            if (fabs(delta) > gkDBL_epsilon)
                fog_density = (atan(start) - atan(end)) / delta;
            else
                fog_density = 1.0 / (Sqr(start) + 1.0);
        }
    }

    if (fog->Turb != nullptr)
    {
        p = (p1 + p2) * 0.5;
        p *= fog->Turb->Turbulence;  // We are 'scaling' turbulence here relative to scene x,y,z.

                                                                 // The further away the less influence turbulence has.
        double k = exp(-fogDepthAtIntersection / fog->Distance); // At fogDepthAtIntersection == fog->Distance k=0.368

        double turbValue = Turbulence(p, fog->Turb, sceneData->noiseGenerator);  // Centered at 0.0 and at omega 0.5 dist ~= -1 t 1.
        turbValue = max(0.0,(1 + turbValue) * 0.5);                              // 0 - 1+ range. '+' depending on how turbulent.

        potentiallyTurbAdjustedDepth *=
            (1.0 - (k * turbValue * fog->Turb_Depth));

          // Old min clamp below creates odd plateau, which turb depth or turbulence
          // eventually overwhelms - if large enough - in any case.
          //(1.0 - (k * min(1.0,turbValue) * fog->Turb_Depth));
    }

    double value;
    if (useDistanceForAttn)  // Always previous method, but tends to make turbulence nearly invisible.
        value = (exp(-potentiallyTurbAdjustedDepth * fog_density / fog->Distance)); // Hard to get turbulence using fog->Distance.
    else
        value = (exp(-potentiallyTurbAdjustedDepth * fog_density / fog->Alt)); // Use altitude over distance to get turbulence.
    return value;
}


void Trace::ComputeShadowMedia(Ray& light_source_ray, Intersection& isect, MathColour& resultcolour,
        bool media_attenuation_and_interaction)
{
    if (resultcolour.IsNearZero(gkDBL_epsilon))
        return;

    // Calculate participating media effects.
    if (media_attenuation_and_interaction && qualityFlags.media &&
        ((light_source_ray.IsHollowRay() == true) || ((isect.Object != nullptr) && (isect.Object->interior != nullptr))))
    {

        // we're using general-purpose media and fog handling code, which
        // insists on computing a transmissive component (for alpha channel)

        ColourChannel tmpTransm = 0.0;
        media.ComputeMedia(sceneData->atmosphere, light_source_ray, isect, resultcolour, tmpTransm);

        if ((sceneData->fog != nullptr) && (light_source_ray.IsHollowRay() == true) && (light_source_ray.IsPhotonRay() == false))
            ComputeFog(light_source_ray, isect, resultcolour, tmpTransm);

        // discard the transmissive component (alpha channel)
    }

    // If ray is entering from the atmosphere or the ray is currently *not* inside an object add it,
    // but if it is currently inside an object, the ray is leaving the current object and is removed
    if ((isect.Object != nullptr) && ((light_source_ray.GetInteriors().empty()) ||
                (light_source_ray.RemoveInterior(isect.Object->interior.get()) == false)))
    {
        light_source_ray.AppendInterior(isect.Object->interior.get());
    }
}



void Trace::ComputeRainbow(const Ray& ray, const Intersection& isect, MathColour& colour, ColourChannel& transm)
{
    int n;
    double dot1, k, ki, index, x, y, l, angle, fade, f;
    Vector3d Temp;
    TransColour Cr;
    ColourChannel CrFilter, CrTransm;
    MathColour CtColour;
    ColourChannel CtFilter, CtTransm;

    // Why are we here.
    if (sceneData->rainbow == nullptr)
        return;

    // TODO - get rid of the use of the RGBFT colour model

    CtColour = MathColour(0.0);
    CtFilter = 1.0;
    CtTransm = 1.0;

    n = 0;

    for (RAINBOW *Rainbow = sceneData->rainbow; Rainbow != nullptr; Rainbow = Rainbow->Next)
    {
        if ((Rainbow->Pigment != nullptr) && (Rainbow->Distance != 0.0) && (Rainbow->Width != 0.0))
        {
            // Get angle between ray direction and rainbow's up vector.
            x = dot(ray.Direction, Rainbow->Right_Vector);
            y = dot(ray.Direction, Rainbow->Up_Vector);

            l = Sqr(x) + Sqr(y);

            if (l > 0.0)
            {
                l = sqrt(l);
                y /= l;
            }

            angle = fabs(acos(y));

            if (angle <= Rainbow->Arc_Angle)
            {
                // Get dot product between ray direction and antisolar vector.
                dot1 = dot(ray.Direction, Rainbow->Antisolar_Vector);

                if (dot1 >= 0.0)
                {
                    // Get index ([0;1]) into rainbow's colour map.
                    index = (acos(dot1) - Rainbow->Angle) / Rainbow->Width;

                    // Jitter index.
                    if (Rainbow->Jitter > 0.0)
                        index += (2.0 * randomNumberGenerator() - 1.0) * Rainbow->Jitter;

                    if ((index >= 0.0) && (index <= 1.0 - gkDBL_epsilon))
                    {
                        // Get colour from rainbow's colour map.
                        Temp = Vector3d(index, 0.0, 0.0);
                        Compute_Pigment(Cr, Rainbow->Pigment, Temp, &isect, &ray, threadData);
                        Cr.GetFT(CrFilter, CrTransm);

                        // Get fading value for falloff.
                        if ((Rainbow->Falloff_Width > 0.0) && (angle > Rainbow->Falloff_Angle))
                        {
                            fade = (angle - Rainbow->Falloff_Angle) / Rainbow->Falloff_Width;
                            fade = (3.0 - 2.0 * fade) * fade * fade;
                        }
                        else
                            fade = 0.0;

                        // Get attenuation factor due to distance.
                        k = exp(-isect.Depth / Rainbow->Distance);

                        // Colour's transm value is used as minimum attenuation value.
                        k = max(k, fade * (1.0 - CrTransm) + CrTransm);

                        // Now interpolate the colours.
                        ki = 1.0 - k;

                        // Attenuate filter value.
                        f = CrFilter * ki;

                        CtColour += k * colour * ((1.0 - f) + f * Cr.colour()) + ki * Cr.colour();
                        CtFilter *= k * CrFilter;
                        CtTransm *= k * CrTransm;

                        n++;
                    }
                }
            }
        }
    }

    if (n > 0)
    {
        colour =  CtColour / n;
        transm *= CtTransm;
    }
}

bool Trace::TestShadow(const LightSource &lightsource, double& depth, Ray& light_source_ray, const Vector3d& p,
        MathColour& colour)
{
    ComputeOneLightRay(lightsource, depth, light_source_ray, p, colour);

    // There's no need to test for shadows if no light
    // is coming from the light source.
    //
    // Test for PURE zero, because we only want to skip this if we're out
    // of the range of a spot light or cylinder light.  Very dim lights
    // should not be ignored.

    if (colour.IsNearZero(gkDBL_epsilon))
    {
        colour.Clear();
        return true;
    }

    // Test for shadows.
    if (qualityFlags.shadows && ((lightsource.Projected_Through_Object != nullptr) ||
                (lightsource.Light_Type != FILL_LIGHT_SOURCE)))
    {
        TraceShadowRay(lightsource, depth, light_source_ray, p, colour);

        if (colour.IsNearZero(gkDBL_epsilon))
        {
            colour.Clear();
            return true;
        }
    }

    return false;
}

bool Trace::IsObjectInCSG(ConstObjectPtr object, ConstObjectPtr parent)
{
    bool found = false;

    if (object == parent)
        return true;

    if (parent->Type & IS_COMPOUND_OBJECT)
    {
        for (auto Sib = (reinterpret_cast<const CSG *>(parent))->children.cbegin();
                Sib != (reinterpret_cast<const CSG *>(parent))->children.cend(); Sib++)
        {
            if (IsObjectInCSG(object, *Sib))
                found = true;
        }
    }

    return found;
}

// SSLT code by Sarah Tariq and Lawrence (Lorenzo) Ibarria

double Trace::ComputeFt(double phi, double eta)
{
#if 0
    double sin_phi   = sin(phi);
    double sin_theta = sin_phi / eta;
    if ((sin_theta < -1.0) || (sin_theta > 1.0))
        return 0; // total reflection, i.e. no transmission at all

    double theta = asin(sin_theta);

    return 1 - 0.5 * (Sqr(sin(phi-theta)) / Sqr(sin(phi+theta)) + Sqr(tan(phi-theta)) / Sqr(tan(phi+theta)));
#else
    double cos_angle = cos(phi);
    double g = sqrt(Sqr(eta) + Sqr(cos_angle) - 1);
    double F = 0.5 * (Sqr(g - cos_angle) / Sqr(g + cos_angle));
    F = F * (1 + Sqr(cos_angle * (g + cos_angle) - 1) / Sqr(cos_angle * (g - cos_angle) + 1));

    return 1.0 - min(1.0,max(0.0,F));
#endif
}

void Trace::ComputeSurfaceTangents(const Vector3d& normal, Vector3d& u, Vector3d& v)
{
#if 1
    if (fabs(normal[0]) <= fabs(normal[1]) && fabs(normal[0]) <= fabs(normal[2]))
        // if x co-ordinate is smallest, creating a tangent in the yz plane is a piece of cake;
        // the following code is equivalent to u = cross(normal, Vector3d(1,0,0)).normalized();
        u = Vector3d(0, normal[2], -normal[1]).normalized();
    else if (fabs(normal[1]) <= fabs(normal[2]))
        // if y co-ordinate is smallest, creating a tangent in the xz plane is a piece of cake;
        // the following code is equivalent to u = cross(normal, Vector3d(0,1,0)).normalized();
        u = Vector3d(-normal[2], 0, normal[0]).normalized();
    else
        // if z co-ordinate is smallest, creating a tangent in the xy plane is a piece of cake;
        // the following code is equivalent to u = cross(normal, Vector3d(0,0,1)).normalized();
        u = Vector3d(normal[1], -normal[0], 0).normalized();
#else
    if (fabs(normal[0]) <= fabs(normal[1]) && fabs(normal[0]) <= fabs(normal[2]))
        // if x co-ordinate is smallest, creating a tangent in the yz plane is a piece of cake
        u = cross(normal, Vector3d(1,0,0)).normalized();
    else if (fabs(normal[1]) <= fabs(normal[0]) && fabs(normal[1]) <= fabs(normal[2]))
        // if y co-ordinate is smallest, creating a tangent in the xz plane is a piece of cake
        u = cross(normal, Vector3d(0,1,0)).normalized();
    else
        // if z co-ordinate is smallest, creating a tangent in the xy plane is a piece of cake
        u = cross(normal, Vector3d(0,0,1)).normalized();
#endif

    v = cross(normal, u);
}

void Trace::ComputeSSLTNormal(Intersection& Ray_Intersection)
{
    Vector3d Raw_Normal;

    /* Get the normal to the surface */
    Ray_Intersection.Object->Normal(Raw_Normal, &Ray_Intersection, threadData);
    Ray_Intersection.INormal = Raw_Normal;
    Ray_Intersection.PNormal = Raw_Normal; // TODO FIXME - we should possibly take normal pertubation into account
}

bool Trace::IsSameSSLTObject(ConstObjectPtr obj1, ConstObjectPtr obj2)
{
    // TODO maybe use something smarter
    return (obj1 && obj2 && obj1->interior == obj2->interior);
}

void Trace::ComputeDiffuseSampleBase(Vector3d& basePoint, const Intersection& out, const Vector3d& vOut,
        double avgFreeDist, TraceTicket& ticket)
{
    Vector3d pOut = out.IPoint;
    Vector3d nOut = out.INormal;

    // make sure to get the normal right; obviously, the observer must be "outside".
    // for algorithm simplicity, we want the normal to point inward
    double cos_phi = dot(nOut, vOut);
    if (cos_phi > 0)
        nOut.invert();

    // typically, place the base point the average free distance below the surface;
    // however, never place it closer to the "back side" than to the front
    Intersection backSide;
    Ray ray(ticket, pOut, nOut); // we're shooting from the surface, so SubsurfaceRay would do us no good (as it would potentially "re-discover" the current surface)
    backSide.Depth = avgFreeDist * 2; // max distance we're looking at
    bool found = FindIntersection(backSide, ray);
    if (found)
    {
        if (IsSameSSLTObject(out.Object, backSide.Object))
            basePoint = pOut + nOut * (backSide.Depth / 2);
        else
            basePoint = pOut + nOut * min(avgFreeDist, backSide.Depth - gkDBL_epsilon);
    }
    else
        basePoint = pOut + nOut * avgFreeDist;
}

void Trace::ComputeDiffuseSamplePoint(const Vector3d& basePoint, Intersection& in, double& sampleArea, TraceTicket& ticket)
{
    // generate a vector in a random direction
    // TODO FIXME - a suitably weighted distribution (oriented according to the surface normal) would possibly be better
    while (ssltUniformDirectionGenerator.size() <= ticket.subsurfaceRecursionDepth)
        ssltUniformDirectionGenerator.push_back(GetSubRandomDirectionGenerator(0, 32767));
    Vector3d v = (*(ssltUniformDirectionGenerator[ticket.subsurfaceRecursionDepth]))();

    Ray ray(ticket, basePoint, v, Ray::SubsurfaceRay);
    bool found = FindIntersection(in, ray);

    if (found)
    {
        ComputeSSLTNormal(in);

        Vector3d vDelta = in.IPoint - basePoint;
        double dist = vDelta.length();
        double cos_phi = abs(dot(vDelta / dist, in.INormal));
        if (cos_phi < 0)
        {
            in.INormal.invert();
            cos_phi = -cos_phi;
        }
        if (cos_phi < 0.01)
            cos_phi = 0.01; // TODO FIXME - rather arbitrary limit
        sampleArea = 4.0 * M_PI * Sqr(dist * sceneData->mmPerUnit) / cos_phi;
    }
    else
    {
        sampleArea = 0.0;
    }
}

void Trace::ComputeOneSingleScatteringContribution(const LightSource& lightsource, const Intersection& out,
        double sigma_t_xo, double sigma_s, double s_prime_out, MathColour& Lo, double eta, const Vector3d& bend_point,
        double phi_out, double cos_out_prime, TraceTicket& ticket)
{
    // TODO FIXME - part of this code is very alike to ComputeOneDiffuseLight()

    // Do Light source to get the correct lightsourceray
    // (note that for now we're mainly interested in the direction)
    Ray lightsourceray(ticket, Ray::SubsurfaceRay);
    double lightsourcedepth;
    ComputeOneWhiteLightRay(lightsource, lightsourcedepth, lightsourceray, bend_point);

    // We're below the surface; determine where a light ray from the source would be intersecting this object's surface
    // (and, more importantly, what the surface normal is there; notice that this intersection is an approximation,
    // ignoring refraction)
    Intersection xi;
    if (!FindIntersection(xi, lightsourceray))
        return;

    if (!IsSameSSLTObject(xi.Object, out.Object))
        return; // TODO - what if the other object is transparent?

    ComputeSSLTNormal(xi);

    // Get a colour and a ray (also recomputes all the lightsourceray stuff).
    MathColour lightcolour;
    ComputeOneLightRay(lightsource, lightsourcedepth, lightsourceray, xi.IPoint, lightcolour, true);

    // Don't calculate spotlights when outside of the light's cone.
    if (lightcolour.IsNearZero(gkDBL_epsilon))
        return;

    // See if light on far side of surface from camera.
    // [CLi] double_illuminate and diffuse backside illumination don't seem to make much sense with BSSRDF, so we ignore them here.
    // [CLi] BSSRDF always does "full area lighting", so we ignore it here.
    double cos_in = dot(xi.INormal, lightsourceray.Direction);
    // [CLi] we're coming from inside the object, so the surface /must/ be properly oriented towards the camera; if it isn't,
    // it must be the normal's fault
    if (cos_in < 0)
    {
        xi.INormal.invert();
        cos_in = -cos_in;
    }
    // [CLi] light coming in almost parallel to the surface is a problem though
    if (cos_in < gkDBL_epsilon)
        return;

    // If light source was not blocked by any intervening object, then
    // calculate it's contribution to the object's overall illumination.
    if (qualityFlags.shadows && ((lightsource.Projected_Through_Object != nullptr) ||
                (lightsource.Light_Type != FILL_LIGHT_SOURCE)))
    {
        // [CLi] Not using lightColorCache because it's unsuited for BSSRDF
        TraceShadowRay(lightsource, lightsourcedepth, lightsourceray, xi.IPoint, lightcolour);
    }

    // Don't calculate anything more if we're in full shadow
    if (lightcolour.IsNearZero(gkDBL_epsilon))
        return;

    double sigma_t_xi = sigma_t_xo; // TODO FIXME - theoretically this should be taken from point where light comes in

    double cos_in_sqr       = Sqr(cos_in);
    double sin_in_sqr       = 1 - cos_in_sqr;
    double eta_sqr          = Sqr(eta);
    double sin_in_prime_sqr = sin_in_sqr / eta_sqr;
    double cos_in_prime_sqr = 1 - sin_in_prime_sqr;
    if (cos_in_prime_sqr < 0.0)
        return; // total reflection
    double cos_in_prime  = sqrt(cos_in_prime_sqr);

    if (cos_in_prime <= gkMinIsectDepthReturned)  // gkMinIsectDepthReturned has relation to sqrt of gkDBL_epsilon
        return; // close enough to total reflection to give us trouble

    //MathColour lightColour = MathColour(lightsource.colour);
    lightcolour *= cos_in; // TODO VERIFY - is this right? Where does this term come from?

    // compute si
    double si = (bend_point - xi.IPoint).length() * sceneData->mmPerUnit;

    // calculate s_prime_i
    double s_prime_i = si * cos_in / cos_in_prime;

    // calculate F
    double phi_in  = acos(cos_in);
    double F = ComputeFt(phi_in, eta) * ComputeFt(phi_out, eta);

    // calculate sigma_tc
    double G = fabs(cos_out_prime / cos_in_prime); // TODO FIXME - theoretically this is only valid for comparatively flat surfaces
    double sigma_tc = sigma_t_xo + G * sigma_t_xi;

    // calculate the phase function
    // NOTE: We're leaving out the 1/pi factor because in POV-Ray, by convention,
    // light intensity is normalized to imply this factor already.
    double p = 1.0 / 4.0; // asume isotropic scattering (normally this would be 1/(4*M_PI))

    // multiply with the e terms
    double eTerms = exp(-s_prime_i * sigma_t_xi) * exp(-s_prime_out * sigma_t_xo);    // TODO FIXME - theoretically first sigma_t should be taken from from xi.IPoint

    double factor = (sigma_s * F * p / sigma_tc) * eTerms;
    if (factor >= DBL_MAX)
        factor = DBL_MAX;
    POV_SUBSURFACE_ASSERT((factor >= 0.0) && (factor <= DBL_MAX)); // verify factor is a non-negative, finite value (no #INF, no #IND, no #NAN)

    lightcolour *= factor;

    // add up the contribution
    Lo += lightcolour;
}

// call this once for each color
// out.INormal is calculated
void Trace::ComputeSingleScatteringContribution(const Intersection& out, double dist, double theta_out,
        double cos_out_prime, const Vector3d& refractedREye, double sigma_t_xo, double sigma_s, MathColour& Lo,
        double eta, TraceTicket& ticket)
{
    Lo.Clear();

    // calculate s_prime_out
    while (ssltUniformNumberGenerator.size() <= ticket.subsurfaceRecursionDepth)
        ssltUniformNumberGenerator.push_back(GetRandomDoubleGenerator(0.0, 1.0, 32767));
    double epsilon = 1.0 - (*(ssltUniformNumberGenerator[ticket.subsurfaceRecursionDepth]))(); // epsilon is a random floating point value in the range (0,1] (i.e. not including 0, but including 1)
    double s_prime_out = fabs(log(epsilon)) / sigma_t_xo;

    if (s_prime_out >= dist)
        return; // not within the object - this is covered by a "zero scattering" term instead

    //compute bend_point wihch is s_prime_out distance away on refractedREye
    Vector3d bend_point = out.IPoint + refractedREye * (s_prime_out / sceneData->mmPerUnit);

    // global light sources, if not turned off for this object
    if ((out.Object->Flags & NO_GLOBAL_LIGHTS_FLAG) != NO_GLOBAL_LIGHTS_FLAG)
    {
        for (auto i = 0ul; i < threadData->lightSources.size(); i++)
        {
            ComputeOneSingleScatteringContribution(*threadData->lightSources[i], out, sigma_t_xo,
                    sigma_s, s_prime_out, Lo, eta, bend_point, theta_out, cos_out_prime, ticket);
        }
    }

    // local light sources from a light group, if any
    if (!out.Object->LLights.empty())
    {
        for (auto i = 0ul; i < out.Object->LLights.size(); i++)
        {
            ComputeOneSingleScatteringContribution(*out.Object->LLights[i], out, sigma_t_xo,
                    sigma_s, s_prime_out, Lo, eta, bend_point, theta_out, cos_out_prime, ticket);
        }
    }

    // TODO FIXME - radiosity should also be taken into account
}

bool Trace::SSLTComputeRefractedDirection(const Vector3d& v, const Vector3d& n, double eta, Vector3d& refracted)
{
    // Phi: angle between normal and -incoming_ray (REye in this case, since it points to Eye)
    // Theta: angle between -normal and outgoing_ray
    double   cosPhi;
    Vector3d unitV = v.normalized();
    Vector3d unitN = n.normalized();

    cosPhi = dot(unitV, unitN);
    if (cosPhi > 0)
        unitN.invert();
    else
        cosPhi = -cosPhi;

    double cosThetaSqr = 1.0 + Sqr(eta) * (Sqr(cosPhi) - 1.0);
    if (cosThetaSqr < 0.0)
        return false;

    double cosTheta = sqrt(cosThetaSqr);
    refracted = (unitV * eta + unitN * (eta * cosPhi - cosTheta)).normalized();

    return true;
}

void Trace::ComputeDiffuseContribution(const Intersection& out, const Vector3d& vOut, const Vector3d& pIn,
        const Vector3d& nIn, const Vector3d& vIn, double& sd, double sigma_prime_s, double sigma_a, double eta)
{
    // TODO FIXME - a great deal of this can be precomputed
    double sigma_prime_t = sigma_prime_s + sigma_a;
    double alpha_prime = sigma_prime_s / sigma_prime_t;
    double F_dr = FresnelDiffuseReflectance(eta);
    double Aconst = ((1 + F_dr) / (1 - F_dr));
    double Rd;

    double cos_phi_in  = clip(dot(vIn,  nIn),         -1.0, 1.0); // (clip values to not run into trouble due to petty precision issues)
    double cos_phi_out = clip(dot(vOut, out.INormal), -1.0, 1.0);
    double phi_in  = acos(cos_phi_in);
    double phi_out = acos(cos_phi_out);

    double F = ComputeFt(phi_in, eta) * ComputeFt(phi_out, eta);

#if 1
    // full BSSRDF model

    double distSqr = (pIn - out.IPoint).lengthSqr() * Sqr(sceneData->mmPerUnit);

    double Dconst = 1 / (3 * sigma_prime_t);
    double sigma_tr = sqrt(3 * sigma_a * sigma_prime_t);

    double z_r = 1.0 / sigma_prime_t;
    double dSqr_r = Sqr(z_r) + distSqr;
    double d_r = sqrt(dSqr_r);
    double z_v = z_r * (1.0 + Aconst * 4.0/3.0);
    double dSqr_v = Sqr(z_v) + distSqr;
    double d_v = sqrt(dSqr_v);

    double common_term  = alpha_prime / (4.0 * M_PI);          // dimensionless
    double C1           = z_r * (sigma_tr + 1.0/d_r);          // dimensionless
    double C2           = z_v * (sigma_tr + 1.0/d_v);          // dimensionless
    double r_term       = C1 * exp(-sigma_tr * d_r) / dSqr_r;  // dimension 1/area
    double v_term       = C2 * exp(-sigma_tr * d_v) / dSqr_v;  // dimension 1/area

    Rd = common_term * (r_term + v_term);

#else
    // uniform illumination BRDF approximation
    // TODO - use this for radiosity?

    // calculate Rd
    double root_term = sqrt(3.0 * (1.0 - alpha_prime));
    Rd = (alpha_prime / 2.0) * (1 + exp((-4.0/3.0) * Aconst * root_term)) * exp(-root_term);

#endif

    // NOTE: We're leaving out the 1/pi factor because in POV-Ray, by convention,
    // light intensity is normalized to imply this factor already.
    sd = F * Rd; // (normally this would be F*Rd/M_PI)
    POV_SUBSURFACE_ASSERT((sd >= 0.0) && (sd <= DBL_MAX)); // verify sd is a non-negative, finite value (no #INF, no #IND, no #NAN)
}

void Trace::ComputeDiffuseContribution1(const LightSource& lightsource, const Intersection& out,
        const Vector3d& vOut, const Intersection& in, MathColour& Total_Colour,
        const PreciseMathColour& sigma_prime_s, const PreciseMathColour& sigma_a, double eta,
        double weight, TraceTicket& ticket)
{
    // TODO FIXME - part of this code is very alike to ComputeOneDiffuseLight()

    // Get a colour and a ray.
    Ray lightsourceray(ticket);
    double lightsourcedepth;
    MathColour lightcolour;
    ComputeOneLightRay(lightsource, lightsourcedepth, lightsourceray, in.IPoint, lightcolour, true);

    // Don't calculate spotlights when outside of the light's cone.
    if (lightcolour.IsNearZero(gkDBL_epsilon))
        return;

    Vector3d nIn = in.INormal;

    // See if light on far side of surface from camera.
    // [CLi] double_illuminate and diffuse backside illumination don't seem to make much sense with BSSRDF, so we ignore them here.
    // [CLi] BSSRDF always does "full area lighting", so we ignore it here.
    double cos_in = dot(nIn, lightsourceray.Direction);
    // [CLi] we're coming from inside the object, so the surface /must/ be properly oriented towards the camera; if it isn't,
    // it must be the normal's fault
    if (cos_in < 0)
    {
        nIn.invert();
        cos_in = -cos_in;
    }
    // [CLi] light coming in almost parallel to the surface is a problem though
    if (cos_in < gkDBL_epsilon)
        return;

    // If light source was not blocked by any intervening object, then
    // calculate it's contribution to the object's overall illumination.
    if (qualityFlags.shadows && ((lightsource.Projected_Through_Object != nullptr) || (lightsource.Light_Type != FILL_LIGHT_SOURCE)))
    {
        // [CLi] Not using lightColorCache because it's unsuited for BSSRDF
        TraceShadowRay(lightsource, lightsourcedepth, lightsourceray, in.IPoint, lightcolour);
    }

    // Don't calculate anything more if we're in full shadow
    if (lightcolour.IsNearZero(gkDBL_epsilon))
        return;

    lightcolour *= cos_in;
    for (int j = 0; j < MathColour::channels; j++)
    {
        double sd;
        ComputeDiffuseContribution(out, vOut, in.IPoint, nIn, lightsourceray.Direction, sd, sigma_prime_s[j], sigma_a[j], eta);
        sd *= weight;
        POV_SUBSURFACE_ASSERT(sd >= 0);
        lightcolour[j] *= sd;
        Total_Colour[j] += lightcolour[j];
    }
}

void Trace::ComputeDiffuseAmbientContribution1(const Intersection& out, const Vector3d& vOut, const Intersection& in,
        MathColour& Total_Colour, const PreciseMathColour& sigma_prime_s, const PreciseMathColour& sigma_a,
        double eta, double weight, TraceTicket& ticket)
{
#if 0
    // generate a random direction vector (using a distribution cosine-weighted along the normal)
    Vector3d axisU, axisV;
    ComputeSurfaceTangents(in.INormal, axisU, axisV);
    while (ssltCosWeightedDirectionGenerator.size() <= ticket.subsurfaceRecursionDepth)
        ssltCosWeightedDirectionGenerator.push_back(GetSubRandomCosWeightedDirectionGenerator(2, 32767));
    Vector3d direction = (*(ssltCosWeightedDirectionGenerator[ticket.subsurfaceRecursionDepth]))();
    double cos_in = direction.y(); // cosine of angle between normal and random vector
    Vector3d vIn = in.INormal*cos_in + axisU*direction.x() + axisV*direction.z();

    POV_SUBSURFACE_ASSERT(fabs(dot(in.INormal, axisU)) < gkDBL_epsilon);
    POV_SUBSURFACE_ASSERT(fabs(dot(in.INormal, axisV)) < gkDBL_epsilon);
    POV_SUBSURFACE_ASSERT(fabs(dot(axisU, axisV)) < gkDBL_epsilon);

    // [CLi] light coming in almost parallel to the surface is a problem
    if (cos_in < gkDBL_epsilon)
        return;

    Ray ambientray = Ray(ticket, in.IPoint, vIn, Ray::OtherRay); // TODO FIXME - [CLi] check whether ray type is suitable
    MathColour ambientcolour;
    ColourChannel dummyTransm;
    TraceRay(ambientray, ambientcolour, dummyTransm, weight, false);

    // Don't calculate anything more if there's no light input
    if (ambientcolour.IsNearZero(gkDBL_epsilon))
        return;

    for (int j = 0; j < 3; j++)
    {
        double sd;
        // Note: radiosity data is already cosine-weighted, so we're passing the surface normal as incident light direction
        ComputeDiffuseContribution(out, vOut, in.IPoint, in.INormal,  vIn, sd, sigma_prime_s[j], sigma_a[j], eta);
        sd *= 0.5/cos_in; // the distribution is cosine-weighted, but sd was computed assuming neutral weighting, so compensate
        sd *= weight;
        POV_SUBSURFACE_ASSERT(sd >= 0);
        ambientcolour[j] *= sd;
        POV_SUBSURFACE_ASSERT(ambientcolour[j] >= 0);
        Total_Colour[j] += ambientcolour[j];
    }
#else
    MathColour ambientcolour;
    // TODO FIXME - should support pertubed normals
    radiosity.ComputeAmbient(in.IPoint, in.INormal, in.INormal, 1.0 /* TODO - brilliance */, ambientcolour, weight, ticket);
    for (int j = 0; j < MathColour::channels; j++)
    {
        double sd;
        // Note: radiosity data is already cosine-weighted, so we're passing the surface normal as incident light direction
        ComputeDiffuseContribution(out, vOut, in.IPoint, in.INormal, in.INormal, sd, sigma_prime_s[j], sigma_a[j], eta);
        sd *= weight;
        POV_SUBSURFACE_ASSERT(sd >= 0);
        ambientcolour[j] *= sd;
        POV_SUBSURFACE_ASSERT(ambientcolour[j] >= 0);
        Total_Colour[j] += ambientcolour[j];
    }
#endif
}

void Trace::ComputeSubsurfaceScattering(const FINISH *Finish, const MathColour& layer_pigment_colour,
        const Intersection& out, Ray& Eye, const Vector3d& Layer_Normal, MathColour& Final_Colour,
        double Attenuation)
{
    int NumSamplesDiffuse = sceneData->subsurfaceSamplesDiffuse;
    int NumSamplesSingle  = sceneData->subsurfaceSamplesSingle;

    // TODO FIXME - this is hard-coded for now
    if (Eye.GetTicket().subsurfaceRecursionDepth >= 2)
        return;
    else if (Eye.GetTicket().subsurfaceRecursionDepth == 1)
    {
        NumSamplesDiffuse = 1;
        NumSamplesSingle  = 1;
        //NumSamplesDiffuse = (int)ceil(sqrt(NumSamplesDiffuse));
        //NumSamplesSingle  = (int)ceil(sqrt(NumSamplesSingle));
    }

    Eye.GetTicket().subsurfaceRecursionDepth++;

    LightSource Light_Source;

    Vector3d vOut = -Eye.Direction;

    MathColour Total_Colour;

    double eta;

    ComputeRelativeIOR(Eye, out.Object->interior.get(), eta);

#if 0
    // user setting specifies mean free path
    PreciseMathColour   alpha_prime     = object->interior->subsurface->GetReducedAlbedo(layer_pigment_colour * Finish->Diffuse);
    PreciseMathColour   sigma_tr        = 1.0 / PreciseMathColour(Finish->SubsurfaceTranslucency);

    PreciseMathColour   sigma_prime_t   = sigma_tr / sqrt(3*(1.0-alpha_prime));
    PreciseMathColour   sigma_prime_s   = alpha_prime * sigma_prime_t;
    PreciseMathColour   sigma_a         = sigma_prime_t - sigma_prime_s;
    PreciseMathColour   sigma_tr_sqr    = sigma_tr * sigma_tr;
#else
    // user setting specifies reduced scattering coefficient
    PreciseMathColour   alpha_prime     = out.Object->interior->subsurface->GetReducedAlbedo(layer_pigment_colour * Finish->Diffuse);
    PreciseMathColour   sigma_prime_s   = 1.0 / PreciseMathColour(Finish->SubsurfaceTranslucency);

    PreciseMathColour   sigma_prime_t   = sigma_prime_s / alpha_prime;
    PreciseMathColour   sigma_a         = sigma_prime_t - sigma_prime_s;
    PreciseMathColour   sigma_tr_sqr    = sigma_a * sigma_prime_t * 3.0;
    PreciseMathColour   sigma_tr        = Sqrt(sigma_tr_sqr);
#endif

    PreciseMathColour   g(0.0); // the mean cosine of the scattering angle; for isotropic scattering, g = 0
    PreciseMathColour   sigma_t_xo      = sigma_prime_t / (1.0-g);
    PreciseMathColour   sigma_s         = sigma_prime_s / (1.0-g);

#if 1

    // colour dependent diffuse contribution

    double sampleArea;
    double weight;
    double weightSum;
    double sigma_a_mean        = sigma_a.Greyscale(); // TODO FIXME - use a "fair" average of all three color channels
    double sigma_prime_s_mean  = sigma_prime_s.Greyscale(); // TODO FIXME - use a "fair" average of all three color channels
    double sigma_prime_t_mean  = sigma_a_mean + sigma_prime_s_mean;
    double sigma_tr_mean_sqr   = sigma_a_mean * sigma_prime_t_mean * 3.0;
    double sigma_tr_mean       = sqrt(sigma_tr_mean_sqr);
    int    trueNumSamples;

    bool radiosity_needed = (sceneData->radiositySettings.radiosityEnabled == true) &&
                            (sceneData->subsurfaceUseRadiosity == true) &&
                            (radiosity.CheckRadiosityTraceLevel(Eye.GetTicket()) == true) &&
                            (Test_Flag(out.Object, IGNORE_RADIOSITY_FLAG) == false);

    Vector3d sampleBase;
    ComputeDiffuseSampleBase(sampleBase, out, vOut, 1.0 / (sigma_prime_t_mean * sceneData->mmPerUnit), Eye.GetTicket());

    weightSum = 0.0;
    trueNumSamples = 0;

    for (int i = 0; i < NumSamplesDiffuse; i++)
    {
        Intersection in;
        ComputeDiffuseSamplePoint(sampleBase, in, sampleArea, Eye.GetTicket());

        // avoid pathological cases
        if (sampleArea != 0)
        {
            weight = sampleArea;
            weightSum += weight;
            trueNumSamples ++;

            if (IsSameSSLTObject(in.Object, out.Object))
            {
                // radiosity-alike ambient illumination
                if (radiosity_needed)
                {
                    // shoot just one random ray to account for ambient illumination (we're averaging stuff anyway)
                    ComputeDiffuseAmbientContribution1(out, vOut, in, Total_Colour, sigma_prime_s, sigma_a,
                            eta, weight, Eye.GetTicket());
                }

                // global light sources, if not turned off for this object
                if ((out.Object->Flags & NO_GLOBAL_LIGHTS_FLAG) != NO_GLOBAL_LIGHTS_FLAG)
                {
                    for (auto k = 0ul; k < threadData->lightSources.size(); k++)
                    {
                        ComputeDiffuseContribution1(*threadData->lightSources[k], out, vOut, in, Total_Colour,
                                sigma_prime_s, sigma_a, eta, weight, Eye.GetTicket());
                    }
                }

                // local light sources from a light group, if any
                if (!out.Object->LLights.empty())
                {
                    for (auto k = 0ul; k < out.Object->LLights.size(); k++)
                    {
                        ComputeDiffuseContribution1(*out.Object->LLights[k], out, vOut, in, Total_Colour,
                                sigma_prime_s, sigma_a, eta, weight, Eye.GetTicket());
                    }
                }
            }
            else
            {
                // TODO - what's the proper thing to do?
            }
        }
    }
    if (trueNumSamples > 0)
        Total_Colour /= trueNumSamples;

#endif

    Vector3d refractedEye;
    if (SSLTComputeRefractedDirection(Eye.Direction, out.INormal, 1.0/eta, refractedEye))
    {
        Ray refractedEyeRay(Eye.GetTicket(), out.IPoint, refractedEye);
        Intersection unscatteredIn;

        double dist;

        // find the intersection of the refracted ray with the object
        // find the distance to this intersection
        bool found = FindIntersection(unscatteredIn, refractedEyeRay);
        if (found)
            dist = (out.IPoint - unscatteredIn.IPoint).length() * sceneData->mmPerUnit;
        else
            dist = HUGE_VAL;

        double cos_out = dot(vOut, out.INormal);
        double cos_out_prime = sqrt(1 - ((Sqr(1.0 / eta)) * (1 - Sqr(cos_out))));
        double theta_out = acos(cos_out);

#if 1

        // colour dependent single scattering contribution

        for (int i = 0; i < NumSamplesSingle; i++)
        {
            for (int j = 0; j < MathColour::channels; j ++)
            {
                MathColour temp;
                ComputeSingleScatteringContribution(out, dist, theta_out, cos_out_prime, refractedEye,
                        sigma_t_xo[j], sigma_s[j], temp, eta, Eye.GetTicket());
                Total_Colour[j] += temp[j] / NumSamplesSingle;
            }
        }

#endif

#if 1

        // colour dependent unscattered contribution

        // Trace refracted ray.
        MathColour tempColour;
        ColourChannel tempTransm;

        // TODO FIXME - account for fresnel attenuation at interfaces
        PreciseMathColour att = Exp(-sigma_prime_t * dist); // TODO should be sigma_t
        weight = att.WeightMax();
        if (weight > Eye.GetTicket().adcBailout)
        {
            if (!found)
            {
                // TODO - trace the ray to the background?
            }
            else if (IsSameSSLTObject(unscatteredIn.Object, out.Object))
            {
                unscatteredIn.Object->Normal(unscatteredIn.INormal, &unscatteredIn, threadData);
                if (dot(refractedEye, unscatteredIn.INormal) > 0)
                    unscatteredIn.INormal.invert();
                Vector3d doubleRefractedEye;
                if (SSLTComputeRefractedDirection(refractedEye, unscatteredIn.INormal, eta, doubleRefractedEye))
                {
                    Ray doubleRefractedEyeRay(refractedEyeRay);
                    doubleRefractedEyeRay.SetFlags(Ray::RefractionRay, refractedEyeRay);
                    doubleRefractedEyeRay.Origin = unscatteredIn.IPoint;
                    doubleRefractedEyeRay.Direction = doubleRefractedEye;
                    TraceRay(doubleRefractedEyeRay, tempColour, tempTransm, weight, false);
                    Total_Colour += MathColour(PreciseMathColour(tempColour) * att);
                }
            }
            else
            {
                // TODO - trace the ray into that object (if it is transparent)
            }
        }

#endif

    }

    Final_Colour += Total_Colour;

    Eye.GetTicket().subsurfaceRecursionDepth--;
}

}
// end of namespace pov
