//******************************************************************************
///
/// @file core/shape/disc.cpp
///
/// Implementation of the disc geometric primitive.
///
/// @author Alexander Enzmann
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2021 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//------------------------------------------------------------------------------
// SPDX-License-Identifier: AGPL-3.0-or-later
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/shape/disc.h"

// C++ variants of C standard header files
// C++ standard header files
//  (none at the moment)

// POV-Ray header files (base module)
//  (none at the moment)

// POV-Ray header files (core module)
#include "core/bounding/boundingbox.h"
#include "core/math/matrix.h"
#include "core/render/ray.h"
#include "core/scene/tracethreaddata.h"
#include "core/support/statistics.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

const double DEPTH_TOLERANCE = 1.0e-6;


bool Disc::All_Intersections(const Ray& ray, IStack& Depth_Stack, TraceThreadData *Thread)
{
    int Intersection_Found;
    double Depth;
    Vector3d IPoint;

    Intersection_Found = false;

    Thread->Stats()[Ray_Disc_Tests]++;
    if (Intersect(ray, &Depth))
    {
        Thread->Stats()[Ray_Disc_Tests_Succeeded]++;
        IPoint = ray.Evaluate(Depth);

        if (Clip.empty() || Point_In_Clip(IPoint, Clip, Thread))
        {
            Depth_Stack->push(Intersection(Depth,IPoint,this));
            Intersection_Found = true;
        }
    }

    return (Intersection_Found);
}


bool Disc::Intersect(const BasicRay& ray, double *Depth) const
{
    double t, u, v, r2, len;
    Vector3d P, D;

    /* Transform the point into the discs space */

    MInvTransPoint(P, ray.Origin, Trans);
    MInvTransDirection(D, ray.Direction, Trans);

    len = D.length();
    D /= len;

    if (fabs(D[Z]) > EPSILON)
    {
        t = -P[Z] / D[Z];

        if (t >= 0.0)
        {
            u = P[X] + t * D[X];
            v = P[Y] + t * D[Y];

            r2 = Sqr(u) + Sqr(v);

            if ((r2 >= iradius2) && (r2 <= oradius2))
            {
                *Depth = t / len;

                if ((*Depth > DEPTH_TOLERANCE) && (*Depth < MAX_DISTANCE))
                    return (true);
            }
        }
    }

    return (false);
}


bool Disc::Inside(const Vector3d& IPoint, TraceThreadData *Thread) const
{
    Vector3d New_Point;

    MInvTransPoint(New_Point, IPoint, Trans);

    if (New_Point[Z] >= 0.0)
    {
        // We are outside.
        return (Test_Flag(this, INVERTED_FLAG));
    }
    else
    {
        // We are inside.
        return (!Test_Flag(this, INVERTED_FLAG));
    }
}


void Disc::Normal(Vector3d& Result, Intersection *, TraceThreadData *) const
{
    Result = normal;
}


void Disc::Translate(const Vector3d&, const TRANSFORM *tr)
{
    Transform(tr);
}


void Disc::Rotate(const Vector3d&, const TRANSFORM *tr)
{
    Transform(tr);
}


void Disc::Scale(const Vector3d&, const TRANSFORM *tr)
{
    Transform(tr);
}


void Disc::Transform(const TRANSFORM *tr)
{
    MTransNormal(normal, normal, tr);

    normal.normalize();

    Compose_Transforms(Trans, tr);

    // Recalculate the bounds
    Compute_BBox();
}


Disc::Disc() : ObjectBase(DISC_OBJECT)
{
    center = Vector3d(0.0, 0.0, 0.0);
    normal = Vector3d(0.0, 0.0, 1.0);

    iradius2 = 0.0;
    oradius2 = 1.0;

    d = 0.0;

    Trans = Create_Transform();

    /* Default bounds */

    Make_BBox(BBox, -1.0, -1.0, -gkMinIsectDepthReturned, 2.0,  2.0, 2.0 * gkMinIsectDepthReturned);
}


// Sep 1994 : Fixed memory leakage [DB]
//
ObjectPtr Disc::Copy()
{
    Disc *New = new Disc();
    Destroy_Transform(New->Trans);
    *New = *this;
    New->Trans = Copy_Transform(Trans);

    return (New);
}


Disc::~Disc()
{}


// Author Dieter Bayer
//
// Calculate the transformation that scales, rotates, and translates
// the disc to the desired location and orientation.
//
//  Aug 1994 : Creation.
//
void Disc::Compute_Disc()
{
    Compute_Coordinate_Transform(Trans, center, normal, 1.0, 1.0);

    Compute_BBox();
}


// Author Dieter Bayer
//
// Aug 1994 : Creation.
//
void Disc::Compute_BBox()
{
    double rad;

    rad = sqrt(oradius2);

    Make_BBox(BBox, -rad, -rad, -gkMinIsectDepthReturned,
              2.0*rad, 2.0*rad, 2.0*gkMinIsectDepthReturned);

    Recompute_BBox(&BBox, Trans);
}

}
// end of namespace pov
