//******************************************************************************
///
/// @file core/shape/spheresweep.cpp
///
/// Implementation of the sphere sweep geometric primitive.
///
/// Idea: J.J. Van Wijk, "Raytracing Objects Defined by Sweeping a Sphere",
/// Eurographics 1984.
///
/// @author Jochen Lippert
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2021 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//------------------------------------------------------------------------------
// SPDX-License-Identifier: AGPL-3.0-or-later
//******************************************************************************

//  Version history (most recent changes first):
//
//  24. Feb. 1998: Fixed a bug in Compute_Sphere_Sweep_BBox that could
//                 result in much too large bounding boxes for
//                 Catmull-Rom-Spline Sphere Sweeps. Added statistics
//                 output for Sphere Sweep primitive.
//
//  19. Jan. 1998: Corrected a problem with the calculation of single
//                 spheres in Compute_Sphere_Sweep which I introduced
//                 in the previous version, and cleaned up the code in
//                 this function a bit. All_Sphere_Sweep_Intersections
//                 now uses QSORT to sort the intersection list.
//
//  6.  Jan. 1998: Fixed a bug that affected the geometry of
//                 Catmull-Rom-Spline Sphere Sweeps, fixed a bug in
//                 Inside_Sphere_Sweep which affected non-proportionally
//                 scaled Sphere Sweeps.
//
//  22. Dec. 1997: Simplyfied code for finding invalid intersections,
//                 fixed some bugs related to this.
//
//  8.  Dec. 1997: Some cleanup & bug fixes.
//
//  1.  Dec. 1997: Made depth tolerance user-selectable.
//
//  29. Nov. 1997: Added support for Catmull-Rom-Splines.
//
//  26. Nov. 1997: Fixed a bug where the wrong value was returned by
//                 All_Sphere_Sweep_Intersections when the clipped_by
//                 statement was used with a sphere sweep.
//
//  24. Nov. 1997: Fixed a pretty silly bug in Copy_Sphere_Sweep where
//                 the sphere sweep wasn't copied completely.
//
//  21. Nov. 1997: Origin release.
//

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/shape/spheresweep.h"

// C++ variants of C standard header files
#include <cstdlib>

// C++ standard header files
#include <algorithm>

// POV-Ray header files (base module)
#include "base/povassert.h"

// POV-Ray header files (core module)
#include "core/bounding/boundingbox.h"
#include "core/math/matrix.h"
#include "core/math/polynomialsolver.h"
#include "core/render/ray.h"
#include "core/scene/tracethreaddata.h"
#include "core/support/statistics.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

// Local preprocessor defines

const double ZERO_TOLERANCE   = 1.0e-4;
const double RADIUS_TOLERANCE = 1.0e-6;

// TODO - Technically this should be `(Num_Spheres + Num_Segments * order) * 2` [CLi]
#define SPHSWEEP_MAX_ISECT  ((Num_Spheres + Num_Segments) * 2 + 64)


// Local variables

const MATRIX Catmull_Rom_Matrix =
{
    { 0.0 / 2.0,  2.0 / 2.0,  0.0 / 2.0,  0.0 / 2.0},
    {-1.0 / 2.0,  0.0 / 2.0,  1.0 / 2.0,  0.0 / 2.0},
    { 2.0 / 2.0, -5.0 / 2.0,  4.0 / 2.0, -1.0 / 2.0},
    {-1.0 / 2.0,  3.0 / 2.0, -3.0 / 2.0,  1.0 / 2.0}
};

const MATRIX B_Matrix =
{
    { 1.0 / 6.0,  4.0 / 6.0,  1.0 / 6.0,  0.0 / 6.0},
    {-3.0 / 6.0,  0.0 / 6.0,  3.0 / 6.0,  0.0 / 6.0},
    { 3.0 / 6.0, -6.0 / 6.0,  3.0 / 6.0,  0.0 / 6.0},
    {-1.0 / 6.0,  3.0 / 6.0, -3.0 / 6.0,  1.0 / 6.0}
};


bool SphereSweep::All_Intersections(const Ray& ray, IStack& Depth_Stack, TraceThreadData *Thread)
{
    // TODO - To improve performance, we might use thread-local buffers for all sphere sweeps.
    SPHSWEEP_INT *Isect = reinterpret_cast<SPHSWEEP_INT *>(POV_MALLOC(sizeof(SPHSWEEP_INT) * SPHSWEEP_MAX_ISECT, "sphere sweep intersections"));
    SPHSWEEP_INT *Sphere_Isect = reinterpret_cast<SPHSWEEP_INT *>(POV_MALLOC(sizeof(SPHSWEEP_INT) * 2 * Num_Spheres, "Sphere sweep sphere intersections"));
    SPHSWEEP_INT *Segment_Isect = reinterpret_cast<SPHSWEEP_INT *>(POV_MALLOC(sizeof(SPHSWEEP_INT) * 12 * Num_Segments, "Sphere sweep segment intersections"));
    BasicRay     New_Ray;
    double       len;
    bool         Intersection_Found = false;
    int          Num_Isect = 0;
    int          Num_Seg_Isect;
    int          i, j;

    Thread->Stats()[Ray_Sphere_Sweep_Tests]++;

    if (Trans == nullptr)
    {
        New_Ray.Origin = ray.Origin;
        New_Ray.Direction = ray.Direction;
        len = 1.0;
    }
    else
    {
        MInvTransRay(New_Ray, ray, Trans);

        len = New_Ray.Direction.length();
        New_Ray.Direction /= len;
    }

    // Intersections with single spheres
    for (i = 0; i < Num_Spheres; i++)
    {
        // Test for end of vector
        if (Num_Isect + 2 <= SPHSWEEP_MAX_ISECT)
        {
            // Are there intersections with this sphere?
            if (Intersect_Sphere(New_Ray, &Sphere[i], Isect + Num_Isect))
                Num_Isect += 2;
        }
    }

    // Intersections with segments
    for (i = 0; i < Num_Segments; i++)
    {
        // Are there intersections with this segment?
        Num_Seg_Isect = Intersect_Segment(New_Ray, &Segment[i], Segment_Isect, Thread->Stats());

        // Test for end of vector
        if (Num_Isect + Num_Seg_Isect <= SPHSWEEP_MAX_ISECT)
        {
            for (j = 0; j < Num_Seg_Isect; j++)
            {
                // Add intersection
                Isect[Num_Isect] = Segment_Isect[j];
                Num_Isect++;
            }
        }
    }

    // Any intersections?
    if (Num_Isect > 0)
    {
        // Sort intersections
        std::qsort(Isect, Num_Isect, sizeof(SPHSWEEP_INT), Comp_Isects);

        // Delete invalid intersections inside the sphere sweep
        Num_Isect = Find_Valid_Points(Isect, Num_Isect, New_Ray);

        // Push valid intersections
        for (i = 0; i < Num_Isect; i++)
        {
            if ((Isect[i].t > gkMinIsectDepthReturned) && (Isect[i].t < MAX_DISTANCE))
            {
                // Was the ray transformed?
                if (Trans != nullptr)
                {
                    // Yes, invert the transformation
                    Isect[i].t /= len;
                    MTransPoint(Isect[i].Point, Isect[i].Point, Trans);
                    MTransNormal(Isect[i].Normal, Isect[i].Normal, Trans);
                    Isect[i].Normal.normalize();
                }

                // Test for clipping volume
                if (Clip.empty() || Point_In_Clip(Isect[i].Point, Clip, Thread))
                {
                    Depth_Stack->push(Intersection(Isect[i].t, Isect[i].Point, Isect[i].Normal, this));
                    Intersection_Found = true;
                }
            }
        }

        if (Intersection_Found)
            Thread->Stats()[Ray_Sphere_Sweep_Tests_Succeeded]++;
    }

    POV_FREE(Isect);
    POV_FREE(Sphere_Isect);
    POV_FREE(Segment_Isect);

    return Intersection_Found;
}



bool SphereSweep::Intersect_Sphere(const BasicRay &ray, const SPHSWEEP_SPH *Sphere,
        SPHSWEEP_INT *Inter)
{
    double   Radius2;
    double   OCSquared;
    double   t_Closest_Approach;
    double   Half_Chord;
    double   t_Half_Chord_Squared;
    Vector3d Origin_To_Center;

    Radius2 = Sqr(Sphere->Radius);

    Origin_To_Center = Sphere->Center - ray.Origin;

    OCSquared = Origin_To_Center.lengthSqr();

    t_Closest_Approach = dot(Origin_To_Center, ray.Direction);

    if ((OCSquared >= Radius2) && (t_Closest_Approach < gkDBL_epsilon))
        return false;

    t_Half_Chord_Squared = Radius2 - OCSquared + Sqr(t_Closest_Approach);

    // TODO Get clipping if parse time scale values for sweep small. Use gkDBL_epsilon instead?
    if (t_Half_Chord_Squared > gkMinIsectDepthReturned)
    {
        Half_Chord = sqrt(t_Half_Chord_Squared);

        // Calculate smaller depth

        Inter[0].t = t_Closest_Approach - Half_Chord;

        // Calculate point
        Inter[0].Point = ray.Evaluate(Inter[0].t);

        // Calculate normal
        Inter[0].Normal = (Inter[0].Point - Sphere->Center) / fabs(Sphere->Radius);

        // Calculate bigger depth
        Inter[1].t = t_Closest_Approach + Half_Chord;

        // Calculate point
        Inter[1].Point = ray.Evaluate(Inter[1].t);

        // Calculate normal
        Inter[1].Normal = (Inter[1].Point - Sphere->Center) / fabs(Sphere->Radius);

        return true;
    }

    return false;
}


/// @bug
///     The current implementation exhibits numeric instabilities for 3rd order polynomial splines.
///     (See GitHub issue #147.)
///
int SphereSweep::Intersect_Segment(const BasicRay &ray, const SPHSWEEP_SEG *Segment,
        SPHSWEEP_INT *Isect, RenderStatistics& stats)
{
    int      Isect_Count;
    double   Dot1, Dot2;
    double   t1, t2;
    Vector3d Vector;
    Vector3d IPoint;
    Vector3d DCenter;
    double   DCenterDot;
    double   temp;
    double   b, c, d, e, f, g, h, i, j, k, l;
    double   Coef[11];
    double   Root[10];
    int      Num_Poly_Roots, m, n;
    double   fp0, fp1;
    double   t;

    SPHSWEEP_SPH Temp_Sphere;

    Isect_Count = 0;

    // Calculate intersections with closing discs near the ends of the spline segment.
    // Note that the positions of the closing discs account for the rate of change in radius
    // at the respective end of the spline segment.

    // We need to compute these intersections even though we will inevitably discard them later,
    // because the algorithm downstream expects intersections to always come in pairs, one entering
    // the segment and the other one exiting it. The intersections with the closing discs will
    // take the place of any intersections we discard because they fall outside the interval [0,1].
    // (Also, if the spline happens to be parallel to the ray direction, we will get no roots
    // at all but still need to have these intersections.)

    // Closing disk near u=1.
    Dot1 = dot(ray.Direction, Segment->Center_Deriv[0]);
    if (fabs(Dot1) > EPSILON)
    {
        Vector = ray.Origin - Segment->Closing_Sphere[0].Center;
        Dot2 = dot(Vector, Segment->Center_Deriv[0]);
        t1 = -(Dot2 + Segment->Closing_Sphere[0].Radius * Segment->Radius_Deriv[0]) / Dot1;
        if ((t1 > -MAX_DISTANCE) && (t1 < MAX_DISTANCE))
        {
            // Calculate point
            IPoint = ray.Evaluate(t1);

            // Is the point inside the closing sphere?
            DCenter = IPoint - Segment->Closing_Sphere[0].Center;
            DCenterDot = DCenter.lengthSqr();
            if (DCenterDot < Sqr(Segment->Closing_Sphere[0].Radius))
            {
                // Add intersection
                Isect[Isect_Count].t = t1;

                // Copy point
                Isect[Isect_Count].Point = IPoint;

                // Calculate normal
                Isect[Isect_Count].Normal = -Segment->Center_Deriv[0].normalized();

                Isect_Count++;
            }
        }
    }

    // Closing disk near u=1.
    Dot1 = dot(ray.Direction, Segment->Center_Deriv[1]);
    if (fabs(Dot1) > EPSILON)
    {
        Vector = ray.Origin - Segment->Closing_Sphere[1].Center;
        Dot2 = dot(Vector, Segment->Center_Deriv[1]);
        t2 = -(Dot2 + Segment->Closing_Sphere[1].Radius * Segment->Radius_Deriv[1]) / Dot1;
        if ((t2 > -MAX_DISTANCE) && (t2 < MAX_DISTANCE))
        {
            // Calculate point
            IPoint = ray.Evaluate(t2);

            // Is the point inside the closing sphere?
            DCenter = IPoint - Segment->Closing_Sphere[1].Center;
            DCenterDot = DCenter.lengthSqr();
            if (DCenterDot < Sqr(Segment->Closing_Sphere[1].Radius))
            {
                // Add intersection
                Isect[Isect_Count].t = t2;

                // Copy point
                Isect[Isect_Count].Point = IPoint;

                // Calculate normal
                Isect[Isect_Count].Normal = Segment->Center_Deriv[1].normalized();

                Isect_Count++;
            }
        }
    }

    auto lambdaQuadraticPolyVal = [](const double t, const double *x) -> double
    {
         double pv = x[0];
         for (int k=1; k<=2; k++)
         {
             pv = pv * t + x[k];
         }
         return(pv);
    };

    auto lambdaTenthOrderPolyVal = [](const double t, const double *x) -> double
    {
         double pv = x[0];
         for (int k=1; k<=10; k++)
         {
             pv = pv * t + x[k];
         }
         return(pv);
    };

    auto lambdaPolynomialFwdDeflation = [](const int order, const double t, const double *x, double *y) -> void
    {
         y[0] = x[0];
         for (int k=0; k<order; k++)
         {
             y[k+1] = t * y[k] + x[k+1];
         }
    };

    // Calculate intersections with sides of the segment

    switch (Segment->Num_Coefs)
    {
        case 2:   // First order Polynomial

            Vector = ray.Origin - Segment->Center_Coef[0];

            // a is always 1.0

            b = dot(Segment->Center_Coef[1], ray.Direction);
            b *= -2.0;

            c = dot(Vector, ray.Direction);
            c *= 2.0;

            d = Segment->Center_Coef[1].lengthSqr();
            d -= Sqr(Segment->Radius_Coef[1]);

            e = dot(Vector, Segment->Center_Coef[1]);
            e += Segment->Radius_Coef[0] * Segment->Radius_Coef[1];
            e *= -2.0;

            f = Vector.lengthSqr();
            f -= Sqr(Segment->Radius_Coef[0]);

            Coef[0] = 4.0 * Sqr(d) - Sqr(b) * d;
            Coef[1] = 4.0 * d * e - 2.0 * b * c * d;
            Coef[2] = Sqr(e) - b * c * e + Sqr(b) * f;

            if (Test_Flag(this, STURM_FLAG))
            {
                Num_Poly_Roots = polysolve(2, Coef, Root, 0.0, 1.0);

                // First normalize coefficients relative to their range. Differences
                // in scene set up and scale create differences in evaluated values
                // as the solvers run. Normalizing to range for better solver accuracy
                // where determinations based on polynomial value at given t and not
                // distance along ray / change in t.

                double coefMax   = std::fmax(std::fmax(Coef[0],Coef[1]),Coef[2]);
                double coefMin   = std::fmin(std::fmin(Coef[0],Coef[1]),Coef[2]);
                double coefRange = (coefMax - coefMin) / 2.0;
                Coef[0] /= coefRange;
                Coef[1] /= coefRange;
                Coef[2] /= coefRange;

                // Use 1st derivative of the polynomial to look for double roots collapsed
                // to mutiplicity of two. In other words where the polynomial parabola
                // just touches, but does not cross so as to create both +- values. This degeneracy
                // happens given certain spacial and ray alignments with respect to the spline.

                double CoefD[2];
                double tD;
                CoefD[0] = 2.0 * Coef[0];
                CoefD[1] = Coef[1];

                if (CoefD[0] != 0.0)
                {
                    tD = -CoefD[1] / CoefD[0];
                    if ((tD >= 0.0) && (tD <= 1.0))
                    {
                        temp = lambdaQuadraticPolyVal(tD,Coef);
                        if (fabs(temp) < gkDBL_epsilon*10.0)
                        {
                            Root[0] = tD;
                            Num_Poly_Roots = 1;
                        }
                    }
                }
            }
            else
            {
                Num_Poly_Roots = solve_quadratic(Coef, Root);
            }
            break;

        case 4:   // Third order polynomial

            Vector = ray.Origin - Segment->Center_Coef[0];

            // a is always 1.0

            b = -2.0 * dot(Segment->Center_Coef[3], ray.Direction);

            c = -2.0 * dot(Segment->Center_Coef[2], ray.Direction);

            d = -2.0 * dot(Segment->Center_Coef[1], ray.Direction);

            e =  2.0 * dot(Vector, ray.Direction);

            f = Segment->Center_Coef[3].lengthSqr();
            f -= Sqr(Segment->Radius_Coef[3]);

            g = dot(Segment->Center_Coef[3], Segment->Center_Coef[2]);
            g -= Segment->Radius_Coef[3] * Segment->Radius_Coef[2];
            g *= 2.0;

            h = dot(Segment->Center_Coef[3], Segment->Center_Coef[1]);
            h *= 2.0;
            temp = dot(Segment->Center_Coef[2], Segment->Center_Coef[2]);
            h += temp;
            h -= 2.0 * Segment->Radius_Coef[3] * Segment->Radius_Coef[1];
            h -= Sqr(Segment->Radius_Coef[2]);

            i = dot(Segment->Center_Coef[3], Vector);
            temp = dot(Segment->Center_Coef[2], Segment->Center_Coef[1]);
            i -= temp;
            i += Segment->Radius_Coef[3] * Segment->Radius_Coef[0];
            i += Segment->Radius_Coef[2] * Segment->Radius_Coef[1];
            i *= -2.0;

            j = dot(Segment->Center_Coef[2], Vector);
            j += Segment->Radius_Coef[2] * Segment->Radius_Coef[0];
            j *= -2.0;
            temp = dot(Segment->Center_Coef[1], Segment->Center_Coef[1]);
            j += temp;
            j -= Sqr(Segment->Radius_Coef[1]);

            k = dot(Segment->Center_Coef[1], Vector);
            k += Segment->Radius_Coef[1] * Segment->Radius_Coef[0];
            k *= -2.0;

            l = Vector.lengthSqr();
            l -= Sqr(Segment->Radius_Coef[0]);

            Coef[0] = 36.0 * Sqr(f) - 9.0 * f * Sqr(b);
            Coef[1] = 60.0 * f * g - 6.0 * g * Sqr(b) - 18.0 * b * c * f;
            Coef[2] = 48.0 * f * h + 25.0 * Sqr(g) - 3.0 * h * Sqr(b)
                    - 13.0 * b * c * g - 8.0 * f * Sqr(c) - 18.0 * b * d * f;
            Coef[3] = 36.0 * f * i + 40.0 * g * h - 18.0 * b * f * e - 8.0 * b * c * h
                    - 6.0 * g * Sqr(c) - 14.0 * b * d * g - 14.0 * c * d * f;
            Coef[4] = 24.0 * f * j + 30.0 * g * i + 16.0 * Sqr(h) - 15.0 * b * g * e
                    - 12.0 * c * f * e + 3.0 * j * Sqr(b) - 3.0 * b * c * i - 4.0 * h * Sqr(c)
                    - 10.0 * b * d * h - 11.0 * c * d * g - 5.0 * f * Sqr(d);
            Coef[5] = 12.0 * f * k + 20.0 * g * j + 24.0 * h * i - 12.0 * b * h * e
                    - 10.0 * c * g * e - 6.0 * d * f * e + 6.0 * k * Sqr(b) + 2.0 * b * c * j
                    - 2.0 * i * Sqr(c) - 6.0 * b * d * i - 8.0 * c * d * h - 4.0 * g * Sqr(d);
            Coef[6] = 10.0 * g * k + 16.0 * h * j + 9.0 * Sqr(i) - 9.0 * b * i * e
                    - 8.0 * c * h * e - 5.0 * d * g * e + 9.0 * l * Sqr(b) + 7.0 * b * c * k
                    - 2.0 * b * d * j - 5.0 * c * d * i - 3.0 * h * Sqr(d);
            Coef[7] = 8.0 * h * k + 12.0 * i * j - 6.0 * b * j * e - 6.0 * c * i * e
                    - 4.0 * d * h * e + 12.0 * b * c * l + 2.0 * k * Sqr(c) + 2.0 * b * d * k
                    - 2.0 * c * d * j - 2.0 * i * Sqr(d);
            Coef[8] = 6.0 * i * k + 4.0 * Sqr(j) - 3.0 * b * k * e - 4.0 * c * j * e
                    - 3.0 * d * i * e + 4.0 * l * Sqr(c) + 6.0 * b * d * l
                    + c * d * k - j * Sqr(d);
            Coef[9] = 4.0 * j * k - 2.0 * c * k * e - 2.0 * d * j * e + 4.0 * c * d * l;
            Coef[10] = Sqr(k) - d * k * e + l * Sqr(d);

            if (Test_Flag(this, STURM_FLAG))
            {
                // See case2 comments above for following range normalization reasons.
                double coefMax, coefMin, coefRange, CoefN[11];
                coefMin = coefMax = Coef[0];
                for (int z = 1; z <= 10; z++)
                {
                    coefMax = std::fmax(coefMax,Coef[z]);
                    coefMin = std::fmin(coefMin,Coef[z]);
                }
                coefRange = (coefMax - coefMin) / 2.0;
                for (int z = 0; z <= 10; z++)
                {
                    CoefN[z] = Coef[z] / coefRange;
                }

                int Num_Poly_RootsD = 0;
                double CoefD[10], RootD[9];
                double vala, valb;

                // Look for collapsed roots using polynomial's derivative...
                CoefD[9] = CoefN[9];
                for (int z = 0; z < 9; z++)
                {
                    CoefD[z] = (double)(10 - z) * CoefN[z];
                }

                Num_Poly_Roots = 0;
                Num_Poly_RootsD = polysolve(9, CoefD, RootD, 0.0, 1.0);
                double lastRootDPolyVal = 0.0;
                bool sawDerivativeRootValSignChange = false;
                for (int rn = 0; rn < Num_Poly_RootsD; rn++)
                {
                    temp = lambdaTenthOrderPolyVal(RootD[rn],CoefN);

                    if ((lastRootDPolyVal * temp) < 0.0)
                    {
                        sawDerivativeRootValSignChange = true;
                    }
                    lastRootDPolyVal = temp;

                    if (fabs(temp) < gkDBL_epsilon*10.0)
                    {
                     // vala = lambdaTenthOrderPolyVal(RootD[rn]-gkDBL_epsilon,CoefN);
                     // valb = lambdaTenthOrderPolyVal(RootD[rn]+gkDBL_epsilon,CoefN);

                     // if ((fabs(val) > fabs(temp)) && ((val * temp) >= 0.0))
                     // {
                     //     val = lambdaTenthOrderPolyVal(RootD[rn]+gkDBL_epsilon,CoefN);
                     //     if ((fabs(val) > fabs(temp)) && ((val * temp) >= 0.0))
                     //     {
                                Root[Num_Poly_Roots] = RootD[rn];
                                Num_Poly_Roots++;
                     //     }
                     // }
                    }
                }

                // Look to deflate polynomial where roots found by derivative roots.
                int order = 10;
                for (int rn = 0; rn < Num_Poly_Roots; rn++)
                {
                   lambdaPolynomialFwdDeflation(order,Root[rn],Coef,CoefN);
                   order--;
                   for (int ii = 0; ii <= order; ii++)
                   {
                       Coef[ii] = CoefN[ii];
                   }
                }

                // @todo look to restore bezier_01 like optimization for arbitrary orders.
                //
                Num_Poly_Roots = polysolve(order, Coef, Root, 0.0, 1.0);

                // @todo Derivative roots sign change seen not handled... Do something smarter.
                //
             // if ((Num_Poly_Roots == 0) && (sawDerivativeRootValSignChange))
             // {
             //     lastRootDPolyVal = lambdaTenthOrderPolyVal(RootD[0],CoefN);
             //     for (int rn = 1; rn < Num_Poly_RootsD; rn++)
             //     {
             //         temp = lambdaTenthOrderPolyVal(RootD[rn],CoefN);
             //
             //         if ((lastRootDPolyVal * temp) < 0.0)
             //         {
             //             // What might that smarter thing be... regula falsi or...
             //         }
             //         lastRootDPolyVal = temp;
             //     }
             // }
            }
            else
            {
                Num_Poly_Roots = bezier_01(10, Coef, Root);
              //Num_Poly_Roots = polysolve(10, Coef, Root, 0.0, 1.0);  // Use to bypass bezier_01
            }

            break;

        default:
            POV_SHAPE_ASSERT(false);
            break;
    }

    // Remove roots not in interval [0, 1].

    // Note that the corresponding intersections are either removed in balanced pairs, or are
    // balanced by intersections with the closing discs.

    m = 0;
    while (m < Num_Poly_Roots)
    {
        if (Root[m] < 0.0 || Root[m] > 1.0)
        {
            for (n = m; n < Num_Poly_Roots - 1; n++)
                Root[n] = Root[n + 1];
            Num_Poly_Roots--;
        }
        else
            m++;
    }

    for (m = 0; m < Num_Poly_Roots; m++)
    {
        // Calculate center and radius of sphere at root

        double rootPwr = 1;
        Temp_Sphere.Center = Segment->Center_Coef[0];
        Temp_Sphere.Radius = Segment->Radius_Coef[0];
        for (n = 1; n < Segment->Num_Coefs; ++n)
        {
            rootPwr *= Root[m];
            Temp_Sphere.Center += Segment->Center_Coef[n] * rootPwr;
            Temp_Sphere.Radius += Segment->Radius_Coef[n] * rootPwr;
        }

        switch (Segment->Num_Coefs)
        {
            case 2:

                fp0 = 2.0 * d * Root[m]
                    + e;
                fp1 = b;
                break;

            case 4:

                fp0 = 6.0 * f * Root[m] * Root[m] * Root[m] * Root[m] * Root[m]
                    + 5.0 * g * Root[m] * Root[m] * Root[m] * Root[m]
                    + 4.0 * h * Root[m] * Root[m] * Root[m]
                    + 3.0 * i * Root[m] * Root[m]
                    + 2.0 * j * Root[m]
                    + k;
                fp1 = 3.0 * b * Root[m] * Root[m]
                    + 2.0 * c * Root[m]
                    + d;
                break;

            default:
                POV_SHAPE_ASSERT(false);
                break;
        }

        bool doubleIsect = (fabs(fp1) < ZERO_TOLERANCE);
        if (!doubleIsect)
        {
            // Looks like a single-intersection root at first glance

            t = -fp0 / fp1;

            if ((t > -MAX_DISTANCE) && (t < MAX_DISTANCE))
            {
                // Calculate intersection point
                Isect[Isect_Count].Point = ray.Evaluate(t);

                // Calculate normal
                Isect[Isect_Count].Normal = (Isect[Isect_Count].Point - Temp_Sphere.Center).normalized();

                // Sanity-check intersection point for proper radius
                double rIsect = (Isect[Isect_Count].Point - Temp_Sphere.Center).length();
                double rRoot  = Temp_Sphere.Radius;

                // Check for numeric instability; this may (only?) happen if the root corresponds to two intersections
                doubleIsect = fabs(rIsect - rRoot) > RADIUS_TOLERANCE;

                if (!doubleIsect)
                {
                    // Add intersection
                    Isect[Isect_Count].t = t;
                    Isect_Count++;
                }
            }
        }

        if (doubleIsect)
        {
            // Calculate intersections
            if (Intersect_Sphere(ray, &Temp_Sphere, Isect + Isect_Count))
                Isect_Count += 2;
        }
    }

    return Isect_Count;
}


bool SphereSweep::Inside(const Vector3d& IPoint, TraceThreadData *Thread) const
{
    int      inside;
    Vector3d New_Point;
    int      i, j;
    Vector3d Vector;
    double   temp;
    double   Coef[7];
    double   Root[6];
    int      Num_Poly_Roots;

    inside = false;

    if (Trans == nullptr)
        New_Point = IPoint;
    else
        MInvTransPoint(New_Point, IPoint, Trans);

    switch (Interpolation)
    {
        case LINEAR_SPHERE_SWEEP:
            // For each segment...
            for (i = 0; i < Num_Segments; i++)
            {
                // Pre-calculate vector
                Vector = New_Point - Segment[i].Center_Coef[0];

                // Coefficient for u^2
                Coef[0] = Segment[i].Center_Coef[1].lengthSqr();
                Coef[0] -= Sqr(Segment[i].Radius_Coef[1]);

                // Coefficient for u^1
                Coef[1] = dot(Vector, Segment[i].Center_Coef[1]);
                Coef[1] += Segment[i].Radius_Coef[0]
                         * Segment[i].Radius_Coef[1];
                Coef[1] *= -2.0;

                // Coefficient for u^0
                Coef[2] = Vector.lengthSqr();
                Coef[2] -= Sqr(Segment[i].Radius_Coef[0]);

                // Find roots
                Num_Poly_Roots = solve_quadratic(Coef, Root);

                // Test for interval [0, 1]
                for (j = 0; j < Num_Poly_Roots; j++)
                {
                    if (Root[j] >= 0.0 && Root[j] <= 1.0)
                    {
                        // At least one root inside interval, so we are inside sphere sweep
                        inside = true;
                        break;
                    }

                    if (inside)
                    {
                        break;
                    }
                }
            }
            break;
        case CATMULL_ROM_SPLINE_SPHERE_SWEEP:
        case B_SPLINE_SPHERE_SWEEP:
            // For each segment...
            for (i = 0; i < Num_Segments; i++)
            {
                // Pre-calculate vector
                Vector = New_Point - Segment[i].Center_Coef[0];

                // Coefficient for u^6
                Coef[0] = Segment[i].Center_Coef[3].lengthSqr();
                Coef[0] -= Sqr(Segment[i].Radius_Coef[3]);

                // Coefficient for u^5
                Coef[1] = dot(Segment[i].Center_Coef[3],
                              Segment[i].Center_Coef[2]);
                Coef[1] -= Segment[i].Radius_Coef[3]
                         * Segment[i].Radius_Coef[2];
                Coef[1] *= 2.0;

                // Coefficient for u^4
                Coef[2] = dot(Segment[i].Center_Coef[3],
                              Segment[i].Center_Coef[1]);
                Coef[2] *= 2.0;
                temp = Segment[i].Center_Coef[2].lengthSqr();
                Coef[2] += temp;
                Coef[2] -= 2.0 * Segment[i].Radius_Coef[3]
                               * Segment[i].Radius_Coef[1];
                Coef[2] -= Sqr(Segment[i].Radius_Coef[2]);

                // Coefficient for u^3
                Coef[3] = dot(Segment[i].Center_Coef[3], Vector);
                temp = dot(Segment[i].Center_Coef[2],
                           Segment[i].Center_Coef[1]);
                Coef[3] -= temp;
                Coef[3] += Segment[i].Radius_Coef[3]
                         * Segment[i].Radius_Coef[0];
                Coef[3] += Segment[i].Radius_Coef[2]
                         * Segment[i].Radius_Coef[1];
                Coef[3] *= -2.0;

                // Coefficient for u^2
                Coef[4] = dot(Segment[i].Center_Coef[2], Vector);
                Coef[4] += Segment[i].Radius_Coef[2]
                         * Segment[i].Radius_Coef[0];
                Coef[4] *= -2.0;
                temp = Segment[i].Center_Coef[1].lengthSqr();
                Coef[4] += temp;
                Coef[4] -= Sqr(Segment[i].Radius_Coef[1]);

                // Coefficient for u^1
                Coef[5] = dot(Segment[i].Center_Coef[1], Vector);
                Coef[5] += Segment[i].Radius_Coef[1]
                         * Segment[i].Radius_Coef[0];
                Coef[5] *= -2.0;

                // Coefficient for u^0
                Coef[6] = Vector.lengthSqr();
                Coef[6] -= Sqr(Segment[i].Radius_Coef[0]);

                // Find roots
                Num_Poly_Roots = bezier_01(6, Coef, Root);
              //Num_Poly_Roots = polysolve(6, Coef, Root, 0.0, 1.0); // Use to bypass bezier_01

                // Test for interval [0, 1]
                for (j = 0; j < Num_Poly_Roots; j++)
                {
                    if (Root[j] >= 0.0 && Root[j] <= 1.0)
                    {
                        // At least one root inside interval, so we are inside the sphere sweep
                        inside = true;
                        break;
                    }
                }

                if (inside)
                {
                    break;
                }
            }
            break;
    }

    if (Test_Flag(this, INVERTED_FLAG))
        inside = !inside;

    return inside;
}


void SphereSweep::Normal(Vector3d& Result, Intersection *Inter, TraceThreadData *) const
{
    Result = Inter->INormal;
}


ObjectPtr SphereSweep::Copy()
{
    SphereSweep *New = new SphereSweep();
    int         i;

    Destroy_Transform(New->Trans);
    *New = *this;

    New->Segment = nullptr;
    New->Sphere = nullptr;
    New->Interpolation = Interpolation;

    New->Num_Modeling_Spheres = Num_Modeling_Spheres;
    New->Modeling_Sphere = reinterpret_cast<SPHSWEEP_SPH *>(POV_MALLOC(Num_Modeling_Spheres * sizeof(SPHSWEEP_SPH), "modeling sphere"));
    for (i = 0; i < New->Num_Modeling_Spheres; i++)
        New->Modeling_Sphere[i] = Modeling_Sphere[i];

    // @note  Look to remove this control. It should be constant for numerical accuracy.
    New->Depth_Tolerance = gkMinIsectDepthReturned;

    New->Compute();

    New->Trans = Copy_Transform(Trans);

    return New;
}


void SphereSweep::Translate(const Vector3d& Vector, const TRANSFORM *tr)
{
    if (Trans == nullptr)
    {
        for (int i = 0; i < Num_Modeling_Spheres; i++)
            Modeling_Sphere[i].Center += Vector;
        Compute();
        Compute_BBox();
    }
    else
    {
        Transform(tr);
    }
}


void SphereSweep::Rotate(const Vector3d&, const TRANSFORM *tr)
{
    if (Trans == nullptr)
    {
        for (int i = 0; i < Num_Modeling_Spheres; i++)
            MTransPoint(Modeling_Sphere[i].Center, Modeling_Sphere[i].Center, tr);
        Compute();
        Compute_BBox();
    }
    else
    {
        Transform(tr);
    }
}


void SphereSweep::Scale(const Vector3d& Vector, const TRANSFORM *tr)
{
    if ((Vector[X] != Vector[Y]) || (Vector[X] != Vector[Z]))
    {
        if (Trans == nullptr)
            Trans = Create_Transform();
    }

    if (Trans == nullptr)
    {
        for (int i = 0; i < Num_Modeling_Spheres; i++)
        {
            Modeling_Sphere[i].Center *= Vector[X];
            Modeling_Sphere[i].Radius *= Vector[X];
        }
        Compute();
        Compute_BBox();
    }
    else
    {
        Transform(tr);
    }
}


// Create a new, "empty" sphere sweep (no modeling spheres).
//
SphereSweep::SphereSweep() : ObjectBase(SPHERE_SWEEP_OBJECT)
{
    Interpolation = -1;

    Num_Modeling_Spheres = 0;
    Modeling_Sphere = nullptr;

    Num_Spheres = 0;
    Sphere = nullptr;

    Num_Segments = 0;
    Segment = nullptr;

    //@todo  Should hard code and remove this variable for numerical accuracy.
    Depth_Tolerance = gkMinIsectDepthReturned;

    Trans = nullptr;
}


void SphereSweep::Transform(const TRANSFORM *tr)
{
    if (Trans == nullptr)
        Trans = Create_Transform();

    Compose_Transforms(Trans, tr);

    Compute_BBox();
}


SphereSweep::~SphereSweep()
{
    POV_FREE(Modeling_Sphere);
    POV_FREE(Sphere);
    POV_FREE(Segment);
}


void SphereSweep::Compute_BBox()
{
    Vector3d    mins;
    Vector3d    maxs;
    int         i;

    mins[X] = mins[Y] = mins[Z] = BOUND_HUGE;
    maxs[X] = maxs[Y] = maxs[Z] = -BOUND_HUGE;

    if (Interpolation == CATMULL_ROM_SPLINE_SPHERE_SWEEP)
    {
        // With Catmull-Rom splines, the spline may overshoot the space defined by the control spheres.
        // For now, we solve this by translating each segment of the sphere sweep into (temporary) corresponding
        // Bezier-style control spheres, and computing the bounding box from those.

        for (i = 1; i < Num_Modeling_Spheres-2; i++)
        {
            // Compute temporary Bezier-style control spheres for the segment.

            SPHSWEEP_SPH tempSphere[4];
            tempSphere[0]        = Modeling_Sphere[i];
            tempSphere[1].Center = Modeling_Sphere[i].Center + (Modeling_Sphere[i+1].Center - Modeling_Sphere[i-1].Center) / 6;
            tempSphere[1].Radius = Modeling_Sphere[i].Radius + (Modeling_Sphere[i+1].Radius - Modeling_Sphere[i-1].Radius) / 6;
            tempSphere[2].Center = Modeling_Sphere[i+1].Center + (Modeling_Sphere[i].Center - Modeling_Sphere[i+2].Center) / 6;
            tempSphere[2].Radius = Modeling_Sphere[i+1].Radius + (Modeling_Sphere[i].Radius - Modeling_Sphere[i+2].Radius) / 6;
            tempSphere[3]        = Modeling_Sphere[i+1];

            // From the Bezier-style control spheres, compute the bounding box.

            for (int j = 0; j < 4; ++j)
            {
                mins = min(mins, tempSphere[j].Center - Vector3d(fabs(tempSphere[j].Radius)));
                maxs = max(maxs, tempSphere[j].Center + Vector3d(fabs(tempSphere[j].Radius)));
            }
        }
    }
    else
    {
        for (i = 0; i < Num_Modeling_Spheres; i++)
        {
            mins = min(mins, Modeling_Sphere[i].Center - Vector3d(fabs(Modeling_Sphere[i].Radius)));
            maxs = max(maxs, Modeling_Sphere[i].Center + Vector3d(fabs(Modeling_Sphere[i].Radius)));
        }
    }

    Make_BBox_from_min_max(BBox, mins, maxs);

    if (Trans != nullptr)
        Recompute_BBox(&BBox, Trans);
}


void SphereSweep::Compute()
{
    size_t size;
    int    i;
    int    coef;
    int    msph;
    int    last_sph;
    int    last_seg;

    switch (Interpolation)
    {
        case LINEAR_SPHERE_SWEEP:
            // Allocate memory if necessary
            if (Segment == nullptr)
            {
                Num_Segments = Num_Modeling_Spheres - 1;
                size = Num_Segments * sizeof(SPHSWEEP_SEG);
                Segment = reinterpret_cast<SPHSWEEP_SEG *>(POV_MALLOC(size, "sphere sweep segments"));
            }

            // Calculate polynomials for each segment
            for (i = 0; i < Num_Segments; i++)
            {
                // Polynomial has two coefficients
                Segment[i].Num_Coefs = 2;

                // Coefficients for u^1
                Segment[i].Center_Coef[1] =
                     Modeling_Sphere[i + 1].Center -
                     Modeling_Sphere[i].Center;

                Segment[i].Radius_Coef[1] =
                     Modeling_Sphere[i + 1].Radius -
                     Modeling_Sphere[i].Radius;

                // Coefficients for u^0
                Segment[i].Center_Coef[0] =
                     Modeling_Sphere[i].Center;

                Segment[i].Radius_Coef[0] =
                     Modeling_Sphere[i].Radius;
            }
            break;
        case CATMULL_ROM_SPLINE_SPHERE_SWEEP:
            // Allocate memory if necessary
            if (Segment == nullptr)
            {
                Num_Segments = Num_Modeling_Spheres - 3;
                size = Num_Segments * sizeof(SPHSWEEP_SEG);
                Segment = reinterpret_cast<SPHSWEEP_SEG *>(POV_MALLOC(size, "sphere sweep segments"));
            }

            // Calculate polynomials for each segment
            for (i = 0; i < Num_Segments; i++)
            {
                // Polynomial has four coefficients
                Segment[i].Num_Coefs = 4;

                // Calculate coefficients
                for (coef = 0; coef < 4; coef++)
                {
                    // Center
                    Segment[i].Center_Coef[coef] =
                           Modeling_Sphere[i].Center *
                           Catmull_Rom_Matrix[coef][0];

                    // Radius
                    Segment[i].Radius_Coef[coef] =
                           Modeling_Sphere[i].Radius *
                           Catmull_Rom_Matrix[coef][0];

                    for (msph = 1; msph < 4; msph++)
                    {
                        // Center
                        Segment[i].Center_Coef[coef] +=
                                     Catmull_Rom_Matrix[coef][msph] *
                                     Modeling_Sphere[i + msph].Center;

                        // Radius
                        Segment[i].Radius_Coef[coef] +=
                                     Catmull_Rom_Matrix[coef][msph] *
                                     Modeling_Sphere[i + msph].Radius;
                    }
                }
            }
            break;
        case B_SPLINE_SPHERE_SWEEP:
            // Allocate memory if necessary
            if (Segment == nullptr)
            {
                Num_Segments = Num_Modeling_Spheres - 3;
                size = Num_Segments * sizeof(SPHSWEEP_SEG);
                Segment = reinterpret_cast<SPHSWEEP_SEG *>(POV_MALLOC(size, "sphere sweep segments"));
            }

            // Calculate polynomials for each segment
            for (i = 0; i < Num_Segments; i++)
            {
                // Polynomial has four coefficients
                Segment[i].Num_Coefs = 4;

                // Calculate coefficients
                for (coef = 0; coef < 4; coef++)
                {
                    // Center
                    Segment[i].Center_Coef[coef] =
                           Modeling_Sphere[i].Center *
                           B_Matrix[coef][0];

                    // Radius
                    Segment[i].Radius_Coef[coef] =
                           Modeling_Sphere[i].Radius *
                           B_Matrix[coef][0];

                    for (msph = 1; msph < 4; msph++)
                    {
                        // Center
                        Segment[i].Center_Coef[coef] +=
                                     B_Matrix[coef][msph] *
                                     Modeling_Sphere[i + msph].Center;

                        // Radius
                        Segment[i].Radius_Coef[coef] +=
                                     B_Matrix[coef][msph] *
                                     Modeling_Sphere[i + msph].Radius;
                    }
                }
            }
            break;

        default:
            POV_SHAPE_ASSERT (false);
            break;
    }

    // Pre-calculate several constants

    for (i = 0; i < Num_Segments; i++)
    {
        // Calculate closing sphere for u = 0

        // Center
        Segment[i].Closing_Sphere[0].Center =
                      Segment[i].Center_Coef[0];

        // Radius
        Segment[i].Closing_Sphere[0].Radius =
                      Segment[i].Radius_Coef[0];

        // Calculate derivatives for u = 0

        // Center
        Segment[i].Center_Deriv[0] =
                      Segment[i].Center_Coef[1];

        // Radius
        Segment[i].Radius_Deriv[0] =
                      Segment[i].Radius_Coef[1];

        // Calculate closing sphere for u = 1

        // Center
        Segment[i].Closing_Sphere[1].Center =
                      Segment[i].Center_Coef[0];

        // Radius
        Segment[i].Closing_Sphere[1].Radius =
                      Segment[i].Radius_Coef[0];

        for (coef = 1; coef < Segment[i].Num_Coefs; coef++)
        {
            // Center
            Segment[i].Closing_Sphere[1].Center +=
                   Segment[i].Center_Coef[coef];

            // Radius
            Segment[i].Closing_Sphere[1].Radius +=
                   Segment[i].Radius_Coef[coef];
        }

        // Calculate derivatives for u = 1

        // Center
        Segment[i].Center_Deriv[1] =
                      Segment[i].Center_Coef[1];

        // Radius
        Segment[i].Radius_Deriv[1] =
                      Segment[i].Radius_Coef[1];

        for (coef = 2; coef < Segment[i].Num_Coefs; coef++)
        {
            // Center
            Segment[i].Center_Deriv[1] += double(coef) *
                         Segment[i].Center_Coef[coef];

            // Radius
            Segment[i].Radius_Deriv[1] += coef *
                         Segment[i].Radius_Coef[coef];
        }
    }

    // Calculate single spheres

    // Allocate memory if necessary
    if (Sphere == nullptr)
    {
        Num_Spheres = Num_Segments + 1;
        size = Num_Spheres * sizeof(SPHSWEEP_SPH);
        Sphere = reinterpret_cast<SPHSWEEP_SPH *>(POV_MALLOC(size, "sphere sweep spheres"));
    }

    // Calculate first sphere of every segment

    for (i = 0; i < Num_Segments; i++)
    {
        // Center
        Sphere[i].Center =
                      Segment[i].Center_Coef[0];

        // Radius
        Sphere[i].Radius =
                      Segment[i].Radius_Coef[0];
    }

    // Calculate last sphere of last segment

    last_sph = Num_Spheres - 1;
    last_seg = Num_Segments - 1;

    // Center
    Sphere[last_sph].Center =
                  Segment[last_seg].Center_Coef[0];
    // Radius
    Sphere[last_sph].Radius =
                  Segment[last_seg].Radius_Coef[0];

    for (coef = 1; coef < Segment[last_seg].Num_Coefs; coef++)
    {
        // Center
        Sphere[last_sph].Center +=
               Segment[last_seg].Center_Coef[coef];

        // Radius
        Sphere[last_sph].Radius +=
               Segment[last_seg].Radius_Coef[coef];
    }
}


int SphereSweep::Find_Valid_Points(SPHSWEEP_INT *Inter, int Num_Inter, const BasicRay &ray)
{
    // NB: First intersection point always enters, last one always leaves, so we're not explicitly testing them.

    int Inside = 1;
    int i      = 1;
    double NormalDotDirection;
    bool   Keep;
    while (i < Num_Inter - 1)
    {
        // Angle between normal and ray
        NormalDotDirection = dot(Inter[i].Normal, ray.Direction);

        // Does the ray enter the part?
        if (NormalDotDirection < 0.0)
        {
            // Ray enters part, keep intersection if ray was outside any part
            Keep = (Inside == 0);

            // increase inside counter
            Inside++;
        }
        else
        {
            // Ray exits part, keep intersection if ray was inside one part
            Keep = (Inside == 1);

            // decrease inside counter
            Inside--;
        }

        // Keep intersection?
        if (Keep)
            i++; // Yes, advance to next one
        else
        {
            // No, delete it
            for (int j = i + 1; j < Num_Inter; j++)
                Inter[j - 1] = Inter[j];
            Num_Inter--;
        }
    }

    return Num_Inter;
}


// Compare the depths of two sphere sweep intersections.
//
int SphereSweep::Comp_Isects(const void *Intersection_1, const void *Intersection_2)
{
    const SPHSWEEP_INT *Int_1;
    const SPHSWEEP_INT *Int_2;

    Int_1 = reinterpret_cast<const SPHSWEEP_INT *>(Intersection_1);
    Int_2 = reinterpret_cast<const SPHSWEEP_INT *>(Intersection_2);

    if (Int_1->t < Int_2->t)
        return -1;

    if (Int_1->t == Int_2->t)
        return 0;
    else
        return 1;
}


// Author Massimo Valentini
//
// Optimization to exclude the presence of a root of a polynomial in (0 1).
// The rendering time decreases impressively.
//
const int lcm_bezier_01[] =
{
    60,  10,  4,  3,  4, 10, 60,
    2520, 252, 56, 21, 12, 10, 12, 21, 56, 252, 2520
};


int SphereSweep::bezier_01(int degree, const double* Coef, double* Roots)
{
    double d[11];
    bool non_negative = true, non_positive = true;
    int i, j;
    const int *lcm = &(lcm_bezier_01[degree == 6 ? 0 : 7]);

    for (i = 0; i <= degree; ++i)
        d[i] = Coef[i] * lcm[i];

    for (i = 0; i <= degree; ++i)
    {
        non_negative = (non_negative && (d[degree - i] >= 0));
        non_positive = (non_positive && (d[degree - i] <= 0));

        if (!(non_negative || non_positive))
        {
            return polysolve(degree, Coef, Roots, 0.0, 1.0);
        }

        for (j = 0; j < degree - i; ++j)
            d[j] += d[j+1];
    }

    return 0;
}

}
// end of namespace pov
