//******************************************************************************
///
/// @file vm/fnintern.cpp
///
/// This module implements built-in render-time functions.
///
/// This module is based on code by D. Skarda, T. Bily and R. Suzuki.
/// It includes functions based on code first introduced by many other
/// contributors.
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2022 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//------------------------------------------------------------------------------
// SPDX-License-Identifier: AGPL-3.0-or-later
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "vm/fnintern.h"

// C++ variants of C standard header files
//  (none at the moment)

// C++ standard header files
#include <algorithm>
#include <ctime>

// POV-Ray header files (base module)
#include "base/mathutil.h"

// POV-Ray header files (core module)
#include "core/material/noise.h"
#include "core/material/pattern.h"
#include "core/material/pigment.h"
#include "core/material/warp.h"
#include "core/math/matrix.h"
#include "core/math/spline.h"
#include "core/scene/scenedata.h"
#include "core/scene/tracethreaddata.h"

// POV-Ray header files (VM module)
#include "vm/fnpovfpu.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

static RandomDoubleSequence gPatternRands(0.0, 1.0, 32768);

static double fnHandle1to8ReturnCalc(const double ary[][4], const int valset, const int width, const int count,
        const int ReturnTreatment)
{
    int minValIndex = 0;
    int localReturnTreatment = ReturnTreatment & 0x07;
    bool returnMapIndexValue = (ReturnTreatment & 0x0A) ? true : false;

    double value = ary[0][valset];
    if (localReturnTreatment == 1)
    {
        for (int i=1; i<std::min(width,count); i++)
        {
            value = std::max(value,ary[i][valset]);
        }
    }
    else if (!returnMapIndexValue)  // (0) min of all the default.
    {
        for (int i=1; i<std::min(width,count); i++)
        {
            value = std::min(value,ary[i][valset]);
        }
    }

    if (returnMapIndexValue)
    {
        double tmpMinValue = ary[0][valset];
        for (int i=1; i<std::min(width,count); i++)
        {
            if (ary[i][valset] < tmpMinValue)
            {
                tmpMinValue = ary[i][valset];
                minValIndex = i;
            }
        }
        return (double)(minValIndex+1)/10.0;
    }
    else
        return value;
}

// Versions of following two functions in textures.cpp too. Local static for
// performance.  The sin(value*TWO_M_PI) version of fnCycloidal_Wave in testing
// returns exactly the same values to double accuracy.

static double fnCycloidal_Wave(double value)
{
//  if (value >= 0.0)
//      return sin((double) (((value - floor(value)) * 50000.0)) / 50000.0 * TWO_M_PI);
//  else
//      return 0.0-sin((double) (((0.0 - (value + floor(0.0 - value))) * 50000.0)) / 50000.0 * TWO_M_PI);
    return sin(value*TWO_M_PI);
}

static double fnTriangle_Wave(double value)
{
    double offset;
    if (value >= 0.0)
        offset = value - floor(value);
    else
        offset = value + 1.0 + floor(fabs(value));
    if (offset >= 0.5)
        return (2.0 * (1.0 - offset));
    else
        return (2.0 * offset);
}

// The fnEnc2x_f32 and fnDec2x_f32 functions just below are used to encode and
// decode two 32 bit floats to and from the space for a double (8 bytes). The
// encoding and decoding uses standard cast to and from float and so unlike the
// f21 routines no rounding is done on encoding, but these should be faster.
// Normally the number of significant digits on encoding of two double values
// as two floats in a double space is a reduction from 17 decimal digits to 9,
// but it can vary.

static double fnEnc2x_f32(double one, double two) {
    static_assert(sizeof(float)  == 4);    // These I think redundant for c++11
    static_assert(sizeof(double) == 8);
    static_assert(sizeof(long long unsigned) == 8);

    union uC {
        float buffer[2];
        double value;
    } u;
    u.value = 0.0f;

    float tmpFlt = (float)one;
    memcpy(&u.buffer[0], &tmpFlt, sizeof(u.buffer[0]));
    tmpFlt = (float)two;
    memcpy(&u.buffer[1], &tmpFlt, sizeof(u.buffer[1]));

    return u.value;
}

static double fnDec2x_f32(double f32_encoded, double which) {
    uint_fast32_t pickWhich = ((uint8_t)fabs(which)) & 0x03;

    union uC {
        float buffer[2];
        double value;
    } u;
    u.value = 0.0f;

    memcpy(&u.value, &f32_encoded, sizeof(u.value));
    double value = 1e-35;

    if (pickWhich == 0)
    {
        value  = ((double)u.buffer[0]);
    }
    else if (pickWhich == 1)
    {
        value  = ((double)u.buffer[1]);
    }
    else
    {
        throw POV_EXCEPTION_STRING("Error. fnDec2x_f32\nDecode selection values must be 0 or 1.");
    }

    return value;
}

// The fnEnc3x_f21 and fnDec3x_f21 functions just below are used to encode and
// decode three 21 bit signed doubles to and from the space for a double (8
// bytes). This encoding permits only 6 significant digits of representation,
// but for many things like normals it's enough. The functions allow the
// specification of decimal places on encoding and decoding. To get the same
// values back on decoding these must match, but one can play games to
// essentially multiply or divide by multiples of ten on the decode.
//

static double fnEnc3x_f21(double one, double two, double three, double decplaces) {
    static_assert(sizeof(double) == 8);    // These, I think redundant for c++11, but...
    static_assert(sizeof(long long unsigned) == 8);

    union uC {
        uint64_t buffer;
        double value;
    } u;
    u.buffer = 0ul;

    uint_fast32_t decimalPlaces = std::min(6,(uint8_t)fabs(decplaces) & 0x7);

    double offset;
    switch (decimalPlaces) {
        case 0 : offset = 1e0;
                break;
        case 1 : offset = 1e1;
                break;
        case 2 : offset = 1e2;
                break;
        case 3 : offset = 1e3;
                break;
        case 4 : offset = 1e4;
                break;
        case 5 : offset = 1e5;
                break;
        case 6 : offset = 1e6;
                break;
        default : offset = 1e6; // To avoid -Wmaybe-uninitialized warning.
                break;
    }

    //---
    uint64_t tmp = std::min(1e6,fabs(round(one*offset)));
    // 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0001 0000 0000 0000 0000 0000
    //        00        00        00        00        00        10        00        00
    if (one < 0.0) u.buffer |= 0x0000000000100000;
    tmp &= 0x00000000000fffff;
    u.buffer += tmp;
    u.buffer = u.buffer << 21;
    //---
    tmp = std::min(1e6,fabs(round(two*offset)));
    if (two < 0.0) u.buffer |= 0x0000000000100000;
    tmp &= 0x00000000000fffff;
    u.buffer += tmp;
    u.buffer = u.buffer << 21;
    //---
    tmp = std::min(1e6,fabs(round(three*offset)));
    if (three < 0.0) u.buffer |= 0x0000000000100000;
    tmp &= 0x00000000000fffff;
    u.buffer += tmp;

    return u.value;
}

static double fnDec3x_f21(double f21_encoded, double decplaces, double which) {
    uint_fast32_t pickWhich = ((uint8_t)fabs(which)) & 0x03;
    uint_fast32_t decimalPlaces = std::min(6,(uint8_t)fabs(decplaces) & 0x7);

    double offset;
    switch (decimalPlaces) {
        case 0 : offset = 1e0;
                 break;
        case 1 : offset = 1e-1;
                 break;
        case 2 : offset = 1e-2;
                 break;
        case 3 : offset = 1e-3;
                 break;
        case 4 : offset = 1e-4;
                 break;
        case 5 : offset = 1e-5;
                 break;
        case 6 : offset = 1e-6;
                 break;
        default : offset = 1e6; // To avoid -Wmaybe-uninitialized warning.
                break;
    }

    uint64_t buffer;
    memcpy(&buffer, &f21_encoded, sizeof(buffer));
    double sign = 1;
    double value = 1e-35;

    if (pickWhich == 0)
    {
        buffer = buffer >> 42;
        if ((buffer & 0x0000000000100000)>0)
            sign = -1;
        value = sign * (((double)(buffer & 0x00000000000fffff))*offset);
    }
    else if (pickWhich == 1)
    {
        buffer = buffer >> 21;
        if ((buffer & 0x0000000000100000)>0)
            sign = -1;
        value = sign * (((double)(buffer & 0x00000000000fffff))*offset);
    }
    else if (pickWhich == 2)
    {
        if ((buffer & 0x0000000000100000)>0)
            sign = -1;
        value = sign * (((double)(buffer & 0x00000000000fffff))*offset);
    }
    else
    {
        throw POV_EXCEPTION_STRING("Error. fnDec3x_f21\nDecode selection values must be 0,1 or 2.");
    }

    return value;
}

// FnRotate0 and FnRotate1 below both do the pattern or isosurface inverse,
// left handed 2d rotation (right handed) along the positive x,y or z axis. The
// FnRotate1 version is currently faster in testing, but keeping the '0' as who
// knows what the future or other architectures might hold.

//  static double FnRotate0(double x, double y, double z, double axis, double angle) {
//      uint_fast32_t xyz   = std::min(2,(uint8_t)fabs(axis) & 0x03);
//                    angle = fmod(angle,360.0);
//      double theta        = angle * M_PI_180;
//      double r,phi;
//      double value =0.0;
//
//      if (xyz == 0)
//      {
//          r   = sqrt(y*y + z*z); // Cartesian to polar
//          phi = atan2(-z,y);
//
//          phi += theta;          // Adjust angle
//
//          y = r*cos(phi);        // Polar to cartesian
//          z = r*sin(phi);
//
//          value = fnEnc2x_f32(y,-z);  // Encode two 32 bit floats into double space.
//      }
//      else if (xyz ==1)
//      {
//          r   = sqrt(x*x + z*z);
//          phi = atan2(z,x);
//
//          phi += theta;
//
//          x = r*cos(phi);
//          z = r*sin(phi);
//
//          value = fnEnc2x_f32(x,z);
//      }
//      else if (xyz ==2)
//      {
//          r   = sqrt(x*x + y*y);
//          phi = atan2(-y,x);
//
//          phi += theta;
//
//          x = r*cos(phi);
//          y = r*sin(phi);
//
//          value = fnEnc2x_f32(x,-y);
//      }
//      else
//      {
//          throw POV_EXCEPTION_STRING("Error. fnRotate0\nAxis selection values must be 0,1 or 2.");
//      }
//
//      return value;
//  }

// FnRotate1 is 20+% faster than the FnRotate0 variation.
// 28.03s -> 22.23s ---> -20.69% (Standalone compile and test 200 million calls)

static double FnRotate1(double x, double y, double z, double axis, double angle) {
    uint_fast32_t xyz = std::min(2,(uint8_t)fabs(axis) & 0x03);
    double theta      = angle * M_PI_180;

    double xr,yr,zr;
    double value =0.0;
    if (xyz == 0)
    {
        yr = z*sin(theta) + y*cos(theta);
        zr = z*cos(theta) - y*sin(theta);
        value = fnEnc2x_f32(yr,zr);
    }
    else if (xyz ==1)
    {
        xr = +x*cos(theta) - z*sin(theta);
        zr = +x*sin(theta) + z*cos(theta);
        value = fnEnc2x_f32(xr,zr);
    }
    else if (xyz ==2)
    {
        xr = +x*cos(theta) + y*sin(theta);
        yr = -x*sin(theta) + y*cos(theta);
        value = fnEnc2x_f32(xr,yr);
    }
    else
    {
        // throw
    }

    return value;
}


// Local preprocessor defines

// IMPORTANT NOTICE
// Never ever use any PARAM_xxx macro after a call to POVFPU_SetLocal because
// POVFPU_SetLocal may change the stack base address and thus the "ptr" passed
// to the functions may be invalid after the call! [trf]

#define PARAM_X (ptr[0])
#define PARAM_Y (ptr[1])
#define PARAM_Z (ptr[2])
#define PARAM(index) (ptr[index + 3])
#define PARAM_N_X(offset) (ptr[offset])
#define PARAM_N_Y(offset) (ptr[offset + 1])
#define PARAM_N_Z(offset) (ptr[offset + 2])
#define PARAM_N(index,offset) (ptr[index + offset + 3])

// The ROT2D macro is no longer used. On implementing f_revolution discovered
// the z axis rotation element was for general input functions unstable. It
// might be it worked OK within the context of the four replaced f_algbr_cyl<n>
// functions.  Did not go back and check.

/* This style comment to avoid -Wcomment warnings on continuations back slashes.
   No longer used. Later delete.
#define ROT2D(p,d,ang) if (p>0) {x2=sqrt(x2+PARAM_Z*PARAM_Z)- d; th=ang*M_PI_180; \
    if (th!=0){ PARAM_X= x2*cos(th)-PARAM_Y*sin(th); PARAM_Y= x2*sin(th)+PARAM_Y*cos(th);} else PARAM_X=x2; \
    x2=PARAM_X*PARAM_X; y2=PARAM_Y*PARAM_Y;}

*/

// Global functions

double f_gradient(FPUContext *ctx, double *ptr, unsigned int fn); // 0
double f_noise_cubed(FPUContext *ctx, double *ptr, unsigned int fn); // 1
double f_radial(FPUContext *ctx, double *ptr, unsigned int fn); // 2
double f_remap(FPUContext *ctx, double *ptr, unsigned int fn); // 3
double f_cylinder(FPUContext *ctx, double *ptr, unsigned int fn); // 4
double f_cone(FPUContext *ctx, double *ptr, unsigned int fn); // 5
double f_granite23(FPUContext *ctx, double *ptr, unsigned int fn); // 6
double f_mult1to8pairs(FPUContext *ctx, double *ptr, unsigned int fn); // 7
double f_max1to8pairs(FPUContext *ctx, double *ptr, unsigned int fn); // 8
double f_min1to8pairs(FPUContext *ctx, double *ptr, unsigned int fn); // 9
double f_morph2to9fncts(FPUContext *ctx, double *ptr, unsigned int fn); // 10
double f_labsweep(FPUContext *ctx, double *ptr, unsigned int fn); // 11
double f_agate(FPUContext *ctx, double *ptr, unsigned int fn); // 12
double f_onion(FPUContext *ctx, double *ptr, unsigned int fn); // 13
double f_sign(FPUContext *ctx, double *ptr, unsigned int fn); // 14
double f_sign0(FPUContext *ctx, double *ptr, unsigned int fn); // 15
double f_cbrt(FPUContext *ctx, double *ptr, unsigned int fn); // 16
double f_ellipsoid(FPUContext *ctx, double *ptr, unsigned int fn); // 17
double f_polyhedron(FPUContext *ctx, double *ptr, unsigned int fn); // 18
double f_lame(FPUContext *ctx, double *ptr, unsigned int fn); // 19
double f_signpow(FPUContext *ctx, double *ptr, unsigned int fn); // 20
double f_signsqrt(FPUContext *ctx, double *ptr, unsigned int fn); // 21
double f_vangle(FPUContext *ctx, double *ptr, unsigned int fn); // 22
double f_turbulence(FPUContext *ctx, double *ptr, unsigned int fn); // 23
double f_bump(FPUContext *ctx, double *ptr, unsigned int fn); // 24
double f_helix1(FPUContext *ctx, double *ptr, unsigned int fn); // 25
double f_helix2(FPUContext *ctx, double *ptr, unsigned int fn); // 26
double f_length(FPUContext *ctx, double *ptr, unsigned int fn); // 27
double f_normal(FPUContext *ctx, double *ptr, unsigned int fn); // 28
double f_hetero_mf(FPUContext *ctx, double *ptr, unsigned int fn); // 29
double f_normalized(FPUContext *ctx, double *ptr, unsigned int fn); // 30
double f_boom(FPUContext *ctx, double *ptr, unsigned int fn); // 31
double f_box(FPUContext *ctx, double *ptr, unsigned int fn); // 32
double f_length_sqr(FPUContext *ctx, double *ptr, unsigned int fn); // 33
double f_triangle_wave(FPUContext *ctx, double *ptr, unsigned int fn); // 34
double f_sine_wave(FPUContext *ctx, double *ptr, unsigned int fn); // 35
double f_scallop_wave(FPUContext *ctx, double *ptr, unsigned int fn); // 36
double f_sine_wavef(FPUContext *ctx, double *ptr, unsigned int fn); // 37
double f_scallop_wavef(FPUContext *ctx, double *ptr, unsigned int fn); // 38
double f_cubic_wave(FPUContext *ctx, double *ptr, unsigned int fn); // 39
double f_boxb(FPUContext *ctx, double *ptr, unsigned int fn); // 40
double f_wrap(FPUContext *ctx, double *ptr, unsigned int fn); // 41
double f_elliptical_cylinder(FPUContext *ctx, double *ptr, unsigned int fn); // 42
double f_odd(FPUContext *ctx, double *ptr, unsigned int fn); // 43
double f_planar(FPUContext *ctx, double *ptr, unsigned int fn); // 44
double f_paraboloid(FPUContext *ctx, double *ptr, unsigned int fn); // 45
double f_quilt_cubic(FPUContext *ctx, double *ptr, unsigned int fn); // 46
double f_ph(FPUContext *ctx, double *ptr, unsigned int fn); // 47
double f_dec_nbit3c(FPUContext *ctx, double *ptr, unsigned int fn); // 48
double f_enc_nbit3c(FPUContext *ctx, double *ptr, unsigned int fn); // 49
double f_hypot(FPUContext *ctx, double *ptr, unsigned int fn); // 50
double f_upoly(FPUContext *ctx, double *ptr, unsigned int fn); // 51
double f_radial_tubes(FPUContext *ctx, double *ptr, unsigned int fn); // 52
double f_dturbulence(FPUContext *ctx, double *ptr, unsigned int fn); // 53
double f_rotate2d(FPUContext *ctx, double *ptr, unsigned int fn); // 54
double f_hessian_dist(FPUContext *ctx, double *ptr, unsigned int fn); // 55
double f_quartic_cylinder(FPUContext *ctx, double *ptr, unsigned int fn); // 56
double f_r(FPUContext *ctx, double *ptr, unsigned int fn); // 57
double f_ridge(FPUContext *ctx, double *ptr, unsigned int fn); // 58
double f_ridged_mf(FPUContext *ctx, double *ptr, unsigned int fn); // 59
double f_rounded_box(FPUContext *ctx, double *ptr, unsigned int fn); // 60
double f_sphere(FPUContext *ctx, double *ptr, unsigned int fn); // 61
double f_enc3x_f21(FPUContext *ctx, double *ptr, unsigned int fn); // 62
double f_dec3x_f21(FPUContext *ctx, double *ptr, unsigned int fn); // 63
double f_spiral(FPUContext *ctx, double *ptr, unsigned int fn); // 64
double f_range2range(FPUContext *ctx, double *ptr, unsigned int fn); // 65
double f_toupie(FPUContext *ctx, double *ptr, unsigned int fn); // 66
double f_supertorus(FPUContext *ctx, double *ptr, unsigned int fn); // 67
double f_superellipsoid(FPUContext *ctx, double *ptr, unsigned int fn); // 68
double f_th(FPUContext *ctx, double *ptr, unsigned int fn); // 69
double f_torus(FPUContext *ctx, double *ptr, unsigned int fn); // 70
double f_repeat(FPUContext *ctx, double *ptr, unsigned int fn); // 71
double f_marble(FPUContext *ctx, double *ptr, unsigned int fn); // 72
double f_wood(FPUContext *ctx, double *ptr, unsigned int fn); // 73
double f_enc2x_f32(FPUContext *ctx, double *ptr, unsigned int fn); // 74
double f_dec2x_f32(FPUContext *ctx, double *ptr, unsigned int fn); // 75
double f_noise3d(FPUContext *ctx, double *ptr, unsigned int fn); // 76
double f_pattern(FPUContext *ctx, double *ptr, unsigned int fn); // 77
double f_snoise3d(FPUContext *ctx, double *ptr, unsigned int fn); // 78
double f_npmod(FPUContext *ctx, double *ptr, unsigned int fn); // 79
double f_eblob(FPUContext *ctx, double *ptr, unsigned int fn); // 80
double f_revolution(FPUContext *ctx, double *ptr, unsigned int fn); // 81
double f_hash(FPUContext *ctx, double *ptr, unsigned int fn); // 82
double f_interior_fade(FPUContext *ctx, double *ptr, unsigned int fn); // 83
double f_distribution(FPUContext *ctx, double *ptr, unsigned int fn); // 84
double f_clock(FPUContext *ctx, double *ptr, unsigned int fn); // 85
double f_randn(FPUContext *ctx, double *ptr, unsigned int fn); // 86
double f_null(FPUContext *ctx, double *ptr, unsigned int fn); // 87
double f_elliptical_sphrswp(FPUContext *ctx, double *ptr, unsigned int fn); // 88

void f_pigment(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp); // 0
void f_transform(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp); // 1
void f_spline(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp); // 2
void f_patmods(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp); // 3


// Global variables

constexpr Trap POVFPU_TrapTable[] =
{
    { f_gradient,                3 + 3 }, // 0
    { f_noise_cubed,             3 + 3 }, // 1
    { f_radial,                  0 + 3 }, // 2
    { f_remap,                   8 + 3 }, // 3
    { f_cylinder,                1 + 3 }, // 4
    { f_cone,                    1 + 3 }, // 5
    { f_granite23,               6 + 3 }, // 6
    { f_mult1to8pairs,          16 + 3 }, // 7
    { f_max1to8pairs,           16 + 3 }, // 8
    { f_min1to8pairs,           16 + 3 }, // 9
    { f_morph2to9fncts,         18 + 3 }, // 10
    { f_labsweep,                7 + 3 }, // 11
    { f_agate,                   7 + 3 }, // 12
    { f_onion,                   3 + 3 }, // 13
    { f_sign,                    0 + 1 }, // 14
    { f_sign0,                   0 + 3 }, // 15
    { f_cbrt,                    0 + 3 }, // 16
    { f_ellipsoid,               6 + 3 }, // 17
    { f_polyhedron,             13 + 3 }, // 18
    { f_lame,                   16 + 3 }, // 19
    { f_signpow,                 0 + 3 }, // 20
    { f_signsqrt,                0 + 1 }, // 21
    { f_vangle,                  4 + 3 }, // 22
    { f_turbulence,              8 + 3 }, // 23
    { f_bump,                    0 + 3 }, // 24
    { f_helix1,                  7 + 3 }, // 25
    { f_helix2,                  5 + 3 }, // 26
    { f_length,                  0 + 3 }, // 27
    { f_normal,                  3 + 3 }, // 28
    { f_hetero_mf,               7 + 3 }, // 29
    { f_normalized,              1 + 3 }, // 30
    { f_boom,                    3 + 3 }, // 31
    { f_box,                     4 + 3 }, // 32
    { f_length_sqr,              0 + 3 }, // 33
    { f_triangle_wave,           0 + 2 }, // 34
    { f_sine_wave,               0 + 2 }, // 35
    { f_scallop_wave,            0 + 2 }, // 36
    { f_sine_wavef,              0 + 2 }, // 37
    { f_scallop_wavef,           0 + 2 }, // 38
    { f_cubic_wave,              0 + 2 }, // 39
    { f_boxb,                    8 + 3 }, // 40
    { f_wrap,                    1 + 3 }, // 41
    { f_elliptical_cylinder,     3 + 3 }, // 42
    { f_odd,                     1 + 3 }, // 43
    { f_planar,                  0 + 3 }, // 44
    { f_paraboloid,              2 + 3 }, // 45
    { f_quilt_cubic,             4 + 3 }, // 46
    { f_ph,                      0 + 3 }, // 47
    { f_dec_nbit3c,              2 + 3 }, // 48
    { f_enc_nbit3c,              1 + 3 }, // 49
    { f_hypot,                   0 + 2 }, // 50
    { f_upoly,                  16 + 3 }, // 51
    { f_radial_tubes,            4 + 3 }, // 52
    { f_dturbulence,             7 + 3 }, // 53
    { f_rotate2d,                2 + 3 }, // 54
    { f_hessian_dist,            2 + 3 }, // 55
    { f_quartic_cylinder,        2 + 3 }, // 56
    { f_r,                       0 + 3 }, // 57
    { f_ridge,                   7 + 3 }, // 58
    { f_ridged_mf,               7 + 3 }, // 59
    { f_rounded_box,             4 + 3 }, // 60
    { f_sphere,                  1 + 3 }, // 61
    { f_enc3x_f21,               1 + 3 }, // 62
    { f_dec3x_f21,               0 + 3 }, // 63
    { f_spiral,                  4 + 3 }, // 64
    { f_range2range,             2 + 3 }, // 65
    { f_toupie,                  4 + 3 }, // 66
    { f_supertorus,              5 + 3 }, // 67
    { f_superellipsoid,          3 + 3 }, // 68
    { f_th,                      0 + 3 }, // 69
    { f_torus,                   2 + 3 }, // 70
    { f_repeat,                 13 + 3 }, // 71
    { f_marble,                  6 + 3 }, // 72
    { f_wood,                    5 + 3 }, // 73
    { f_enc2x_f32,               0 + 2 }, // 74
    { f_dec2x_f32,               0 + 2 }, // 75
    { f_noise3d,                 1 + 3 }, // 76
    { f_pattern,                 0 + 3 }, // 77
    { f_snoise3d,                1 + 3 }, // 78
    { f_npmod,                   0 + 3 }, // 79
    { f_eblob,                  16 + 3 }, // 80
    { f_revolution,              3 + 3 }, // 81
    { f_hash,                    1 + 3 }, // 82
    { f_interior_fade,           3 + 3 }, // 83
    { f_distribution,            4 + 3 }, // 84
    { f_clock,                   0 + 3 }, // 85
    { f_randn,                   2 + 3 }, // 86
    { f_null,                    0 + 1 }, // 87
    { f_elliptical_sphrswp,      6 + 3 }, // 88
    { nullptr, 0 }
};

constexpr TrapS POVFPU_TrapSTable[] =
{
    { f_pigment,                 0 + 3 }, // 0
    { f_transform,               0 + 3 }, // 1
    { f_spline,                  0 + 1 }, // 2
    { f_patmods,                 0 + 3 }, // 3
    { nullptr, 0 }
};

const unsigned int POVFPU_TrapTableSize = 89;
const unsigned int POVFPU_TrapSTableSize = 4;


// Functions

double f_gradient(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 0   was f_algbr_cyl1
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[0].parameter_cnt == 6u);
    double gx=PARAM(0), gy=PARAM(1), gz=PARAM(2);
    double value = dot(Vector3d(PARAM_X,PARAM_Y,PARAM_Z),Vector3d(gx,gy,gz).normalized());
    return (value);
}

double f_noise_cubed(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 1   was f_algbr_cyl2 -> f_dents
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[1].parameter_cnt == 6u);
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int noiseGen = std::max(1,(int)PARAM(0) & 0x03);  // 1, 2 or 3 only allowed values.
    int mode     = std::max(0,(int)PARAM(1) & 0x01);  // 0, 1, 2 or 3 only allowed values.
    double scale = PARAM(2);                     // Scale the 0 to 1, 0.5 centered noise dist.

    auto lambdaScaleNoiseDistVal = [&scale] (double val) -> double
    {
        val = val - 0.5;
        val *= scale;
        val += 0.5;
        return (val);
    };

    auto lambdaModePowShapeVal = [] (double val) -> double
    {
        val = std::max(0.0+gkMinIsectDepthReturned,std::min(1.0-gkMinIsectDepthReturned,val));
        val = pow(val,0.95/val);
        val = pow(val,0.50/val);
        return (val);
    };

    double value = Noise(Vec,noiseGen);
    if (scale != 1.0)
        value = lambdaScaleNoiseDistVal(value);

    if (mode <= 1)
    {
        if (mode == 1)
            value = lambdaModePowShapeVal(value);

        value = value * value * value;
    }
    else
    {
        double value2 = Noise(Vector3d(PARAM_X+PARAM_Y,PARAM_Y+PARAM_Z,PARAM_Z+PARAM_X),noiseGen);
        double value3 = Noise(Vector3d(PARAM_X-PARAM_Z,PARAM_Y-PARAM_X,PARAM_Z-PARAM_Y),noiseGen);

        if (scale != 1.0)
        {
            value2 = lambdaScaleNoiseDistVal(value2);
            value3 = lambdaScaleNoiseDistVal(value3);
        }

        if (mode == 3)
        {
            value  = lambdaModePowShapeVal(value);
            value2 = lambdaModePowShapeVal(value2);
            value3 = lambdaModePowShapeVal(value3);
        }
        value = value * value2 * value3;

    }

    if (mode & 0x01)  // If mode is 1 or 3
       value = pow(value,0.125/value);

    return (value);
}

double f_radial(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 2   was f_algbr_cyl3
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[2].parameter_cnt == 3u);
    double angle = atan2(PARAM_Z,PARAM_X);
    if (angle > 0.0)
    {
        angle = TWO_M_PI - angle;
    }
    double value = fabs(angle)/TWO_M_PI;
    return value;
}

double f_remap(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 3   was f_algbr_cyl4
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[3].parameter_cnt == 11u);
    const int width=8, height=3;
    double returnValue=1.0, ary[width][height];

    double value = PARAM_X;                       // Incoming value.
    int pCount = std::max(2,(int)PARAM_Y & 0x0F); // Number of 0 crossings.
    double thickness = fabs(PARAM_Z);             // value thickness.

    int i,j;
    for (i=0; i<std::min(width,pCount); i++)
    {
        j = i - 1;
        ary[i][0] = PARAM(i);               // Additional triangle peaks.

        if (i > 0)
        {
            if (ary[i][0] <= ary[j][0])
            {
                 ary[i][0] = ary[j][0] + 1.0;  // Force increasing values.
            }

            ary[i][1] = ary[i][0] - (ary[i][0] - ary[j][0])/2.0;     // Triangle tip below this i threshold.
            ary[j][2] = ary[i][1];                                   // Triangle tip above prev j threshold.

            if (i == 1)
                ary[j][1] = ary[j][0] - (ary[i][0] - ary[j][0])/2.0; // Triangle tip below j=0 threshold.
        }
    }
    i -= 1;
    j = i - 1;
    ary[i][2] = ary[i][0] + (ary[i][0] - ary[j][0])/2.0; // Triangle tip above the last threshold.

    // If outside the ranges of interest return appropriate -1 or 1 value.

    if (value <= ary[0][1])
        return (thickness > gkDBL_epsilon) ? +1.0 : -1.0;
    if (value >= ary[pCount-1][2])
        return ((thickness > gkDBL_epsilon) || (pCount & 0x01)) ? +1.0 : -1.0;

    // To normalize the thickness must scan for maximum half delta.
    double minAboveBelowDelta = HUGE_VAL;
    if (thickness > gkDBL_epsilon)
    {
        for (i=0; i<std::min(width,pCount); i++)
        {
            double belowDelta = ary[i][0] - ary[i][1];
            double aboveDelta = ary[i][2] - ary[i][0];
            minAboveBelowDelta = std::min(minAboveBelowDelta,std::min(belowDelta,aboveDelta));
        }
    }
    //
    for (i=0; i<std::min(width,pCount); i++)
    {
        double belowDelta = ary[i][0] - ary[i][1];
        double aboveDelta = ary[i][2] - ary[i][0];
        if ((value > ary[i][1]) && (value <= ary[i][2]))
        {
            double tmpVal = value - ary[i][1];
            if (tmpVal <= belowDelta)
            {
                if (i & 0x01)
                    returnValue = (0.5 - (0.5 * (tmpVal/belowDelta))) / +0.5;
                else
                    returnValue = (0.5 - (0.5 * (tmpVal/belowDelta))) / -0.5;
            }
            else
            {
                tmpVal = value - ary[i][0];
                if (i & 0x01)
                    returnValue = ((0.5 * (tmpVal/aboveDelta))) / -0.5;
                else
                    returnValue = ((0.5 * (tmpVal/aboveDelta))) / +0.5;
            }
            if (thickness > gkDBL_epsilon)
            {
                if (i & 0x01)
                {
                    returnValue = std::max(+returnValue-(thickness/2.0)*(minAboveBelowDelta/belowDelta),
                            -returnValue-(thickness/2.0)*(minAboveBelowDelta/aboveDelta));
                }
                else
                {
                    returnValue = std::max(+returnValue-(thickness/2.0)*(minAboveBelowDelta/belowDelta),
                            -returnValue-(thickness/2.0)*(minAboveBelowDelta/aboveDelta));
                }
            }
        }
    }
    return returnValue;
}

double f_cylinder(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 4   was f_bicorn
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[4].parameter_cnt == 4u);

  //--- In testing this encoding +9.5% slower
  //double value = std::hypot(PARAM_X,PARAM_Z) - PARAM(0);

    double value = sqrt(PARAM_X*PARAM_X + PARAM_Z*PARAM_Z) - PARAM(0);
    return value;
}

double f_cone(FPUContext* /*ctx*/, double *ptr, unsigned int) // 5   was f_bifolia
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[5].parameter_cnt == 4u);

  //--- In testing this encoding +9.5% slower
  //double value = std::hypot(PARAM_X,PARAM_Z) - PARAM(0) + fabs(PARAM_Y);

    double value = sqrt(PARAM_X*PARAM_X + PARAM_Z*PARAM_Z) - PARAM(0) + fabs(PARAM_Y);
    return value;
}

double f_granite23(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 6   was f_blob
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[6].parameter_cnt == 9u);
    int i;
    double value = 0.0, freq = 1.0;
    Vector3d tv1, tv2;

    tv1 = Vector3d(PARAM_X,PARAM_Y,PARAM_Z) * 4.0;

    for (i = 0; i < 6; freq *= 2.0, i++)
    {
        tv2 = tv1 * freq;
        // In the block below tried the idea of using Perlin(3) and NoiseGen(2)
        // alternatively to better avoid artefacts. It looks good for artefacts
        // and coding up the alternative as f_granite23 and deleting f_granite

        if (i & 0x1)
        {
          //value += Noise(tv2,2) * (0.5/freq); // Fractional summed noise (wrinkle like?)
            value += std::min(0.5,fabs(1.0 - 2.0 * Noise(tv2,2))) * (0.5/freq);  // Granite hack, looks ok but a little different...
        }
        else
        {
          //value += Noise(tv2,3) * (0.50/freq);
            value += std::min(0.5,fabs(1.0 - 2.0 * Noise(tv2,3))) * (0.50/freq);
        }
    }

    value = (value <= 0.5) ? value : value + ((value-0.5)*0.70); // Granite hack? (YES and with inversion get puffy cloud look)

    // Comment these checks unless working on a better fit.
  //if (value > 1.0)
  //    throw POV_EXCEPTION_STRING("value > 1.0");
  //if (value < 0.0)
  //    throw POV_EXCEPTION_STRING("value < 0.0");

    value = std::max(0.0,std::min(1.0,value));

    if (PARAM(0) != 0.0)
        value = 1.0 - value;

    if (PARAM(1) != 0.0)
        value = value * value * value;  // A dents like treatment if passed unclamped sum...  (Use with cbrts below interesting...)

    if (PARAM(2) != 0.0)
        value = 1.0 - value;

    if (PARAM(3) != 0.0)                   // Amazing mountain range scale <1,5,1> at small gradient
        value = cbrt(cbrt(value) - value); // Really sharp peaks if also inverted...

    if (PARAM(4) != 0.0)
        value = 1.0 - value;

    if (PARAM(5) != 0.0)
        value *= PARAM(5);

    return value;
}

double f_mult1to8pairs(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 7   was f_blob2
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[7].parameter_cnt == 19u);
    const int width=8, height=4;
    double ary[width][height];
    int count = std::min(width,std::max(1,(int)PARAM_X & 0x0F)); // count of incoming pairs.
    int Method = (int)PARAM_Y & 0x0F;           // Method of multiplication.
    int ReturnTreatment = (int)PARAM_Z & 0x0F;  // How to combine all the pair results. 0 = min. 1 = max.
                                                // Add 10 to treatment to get map value back.

  //for (int i=0; i<std::min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);       // incoming first in pair values.
        ary[i][1] = PARAM(width+i); // incoming second in pair values.

        if (Method == 0.0)
        {
            ary[i][2] = ((PARAM(i) < 0.0) && (PARAM(width+i) < 0.0)) ? -1.0 : 1.0;
        }
        else
        {
          //ary[i][2] = 1.0;
        }
    }

    if (Method == 0.0)
    {
      //for (int i=0; i<std::min(width,count); i++) // As above not faster...
        for (int i=0; i<width; i++)
        {
           ary[i][3] = sqrt(fabs(ary[i][0]) * fabs(ary[i][1])) * ary[i][2]; // sqrt controls values and gradients.
        }
    }
    else // Grouping other in else as 0 mode expected most common use.
    {
        if (Method == 2.0) // Traditional multiply followed by sqrt as with method 0, but after restoring polarity.
        {
            for (int i=0; i<width; i++)
            {
                ary[i][3] = ary[i][0] * ary[i][1];
                ary[i][2] = (ary[i][3] < 0.0) ? -1.0 : 1.0;
                ary[i][3] = sqrt(fabs(ary[i][3])) * ary[i][2];
            }
        }
        else
        {
            for (int i=0; i<width; i++)
            {
               ary[i][3] = ary[i][0] * ary[i][1];
            }
        }
    }

    double value = fnHandle1to8ReturnCalc(ary,3,width,count,ReturnTreatment);
    return value;
}

// At the full 8 pairs, below 21% faster than 15 'vm' coded max()s.
double f_max1to8pairs(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 8   was f_boy_surface
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[8].parameter_cnt == 19u);
    const int width=8, height=4;
    double ary[width][height];
    int count = std::min(width,std::max(1,(int)PARAM_X & 0x0F)); // Count of incoming pairs.
  //int Method = (int)PARAM_Y & 0x0F;                  // Method of maxes. Currently unused.
    int ReturnTreatment = (int)PARAM_Z & 0x0F;         // How to combine all the pair results. 0 = min. 1 = max.
                                                       // Add 10 to treatment to get map value back.

  //for (int i=0; i<std::min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);       // incoming first in pair values.
        ary[i][1] = PARAM(width+i); // incoming second in pair values.
    }

    for (int i=0; i<std::min(width,count); i++)
    {
        ary[i][2] = std::max(ary[i][0],ary[i][1]);
    }

    double value = fnHandle1to8ReturnCalc(ary,2,width,count,ReturnTreatment);
    return value;
}

// At the full 8 pairs, below 21% faster than 15 'vm' coded min()s.
double f_min1to8pairs(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 9   was f_comma
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[9].parameter_cnt == 19u);
    const int width=8, height=4;
    double ary[width][height];
    int count = std::min(width,std::max(1,(int)PARAM_X & 0x0F)); // Count of incoming pairs.
  //int Method = (int)PARAM_Y & 0x0F;           // Method of mins. Currently unused.
    int ReturnTreatment = (int)PARAM_Z & 0x0F;  // How to combine all the pair results. 0 = min. 1 = max.
                                                // Add 10 to treatment to get map value back.

  //for (int i=0; i<std::min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);       // incoming first in pair values.
        ary[i][1] = PARAM(width+i); // incoming second in pair values.
    }

    for (int i=0; i<std::min(width,count); i++)
    {
        ary[i][2] = std::min(ary[i][0],ary[i][1]);
    }

    double value = fnHandle1to8ReturnCalc(ary,2,width,count,ReturnTreatment);
    return value;
}

double f_morph2to9fncts(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 10   was f_cross_ellipsoids
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[10].parameter_cnt == 21u);
    const int width=9, height=2;
    double ary[width][height];
    double valmin,valmax;
    int minValIndex;

    bool returnMapIndexValue = (PARAM_X == 0.0) ? true : false; // If count is zero -> map

    // First a prefilter to min of all values while doing some work for possible
    // map index return too. Note! The prefilter causes false max gradient warnings.
    valmax = std::max(+1e-6,PARAM_Z); // Positive value clamp.
    minValIndex = 0;
    valmin = PARAM(0);
    for (int i=1; i<width; i++)
        if (PARAM(i) < valmin)
        {
            if (PARAM(i+width) > gkDBL_epsilon)  // Use this value if weight > 0.0
            {
                minValIndex = i;
                valmin = PARAM(i);
            }
        }
    if (returnMapIndexValue)
        return (double)(minValIndex+1)/10.0;
    else if (valmin > valmax)
        return valmin;

    // OK. Past prefilter and not returning map index so do all of it.
    int count = std::min(width,std::max(2,(int)PARAM_X & 0x0F)); // count of value, weight pairs.
    valmin = std::min(-1e-6,PARAM_Y); // Negative value clamp.

    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);                     // incoming values.
        ary[i][1] = std::max(0.0,PARAM(width+i)); // incoming weights.
    }

    int positiveWeightCount = 0;
    double weightTotal = 0.0;
    for (int i=0; i<count; i++)
    {
        if (ary[i][1] > gkDBL_epsilon) // inc weight count if weight > 0.0
            positiveWeightCount += 1;
        weightTotal += ary[i][1];
    }
    if (weightTotal <= 0.0)
        return -1;
    for (int i=0; i<count; i++)
    {
        ary[i][1] /= weightTotal;  // normalize incoming weights.
    }

    valmax /= positiveWeightCount; // Done so positive values do not overwhelm negative shapes...
    double value = 0.0;
    for (int i=0; i<count; i++)
    {
        double tmpvalue = std::min(valmax,std::max(valmin,ary[i][0])); // Clamp to enable more morphing...
        value += (tmpvalue * (ary[i][1]));                             // Sum values by normalized weights.
    }
    return value;
}

double f_labsweep(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 11   was f_crossed_trough
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[11].parameter_cnt == 10u);
    double value=0.0, x=PARAM_X, y=PARAM_Y, z=PARAM_Z;
    int xyOrz = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z.

    // Method here modeled on one posted by Tor Olav Kristensen which
    // implemented arbitrary point a to point b rounded cap cylinders.

    Vector3d pA = Vector3d(PARAM(1),PARAM(2),PARAM(3));
    Vector3d vL = Vector3d(PARAM(4),PARAM(5),PARAM(6)) - pA;
    double   LL = dot(vL,vL);
    Vector3d vD = vL/LL;
    double   AD = dot(pA,vD);

    if (xyOrz == 0)
        value = x - pA.x() - std::max(0.0,std::min(x*vD.x() + y*vD.y() + z*vD.z() - AD,1.0))*vL.x();
    else if (xyOrz == 1)
        value = y - pA.y() - std::max(0.0,std::min(x*vD.x() + y*vD.y() + z*vD.z() - AD,1.0))*vL.y();
    else if (xyOrz == 2)
        value = z - pA.z() - std::max(0.0,std::min(x*vD.x() + y*vD.y() + z*vD.z() - AD,1.0))*vL.z();

    return value;
}

double f_agate(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 12   was f_cubic_saddle
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[12].parameter_cnt == 10u);
    // f_agate(x,y,z,2,1.0,0.77,6,0.5,2.0,2)  <-- Matches pattern agate.
    double value    = 0.0;
    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int xyOrz       = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z. (Note. pattern agate always uses z)
    double scaleTurb   = fabs(PARAM(1));     // pattern agate default is 1.0.
    double powExponent = fabs(PARAM(2));     // pattern agate always uses 0.77.

    TurbulenceWarp Turb;
    Turb.Octaves    = std::min(10,std::max(1,(int)PARAM(3) & 0x07)); // pattern agate defaults to 6;
    Turb.Omega      = PARAM(4);                                      // pattern agate defaults to 0.5;
    Turb.Lambda     = std::min(10.0,std::max(-10.0,PARAM(5)));       // pattern agate degaults to 2.0;

    int noiseGen    = std::max(1,(int)PARAM(6) & 0x03);  // Currently 1-3 only allowed values.

    // Note. Agate default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    value    = scaleTurb * Turbulence(EPoint, &Turb, noiseGen);

    double xyOrzValue = PARAM_Z;
    if (xyOrz == 0)
        xyOrzValue = PARAM_X;
    else if (xyOrz == 1)
        xyOrzValue = PARAM_Y;

    value = 0.5 * (fnCycloidal_Wave(1.3 * value + 1.1 * xyOrzValue) + 1.0);

    value = std::max(0.0,std::min(1.0,value));

    value = pow(value, powExponent);

    return(value);
}

double f_onion(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 13   was f_cushion
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[13].parameter_cnt == 6u);
    Vector3d EPoint   = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    double layerDepth = std::max(gkMinIsectDepthReturned,fabs(PARAM(0)));  // Min 4.4e-8
    int wavetype      = (int)PARAM(1) & 0x01;  // 0 ramp, 1 triangular
    bool functionInterval = (PARAM(2) != 1.0) ? false : true;

    double value;
    if ( wavetype == 0 )
    {
        value = fmod(EPoint.length(),layerDepth);
    }
    else
    {
        value = (fmod(EPoint.length(),layerDepth)-(layerDepth/2.0));
        value = std::min(1.0,std::max(0.0,1.0-fabs(value)*2.0));
    }

    if ( functionInterval )
    {
        value = (2.0*value) - 1.0;
    }
    return value;
}

double f_sign(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 14   was f_devils_curve
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[14].parameter_cnt == 1u);
    if (PARAM_X < 0.0)
        return -1.0;
    else
        return +1.0;
}

double f_sign0(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 15   was f_devils_curve2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[15].parameter_cnt == 3u);
    double value = PARAM_X + PARAM_Z;  // x the value, z an offset (usually 0.0)

    if (value < 0.0)
        return -1.0;
    else if (value > 0.0)
        return +1.0;
    else
    {
        if (PARAM_Y == 0.0)
            return 0.0;
        else if (PARAM_Y < 0.0)
            return -1.0;
        else
            return +1.0;
    }
}

double f_cbrt(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 16   was f_dupin_cyclid
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[16].parameter_cnt == 3u);
    double value = PARAM_X;
    int mode     = (int)PARAM_Y & 0x01; // 0 = just cbrt(s), If 1 only magnitude >1 values.
    int extras   = (int)PARAM_Z & 0x0F; // Additional cbrt()s of input value.
    if (mode == 1)
        for (int i=0; i<=extras; i++)
        {
            if (fabs(value) > 1.0)
                value = std::cbrt(value);
        }
    else
        for (int i=0; i<=extras; i++)
        {
            value = std::cbrt(value);
        }

    return (fabs(value) <= gkDBL_epsilon) ? 0.0 : value;
}

// Previously f_ellipsoid() only worked for isosurfaces of threshold 1.0 and
// the axis scaling was inverted. In other words, it was not a general function
// thought relatively fast. Changed in povr May 14, 2020 to align with standard
// root/threshold behavior.  Further, extended to pass exponent as this enables
// some f_superellipsoid results at a much lower compute cost.
double f_ellipsoid(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 17
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[17].parameter_cnt == 9u);
    int sqrtCount = (int)PARAM(5) & 0x07; // How many times to take sqrt to reduce gradient.
    if (PARAM(0) == 0.0)
    {
        double x = PARAM_X,  y = PARAM_Y,  z = PARAM_Z;
        double a = PARAM(1), b = PARAM(2), c = PARAM(3);
        double e = PARAM(4);
        double value = pow((fabs(x/a)),e)+pow((fabs(y/b)),e)+pow((fabs(z/c)),e);
        value = sqrt(value);   // As with sphere form one sqrt() good initial.
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrt(value);   // Lowering gradient.
        }
        return value - 1.0;
    }
    else   // Use single floats. Often much faster at the expense of
           // a little accuarcy in practice. 3.46 -> 1.998s ---> -42.25%
    {
        float x = PARAM_X,  y = PARAM_Y,  z = PARAM_Z;
        float a = PARAM(1), b = PARAM(2), c = PARAM(3);
        float e = PARAM(4);
        float value = powf((fabs(x/a)),e)+powf((fabs(y/b)),e)+powf((fabs(z/c)),e);
        value = sqrtf(value);   // As with sphere form one sqrt() good initial.
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrtf(value);   // Lowering gradient.
        }
        return (double)(value -1.0);
    }

    // Original coding below. It is relatively quick.
    //return (sqrt(PARAM_X*PARAM_X * PARAM(0)*PARAM(0) +
    //             PARAM_Y*PARAM_Y * PARAM(1)*PARAM(1) +
    //             PARAM_Z*PARAM_Z * PARAM(2)*PARAM(2)));
}

// See f_lame for form with 'super' face curvature properties.
double f_polyhedron(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 18   was f_enneper
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[18].parameter_cnt == 16u);
    const int width=4, height=4;
    double ary[width][height];
    bool returnMapIndexValue = (PARAM(0) < 0.0) ? true : false; // If count is less than zero -> map
    int count = std::min(width,std::max(1,(int)fabs(PARAM(0)) & 0x0F)); // count of incoming gradients / values.

    ary[0][0] = PARAM_X; ary[0][1] = PARAM_Y; ary[0][2] = PARAM_Z; ary[0][3] = 0.0;

  //for (int i=1; i<std::min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
    for (int i=0; i<width; i++)
    {
        ary[1][i] = (i<count)         ? PARAM(i+1)         : 0.0;  // incoming first set of 4 values / gradients.
        ary[2][i] = (width+i<count)   ? PARAM(width+i+1)   : 0.0;  // incoming second set of 4 values / gradients.
        ary[3][i] = (width+i+i<count) ? PARAM(width+i+i+1) : 0.0;  // incoming third set of 4 values / gradients.
    }

    int loops = count/width;

    double value = 0.0;
    int minValIndex = -1;
    if (returnMapIndexValue)
    {
        value = -HUGE_VAL;
        for (int i=0; i<=loops; i++)
            for (int j=0; j<4; j++)
                if ( ((fabs(ary[i][j])) > 0.0) && ((fabs(ary[i][j])) >= value) )
                    {
                        value = fabs(ary[i][j]);
                        minValIndex = (i * 4) + j;
                    }
        if (minValIndex<3)  // The ary[0][3] is skipped over to make code above simpler.
            return (double)(minValIndex+1)/16.0;
        else
            return (double)(minValIndex)/16.0;
    }
    else
    {
        for (int i=0; i<=loops; i++)
        {
            value += fabs(ary[i][0]) + fabs(ary[i][1]) + fabs(ary[i][2]) + fabs(ary[i][3]);
        }
    }

    value -= 1;
    return value;
}

// See f_polyhedron for form implementing e==1 flat faces more effeciently.
// SDL example:
// function { sqrt((pow((abs(x)/a),e)+pow((abs(y)/b),e)+pow((abs(z)/c),e) +
// pow((abs(f_gradient(x,y,z,-1,0,-1))/0.9),e) +
// pow((abs(f_gradient(x,y,z,-1,0,+1))/0.9),e)
// )*(3/7)) - 1.0}

double f_lame(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 19   was f_flange_cover
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[19].parameter_cnt == 19u);
    const int width=8, height=2;
    bool returnMapIndexValue = (PARAM_X < 0.0) ? true : false; // If count is less than zero -> map
    int count = std::min(width,std::max(1,(int)fabs(PARAM_X) & 0x0F)); // count of incoming gradient, pow exponent pairs.
    int FloatMethod = (int)PARAM_Y & 0x0F;  // Whether using double (0) or single (!0) floats.
    int sqrtCount   = (int)PARAM_Z & 0x07;  // How many times to take sqrt to reduce gradient.

    int minValIndex = -1;
    if (FloatMethod == 0)
    {
        double ary[width][height];
        double value = 0, maxVal = -HUGE_VAL;
      //for (int i=0; i<std::min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
        for (int i=0; i<width; i++)
        {
            ary[i][0] = PARAM(i);       // incoming first in pair values.
            ary[i][1] = PARAM(width+i); // incoming second in pair values.
        }
        for (int i=0; i<std::min(width,count); i++)
        {
            value += pow(fabs(ary[i][0]),ary[i][1]);
            if (returnMapIndexValue && (fabs(ary[i][0])>=maxVal))
            {
                maxVal = fabs(ary[i][0]);
                minValIndex = i;
            }
        }
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrt(value);
        }
        if (!returnMapIndexValue) return value -1.0;
    }
    else
    {
        float ary[width][height];
        float value = 0, maxVal = -HUGE_VAL;
      //for (int i=0; i<std::min(width,count); i++) // In testing at 2 values this not faster. SIMD optimization below?
        for (int i=0; i<width; i++)
        {
            ary[i][0] = PARAM(i);       // incoming first in pair values.
            ary[i][1] = PARAM(width+i); // incoming second in pair values.
        }
        for (int i=0; i<std::min(width,count); i++)
        {
            value += powf(fabs(ary[i][0]),ary[i][1]);
            if (returnMapIndexValue && (fabs(ary[i][0])>=maxVal))
            {
                maxVal = fabs(ary[i][0]);
                minValIndex = i;
            }
        }
        for (int i=0; i<sqrtCount; i++)
        {
            value = sqrtf(value);
        }
        if (!returnMapIndexValue) return (double)(value -1.0);
    }
    return (double)(minValIndex+1)/10.0; // Only reach when returnMapIndexValue true.
}

double f_signpow(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 20   was f_folium_surface
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[20].parameter_cnt == 3u);
    double value = fabs(PARAM_X);
    if (value == 0.0) return 0.0;
    double sign  = (PARAM_X >= 0.0) ? +1.0 : -1.0;
    double ep    = std::max(PARAM_Y,gkMinIsectDepthReturned);
    double en    = std::max(PARAM_Z,gkMinIsectDepthReturned);
    if (sign == 1.0)
    {
        value = pow(value,ep);
    }
    else
    {
        value = pow(value,en);
    }
    return value*sign;
}

double f_signsqrt(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 21   was f_folium_surface
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[21].parameter_cnt == 1u);
    double value = fabs(PARAM_X);
    if (value == 0.0) return 0.0;
    double sign = (PARAM_X >= 0.0) ? +1.0 : -1.0;
    return sqrt(value)*sign;
}

double f_vangle(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 22   was f_glob
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[22].parameter_cnt == 7u);
    Vector3d v1n = Vector3d(PARAM_X,PARAM_Y,PARAM_Z).normalized();
    Vector3d v2n = Vector3d(PARAM(0),PARAM(1),PARAM(2)).normalized();
    double value;

    if (PARAM(3) == 0.0)
    {
      // Call to dot (no link time opt anyway) is costly.
      //value = std::acos(std::max(-1.0,std::min(1.0,dot(v1n,v2n))));
        double dotv = v1n.x()*v2n.x() + v1n.y()*v2n.y() + v1n.z()*v2n.z();
        value = std::acos(std::max(-1.0,std::min(1.0,dotv)));
    }
    else
        value = (2*atan2((v1n - v2n).length(), (v1n + v2n).length() ));

    return value;
}

double f_turbulence(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 23    was f_heart
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[23].parameter_cnt == 11u);
    double value    = 0.0;
    int xyOrz       = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z. (Use to get different x,y,z values at/near point)
    int noiseGen    = std::max(1,(int)PARAM(1) & 0x03);  // Currently 1-3 only allowed values.
    double freqTurb = std::max(gkMinIsectDepthReturned,fabs(PARAM(2))); // scale magitude of turbulence() after any normalization.
    int normalize   = (int)PARAM(3) & 0x0F;  // 0 = none, 1 = 0 to +1, 2 = -1 to +1.
                                             //           3 clamps to, 4 clamps to range.
                                             //           5,7 triangle, cyloidal to  0 - +1 range.
                                             //           6,8 triangle, cyloidal to -1 - +1 range.
    double deviation = fabs(PARAM(4));       // How far out is tail of distribution for normalizing opts 1 and 2.

    TurbulenceWarp Turb;
    Turb.Octaves    = std::min(10,std::max(2,(int)PARAM(5) & 0x0F)); // pattern turbulence defaults to 6;
    Turb.Omega      = PARAM(6);                                      // pattern turbulence defaults to 0.5;
    Turb.Lambda     = std::min(10.0,std::max(-10.0,PARAM(7)));       // pattern turbulence defaults to 2.0;

    // By axis offsets necessary to avoid returning the same turbulence value
    // to each axis where function used at the same incoming x,y,z values to
    // realize an effective DTurbulence() (vector turbulence) without paying
    // 3x price of those calls for each return x,y or z offset.
    Vector3d EPoint;
    if (xyOrz == 0)
        EPoint = Vector3d(PARAM_X+(M_PI*sqrt(2)), PARAM_Y, PARAM_Z);
    else if (xyOrz == 1)
        EPoint = Vector3d(PARAM_X, PARAM_Y+(M_PI*sqrt(3)), PARAM_Z);
    else
        EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z+(M_PI*sqrt(5)));

    // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    value = Turbulence(EPoint, &Turb, noiseGen);

    if (normalize > 0)
    {
        if (normalize > 4)
        {
            if (normalize == 5)
                value = fnTriangle_Wave(value);
            else if (normalize == 6)
                value = (fnTriangle_Wave(value) - 0.5) * 2.0;
            if (normalize == 7)
                value = (fnCycloidal_Wave(value) + 1.0) / 2.0;
            else if (normalize == 8)
                value = fnCycloidal_Wave(value);
        }
        else
        {
            //---
            if ((normalize == 1) || (normalize == 3))
            {
                value += deviation;       // Zero centered so shift distribution 'all' positive.
                value /= (deviation/0.5); // Value distribution to 0 to 1 range.
            }
            else if ((normalize == 2) || (normalize == 4))
                value /= (deviation/1.0); // Stretch to -1 to 1 range.
            //---
            if (normalize == 3)
                value = std::max(0.0,std::min(1.0,value));
            else if (normalize == 4)
                value = std::max(-1.0,std::min(1.0,value));
        }
    }

 // Vector3d result;
 // DTurbulence(result, EPoint, &Turb);  // !!! What about noise generator?  (At .x() looks OK, using DNoise)
 // DNoise(result, EPoint);              // This is centered and like 0 omega turbulence...
 // value = scaleTurb * result.x();
 // value = result.x();
 // value = Noise(EPoint,noiseGen);
//  void DTurbulence(Vector3d& result, const Vector3d& EPoint, const GenericTurbulenceWarp *Turb)

    return (value * freqTurb);
}

double f_bump(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 24   was f_helical_torus
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[24].parameter_cnt == 3u);
    double value = std::max(-1.0+gkMinIsectDepthReturned,
                   std::min(1.0-gkMinIsectDepthReturned,PARAM_X));
    double sign = (value < 0) ? -1.0 : +1.0;
    int powExpMult = (int)PARAM_Y & 0x7F;  // Usual rounded bump use multiplier of 1. 20 practical top depending.
    int normalize  = (int)PARAM_Z & 0x03;  // If 1, the result is normalized to 0-1 range.
                                           // If 2, result is normalized to -1.0 to 1.0 range with negative inputs
                                           // going toward -1 and positive toward 1.0 respectively. ?
    value = exp(1.0/(pow(value,2.0*std::max(1,powExpMult))-1));

    if (normalize == 0)
        return value;
    else if (normalize <= 2)
    {
        value *= exp(1.0);
        if (normalize == 2)
            value *= sign;
    }

    return value;
}

double f_helix1(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 25
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[25].parameter_cnt == 10u);
    double r, r2, temp, th, ph, x2;
    /* helix                                     *
     *   p[0] : number of helix                  *
     *   p[1] : frequency                        *
     *   p[2] : minor radius                     *
     *   p[3] : major radius                     *
     *   p[4] : shape parameter                  *
     *   p[5] : cross section                    *
     *             p[5] = 1:  circle             *
     *                  = 2:  diamond            *
     *                  < 1:  rectangle(rounded) *
     *   p[6] : rotation angle for p[5]<1        */

    r = sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z);
    if ((PARAM_X == 0) && (PARAM_Z == 0))
        PARAM_X = 0.000001;
    th = atan2(PARAM_Z, PARAM_X);
    th = fmod(th * PARAM(0) + PARAM_Y * PARAM(1) * PARAM(0), TWO_M_PI);
    if (th < 0)
        th += TWO_M_PI;
    PARAM_Z = (th - M_PI) / PARAM(4) / (PARAM(1) * PARAM(0));

    PARAM_X = r - PARAM(3);
    if (PARAM(5) == 1)
        r2 = sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z);
    else
    {
        if (PARAM(6) != 0)
        {
            th = cos(PARAM(6) * M_PI_180);
            ph = sin(PARAM(6) * M_PI_180);
            x2 = PARAM_X * th - PARAM_Z * ph;
            PARAM_Z = PARAM_X * ph + PARAM_Z * th;
            PARAM_X = x2;
        }
        if (PARAM(5) != 0)
        {
            temp = 2. / PARAM(5);
            r2 = pow((pow(fabs(PARAM_X), temp) + pow(fabs(PARAM_Z), temp)), PARAM(5) *.5);
        }
        else
            r2 = std::max(fabs(PARAM_X), fabs(PARAM_Z));
    }

    return (-PARAM(2) + std::min((PARAM(3) + r), r2));
}

double f_helix2(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 26
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[26].parameter_cnt == 8u);
    double th, ph, x2, z2, r2, temp;
    // What does NFUNCTION mean?
    /* helical shape  for (minor radius>major radius  *
     *    cross section   p[4] same as NFUNCTION = 6      */
    th = PARAM_Y * PARAM(0);
    ph = cos(th);
    th = sin(th);
    x2 = PARAM_X - PARAM(2) * ph;
    z2 = PARAM_Z - PARAM(2) * th;
    PARAM_X = x2 * ph + z2 * th;
    PARAM_Z = (-x2 * th + z2 * ph);

    if (PARAM(4) == 1)
        return (sqrt(PARAM_X * PARAM_X + PARAM_Z * PARAM_Z) - PARAM(1));
    if (PARAM(4) != 0)
    {
        temp = 2. / PARAM(4);
        r2 = pow((pow(fabs(PARAM_X), temp) + pow(fabs(PARAM_Z), temp)), PARAM(4) *.5);
    }
    else
        r2 = std::max(fabs(PARAM_X), fabs(PARAM_Z));

    return (r2 - PARAM(1));
}

double f_length(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 27   was f_hex_x
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[27].parameter_cnt == 3u);
    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    return EPoint.length();
}

double f_normal(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 28   was f_hex_y
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[28].parameter_cnt == 6u);
    double nearness = fabs(PARAM(2)); // How close to 0.0 threshold to calculate real values.
    double eptval   = PARAM_X;        // Function value at point.

    double value = 0.0;
    if (fabs(eptval) <= 2.0*nearness)
    {
        double bumpxval = PARAM_Y;  // Function value at point where x bumped up by accuracy.
        double bumpyval = PARAM_Z;  // Function value at point where y bumped.
        double bumpzval = PARAM(0); // Function value at point where z bumped.
        int xyz = std::min(3,(int)PARAM(1) & 0x03);

        Vector3d EPoint = Vector3d(eptval-bumpxval,eptval-bumpyval,eptval-bumpzval);

        if ((EPoint.x() == 0) && (EPoint.y() == 0) && (EPoint.z() == 0))
            EPoint.x() = 1.0;

        EPoint.normalize();

        double fademult = (fabs(eptval) <= nearness) ? 1.0 : std::max(0.0,1.0-(fabs(eptval)-nearness)/nearness);

        if (xyz == 0)
            value = (EPoint.x() * fademult);
        else if (xyz == 1)
            value = (EPoint.y() * fademult);
        else if (xyz == 2)
            value = (EPoint.z() * fademult);
        else if (xyz == 3)
        {
            value = fnEnc3x_f21(
                        (EPoint.x() * fademult),
                        (EPoint.y() * fademult),
                        (EPoint.z() * fademult),
                    6);
        }

        return value;
    }
    else
        return value;
}

double f_hetero_mf(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 29
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[29].parameter_cnt == 10u);
    double Scale  =  PARAM(6); // Scale overall result before return.
    if (fabs(Scale) < gkDBL_epsilon)
        return 0.0;

    double H      = -PARAM(0);   // ? Negative so pow(Lac,-H) acts as 1.0/pow().
    double Lacu   =  PARAM(1);   // Lacunarity (gap).
    int Oct       =  std::max(0,(int)PARAM(2)); // Octaves.
    double OctR   =  std::max(0.0,PARAM(2)) - (double)Oct;  // Octaves remainder.
    double Offset =  PARAM(3);  // Offset.
    int T         = std::max(0,(int)PARAM(4));
    int ngen      = (int)PARAM(5) & 0x03;  // Passed 0 defaults to 1 in Noise().

    Vector3d V1 = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    double signal = (Noise(V1, ngen)*2.0 - 1.0) + Offset;

    V1 *= Lacu;

    double p1_2_mp0 = powf(Lacu,H), ea = p1_2_mp0; // About 6% faster no difference in result seen.

    for (int i = 1; i < Oct; i++)
    {
        // Make a noisy increment and scale it by f^(-H)
        double inc = ((Noise(V1,ngen)*2.0 - 1.0) + Offset) * ea;

        // Scale the increment by (the current signal)^T at V1 so that T=0
        // gives 'straight' 1/f, T=1 gives heterogeneous fractal, etc

        double lastInc = inc;
        for (int p = T; p > 0; --p)
        {
            inc *= signal;
            if (fabs(inc) > 1.0)
            {                   // For larger T values must clamp because high freq noise
                inc = lastInc;  // ~1 or greater out of noise gen can explode in magnitude.
                break;          // This was a effective bug in the original implementation
            }                   // sometimes causing infinite isosurface gradients.
            else
                lastInc = inc;
        }

        // Note. inc *= pow(signal,T); pow(), powf() both much slower than the
        // looping with inc *= signal; given all integer powers;

        signal += inc;
        // Go to next 'octave'
        V1 *= Lacu;
        ea *= p1_2_mp0;
        if (fabs(ea) < gkDBL_epsilon)
            break;
    }

    if (OctR != 0.0)
    {
        // Do something with fraction part of octave
        double inc = ((Noise(V1,ngen)*2.0-1.0) + Offset) * ea;
        signal += OctR * inc * signal;
    }

    signal *= Scale;

    return (fabs(signal) < gkDBL_epsilon) ? 0.0 : signal;
}

double f_normalized(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 30   was f_hunt_surface
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[30].parameter_cnt == 4u);
    int xyz = std::min(2,(int)PARAM(0) & 0x03);
    Vector3d EPoint = Vector3d(PARAM_X,PARAM_Y,PARAM_Z);

    double value = 0.0;
    double len = EPoint.length();

    if (len != 0.0)
    {
        if (xyz == 0)
            value = (EPoint.x() / len);
        else if (xyz == 1)
            value = (EPoint.y() / len);
        else
            value = (EPoint.z() / len);
    }

    return value;
}

double f_boom(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 31   was f_hyperbolic_torus
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[31].parameter_cnt == 6u);
    char str[1024];
    snprintf(str,sizeof(str),
            "\nf_boom\n1(x) -> %.16g,\n2(y) -> %.16g,\n3(z) -> %.16g,\n4(0) -> %.16g,\n5(1) -> %.16g,\n6(2) -> %.16g\n",
            PARAM_X,
            PARAM_Y,
            PARAM_Z,
            PARAM(0),
            PARAM(1),
            PARAM(2));
    throw POV_EXCEPTION_STRING(str);
    return 0.0;
}

double f_box(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 32   was f_isect_ellipsoids
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[32].parameter_cnt == 7u);
    int mode = (int)PARAM(0) & 0x01; // If 0, normalizes values to scale.
                                     // If 1, Smallest scale dimension sets value range.
    double x  = fabs(PARAM_X);
    double y  = fabs(PARAM_Y);
    double z  = fabs(PARAM_Z);
    double sx = std::max(gkMinIsectDepthReturned,PARAM(1));
    double sy = std::max(gkMinIsectDepthReturned,PARAM(2));
    double sz = std::max(gkMinIsectDepthReturned,PARAM(3));

    if (mode == 0)
    {
        x = x/sx - 1.0;
        y = y/sy - 1.0;
        z = z/sz - 1.0;
    }
    else
    {
        x  -= sx;
        y  -= sy;
        z  -= sz;
    }

    double value = std::max(x,std::max(y,z));
    return (fabs(value) < gkDBL_epsilon) ? 0.0 : value;
}

double f_length_sqr(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 33   was f_kampyle_of_eudoxus
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[33].parameter_cnt == 3u);
    return ( PARAM_X*PARAM_X + PARAM_Y*PARAM_Y + PARAM_Z*PARAM_Z );
}

double f_triangle_wave(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 34   was f_kampyle_of_eudoxus_2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[34].parameter_cnt == 2u);
    double value = fnTriangle_Wave(PARAM_X);
    if (PARAM_Y != 0)
        value = value * 2.0 - 1.0;
    return value;
}

double f_sine_wave(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 35   was f_klein_bottle
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[35].parameter_cnt == 2u);
    double value = PARAM_X;
    int mode = (int)PARAM_Y;

    // A conditional to flip to float costing 3% for double case so created
    // f_sine_wavef version.

    // Instead of calling fnCycloidal_Wave() trying simpler raw calcs
    // Partly because we have sin and sinf versions.
    if (mode == 0)
        value = (sin(value*TWO_M_PI) + 1.0) * 0.5; // 0 - 1 magnitude within 0 to 1 interval
    else if (mode == 1)
        value = sin(value*TWO_M_PI);               // -1 - 1 magnitude within 0 to 1 interval
    else if (mode == 2)
        value = sin(value*M_PI);                   // -1 to 1 magnitude within -1 to 1 interval

    return value;
}

double f_scallop_wave(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 36  was f_kummer_surface_v1
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[36].parameter_cnt == 2u);
    double value = PARAM_X;
    int mode = (int)PARAM_Y;

    // Instead of calling fnCycloidal_Wave() trying simpler raw calcs
    // Partly because we have sin and sinf version.
    if (mode == 0)
        value = fabs(sin(value*M_PI));                 // 0 to 1 magnitude within 0 to 1 interval
    else if (mode == 1)
        value = fabs(sin(value*M_PI)) * 2.0 - 1.0;     // -1 to 1 magnitude within 0 to 1 interval
    else if (mode == 2)
        value = fabs(sin(value*TWO_M_PI));             // 0 to 1 magnitude twice within 0 to 1 interval
    else if (mode == 3)
        value = fabs(sin(value*TWO_M_PI)) * 2.0 - 1.0; // -1 to 1 magnitude twice within 0 to 1 interval

    return value;
}

double f_sine_wavef(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 37   was f_kummer_surface_v2
{   // Float version of f_sine_wave().  11% faster 2 threaded cores.
    POV_STATIC_ASSERT(POVFPU_TrapTable[37].parameter_cnt == 2u);
    double value = PARAM_X;
    int mode = (int)PARAM_Y;

    // Instead of calling fnCycloidal_Wave() trying simpler raw calcs
    // Partly because we have sin and sinf versions.
    if (mode == 0)
        value = (sinf(value*TWO_M_PI) + 1.0) * 0.5; // 0 - 1 magnitude within 0 to 1 interval
    else if (mode == 1)
        value = sinf(value*TWO_M_PI);               // -1 - 1 magnitude within 0 to 1 interval
    else if (mode == 2)
        value = sinf(value*M_PI);                   // -1 to 1 magnitude within -1 to 1 interval

    return value;
}

double f_scallop_wavef(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 38  was f_lemniscate_of_gerono
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[38].parameter_cnt == 2u);
    double value = PARAM_X;
    int mode = (int)PARAM_Y;

    // Instead of calling fnCycloidal_Wave() trying simpler raw calcs
    // Partly because we have sin and sinf versions.
    if (mode == 0)
        value = fabs(sinf(value*M_PI));                 // 0 to 1 magnitude within 0 to 1 interval
    else if (mode == 1)
        value = fabs(sinf(value*M_PI)) * 2.0 - 1.0;     // -1 to 1 magnitude within 0 to 1 interval
    else if (mode == 2)
        value = fabs(sinf(value*TWO_M_PI));             // 0 to 1 magnitude twice within 0 to 1 interval
    else if (mode == 3)
        value = fabs(sinf(value*TWO_M_PI)) * 2.0 - 1.0; // -1 to 1 magnitude twice within 0 to 1 interval

    return value;
}

double f_cubic_wave(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 39  was f_lemniscate_of_gerono_2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[39].parameter_cnt == 2u);
    double value = PARAM_X;
    int mode = (int)PARAM_Y;

    if (fabs(value) <= 1.0)
        value = (Sqr(value) * ((-2.0 * fabs(value)) + 3.0));
    else
        value = 1.0;

    if (mode == 1)              // input expected in -1 to 1 range, maintain.
    {
        if (PARAM_X < 0.0)
            value *= -1.0;
    }
    else if (mode == 2)        // input expected in 0 to 1 range, want -1 to 1 out.
        value = (2.0 * value) - 1.0;

    return value;
}

double f_boxb(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 40  was f_mesh1
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[40].parameter_cnt == 11u);
    double x  = PARAM_X;
    double y  = PARAM_Y;
    double z  = PARAM_Z;
    double sx = std::max(gkMinIsectDepthReturned,PARAM(0)); // Scalings
    double sy = std::max(gkMinIsectDepthReturned,PARAM(1));
    double sz = std::max(gkMinIsectDepthReturned,PARAM(2));
    double ox = PARAM(3); // Offsets
    double oy = PARAM(4);
    double oz = PARAM(5);
    double b  = PARAM(6); // Value/Function being bound.
    double sb = std::max(gkMinIsectDepthReturned,PARAM(7)); // Function scaling / skin thickness.

    x = fabs(x - ox) - sx;
    y = fabs(y - oy) - sy;
    z = fabs(z - oz) - sz;

    double value  = std::max(x,std::max(y,z))/2.0;
    double valueb = fabs(std::max(b,value)) - sb;

    value = std::max(value,valueb);

    return (fabs(value) < gkDBL_epsilon) ? 0.0 : value;
}

double f_wrap(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 41  was f_mitre
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[41].parameter_cnt == 4u);
    int mode = (int)PARAM(0);
    double value = PARAM_X;
    double loval = PARAM_Y;
    double hival = PARAM_Z;
    double range = (hival - loval);

    // { mod(x-y,z-y) + select(mod(x-y,z-y),z,y }  <-- Old POV-Ray (f|F)_clamp macro.
    if (mode == 0)
    {
        value -= loval;
        double modvalue = fmod(value,range);

        if (modvalue < 0.0)
            value = modvalue + hival;
        else
            value = modvalue + loval;
    }
    else if (mode == 1)
    {
        if ((value >= loval) && (value <= hival))
            return value;

        value -= loval;
        double modvalue = fmod(value,range);

        if (fabs(modvalue) < gkDBL_epsilon)
        {
            if (value == -range)
               return hival;

            if (fabs(value) > range)
            {
                unsigned long long tmpCount = std::abs(value / range);
                bool iseven = ((tmpCount % 2) == 0) ? true : false;
                if (iseven)
                    return loval;
                else
                    return hival;
            }
            else
            {
                if (modvalue < 0.0)
                    value = modvalue + hival;
                else
                    value = modvalue + loval;
            }
        }
        else
        {
            if (modvalue < 0.0)
                value = modvalue + hival;
            else
                value = modvalue + loval;
        }
    }

    return value;
}

double f_elliptical_cylinder(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 42  was f_nodal_cubic()
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[42].parameter_cnt == 6u);
    double x = PARAM_X,                z = PARAM_Z;  // y = PARAM_Y not used.
    double a = PARAM(0), b = PARAM(1), e = std::max(gkDBL_epsilon,fabs(PARAM(2)));
    double value = pow((fabs(x/a)),e)+pow((fabs(z/b)),e);
    value = pow(value,std::min(1.0,1.0/e));   // As with sphere form one sqrt() good initial.
    return value - 1.0;
}

double f_odd(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 43  Special use in returning 99 to indicate fork.
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[43].parameter_cnt == 4u);
    return(99.0);
}

double f_planar(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 44  was f_ovals_of_cassini
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[44].parameter_cnt == 3u);
    int mode = (int)PARAM_Z & 0x01; // If 0, normalizes values to scale.
                                    // If 1, Smallest scale dimension sets value range.
    double p  = fabs(PARAM_X);      // x, y, z, or whatever gradient.
    double sp = std::max(gkMinIsectDepthReturned,PARAM_Y);

    double value;
    if (mode == 0)
        value = fabs(p)/sp - 1.0;
    else
        value = fabs(p) - sp;

    return (fabs(value) < gkDBL_epsilon) ? 0.0 : value;
}

double f_paraboloid(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 45
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[45].parameter_cnt == 5u);
    // Normally up or down facing depending on incoming y polrity. The wantdual
    // parameter creates paraboloids above and below the x,z plane.
    int wantdual = (int)PARAM(0) & 0x01;  // If 1, create dual paraboloids
    int cbrtoff  = (int)PARAM(1) & 0x01;  // If 1, return values >1 no cbrt().
    double y;
    if (wantdual == 1)
        y = fabs(PARAM_Y);
    else
        y = PARAM_Y;

    double value = (PARAM_X*PARAM_X - y  + PARAM_Z*PARAM_Z);

    if ((fabs(value) > 1.0) && (cbrtoff == 0))
        value = std::cbrt(value);  // Helps the gradient for typical shapes.

    return (fabs(value) <= gkDBL_epsilon) ? 0.0 : value;
}

double f_quilt_cubic(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 46  was f_parabolic_torus
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[46].parameter_cnt == 7u);
    double t = PARAM_X;
    double nx0 = PARAM_Y, ny0 = PARAM_Z;
    double nx1 = PARAM(0),ny1 = PARAM(1);
    double c0  = PARAM(2), c1 = PARAM(3);
    double delta = nx1 - nx0;
    t -= nx0;
    t /= delta;

    // Orginal quilt_cubic function.
    // val= (double)(itcubed*n1+(tcubed)*n2+3*t*(itsqrd)*p1+3*(tsqrd)*(it)*p2);
    double it=(1-t);
    double tsqrd=t*t;
    double tcubed=t*t*t;
    double itsqrd=it*it;
    double itcubed=it*it*it;

    double value = itcubed*ny0 + tcubed*ny1 + 3*t*itsqrd*c0 + 3*tsqrd*it*c1;

    return value;
}

double f_ph(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 47
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[47].parameter_cnt == 3u);

  //---Thiis encoding tested +5.5% slower on my i3.
  //double value = atan2(std::hypot(PARAM_X,PARAM_Z),PARAM_Y);
  //return value;

    double value = atan2(sqrt(PARAM_X*PARAM_X + PARAM_Z*PARAM_Z),PARAM_Y);
    return value;
}

double f_dec_nbit3c(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 48   was f_pillow
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[48].parameter_cnt == 5u);
    int bitDepth    = std::min(17,std::max(1,(int)PARAM_X & 0x1F));
    bool normalized = (PARAM_Y == 1.0) ? true : false;
    double valR     = PARAM_Z;
    double valG     = PARAM(0);
    double valB     = PARAM(1);

    double value;
    double maxVal  = pow(2.0,bitDepth);
    if (normalized)
    {
        valG = valG / maxVal;
        valB = valB / pow(maxVal,2.0);
        value = valR + valG + valB;
    }
    else
    {
        value = valR + valG + valB;
    }

    return value;
}

double f_enc_nbit3c(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 49   was f_piriform
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[49].parameter_cnt == 4u);
    int bitDepth   = std::min(17,std::max(2,(int)PARAM_X & 0x1F));
    int channel    = std::min(3,std::max(0,(int)PARAM_Y & 0x03));
    bool normalize = (PARAM_Z == 1.0) ? true : false;
    double value   = PARAM(0);

    double maxVal  = pow(2.0,bitDepth);
    double adj     = (maxVal-1.0)/maxVal;

    double VRr, G, VGr, B, VBr;
    double R   = adj * value;
  //double VRr = value - R;
  //double G   = adj * VRr;
  //double VGr = VRr - G;
  //double B   = adj * VGr;
  //double VBr = VGr - B;   // If channel =  3. Encoding epsilon for bit depth.
    switch (channel)
    {
        case 0:
                {
                    return R; // Is normalized as is.
                    break;
                }
        case 1:
                {
                    VRr = value - R;
                    G   = adj * VRr;
                    if (normalize)
                        return G * maxVal;
                    else
                        return G;
                    break;
                }
        case 2:
                {
                    VRr = value - R;
                    G   = adj * VRr;
                    VGr = VRr - G;
                    B   = adj * VGr;
                    if (normalize)
                        return B * pow(maxVal,2.0);
                    else
                        return B;
                    break;
                }
        default:  // 3
                {
                    VRr = value - R;
                    G   = adj * VRr;
                    VGr = VRr - G;
                    B   = adj * VGr;
                    VBr = VGr - B;
                    return VBr;
                  //return (R + G + B);  // Debugging. Should return incoming value to bit depth epsilon.
                    break;
                }
    } // channel switch
}

double f_hypot(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 50   was f_piriform_2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[50].parameter_cnt == 2u);
    // Is std::hypot() version always much slower than raw encoding...
    double value = sqrt(PARAM_X*PARAM_X + PARAM_Y*PARAM_Y);
    return value;
}

double f_upoly(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 51  was f_poly4
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[51].parameter_cnt == 19u);

    const int width=16;
    double ary[width];

    double t = PARAM_X;

    int order = std::min(16,std::max(0,(int)PARAM_Y & 0x1F)); // The order of equation to handle up to 16.

    double constant = PARAM_Z; // Univariate polynomial constant.

  //for (int i=1; i<std::min(width,order); i++) // In testing at two values this not faster.
    for (int i=0; i<width; i++)
        ary[i] = (i < order) ? PARAM(i) : 0.0;

    double pv = ary[0];
    for (int i=1; i < order; i++)
        pv = pv * t + ary[i];

    double value = pv * t + constant;

    return value;
}

double f_radial_tubes(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 52   was f_polytubes
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[52].parameter_cnt == 7u);

    double x = PARAM_X, z=PARAM_Z; // Incoming position. Only 'x','z' used for
                                   // radial position about 'y'. y=PARAM_Y

    double surface = PARAM(0);                        // Incoming surface 'value' tubes follow
    int    count   = std::max(1,(int)fabs(PARAM(1))); // The number of tubes about axis.
    double theta   = PARAM(2) * M_PI_180;             // Rotation if any.
    double scale   = PARAM(3);                        // How much of complete rotation to use.

    if (theta != 0.0)
    {
        double r   = sqrt(x*x + z*z); // Cartesian to polar
        double phi = atan2(z,x);
        phi += theta;                 // Adjust angle

        x = r*cos(phi);               // Polar to cartesian
        z = r*sin(phi);
    }

    double x2, z2;
    double value   = -1e4;                       // Initialization value from old code. Better one?
    for (int i=0;i<count;i++)
    {
      //x2 = x - surface*sin((2*M_PI/count*i)*scale); // Old encoding starts at -z and rotates left handed.
      //z2 = z - surface*cos((2*M_PI/count*i)*scale);
        x2 = x + surface*cos((2*M_PI/count*i)*scale); // Start at +x and rotate left handed about y (like radial).
        z2 = z - surface*sin((2*M_PI/count*i)*scale);
        value = std::max(value,-sqrt(x2*x2 + z2*z2));
    }

    return (-value*sqrt(2.0));
}

double f_dturbulence(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 53   was f_quantum
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[53].parameter_cnt == 10u);

    int xyz = (int)PARAM(0) & 0x07;  // 0=x, 1=y, 2=z. 3= x,y,z f21 bit encoded.
    double x = PARAM_X, y = PARAM_Y, z = PARAM_Z;
    Vector3d EPoint = Vector3d(x,y,z);

    double sx = PARAM(1), sy = PARAM(2), sz = PARAM(3);
    TurbulenceWarp Turb;
//  Turb.Turbulence = Vector3d(PARAM(1),PARAM(2),PARAM(3));// 3 x,y,z inbuilt magnitudes. As with Turbulence(), not used.
    Turb.Octaves    = std::min(10,std::max(1,(int)PARAM(4) & 0x0F)); // Commonly defaulted to 6;
    Turb.Omega      = PARAM(5);                                      // Commonly defaulted to 0.5;
    Turb.Lambda     = std::min(10.0,std::max(-10.0,PARAM(6)));       // Commonly defaulted to 2.0;

    double value    = 0.0;
    Vector3d VTurbulence;
    DTurbulence(VTurbulence,EPoint,&Turb);

    if (xyz >= 3)
    {
        value = fnEnc3x_f21(VTurbulence.x()*sx, VTurbulence.y()*sy, VTurbulence.z()*sz,5);
        return value;
    }
    else if (xyz == 0)
        return VTurbulence.x()*sx;
    else if (xyz == 1)
        return VTurbulence.y()*sy;
    else if (xyz == 2)
        return VTurbulence.z()*sz;
    else
        return value;
}

double f_rotate2d(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 54   was f_quartic_paraboloid
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[54].parameter_cnt == 5u);

  //double axis  = PARAM(0); // Conversion to int handled in called function.
  //double angle = PARAM(1);

  //double value = FnRotate0(PARAM_X,PARAM_Y,PARAM_Z,PARAM(0),PARAM(1));
    double value = FnRotate1(PARAM_X,PARAM_Y,PARAM_Z,PARAM(0),PARAM(1));

    return value;
}

double f_hessian_dist(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 55   was f_quartic_saddle
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[55].parameter_cnt == 5u);

    double x = PARAM_X, z = PARAM_Z;
    double theta  = PARAM(0) * M_PI_180;
    double offset = PARAM(1);

    double value = z*cos(theta) + x*sin(theta) - offset;

    return value;
}

double f_quartic_cylinder(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 56
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[56].parameter_cnt == 5u);
    double x2 = PARAM_X*PARAM_X, z2 = PARAM_Z*PARAM_Z;
    double invfalloff = PARAM(1)*PARAM(1);

    double value = ( (x2 + z2) * PARAM_Y*PARAM_Y + invfalloff * (x2 + z2) -
            invfalloff * PARAM(0)*PARAM(0)  );

    if (fabs(value) > 1.0)
        value = std::cbrt(value);  // Helps the gradient for typical shapes.

    return (fabs(value) <= gkDBL_epsilon) ? 0.0 : value;
}

double f_r(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 57
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[57].parameter_cnt == 3u);
    return sqrt(PARAM_X*PARAM_X + PARAM_Y*PARAM_Y + PARAM_Z*PARAM_Z);
}

double f_ridge(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 58
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[58].parameter_cnt == 10u);
    double RtrnScale = PARAM(6); // Scale overall result before return.
    if (fabs(RtrnScale) < gkDBL_epsilon)
        return 0.0;

    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    double Lambda   = PARAM(0);
    int Octaves     = std::max(0,(int)PARAM(1));
    double OctavesR = std::max(0.0,PARAM(1)) - (double)Octaves; // Octaves remainder.
    double Omega    = PARAM(2);
    double Offset   = PARAM(3);
    double Ridge    = PARAM(4);
    int ngen        = (int)PARAM(5) & 0x03;

    double rscale = 1.0 / std::max(Ridge, 1.0 - Ridge);
    double scale  = 1.0 / std::max(Offset, 1.0 - Offset);

    double v     = fabs(Noise(EPoint,ngen) - Ridge) * rscale;
    double value = (v - Offset);
    double tot   = 1.0;
    double l     = Lambda;
    double o     = Omega;

    for (int i = 2; i <= Octaves; i++)
    {
        Vector3d tmpvec = EPoint * l;
        v = fabs(Noise(tmpvec,ngen) - Ridge) * rscale;
        value += o * (v - Offset);
        tot += o;
        l *= Lambda;
        o *= Omega;
    }
    if (0.0 != OctavesR)
    {
        Vector3d tmpvec = EPoint * l;
        v = fabs(Noise(tmpvec,ngen) - Ridge) * rscale;
        value += o * (v - Offset) * OctavesR;
        tot += o * OctavesR;
    }
    value *= ((scale / tot) * RtrnScale);

    return (fabs(value) < gkDBL_epsilon) ? 0.0 : value;
}

double f_ridged_mf(FPUContext* ctx, double *ptr, unsigned int fn) // 59
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[59].parameter_cnt == 10u);
    double RtrnScale = PARAM(6); // Scale overall result before return.
    if (fabs(RtrnScale) < gkDBL_epsilon)
        return 0.0;

    FunctionCode *f = ctx->functionvm->GetFunction(fn);
    double *ea, signal, weight, result;

    Vector3d V1 = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    double Lac  = PARAM(1);               // Lacunarity
    int Octs    = (int)PARAM(2);          // Octaves
    double Offs = PARAM(3);               // Offset
    double Gain = PARAM(4);
    int ngen    = (int)PARAM(5) & 0x03;

    if (f->private_data == nullptr)
    {
        // TODO Is this really thread safe? Suspect might "sometimes" have a
        // small memory leak where multiple threads allocate space below at
        // nearly the same time, but f->private_data ends up pointing only to
        // the last thread done's memory.
        double H = PARAM(0);
        ea = reinterpret_cast<double *>(POV_MALLOC((Octs + 1)*sizeof(double), "exponent array"));
        double freq = 1.0;
        for (int i=0; i<=Octs; i++)
        {
            ea[i]= pow(freq,-H);
            freq *= Lac;
        }
        f->private_data = reinterpret_cast<void *>(ea);
    }
    else
    {
        ea = reinterpret_cast<double *>(f->private_data);
    }
    signal = Noise(V1,ngen)*2.0-1.0;
    if (signal < 0.0 )
        signal = -signal;
    signal = Offs - signal;
    signal *= signal;
    result = signal;
    weight = 1.0;

    for (int i=1; i<Octs; i++)
    {
        V1 *= Lac;
        weight = signal * Gain;
        if (weight > 1.0)
            weight = 1.0;
        if (weight < 0.0)
            weight = 0.0;
        signal = Noise(V1,ngen)*2.0-1.0;
        if (signal < 0.0 )
            signal = -signal;
        signal = Offs - signal;
        signal *= signal;
        signal *= weight;
        result += signal * ea[i];
    }
    result *= RtrnScale;
    return (fabs(result) < gkDBL_epsilon) ? 0.0 : result;
}

double f_rounded_box(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 60
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[60].parameter_cnt == 7u);
    double x2, y2, z2, x3, y3, z3;
    double x  = PARAM_X;
    double y  = PARAM_Y;
    double z  = PARAM_Z;
    double r  = PARAM(0);
    double sx = PARAM(1);
    double sy = PARAM(2);
    double sz = PARAM(3);

    x2 = sx - r;
    y2 = sy - r;
    z2 = sz - r;
    x3 = (x < x2) ? 0 : (x - x2); // Map v3 to origin if < upper scale less radius.
    y3 = (y < y2) ? 0 : (y - y2); // Otherwise adjust coordinates relative to sphere.
    z3 = (z < z2) ? 0 : (z - z2);
    x2 = r - sx;
    y2 = r - sy;
    z2 = r - sz;
    x = (x > x2) ? 0 : (x2 - x); // Map v3 to origin if > lower scale less radius.
    y = (y > y2) ? 0 : (y2 - y); // Otherwise adjust coordinates relative to sphere.
    z = (z > z2) ? 0 : (z2 - z);
    x = std::max(x, x3); // Use the larger 'x|y|z' of pair calculated as sphere center.
    y = std::max(y, y3);
    z = std::max(z, z3);

    // Leave -1e-6 near as is. Isosurface use 1e-6 to 1e-8 at max solver accuracy.
    // Without the subtraction the 'not rounded corner' performance is extremely bad.
    return (-r + sqrt(x*x + y*y + z*z) - 1e-6);
}

double f_sphere(FPUContext* /*ctx*/, double *ptr, unsigned int) // 61
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[61].parameter_cnt == 4u);
  // Alternatives below without sqrt bad gradients lose big with current isosurface solver.
  //double value = (PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z - PARAM(0) * PARAM(0));
  //return std::min(1.0,std::max(-1.0,value));                                                    // 145.06 -> 295.24s ---> +103.53%
  //return (PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z - PARAM(0) * PARAM(0));     // 145.06 -> 288.72s --->  +99.03%
  //return (-PARAM(0) + sqrt(PARAM_X * PARAM_X + PARAM_Y * PARAM_Y + PARAM_Z * PARAM_Z));  // Best.

    double x = PARAM_X, y = PARAM_Y, z = PARAM_Z, r = PARAM(0);
    double value = -r + sqrt(x*x + y*y + z*z);
    return value;
}

double f_enc3x_f21(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 62   was f_spikes
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[62].parameter_cnt == 4u);
                          //   float   float   float   decimal places (0-6)
                          //   |       |       |       |
    double value = fnEnc3x_f21(PARAM_X,PARAM_Y,PARAM_Z,PARAM(0));
                   // Called fn itself handles 0 param as truncated integer

    return value;
}

double f_dec3x_f21(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 63   was f_spikes_2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[63].parameter_cnt == 3u);
                          //   Encoded double
                          //   |       Decimal places
                          //   |       |       Which of three 21 bit floats to decode and return?
                          //   |       |       |       |
    double value = fnDec3x_f21(PARAM_X,PARAM_Y,PARAM_Z);
                   // Called fn itself handles Y and Z as truncated integers

    return value;
}

double f_spiral(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 64
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[64].parameter_cnt == 7u);
    double r, r2, th, temp;
    double x = PARAM_X, y = PARAM_Y, z = PARAM_Z;
    double wspace = PARAM(0);  // Min space (distance) between winding.
    double halfw  = PARAM(1);  // Winding half width (radius for 1, half min width otherwise).
    double sphr   = PARAM(2);  // Radius of the containing sphere for spiral.
    double crossx = PARAM(3);  // Cross section type. 0 Square. 1 Sphere. 2 Diamond. 3 Inward diamond.

    r = sqrt(x*x + z*z);
    if ((x == 0) && (z == 0))
        x = 0.000001;
    th = atan2(z, x);
    r = r + wspace * th / TWO_M_PI;
    r2 = fmod(r, wspace) - wspace * 0.5;

    if (crossx == 1)
        r2 = sqrt(r2 * r2 + y * y);
    else if (crossx != 0)
    {
        temp = 2 / crossx;
        r2 = pow((pow(fabs(r2), temp) + pow(fabs(y), temp)), 1. / temp);
    }
    else
        r2 = std::max(fabs(r2), fabs(y));

    r = sqrt(x*x + y*y + z*z);

    return (-std::min(sphr - r, halfw - std::min(r2, r)));
  //return (-std::min(sphr - r, halfw - std::min(r2, 1.0))); // No sphere at center...
}

double f_range2range(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 65   was f_steiners_roman
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[65].parameter_cnt == 5u);
    double value = PARAM_X;
    double ilo   = PARAM_Y,  ihi = PARAM_Z;
    double olo   = PARAM(0), ohi = PARAM(1);

    (ihi < ilo) ? (ihi = ilo + gkMinIsectDepthReturned) : ihi;
    (ohi < olo) ? (ohi = olo + gkMinIsectDepthReturned) : ohi;

    value = ((value-ilo)/(ihi-ilo))*(ohi-olo) + olo;

    return value;
}

double f_toupie(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 66   was f_strophoid
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[66].parameter_cnt == 7u);
    double value;
    double e = std::max(1,(int)PARAM(3) & 0x0F); // Integer exponents 1-15
    if (PARAM(0) == 0.0)
    {
        double MjR = PARAM(1), MnR = PARAM(2);
      // std::hypot() not used as slower
        value = std::cbrt(1.0 + pow(sqrt(PARAM_X*PARAM_X + PARAM_Z*PARAM_Z)-MjR,e) +
                pow(PARAM_Y,2.0)) - std::cbrt(1.0 + MnR);
        return value;
    }
    else
    {
        float x = PARAM_X, y = PARAM_Y, z = PARAM_Z;
        float MjR = PARAM(1), MnR = PARAM(2);
      // std::hypotf() not used as slower
        value = std::cbrtf(1.0 + powf(sqrtf(x*x + z*z)-MjR,e) +
                powf(y,2.0)) - std::cbrtf(1.0 + MnR);
        return value;
    }
}

double f_supertorus(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 67   was f_strophoid_2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[67].parameter_cnt == 8u);
    double value;

    if (PARAM(0) == 0.0)
    {
        double MjR = PARAM(1), MnR = PARAM(2);
        double cp  = 2 / std::max(gkDBL_epsilon,fabs(PARAM(3)));
        double rp  = 2 / std::max(gkDBL_epsilon,fabs(PARAM(4)));
        value = pow(pow(fabs(pow(pow(fabs(PARAM_X),rp)
                           + pow(fabs(PARAM_Z),rp),1/rp)-MjR),cp)
                    + pow(fabs(PARAM_Y),cp),1/cp)-MnR;
        return value;
    }
    else
    {
        float MjR = PARAM(1), MnR = PARAM(2);
        float cp  = 2 / std::max(gkMinIsectDepthReturned,fabs(PARAM(3)));
        float rp  = 2 / std::max(gkMinIsectDepthReturned,fabs(PARAM(4)));
        value = powf(powf(fabs(powf(powf(fabs(PARAM_X),rp)
                           + powf(fabs(PARAM_Z),rp),1/rp)-MjR),cp)
                     + powf(fabs(PARAM_Y),cp),1/cp)-MnR;
        return (double)value;
    }
}

// Used to return the shape as positive values.
// Flipped in povr Apr 15, 2020 to align with standard behavior.
double f_superellipsoid(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 68
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[68].parameter_cnt == 6u);
    double value;
    if (PARAM(0) == 0.0)
    {
        double p = 2 / PARAM(1), n = 1 / PARAM(2);
        value = -1+pow((pow((pow(fabs(PARAM_X),p)
                            +pow(fabs(PARAM_Y),p)),PARAM(1)*n)
                            +pow(fabs(PARAM_Z),2*n)),PARAM(2)*.5);
    }
    else
    {
        float p1 = PARAM(1), p2 = PARAM(2);
        float p = 2 / p1, n = 1 / p2;
        value = -1+powf((powf((powf(fabs(PARAM_X),p)
                            +powf(fabs(PARAM_Y),p)),p1*n)
                            +powf(fabs(PARAM_Z),2*n)),p2*.5);
    }
    return value;
}

double f_th(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 69
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[69].parameter_cnt == 3u);
    return( atan2(PARAM_X,PARAM_Z) );
}

double f_torus(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 70
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[70].parameter_cnt == 5u);
    // Often see form: pow(sqrt(x*x+z*z)-0.5,2) + y*y - pow(0.05,2) With last
    // minor radius calc as a constant so fewer calculations in total. However,
    // the value gradient is much worse/higher. In POV-Ray the performance
    // therefore worse (2.8x slower).

    //--- std::hypot is 9% slower in testing.
  //PARAM_X = std::hypot(PARAM_X,PARAM_Z) - PARAM(0);
  //return -PARAM(1) + std::hypot(PARAM_X,PARAM_Y);

    PARAM_X = sqrt(PARAM_X*PARAM_X + PARAM_Z*PARAM_Z) - PARAM(0);
    return -PARAM(1) + sqrt(PARAM_X*PARAM_X + PARAM_Y*PARAM_Y);
}

double f_repeat(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 71   was f_torus2
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[71].parameter_cnt == 16u);
    // 0=x, 1=y, 2=z, 3=cells val (0 to 1), 4=cells val (-1 to 1)
    int xyzOrc = std::min(4,(int)PARAM(0) & 0x07);

    double value=0.0, x=PARAM_X, y=PARAM_Y, z=PARAM_Z;

    double xcnt=PARAM(1), ycnt=PARAM(2), zcnt=PARAM(3);
    double xdlt=PARAM(4), ydlt=PARAM(5), zdlt=PARAM(6);
    double xdir=PARAM(7), ydir=PARAM(8), zdir=PARAM(9);

    if (xyzOrc >= 3)
    {
        double xcellsOff=floor(PARAM(10)), ycellsOff=floor(PARAM(11)), zcellsOff=floor(PARAM(12));
        value = std::min(gPatternRands(
                        Hash3d(
                            (int)floor((x + xcellsOff + gkDBL_epsilon)*(1.0/fabs(xdlt))),
                            (int)floor((y + ycellsOff + gkDBL_epsilon)*(1.0/fabs(ydlt))),
                            (int)floor((z + zcellsOff + gkDBL_epsilon)*(1.0/fabs(zdlt)))
                        )
                    ),1.0);
        if (xyzOrc == 4)
            return (value-0.5)*2.0;
        else
            return value;
    }

    auto lambdaRepeatVal = [&value]
         (const double xyz, const double cnt, const double dlt, const double dir,
          const bool possibleInRange) -> void
    {
        if (possibleInRange && (fabs(xyz)<=(cnt*dlt)))
        {
            if (dir == 0.0)
                value = fmod(fabs(xyz),dlt)-dlt/2.0;
            else if ((dir < 0.0) && (xyz < 0.0))
                value = fmod(fabs(xyz),dlt)-dlt/2.0;
            else if ((dir > 0.0) && (xyz > 0.0))
                value = fmod(xyz,dlt)-dlt/2.0;
        }
        else
        {
            value = -dlt/2.0;
        }
        // Should I do fabs of dlt and then value = fabs(dlt) - value on odd floor(xyz) ? to flip x,y,z ?
    };

    bool validx = ((xdir == 0.0) || ((xdir < 0.0) && (x < 0.0)) || ((xdir > 0.0) && (x > 0.0))) ? true : false;
    bool validy = ((ydir == 0.0) || ((ydir < 0.0) && (y < 0.0)) || ((ydir > 0.0) && (y > 0.0))) ? true : false;
    bool validz = ((zdir == 0.0) || ((zdir < 0.0) && (z < 0.0)) || ((zdir > 0.0) && (z > 0.0))) ? true : false;

    if (xyzOrc == 0)
    {
        lambdaRepeatVal(x,xcnt,xdlt,xdir,(validy && validz));
    }
    else if (xyzOrc == 1)
    {
        lambdaRepeatVal(y,ycnt,ydlt,ydir,(validx && validz));
    }
    else if (xyzOrc == 2)
    {
        lambdaRepeatVal(z,zcnt,zdlt,zdir,(validx && validy));
    }

    return value;
}


double f_marble(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 72   was f_torus_gumdrop
{
    // NOTE! Perhaps not the final form. Play with this some more!

    POV_STATIC_ASSERT(POVFPU_TrapTable[72].parameter_cnt == 9u);
    // f_marble(x,y,z,0,(1/3)/10,6,0.5,2.0,2)  <-- Matches pattern marble.
    double value       = 0.0;
    Vector3d EPoint = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int xyOrz       = (int)PARAM(0) & 0x03;  // 0=x, 1=y, 2=z. (Note. pattern marble always uses x)
    double scaleTurb   = fabs(PARAM(1));        // pattern marble default is 1.0/3.0*0.1.

    TurbulenceWarp Turb;
    Turb.Octaves    = std::min(10,std::max(1,(int)PARAM(2) & 0x07)); // pattern marble defaults to 6;
    Turb.Omega      = PARAM(3);                                      // pattern marble defaults to 0.5;
    Turb.Lambda     = std::min(10.0,std::max(-10.0,PARAM(4)));       // pattern marble defaults to 2.0;

    int noiseGen    = std::max(1,(int)PARAM(5) & 0x03);  // Currently 1-3 only allowed values.

    // Note. Default Turb->Turbulence vector values are 0.0 and that is OK.
    // Turbulence() uses only octaves, omega, lambda values.
    value = scaleTurb * Turbulence(EPoint,&Turb,noiseGen);

    double xyOrzValue = PARAM_X;
    if (xyOrz == 1)
        xyOrzValue = PARAM_Y;
    else if (xyOrz == 2)
        xyOrzValue = PARAM_Z;

    value += xyOrzValue;

    // In pattern version wave shape modifier, triangle_wave, does work below.
    value = fnTriangle_Wave(value);

    return (value);
}

double f_wood(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 73   was f_umbrella
{
    // NOTE! While called f_wood this function is NOT a parallel of the wood pattern.
    // See: functions.inc for details.

    POV_STATIC_ASSERT(POVFPU_TrapTable[73].parameter_cnt == 8u);
    int P1type   = (int)std::max(0.0,PARAM_X) & 0x01; // The type. 0=random val offset. 1 = random val to normal offset.
    int Rtype    = (int)std::max(0.0,PARAM_X) & 0x06; // The return type. 0 = thisRing/rings. 2 = thisRingWidth/baseWidth.
                                                      // 4 = sum,sumprev.
    double val   = fabs(PARAM_Y);                     // Distance from center of rings.
    double core  = std::max(gkDBL_epsilon,PARAM_Z);   // Core cylinder around which tree rings 'grow'
    double width = PARAM(0);                          // Ignoring the core, the base ring width.
    double rings = std::max(1.0,PARAM(1));            // The tree age in growth rings excluding core.
    double seed  = PARAM(2);
    double mean  = PARAM(3);
    double sigma = PARAM(4);

    std::size_t hashvalue;
    double gaussV,gaussVV,gaussR2; // Used in lambdaGaussWoodVal

    auto lambdaGaussWoodVal = [&] (double R) -> double
    {
        constexpr double Cn = 4.0*exp(-0.5)/std::sqrt(2.0);
        do
        {
            hashvalue = std::hash<double>{}(R);
            R  = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            hashvalue = std::hash<double>{}(R);
            gaussR2 = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            gaussV  = Cn*(gaussR2-0.5)/R;
            gaussVV = gaussV*gaussV/4.0;;
        } while (gaussVV>-std::log(R));
        return(mean+(gaussV*sigma));
    };

    double value = seed;
    double sumPrev=0, sum = core; // sumPrev Init to avoid -Wmaybe-uninitialized warning.
    double thisWringWidth = core; // Init to avoid -Wmaybe-uninitialized warning.
    POV_LONG cnt = 0;
    if (val <= sum)
        value = val/core;
    else
    {
        do
        {
            sumPrev = sum;
            hashvalue = std::hash<double>{}(value);
            value = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            if (P1type==1) // normal distribution about mean
                sum += std::max(gkMinIsectDepthReturned,(width+(lambdaGaussWoodVal(value)*width)));
            else           // Type 0, [0..1] random value adjusted to [-0.5..0.5].
                sum += std::max(gkMinIsectDepthReturned,(width+((value-0.5)*width)));
            cnt++;
        } while (val > sum);
        thisWringWidth = sum - sumPrev;
        value = (val-sumPrev)/thisWringWidth;
    }

    if      (Rtype==2)
        value = fnEnc2x_f32(value,thisWringWidth/width);
    else if (Rtype==4)
        value = fnEnc2x_f32(sumPrev,sum);
    else
        value = fnEnc2x_f32(value,cnt+fmod(cnt/rings,1.0));

    return value;
}

double f_enc2x_f32(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 74   // was f_witch_of_agnesi
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[74].parameter_cnt == 2u);
    double value = fnEnc2x_f32(PARAM_X, PARAM_Y);
    return (value);
}

double f_dec2x_f32(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 75   // was f_witch_of_agnesi_2d
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[75].parameter_cnt == 2u);
    double value = fnDec2x_f32(PARAM_X, PARAM_Y);
    return (value);
}

double f_noise3d(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 76
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[76].parameter_cnt == 4u);
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int ngen = (int)PARAM(0) & 0x03; // Force explicitness. 1-3 Valid currently. 2-usual use.
    return Noise(Vec, ngen); // Noise handles invalid values in defaulting to 2.
}

double f_pattern(FPUContext* ctx, double *ptr, unsigned int fn) // 77
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[77].parameter_cnt == 3u);
    Vector3d TPoint;
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
        return 0.0;

    Warp_EPoint(TPoint, Vec, reinterpret_cast<const TPATTERN *>(f->private_data));

    return Evaluate_TPat(reinterpret_cast<const TPATTERN *>(f->private_data), TPoint, nullptr, nullptr, ctx->threaddata);
}

double f_snoise3d(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 78
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[78].parameter_cnt == 4u);
    Vector3d Vec = Vector3d(PARAM_X, PARAM_Y, PARAM_Z);
    int ngen = (int)PARAM(0) & 0x03;  // Force explicitness. 1-3 Valid currently 2-usual use.
    double value = 2.0 * Noise(Vec, ngen) -1.0;
    return value;
}

double f_npmod(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 79
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[79].parameter_cnt == 3u);
    double value=PARAM_X, valueDenominator=PARAM_Y, valueMultiplier=PARAM_Z;

    if ((value == 0.0) || (valueDenominator <= 0.0))
    {
        return value;
    }
    else
    {
        if (fabs(value) > valueDenominator)
        {
            double tmpValue = value;
            double tmpCount = (unsigned long long)fabs(value / valueDenominator);
            value = fmod(tmpValue, valueDenominator);
            if (fabs(value) < gkDBL_epsilon)
            {
                return tmpValue / tmpCount * valueMultiplier;
            }
            else
            {
                return value * valueMultiplier;
            }
        }
        else
        {
            return value * valueMultiplier;
        }
    }
}

double f_eblob(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 80
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[80].parameter_cnt == 19u);
    const int width=8, height=4;
    double ary[width][height];
    for (int i=0; i<width; i++)
    {
        ary[i][0] = PARAM(i);                      // incoming values
        ary[i][1] = fabs(PARAM(width+i)) / exp(1); // psuedo normalized strengths
    }

    double eDelta = PARAM_X;                           // offset between 1/exp()s. 0.001 good.
    int eCount    = (int)std::max(1.0,PARAM_Y) & 0x0F; // count of meaningful passed values.
    bool returnAsPatternValue  = (PARAM_Z == 0) ? false : true;
    const double limitValToExp = 38.0;   // Clamp +-value for values passed to exp().
    double threshold = (2.0 * eDelta) + (eDelta / 10.0);

    for (int i=0; i<width; i++)
        ary[i][2] = std::max(-limitValToExp,std::min(limitValToExp,(ary[i][0] * ary[i][1])));

    double value = threshold;
    for (int i=0; i<std::min(width,eCount); i++)
    {
        double valPolarity = (PARAM(width+i) >= 0.0) ? 1.0 : -1.0;
        if (ary[i][1] == 0.0)
        {
            ary[i][3] = ary[i][0];
            value = std::min((value * valPolarity),ary[i][0]);
        }
        else
        {
            double valOneOverEp = 1.0 / exp(ary[i][2] + eDelta);
            double valOneOverEn = 1.0 / exp(ary[i][2] - eDelta);
            ary[i][3] = valOneOverEp - valOneOverEn;
            value += ary[i][3] * valPolarity;
        }
    }

    if ((!returnAsPatternValue) || (value >= 0.0) || (eCount <= 1))
    {
        return value;
    }
    else
    {
        double valPattern = 0.0;
        double mapCnt = 0;
        switch (eCount)
        {
            case 2: mapCnt = 1+ 2; break;
            case 3: mapCnt = 1+ 3; break;
            case 4: mapCnt = 1+ 6; break;
            case 5: mapCnt = 1+10; break;
            case 6: mapCnt = 1+15; break;
            case 7: mapCnt = 1+21; break;
            case 8: mapCnt = 1+28; break;
            default: mapCnt = 1+ 2;
        }

        double valMostNegTwoCombo = HUGE_VAL;
        const double localEpsilon = 1e-6; // Align with map entry formatting to six sig digits.
        int cnt = 0;
        for (int i = 0; i < eCount; i++)
        {
            for (int j = i+1; j < eCount; j++)
            {
                cnt++;
                double pairVal = ary[i][3] + ary[j][3];
                if (pairVal < valMostNegTwoCombo)
                {
                    valMostNegTwoCombo = pairVal;
                    double valMapLo = cnt/mapCnt;
                    double valMapHi = (cnt+1)/mapCnt;
                    double valMapDelta = valMapHi - valMapLo - (2.0*localEpsilon);
                    if ((ary[i][3] < 0.0) && (ary[j][3] < 0.0))
                    {
                        if ((PARAM_Z == 1.1) && (fabs(ary[i][3]) > threshold))
                        {
                            valPattern = valMapLo + localEpsilon;
                        }
                        else if ((PARAM_Z == 1.1) && (fabs(ary[j][3]) > threshold))
                        {
                            valPattern = valMapHi - localEpsilon;
                        }
                        else
                        {
                            double valHiWeightLo = ary[i][3]/(ary[i][3]+ary[j][3]);
                            double valHiWeightHi = ary[j][3]/(ary[i][3]+ary[j][3]);
                            if ((PARAM_Z > 1.1) && (PARAM_Z <= 6.0))
                            {   // poly wave / pow() to speed change from middle to ends.
                                double valPow = PARAM_Z;
                                valPattern = valMapLo + valMapDelta*0.5; // middle
                                if (valHiWeightLo < valHiWeightHi)
                                {
                                    valPattern += ((valMapDelta * 0.5) * (1.0 - pow(valHiWeightLo/0.5,valPow)));
                                    valPattern -= localEpsilon;
                                }
                                else
                                {
                                    valPattern -= ((valMapDelta * 0.5) * (1.0 - pow(valHiWeightHi/0.5,valPow)));
                                    valPattern += localEpsilon;
                                }
                            }
                            else
                            {   // Linear interpolation between two contributing values.
                                if (valHiWeightLo < valHiWeightHi)
                                {
                                    valPattern = valMapLo + localEpsilon + (valMapDelta * valHiWeightHi);
                                }
                                else
                                {
                                    valPattern = valMapHi - localEpsilon - (valMapDelta * valHiWeightLo);
                                }
                            }
                        }
                    }
                    else if (ary[i][3] < 0.0)
                    {
                        valPattern = valMapLo + localEpsilon;
                    }
                    else if (ary[j][3] < 0.0)
                    {
                        valPattern = valMapHi - localEpsilon;
                    }
                    else
                    {
                        valPattern = 0.0;
                    }
                }
            }
        }
        return valPattern;
    }
}


double f_revolution(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 81
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[81].parameter_cnt == 6u);
    double x = PARAM_X, y = PARAM_Y, z = PARAM_Z;
    double d = PARAM(0), a = PARAM(1);
    int rtrnXorY = (int)std::max(0.0,PARAM(2)) & 0x01; // Are we going to return x or y.

    double value = 0.0, xd;

    if (a != 0.0)
    {
        double enc2xf32 = FnRotate1(x,y,z,2,a);
        x  = fnDec2x_f32(enc2xf32,0);
        xd = sqrt(x*x + z*z) - d;
        double enc2xf32a = FnRotate1(xd,y,z,2,a);

        if (rtrnXorY == 0)
            value = fnDec2x_f32(enc2xf32a,0);
        else
            value = fnDec2x_f32(enc2xf32a,1);
    }
    else
    {
        if (rtrnXorY == 0)
        {
            xd = sqrt(x*x + z*z) - d;
            value = xd;
        }
        else
            value = y;
    }

    return value;
}


double f_hash(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 82
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[82].parameter_cnt == 4u);
    double value = PARAM_X;
    double seed  = PARAM_Y;
    double step  = fabs(PARAM_Z); // abs so as not to not change (value+seed) sign.
    int mode  = (int)std::max(0.0,PARAM(0)) & 0x01;

    value += seed;
    std::size_t hashvalue;
    if (step > 0)
    {
        hashvalue = std::hash<double>{}((double)(POV_INT64)(value/step));
    }
    else
    {
        hashvalue = std::hash<double>{}(value);
    }

    value = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
    if (mode == 1)
    {
        value = (value * 2.0) - 1.0;
    }

    return (value);
}


double f_interior_fade(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 83
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[83].parameter_cnt == 6u);
    float Amplify  = PARAM_X;
    float iDepth   = PARAM_Y;
    float FadeCol  = PARAM_Z;
    float FadeDist = PARAM(0);
    float FadePow  = PARAM(1);
    float intenMax = PARAM(2);

    double value = Amplify;
    if (fabs(FadeDist) > 4.4e-8)
    {
        float depth = iDepth / FadeDist;
        float invFadeCol = 1.0f - FadeCol;
        if (FadePow >= 1000.0f)
        {
            value = Amplify * exp(-invFadeCol*depth);
        }
        else
        {
            if (depth > 0.0f)
            {
                float attVal = 1.0f + powf(depth,FadePow);
                attVal = FadeCol + invFadeCol/attVal;
                attVal = std::max(0.0f,attVal);
                value  = attVal * Amplify;
            }
            else
            {
                value = Amplify;
            }
        }
    }
    else
    {
        value = Amplify;
    }

    // intensity_max stand in.
    if (intenMax != 0.0)
    {
        if (intenMax > 0.0)
        {
            value = std::max(0.0,value);
            if (value > intenMax)
                value = intenMax;
        }
        else // Cannot match 'alarm pattern' return specific neg value for alarm.
        {
            if ((value < 0.0) || (value > fabs(intenMax)))
                value = -1.23456789;
        }
    }

    return value;
}


double f_distribution(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 84
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[84].parameter_cnt == 7u);
    int P1type = (int)std::max(0.0,PARAM_X) & 0x0F; // The distribution type.
    double P2flt64 = PARAM_Y;
    double P3flt64 = PARAM_Z;
    double P4flt64 = PARAM(0);
    double P5flt64 = PARAM(1);
    double P6flt64 = PARAM(2);
    double P7flt64 = PARAM(3);

    std::size_t hashvalue;
    double gaussV,gaussVV,gaussR2; // Used in lambdaGaussDistVal

    auto lambdaGaussDistVal = [&] (double R) -> double
    {
        constexpr double Cn = 4.0*exp(-0.5)/std::sqrt(2.0);
        do
        {
            hashvalue = std::hash<double>{}(R);
            R  = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            hashvalue = std::hash<double>{}(R);
            gaussR2 = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            gaussV  = Cn*(gaussR2-0.5)/R;
            gaussVV = gaussV*gaussV/4.0;;
        } while (gaussVV>-std::log(R));
        return(P3flt64+(gaussV*P4flt64));
    };

    double value = 0.0;
    switch (P1type)
    {
        case 0:     // Probability
                {
                    // randOrHashValue < Probability_0to1
                    value = P3flt64<P2flt64;
                    break;
                }
        case 1:     // Bernoulli
                {
                    // (Probability_0to1>=randOrHashValue?true:false)
                    value = P2flt64>=P3flt64?1.0:0.0;
                    break;
                }
        case 2:     // Two point (Bernoulli return two values)
                {
                    // (Probability_0to1>=randOrHashValue?valT:ValF)
                    value = P2flt64>=P3flt64?P4flt64:P5flt64;
                    break;
                }
        case 3:     // Binomial
                {
                    unsigned int n   = std::max(1,(int)P3flt64);
                    unsigned int cnt = 0;
                    value = P4flt64;
                    bool normalizeToCnt = (P5flt64==0)?0:1;
                    for (decltype(n) i=0; i < n; i++)
                    {
                        hashvalue = std::hash<double>{}(value);
                        value = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
                        if (value <= P2flt64)
                            cnt++;
                    }
                    if (normalizeToCnt)
                        value = (double)cnt/(double)n; // This much more useful in patterns and functions.
                    else
                        value = (double)cnt;  // Old macro returned the count.
                    break;
                }
        case 4:     // Geometric (aka Negative Binomial (1.0,0.5))
                {
                    // (floor(ln(rand(Stream))/ln(1-P))) POV-Ray macro.
                    bool ReturnContinuous = (P4flt64==0)?0:1;
                    if (ReturnContinuous)
                        value =       log(P3flt64)/log(1.0-P2flt64);  // Continuous result more useful to patterns/functions.
                    else
                        value = floor(log(P3flt64)/log(1.0-P2flt64)); // Number of trials/tosses to get true.
                    if (P5flt64!=0.0)
                        value /= P5flt64;  // Useful to get result in 0-1 range. (0.5 prob -> 5.0 say. Neg OK.)

                    break;
                }
        case 5:     // Gaussian aka Normal distribution ( 1.0/(sigma*sqrt(2.0*M_PI)) * exp(-0.5*pow(((val-mean)/sigma),2.0)) )
                {   // Code basically rand.inc's Rand_Normal except for optional clamping.
                    value = lambdaGaussDistVal(P2flt64); // Uses: P2flt64(val) P3flt64(mean) P3flt64(sigma)
                    if (P6flt64 > P5flt64)               // P5flt64(clamp lo) P5flt64(clamp hi)
                        value = std::max(P5flt64,std::min(P6flt64,value));
                    break;
                }
        case 6:     // Cauchy distribution. Method here different than rand.inc's Rand_Cauchy.
                {
                    double psqr  = P4flt64 * P4flt64;
                    double gamma = std::atanh((1.0-psqr)/(1+psqr));
                    value = fmod((tan(M_PI*(P2flt64-0.5))*gamma+M_PI), TWO_M_PI) - M_PI;
                    value += P3flt64;
                    if (P6flt64 > P5flt64)
                    {
                        int retryCnt = 0;
                        double R = P2flt64;
                        while ((value<P5flt64 || value>P6flt64) && retryCnt<20) // Short search for value within range.
                        {
                            retryCnt++;
                            hashvalue = std::hash<double>{}(R);
                            R = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
                            value = fmod((tan(M_PI*(R-0.5))*gamma+M_PI), TWO_M_PI) - M_PI;
                            value += P3flt64;
                        }
                        value = std::max(P5flt64,std::min(P6flt64,value));
                    }
                    break;
                }
        case 7:     // Triangular distribution. Method here different than rand.inc's Rand_Triangle.
                {
                    double x = P2flt64;
                    double a = P3flt64;
                    double b = P4flt64;
                    double c = P5flt64;
                    int mode  = (int)std::max(0.0,P6flt64) & 0x03;

                    value = 0.0;
                    if ((x >= a) && (x < c))
                        value = 2.0*(x-a)/((b-a)*(c-a));
                    else if ((x > c) && (x <= b))
                        value = 2.0*(b-x)/((b-a)*(b-c));
                    else if (x == c)
                        value = 2.0/(b-a);

                    if (mode == 1)
                        value = value/(2.0/(b-a));            // [0..1]
                    else if (mode == 2)
                        value = (value/(2.0/(b-a))*2.0)-1.0;  // [-1..1]
                    else
                        value = value/(2.0/(b-a))*x;          // [0..x]
                    break;
                }
        case 8:     // Trapezoidal distribution.
                {
                    double x = P2flt64;
                    double a = P3flt64;
                    double b = P4flt64;
                    double c = P5flt64;
                    double d = P6flt64;
                    int mode  = (int)std::max(0.0,P7flt64) & 0x03;

                    value = 0.0;
                    if ((x >= a) && (x < b))
                        value = (2.0/(d+c-a-b)) * ((x-a)/(b-a));
                    else if ((x >= b) && (x < c))
                        value = 2.0/(d+c-a-b);
                    else if ((x >= c) && (x <= d))
                        value = (2.0/(d+c-a-b)) * ((d-x)/(d-c));

                    if (mode == 1)
                        value = value/(2.0/(d+c-a-b));           // [0..1]
                    else if (mode == 2)
                        value = (value/(2.0/(d+c-a-b))*2.0)-1.0; // [-1..1]
                    else
                        value = value/(2.0/(d+c-a-b))*x;         // [0..x]
                    break;
                }
        case 9:     // Exponential distribution.
                {   // We are not matching rand.inc's Rand_Exp (-ln(rand(Stream))/Lambda) ?
                    //                     value = -log(P2flt64)/P3flt64; // P3 -> Lambda, rate, 1/mean.
                    double scale  = std::max(gkDBL_epsilon,P4flt64);
                    double val    = P2flt64 * scale;
                    double lambda = std::min(1.0,std::max(0.0,P3flt64));
                    bool ReturnInverse = (P5flt64==0)?0:1;
                    value = lambda * std::exp(-lambda*val);
                    if (ReturnInverse)
                        value = lambda-value;
                    break;
                }
        case 10:    // Circular distribution. More like Triangular/Trapezoidal in selecting/shaping values.
                {
                    double v = P2flt64;
                    double r = P3flt64; // circular distribution radius
                    double mean = P4flt64;
                    int mode    = (int)std::max(0.0,P5flt64) & 0x03;

                    value = std::max(-r,std::min(r,v-mean));
                    double angle = acos(value/r);
                    value = sin(angle)*r;

                    switch (mode)
                    {
                        case 1:  // [0..1]
                            value /= r; // [0..1]
                            break;
                        case 2:  // [-1..1]
                            value /= r;
                            value = (value*2.0)-1.0;
                            break;
                        case 3:
                            value /= r;
                            value = (value*2.0)-1.0;
                            value *= r;  // (-r..r)
                            break;
                        default:
                              {
                                  // [0..r]
                                  break;
                              }
                    } // mode

                    break;
                }
        case 11:    // Hacked off Wrapped Cauchy Distribution.
                {   // Likely useful as pattern, but much less sure about isosurfaces
                    double v = P2flt64;
                    double r = P3flt64; // circular distribution radius
                    double mean  = P4flt64;
                    double sigma = P5flt64;
                    int mode     = (int)std::max(0.0,P6flt64) & 0x07;

                    value = std::max(-r,std::min(r,v-mean));
                    double angle    = acos(value/r);
                    double norm     = (1.0/TWO_M_PI) * sinh(sigma)/(cosh(sigma)-1.0);
                    // Probability Density Function (PDF)
                    value = (1.0/TWO_M_PI) * sinh(sigma)/(cosh(sigma)-cos(angle));
                  //double epsilon  = (1.0/TWO_M_PI) * sinh(sigma)/(cosh(sigma)+1.0);
                  //value -= epsilon;  // Adjust min circular value to zero  (Better for patterns without...)

                    switch (mode)
                    {
                        case 1:
                            value /= norm; // After adjustment to zero. PDF normalized [0..1)
                            break;
                        case 2:
                            value /= norm; // After adjustment to zero. PDF normalized [0..1)
                            value = (value*2.0)-1.0; // (-1..1)
                            break;
                        case 3:
                            value /= norm; // After adjustment to zero. PDF normalized [0..1)
                            value *= r;  // [0..v)
                            break;
                        case 4:
                            value /= norm; // After adjustment to zero. PDF normalized [0..1)
                            value = (value*2.0)-1.0; // (-1..1)
                            value *= r;  // (-v..v)
                            break;
                        default:
                              {
                                  // PDF value adjusted to zero at min
                                  break;
                              }
                    } // mode

                    break;
                }

        // #macro Rand_Spline(Spl, Stream)   Might we re-hash the incoming hash.
        // for this type of weighted distribution for the random values. It leaves
        // us though with no way to re-run the spline function. Hmm...
        default:
                {
                    //
                    break;
                }
    } // switch on distributin type.

    return value;
}


double f_clock(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 85
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[85].parameter_cnt == 3u);
    int P1type = (int)std::max(0.0,PARAM_X) & 0x07; // The clock type and return type.
    int P2dst  = (int)PARAM_Y;                      // Account for dst 1 yes. 0 no.

    double value=0.0;

    time_t now;
    struct tm year2K;
    double seconds=0;  // Init to avoid -Wmaybe-uninitialized warning.

    if (P1type==0) // Try to keep the 'timer' version faster. (Note all clock stuff is slow...)
    {   // Returns fractional seconds. Use for timing.
        // This clock not affected by the usual wall clock adjustments
        auto t_nano  = std::chrono::duration_cast<std::chrono::nanoseconds>
            (std::chrono::steady_clock::now().time_since_epoch()).count();
        value = t_nano / 1e9;  // Convert to seconds.
    }
    else
    {
        if (P1type>=1 && P1type<=3)
        {
            time(&now); // Note wall clock based values can change at any time. (ie. Sync to time servers, etc.)
            year2K = *localtime(&now);
            year2K.tm_year = 2000 - 1900; year2K.tm_mon = 0;  year2K.tm_mday = 1;
            year2K.tm_hour = 0; year2K.tm_min = 0; year2K.tm_sec = 0;
            year2K.tm_isdst = P2dst;
            seconds = difftime(now,mktime(&year2K)); // Not used with (3)
        }
        switch (P1type)
        {
            case 1:     // Whole seconds to the systems wall clock.
                    {   // Similar to 'now' except in 'seconds' since Y2K
                        value = seconds;
                        break;
                    }
            case 2:     // Fractional days to the systems wall clock (what 'now' returns)
                    {
                        value = seconds/(24 * 60 * 60);
                        break;
                    }
            case 3:     // Difference between machine epoch and y2k.
                    {   // Expect this to be used with (4 as in 4's input being (1)+(3))
                        time_t epoch = 0;
                        value = difftime(epoch,mktime(&year2K));
                        break;
                    }
            case 4:     // Convert local seconds since epoch to utc
                    {
                        std::time_t localt = PARAM_Z;
                        auto local_field = *std::gmtime(&localt);
                        local_field.tm_isdst = -1;
                        auto utc = std::mktime(&local_field);
                        value = utc;
                        break;
                    }
            default:
                        break;
        } // switch on clock type.
    }

    return value;
}


double f_randn(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 86
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[86].parameter_cnt == 5u);
    int P1type  = (int)std::max(0.0,PARAM_X) & 0x07; // The return type.
    double seed = PARAM_Y;
    POV_LONG N  = std::max(1.0,PARAM_Z);
    double P4flt64 = PARAM(0);
    double P5flt64 = PARAM(1);

    std::size_t hashvalue;
    double gaussV,gaussVV,gaussR2; // Used in lambdaGaussRandnVal

    auto lambdaGaussRandnVal = [&] (double R) -> double
    {
        constexpr double Cn = 4.0*exp(-0.5)/std::sqrt(2.0);
        do
        {
            hashvalue = std::hash<double>{}(R);
            R  = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            hashvalue = std::hash<double>{}(R);
            gaussR2 = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
            gaussV  = Cn*(gaussR2-0.5)/R;
            gaussVV = gaussV*gaussV/4.0;;
        } while (gaussVV>-std::log(R));
        return(P4flt64+(gaussV*P5flt64));
    };

    double value = seed;
    double sumPrev, sum = 0.0;
    POV_LONG cnt = 0;
    do
    {
        sumPrev = sum;
        hashvalue = std::hash<double>{}(value);
        value = (double)hashvalue/(double)std::numeric_limits<std::size_t>::max();
        if (P1type==3)
            sum += lambdaGaussRandnVal(value); // This can be especially expensive
        else
            sum += value; // Only used for type 2, but as cheap as conditional.
        cnt++;
    } while (cnt<N);

    switch (P1type)
    {
        case 0:     // Nth random value.
                {
                    if (P4flt64==1)
                        value = value*2.0 - 1.0;
                    break;
                }
        case 1:     // Nth normal distribution value.
                {
                    value = lambdaGaussRandnVal(value);
                    break;
                }
        case 2:     // Sum of random values through the Nth one.
                    // FALLTHROUGH
        case 3:     // Sum of random normal values through the Nth one.
                {
                    value = sum;
                    break;
                }
        case 4:     // SumPrev, Sum of random values through the Nth one.
                    // FALLTHROUGH
        case 5:     // SumPrev, Sum of random normal values through the Nth one.
                {
                    value = fnEnc2x_f32(sumPrev,sum);
                    break;
                }
        default:
                {
                    //
                    break;
                }
    } // switch on distributin type.

    return value;
}

double f_null(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 87
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[87].parameter_cnt == 1u);
    return (PARAM_Y*0.0);  // Primarily used to establish base line timing.
}

double f_elliptical_sphrswp(FPUContext* /*ctx*/, double *ptr, unsigned int /*fn*/) // 88
{
    POV_STATIC_ASSERT(POVFPU_TrapTable[88].parameter_cnt == 9u);
    const int width=8, height=6; // Rows: 0|3 - Sphere sample vals; 2|4 - Rad at sample; 3|4 - three sequence value sums.
    double ary[width][height];

    double a = std::max(gkDBL_epsilon,fabs(PARAM(0))), b = std::max(gkDBL_epsilon,fabs(PARAM(1))), r = fabs(PARAM(2));

    int Method = (int)PARAM(5) & 0x03;  // 0 = sweep with optimization. 1 = return 0 to 1 map value.
                                        // 2 = skip up front torus filter. (e.g. Using sweep values for map control)
    double value;

    // If not inside slightly larger scaled torus, return quickly to avoid sampling.
    // Note! Metod 0 does create a significant discontinuity in the gradient, but with
    // respect to isosurfaces, this happens away from the surface and so doesn't
    // 'really' matter. The actual gradient near the surface below is always near 1.0.
    // TODO. Need smarter gradient reporting, some ability to turn off max
    // gradient warnings - OR both.
    if (Method != 2)
    {
      //value = std::hypot(PARAM_X*b,PARAM_Z*a) - b*a;  // std::hypot is slower.
      //value = -(r+r/10) + std::hypot(value,PARAM_Y);
        double tmpX = PARAM_X*b, tmpZ = PARAM_Z*a;
        value = sqrt(tmpX*tmpX + tmpZ*tmpZ) - b*a;
        value = -(r+r/10) + sqrt(value*value + PARAM_Y*PARAM_Y);
        if (value > gkDBL_epsilon)
            return value;
    }

    Vector3d Ept = Vector3d(PARAM_X,PARAM_Y,PARAM_Z);

    double degStart = std::min(360.0,std::max(0.0,fabs(PARAM(3))));
    double degOfArc = std::min(360.0,std::max(-360.0,PARAM(4)));     // Plus or minus arc OK.

    double radStart = -degStart * M_PI_180;
    double radOfArc = -degOfArc * M_PI_180;

    int steps = width - 1; // To get width value samples on each pass.
    double thisRadOfArcStep, thisRadOfArcHalfStep;

    auto lambdaEllipticalTorusSphereVal = []
         (const double xt, const double zt, const double r, const Vector3d ept) -> double
    {
        double sx = ept.x() - xt;
        double sz = ept.z() - zt;

        return sqrt(sx*sx + ept.y()*ept.y() + sz*sz)-r;
    };

    auto lambdaEllipticalSamples = [&ary, &thisRadOfArcStep, &thisRadOfArcHalfStep, &lambdaEllipticalTorusSphereVal]
         (const double thisRadStart, const double thisRadOfArc, const int steps,
          const double a, const double b, const double r, const Vector3d ept) -> void
    {
        thisRadOfArcStep     = thisRadOfArc/steps;
        thisRadOfArcHalfStep = thisRadOfArcStep / 2.0;
        double tmpSign = (thisRadOfArc > 0.0) ? +1.0 : -1.0;
        int scnt = 0;
        for (double radDeltaAngle = 0; radDeltaAngle <= tmpSign*(thisRadOfArc + thisRadOfArcHalfStep); radDeltaAngle += thisRadOfArcStep)
        {
            double thisAngle = thisRadStart + radDeltaAngle;
            ary[scnt][0] = lambdaEllipticalTorusSphereVal(a*cos(thisAngle),b*sin(thisAngle),r,ept);
            ary[scnt][1] = thisAngle;
            scnt += 1;
        }
    };

    if (radOfArc == 0.0)
    {
        value = lambdaEllipticalTorusSphereVal(a*cos(radStart),b*sin(radStart),r,Ept);
        return value;
    }

    int cnt;
    int cntMin=0, cntMin0;  // cntMin init suppresses -Wmaybe-uninitialized warning.
    double thisRadStart[2],thisRadOfArc[2];
    thisRadStart[0] = radStart;
    thisRadOfArc[0] = thisRadOfArc[1] = radOfArc/2.0;
    thisRadStart[1] = radStart + thisRadOfArc[0];
    for (int h=0; h<2; h++)          // Handle arc in two halves to avoid numerical confusion in certain arcs >180 and <360. (h)alf.
    {
        for (int i=0; i<22; i++) // Limit at 22 for 'roughly' gkMinIsectDepthReturned accuracy. (Make a parameter? Can often be less)
        {
            lambdaEllipticalSamples(thisRadStart[h],thisRadOfArc[h],steps,a,b,r,Ept);
            if (fabs(thisRadOfArcStep) < gkMinIsectDepthReturned)
                break;

            //--- Calculate six, three value sum spans         //    01234567
            ary[0][2]  = ary[1][2] = (ary[1][0] +  ary[2][0]); // 12          1+2
            ary[0][2] += ary[0][0];                            // 0  012      12+0
            ary[1][2] += ary[3][0];                            // 1   123     12+3
            ary[2][2]  = ary[3][2] = (ary[3][0] +  ary[4][0]); // 34          3+4
            ary[2][2] += ary[2][0];                            // 2    234    34+2
            ary[3][2] += ary[5][0];                            // 3     345   34+5
            ary[4][2]  = ary[5][2] = (ary[5][0] +  ary[6][0]); // 56          5+6
            ary[4][2] += ary[4][0];                            // 4      456  56+4
            ary[5][2] += ary[7][0];                            // 5       567 56+7
          //ary[6][2]  = ary[7][2] = 0;
            //--- Find minimum and maximum 3 value sum.
            double maxVal = ary[0][2];
            double minVal = maxVal;
            for (cnt=1; cnt<width-2; cnt++)
            {
                maxVal = std::max(maxVal,ary[cnt][2]);
                minVal = std::min(minVal,ary[cnt][2]);
            }
            //--- Set cnt to maximum three value sum.
            for (cnt=0; cnt<width-2; cnt++)
                if (ary[cnt][2] == maxVal) break;
            //--- Set cntMin to minimum three value sum.
            for (cntMin=0; cntMin<width-2; cntMin++)
                if (ary[cntMin][2] == minVal) break;
            //--- Updates for next set of 8 samples.
            switch (cnt)
            {
                case 0:                         thisRadStart[h] = ary[2][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; break;
                case 1:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[2][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[2][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 2:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[3][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[3][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 3:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[4][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[4][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 4:
                        {
                            if (cntMin < cnt) { thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[5][1] - thisRadStart[h]; }
                            else              { thisRadStart[h] = ary[5][1]; thisRadOfArc[h] = ary[7][1] - thisRadStart[h]; }
                            break;
                        }
                case 5:                         thisRadStart[h] = ary[0][1]; thisRadOfArc[h] = ary[5][1] - thisRadStart[h]; break;
                default: break;
            }
        }
        if (h == 0)
        {   // Remember results from first half arc.
            cntMin0 = cntMin;
            for (int j=0; j<height/2; j++)
                for (int k=0; k<width; k++)
                    ary[k][3+j] = ary[k][j];
        }
    }

    // Ending with 3 with thoughts of possible blobbing extensions lowering loop
    // depth - and ability to generate and orientation vector. For now min()s.
    if (Method == 1)
    {
        double valHalf0 = std::min(ary[cntMin0][3],std::min(ary[cntMin0+1][3],ary[cntMin0+2][3]));
        double valHalf1 = std::min(ary[cntMin][0],std::min(ary[cntMin+1][0],ary[cntMin+2][0]));
        if (valHalf0 <= valHalf1)
            value = std::max(0.0,std::min(1.0,fabs((ary[cntMin0+1][4] - radStart)/radOfArc)));
        else
            value = std::max(0.0,std::min(1.0,fabs((ary[cntMin+1][1] - radStart)/radOfArc)));
    }
    else
    {
        double valHalf0 = std::min(ary[cntMin0][3],std::min(ary[cntMin0+1][3],ary[cntMin0+2][3]));
        double valHalf1 = std::min(ary[cntMin][0],std::min(ary[cntMin+1][0],ary[cntMin+2][0]));
        value = std::min(valHalf0,valHalf1);
    }

    return value;
}


void f_pigment(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp) // 0
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[0].parameter_cnt == 3u);
    Vector3d Vec = Vector3d(PARAM_N_X(5), PARAM_N_Y(5), PARAM_N_Z(5));
    TransColour Col;
    RGBFTColour rgbftCol;
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + pRED,    0.0);
        ctx->SetLocal(sp + pGREEN,  0.0);
        ctx->SetLocal(sp + pBLUE,   0.0);
        ctx->SetLocal(sp + pFILTER, 0.0);
        ctx->SetLocal(sp + pTRANSM, 0.0);
        return;
    }

    Compute_Pigment(Col, reinterpret_cast<const PIGMENT *>(f->private_data), Vec, nullptr, nullptr, ctx->threaddata);
    rgbftCol = ToRGBFTColour(Col);

    ctx->SetLocal(sp + pRED,    rgbftCol.red());
    ctx->SetLocal(sp + pGREEN,  rgbftCol.green());
    ctx->SetLocal(sp + pBLUE,   rgbftCol.blue());
    ctx->SetLocal(sp + pFILTER, rgbftCol.filter());
    ctx->SetLocal(sp + pTRANSM, rgbftCol.transm());
}

void f_transform(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp) // 1
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[1].parameter_cnt == 3u);
    Vector3d Vec = Vector3d(PARAM_N_X(3), PARAM_N_Y(3), PARAM_N_Z(3));
    Vector3d Result;
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + X, 0.0);
        ctx->SetLocal(sp + Y, 0.0);
        ctx->SetLocal(sp + Z, 0.0);
        return;
    }

    MTransPoint(Result, Vec, reinterpret_cast<const TRANSFORM *>(f->private_data));

    ctx->SetLocal(sp + X, Result[X]);
    ctx->SetLocal(sp + Y, Result[Y]);
    ctx->SetLocal(sp + Z, Result[Z]);
}

void f_spline(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp) // 2
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[2].parameter_cnt == 1u);
    EXPRESS Result;
    FunctionCode *f = ctx->functionvm->GetFunction(fn);
    int Terms;

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + X, 0.0);
        ctx->SetLocal(sp + Y, 0.0);
        ctx->SetLocal(sp + Z, 0.0);
        return;
    }

    Terms = (reinterpret_cast<GenericSpline *>(f->private_data))->Terms;

    Get_Spline_Val(reinterpret_cast<GenericSpline *>(f->private_data), PARAM_N_X(Terms), Result, &Terms);

    ctx->SetLocal(sp + X, Result[X]);
    ctx->SetLocal(sp + Y, Result[Y]);
    if (Terms > 2)
    {
        ctx->SetLocal(sp + Z, Result[Z]);
        if (Terms > 3)
        {
            ctx->SetLocal(sp + T, Result[T]);
            if (Terms > 4)
                ctx->SetLocal(sp + T + 1, Result[T + 1]);
        }
    }
}

void f_patmods(FPUContext *ctx, double *ptr, unsigned int fn, unsigned int sp) // 3
{
    POV_STATIC_ASSERT(POVFPU_TrapSTable[3].parameter_cnt == 3u);
    Vector3d TPoint;
    Vector3d Vec = Vector3d(PARAM_N_X(3), PARAM_N_Y(3), PARAM_N_Z(3));
    FunctionCode *f = ctx->functionvm->GetFunction(fn);

    if (f->private_data == nullptr)
    {
        ctx->SetLocal(sp + X, 0.0);
        ctx->SetLocal(sp + Y, 0.0);
        ctx->SetLocal(sp + Z, 0.0);
        return;
    }

    Warp_EPoint (TPoint, Vec, reinterpret_cast<const TPATTERN *>(f->private_data));

    ctx->SetLocal(sp + X, TPoint.x());
    ctx->SetLocal(sp + Y, TPoint.y());
    ctx->SetLocal(sp + Z, TPoint.z());
    return;
}


}
// end of namespace pov
