// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: munctions.inc
// Last updated: 10 June 2021

// Description: This file contains implementations and commented examples for
// what are being called for short "munctions." These are macro wrapped,
// globally defined and compiled once, re-used functions which were enabled
// with the global pseudo dictionary capability implemented in the v3.8
// versions of POV-Ray.

// Compiled user function equivalents are almost always many times faster than
// macros. Given the original parser set up, it's been difficult to fully
// exploit this capability.

// POV-Ray and earlier versions of the povr branch have shipped with sets of
// pre-compiled functions from functions.inc. See: f_even and F_even (povr's
// name) with code like:

//      #declare F_even = function(x) {select(mod(x,2),0,1,0)}

// There are a significant number of these and when a file like functions.inc
// is included they are all compiled by the parser and so have a negative
// implact on parser time and run time.

// Worse, users often attempt to exploit the speed of compiled functions with
// code like:

//  #version 3.8;
//  #macro Mcr02(_string,_a,_b,_c)
//      #local FnMcr02 = function { x+y+z }
//      #debug "---Compiled FnMcr02\n"
//      #debug concat("02 ",_string," ",str(FnMcr02(_a,_b,_c),6,4),"\n")
//  #end
//  Mcr02("Call 0",1,2,3)
//  Mcr02("Call 1",4,5,6)
//  Mcr02("Call 2",7,8,9)
//  #error "Parsing demonstration.Stop early"

// The function internal to the macro could be compiled and kept for future
// use, but it gets compile on each and every macro call. We can make the use
// of pre-compiled functions from within macro calls much more worthwhile by
// making use of the global pseudo dictionary. Defining and calling functions
// we want to compile once - using the global name space.

//  #version 3.8;
//  #macro Mcr02(_string,_a,_b,_c)
//      #ifndef(global.FnMcr02)
//          #declare FnMcr02 = function { x+y+z }
//          #debug "---Compiled FnMcr02\n"
//      #end
//      #debug concat("02 ",_string," ",
//          str(global.FnMcr02(_a,_b,_c),6,4),"\n")
//  #end
//  Mcr02("Call 0",1,2,3)
//  Mcr02("Call 1",4,5,6)
//  Mcr02("Call 2",7,8,9)
//  #error "Parsing demonstration.Stop early"

// Only functions used, get compiled - and they only get compiled once.

// Generically, the best performance usage would be to call the munction to set
// up and compile the function. Thereafter, using the compiled function name
// directly to avoid the macro call overhead(1).

// (1) - NOTE! The munctions herein are set up to run the compiled function. You
// can code calling, always, the macro / 'munction' name.

// This include file attempts to gather up functions and macros better coded as
// "munctions."


//------------------------------------------------------------------------------

#ifndef(Munctions_Inc_Temp)
#declare Munctions_Inc_Temp = version;
#version 3.8; // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including munction.inc\n"
#end

#include "functions.inc"

//---  (v3.7/v3.8 evaluate_pigment replacement)
// M_evaluate_pigment() --> 'FnToEvalPigmentAtLoc'()

// Optimally defines a function returning the color pigment values at a point
// in 3D space. Meaning the function whether called via this macro or
// explicitly (fastest) returns a full color vector of
// <red,green,blur,filter,transmit>. These which can be accessed via, mostly,
// shorter dot names of <x,y,x,t,transmit>. See the related inbuilt testing
// at the bottom of this file, for example usage.

// Background.

// In POV-Ray proper a macro called eval_pigment has long been pretending to be
// inbuilt functionality by name. The povr branch works to make explicit by
// defined name what is inbuilt (all lower case identifiers) and what is not.

// The eval_pigment macro was originally in functions.inc. The povr branch
// renamed it and moved it first to math.inc, but it now better fits here
// in munctions.inc.

// The old implementation:

//  #macro Eval_Pigment(_pigm, _vec)
//      #local Fn = function { pigment { _pigm } }
//      #local Result = (Fn(_vec.x,_vec.y,_vec.z));
//      Result
//  #end

// It was not efficient in recompiling the pigment function on every call. I
// was also defining a local ID on each call which in this case is complete
// overhead. Further, there was a potential side effect where, if sometime
// during use, the passed pigment ID was updated with a new pigment, one got
// changed results too. The updated code forces the user to be explicit on the
// call whether to recompile to the passed pigment - otherwise, one gets always
// the pigment which existed when the function was first compiled.

// The parameters are:
//  1) The global function name as a string ("FnEvPig00").
//  2) The pigment to use in the compiled function (Pig00).
//  3) The location specified as a 3d vector.
//  4) Passing true or false indicating whether to force recompilation.

// Notes. Setting (4) true, essentially, mimics the old eval_pigment /
// Eval_Pigment behavior. Further, the function name string can be "" to avoid
// collisions with meaningful global functions.

//  #debug concat("Compiling M_evaluate_pigment function ",_functionName,"\n")
//      #debug concat("Re-compiling M_evaluate_pigment function ",_functionName,"\n")
#macro M_evaluate_pigment(_functionName,_pigment,_location,_recompile)
    #ifndef (global[_functionName])
        #declare global[_functionName] = function { pigment { _pigment } }
    #else
        #if (_recompile)
            #undef global[_functionName]
            #declare global[_functionName] = function { pigment { _pigment } }
        #end
    #end
    (global[_functionName](_location.x,_location.y,_location.z))
#end


//---
// Waveform functions. Two, unbounded pattern like waveform functions were long
// implemented in the functions.inc file as all lower case inbuilt looking
// names f_sine_wave and f_scallop_wave.
//
//---  Both take three arguments in (value, amplitude frequency)
// M_sine_wave()    - Creates/compiles F_sine_wave()
// M_scallop_wave() - Creates/compiles F_scallop_wave()

#macro M_sine_wave(_x,_y,_z)
    #ifndef(global.F_sine_wave)
        #declare F_sine_wave = function {sin(x*z*pi)*y}
    #end
    (F_sine_wave(_x,_y,_z))
#end
#macro M_scallop_wave(_x,_y,_z)
    #ifndef(global.F_scallop_wave)
        #declare F_scallop_wave = function {abs(sin(x*z*pi)*y)}
    #end
    (F_scallop_wave(_x,_y,_z))
#end


//---
// Trigonometry related collection with degree arguments and return values.
// POV-Ray had these in math.inc without F_ prefix looking like true SDL
// keywords such as in sind(). The povr branch is strictly allowing only
// true, internally implemented keywords as all lower case.
//
// NOTE! It's probably better to use SDL for domain checking. The v3.7/v3.8
// versions, and so povr too, don't handle trig function exceptions in a
// user friendly way while the pure parser SDL does.
//
//---  (a = angle, v = value)
// M_sind(a)     - Creates F_sind().
// M_cosd(a)     - Creates F_cosd().
// M_tand(a)     - Creates F_tand().
// M_asind(v)    - Creates F_asind().
// M_acosd(v)    - Creates F_acosd().
// M_atand(v)    - Creates F_atand().
// M_atand2(y,x) - Creates F_atand().

#macro M_sind(_x)
    #ifndef(global.F_sind)
        #declare F_sind   = function (x) {sin(radians(x))}
    #end
    (F_sind(_x))
#end
#macro M_cosd(_x)
    #ifndef(global.F_cosd)
        #declare F_cosd   = function (x) {cos(radians(x))}
    #end
    (F_cosd(_x))
#end
#macro M_tand(_x)
    #ifndef(global.F_tand)
        #declare F_tand   = function (x) {tan(radians(x))}
    #end
    (F_tand(_x))
#end
#macro M_asind(_x)
    #ifndef(global.F_asind)
        #declare F_asind  = function (x)   {degrees(asin(x))}
    #end
    (F_asind(_x))
#end
#macro M_acosd(_x)
    #ifndef(global.F_acosd)
        #declare F_acosd  = function (x)   {degrees(acos(x))}
    #end
    (F_acosd(_x))
#end
#macro M_atand(_x)
    #ifndef(global.F_atand)
        #declare F_atand  = function (x)   {degrees(atan(x))}
    #end
    (F_atand(_x))
#end
#macro M_atan2d(_x,_y)
    #ifndef(global.F_atan2d)
        #declare F_atan2d = function (x,y) {degrees(atan2(x,y))}
    #end
    (F_atan2d(_x,_y))
#end


//---
// M_even() - Creates F_even().

// NOTE! Only exact integer (floats, no significant decimal digits work correctly)
// Uses mod(x,2) so any time a remainder you end up 'odd.'

#macro M_even(_x)
    #ifndef(global.F_even)
        #declare F_even = function(x) {select(mod(x,2),0,1,0)}
    #end
    (F_even(_x))
#end

//---
// M_odd() - Creates F_odd().

// NOTE! Only exact integer (floats, no significant decimal digits work correctly)
// Uses mod(x,2) so any time a remainder you end up 'odd.'

#macro M_odd(_x)
    #ifndef(global.F_odd)
        #declare F_odd  = function(x) {select(mod(x,2),1,0,1)}
    #end
    (F_odd(_x))
#end


//---
// M_sgn() - Creates F_sgn().

// Returns the sign of a value. In povr use inbuilt f_sign, f_sign0 instead.

#macro M_sgn(_x)
    #ifndef(global.F_sgn)
        #declare F_sgn = function (x) {select(x,-1,0,1)}
    #end
    (F_sgn(_x))
#end


//---
// M_clip() - Creates F_clip().

// Range handling clips a number (x) to the range [Min, Max] ([y, z]). Values
// above Max return Max, below Min return Min.

#macro M_clip(_x,_y,_z)
    #ifndef(global.F_clip)
        #declare F_clip = function (x,y,z) {min(z,max(x,y))}
    #end
    (F_clip(_x,_y,_z))
#end


//---
// M_clamp() - Creates F_clamp().

// Clamps a number (x) to the range [Min, Max) ([y, z)). Values outside this
// range wrap around. See notes 1-3!

// 1) This function likely should've been called F_wrap / f_wrap - as this more
// accurately what it does. F_clip, clipping and clamping are often used
// synonymously though probably clamping should be used in place of clipping
// which means to cut / select... Yes, aware the usage as muddled outside
// POV-Ray / povr as inside.

// 2) The documentation herein long claimed this clamping was inclusive to
// [Min, Max], but the max value was never in the result even when the input
// value matched the Max value exactly.

// 3) It's better to use the now inbuilt f_wrap. It duplicates this macro for
// mode 0. A mode 1, where even aligned wrapped around values do go all the way
// to Max where the odd aligned wraps go to 0 at the 'Max'.

#macro M_clamp(_x,_y,_z)
    #ifndef(global.F_clamp)
        #declare F_clamp = function (x,y,z) {mod(x-y,z-y) + select(mod(x-y,z-y),z,y)}
    #end
    (F_clamp(_x,_y,_z))
#end


//---
// M_adj_range() - Creates F_adj_range().

// Adjusts input values (x) in the range [0, 1] to output values in range
// [Rmn,Rmx] ([y, z]). Normally (x) in input range, but it need not be.

#macro M_adj_range(_x,_y,_z)
    #ifndef(global.F_adj_range)
        #declare F_adj_range = function (x,y,z) {x*(z-y)+y}
    #end
    (F_adj_range(_x,_y,_z))
#end


//---
// M_adj_range2() - Creates F_adj_range2().

// Adjusts passed (x) values in a specified range [Rmin(y), Rmax(z)] to a new
// value relative the specified range [Min(_mn), Max(_mx)]. Normally (x) in
// input range, but it need not be. See inbuilt f_range2range.

#macro M_adj_range2(_x,_y,_z,_mn,_mx)
    #ifndef(global.F_adj_range2)
        #declare F_adj_range2 = function (x,y,z,_min,_max) { ((x-y)/(z-y))*(_max-_min)+_min }
    #end
    (F_adj_range2(_x,_y,_z,_mn,_mx))
#end

//---
// M_sqr() - Creates F_sqr() for squaring an input value.

// Call once with a float value - M_sqr(1) - to compile a function called
// F_sqr. The include math.inc long always compiled an f_sqr used in shapes.inc
// and elsewhere which is about 20% faster than x*x or pow(x,2). Note! Did
// experiment with an inbuilt f_sqr, but in this case it is not faster than
// letting the vm compile the function.

// Squares the value passed. Usually better to use the inbuilt pow() or just
// value*value.

#macro M_sqr(_x)
    #ifndef(global.F_sqr)
        #declare F_sqr = function (x) {(x*x)}
    #end
    (F_sqr(_x))
#end

//---
// M_dents() - Creates F_dents() as part of true isosurface denting solution.

// A normal vector perturbation dents implementation. The macro implements the type
// of dents executed in the normal{} block and more.

// Example use:

// #declare Fn00 = function (x,y,z) { f_cylinder(x,y,z,0.3) }
// #declare Fn01 = function (x,y,z,_scale_dents,_scale_turb,_bump_depth) {
//     Fn00(x+F_dents(x,y,z,_scale_dents,0,_scale_turb,_bump_depth,
//              Fn00(x,y,z),Fn00(x+1e-4,y,z),Fn00(x,y+1e-4,z),Fn00(x,y,z+1e-4)),
//          y+F_dents(x,y,z,_scale_dents,1,_scale_turb,_bump_depth,
//              Fn00(x,y,z),Fn00(x+1e-4,y,z),Fn00(x,y+1e-4,z),Fn00(x,y,z+1e-4)),
//          z+F_dents(x,y,z,_scale_dents,2,_scale_turb,_bump_depth,
//              Fn00(x,y,z),Fn00(x+1e-4,y,z),Fn00(x,y+1e-4,z),Fn00(x,y,z+1e-4))
//         )+f_noise3d(x*50,y*50,z*50,2)*0.003
// }
// #declare Iso99 = isosurface {
//     function { Fn01(y,x,z,0.11,21.3,0.1) } // A
//   //function { Fn01(y,x,z,0.21,11.3,0.2) } // B
//   //function { Fn01(y,x,z,0.05,12.9,0.1) } // C
// ...

// _fnc is short for (f)_(n)oise_(c)ubed() hard coded additional parameters 1,2 and 3;
// noise generator, mode and scale.
#macro M_dents(_fnc_ng,_fnc_mode,_fnc_scale)
    #ifndef(global.F_dents)
        #declare F_dents = function
            (x,y,z,_scale_dents,_xyz,_scale_turb,_bump_depth,_ptv,_ptxv,_ptyv,_ptzv) {
            f_noise_cubed(x/_scale_dents,y/_scale_dents,z/_scale_dents,_fnc_ng,_fnc_mode,_fnc_scale)
            * _bump_depth *
            abs(f_turbulence(x/_scale_turb,y/_scale_turb,z/_scale_turb,_xyz,
                2,1.0,0,1,6,0.5,2.7)) *
            f_normal(_ptv,_ptxv,_ptyv,_ptzv,_xyz,_bump_depth)
        }
    #end
    (F_dents(2,2,2,1,0,1,0.5,
        f_cylinder(0.548,0,0.548,.3),
        f_cylinder(0.548+1e-4,0,0.548,.3),
        f_cylinder(0.548,0+1e-4,0.548,.3),
        f_cylinder(0.548,0,0.548+1e-4,.3)
    ))
#end


// Set of patterns as functions... Long in POV-Ray functions.inc included a list
// of simple function { pattern{}} wrapped inbuilt patters like f_agate etc which
// looked like inbuilt functions, but which were not. This not insubstantial list
// of pattern functions was always compiled on inclusion of functions.inc.

// Believe the original f_* versions only caused confusion and were mostly of
// limited practical use. Said another way, the ability to use patterns as
// functions is powerful and it's better for users see and learn how this is
// done - it's better for users to just code up what they need.

// The povr branch has implemented true inbuilt variants for some of these as
// indicated in the comments below. Otherwise all of the comments below are
// retained as rough documentation for what is now in the povr branch.

// (Old note for v3.7/v3.8)
// Some of these are not implemented because they require special parameters
// that must be specified in the definition. For this reason, you probably
// would want to define your own versions of these functions.

//--------------------
// f_agate           Replaced with inbuilt f_agate.
// average           Not implemented
// f_boxed           See inbuilt f_box(). pattern{} wrap or function { -max(0.0,f_box(...)} }
// f_bozo            Deleted. It was always only f_noise3d.
// f_brick           Deleted. Not commonly useful. pattern{} wrap user function if needed.
// f_bumps           Deleted. It was always only f_noise3d.
// f_cells           Never wrapped herein... Note! It is now available
//                   as modes(4,5) of the inbuilt f_repeat.
// f_checker         Deleted. Note commonly useful. pattern{} wrap user function if needed.
//#declare F_crackle   = function {pattern {crackle}} // Was f_crackle
// f_cylindrical     See f_cylinder otherwise {pattern {cylindrical}})
// density_file      Not implemented
// f_dents           Replaced with inbuilt f_noise_cubed. dents and F_dents normal/iso versions.
// f_gradientX       Three replaced with single inbuilt f_gradient.
// f_gradientY
// f_gradientZ
// f_granite         Alternate inbuilt f_granite23 otherwise pattern{} wrap.
// image_pattern     Not implemented
// f_hexagon         Deleted. Not commonly useful. pattern{} wrap user function if needed.
//#declare F_leopard   = function {pattern {leopard}} // Was f_leopard
                     //Only the basic mandel pattern is implemented, its variants and
                     //The other fractal patterns are not implemented.
// f_mandel          Deleted. Never useful reason to wrap just one in {pattern {mandel 25}}
// f_marble          Replaced by inbuilt f_marble // also sometime as F_marble
// object            Not implemented
// f_onion           Replaced by inbuilt f_onion.
// pigment_pattern   Not implemented
// f_planar          Replaced by inbuilt f_planar (a shell function)
// f_radial          Replaced wtih inbuilt f_radial.
//#declare F_ripples   = function {pattern {ripples}} // was f_ripples
// f_spherical       See f_sphere (r+fsphere()/r) or otherwise user {pattern {spherical}}
//#declare F_spiral1   = function {pattern {spiral1 2}} // was f_spiral1
//#declare F_spiral2   = function {pattern {spiral2 2}} // was f_spiral2
// f_spotted         Deleted. It was always only f_noise3d. Further, spotted pattern itself deleted.
//#declare F_waves     = function {pattern {waves}} // was f_waves
// f_wood            Replaced by inbuilt f_wood // also sometime as F_wood
//#declare F_wrinkles  = function {pattern {wrinkles}} // was f_wrinkles


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="munctions.inc")

    #fopen FID "SelfTestMunctionsInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n\n")
    // M_sqr() --> F_sqr()
    #write (FID,"---\n")
    #write (FID,concat("M_sqr  4.000000 ",str(M_sqr(2.0),9,-1)," compiling F_sqr\n"))
    #write (FID,concat("F_sqr  4.000000 ",str(global.F_sqr(2.0),9,-1)," running   F_sqr\n"))
    #write (FID,"\n")
    // M_dents() --> F_dents()
    #write (FID,"---\n")
    #write (FID,concat("M_dents  -0.034739 ",str(M_dents(2,0,1),9,-1)," compiling F_dents\n"))
    #declare Val = global.F_dents(2,2,2,1,0,1,0.5,
        f_cylinder(0.548,0,0.548,.3),
        f_cylinder(0.548+1e-4,0,0.548,.3),
        f_cylinder(0.548,0+1e-4,0.548,.3),
        f_cylinder(0.548,0,0.548+1e-4,.3)
    );
    #write (FID,concat("F_dents  -0.034739 ",str(Val,9,-1)," running   F_dents\n"))
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // F_adj_range
    #write (FID,"---\n")
    #write (FID,concat("M_adj_range   0.000000 ",str(M_adj_range(+0.5,-1.0,1.0),9,6)," (+0.5 -> ) Create F_adj_range\n"))
    #write (FID,concat("F_adj_range   0.000000 ",str(F_adj_range(+0.5,-1.0,1.0),9,6)," (+0.5 -> ) val --> -1,+1\n"))
    #write (FID,concat("F_adj_range  -1.000000 ",str(F_adj_range(+0.0,-1.0,1.0),9,6)," (+0.0 -> ) \n"))
    #write (FID,concat("F_adj_range  +1.000000 ",str(F_adj_range(+1.0,-1.0,1.0),9,6)," (+1.0 -> ) \n"))
    #write (FID,concat("F_adj_range  -2.000000 ",str(F_adj_range(-0.5,-1.0,1.0),9,6)," (-0.5 -> ) \n"))
    #write (FID,concat("F_adj_range  +2.000000 ",str(F_adj_range(+1.5,-1.0,1.0),9,6)," (+1.5 -> ) \n"))
    #write (FID,"\n")
    // F_adj_range2, f_range2range (from functions.inc)
    #write (FID,"---\n")
    #write (FID,concat("M_adj_range2   0.000000 ",str(M_adj_range2( +0.5,0.0,1.0,-1.0,1.0),9,6)," (+0.5 -> ) Create F_adj_range2\n"))
    #write (FID,concat("f_range2range  0.000000 ",str(f_range2range(+0.5,0.0,1.0,-1.0,1.0),9,6)," (+0.5 -> ) val, 0-1 --> -1,+1\n"))
    #write (FID,concat("F_adj_range2   0.000000 ",str(F_adj_range2( +0.5,0.0,1.0,-1.0,1.0),9,6)," (+0.5 -> ) \n"))
    #write (FID,concat("f_range2range -1.000000 ",str(f_range2range(+0.0,0.0,1.0,-1.0,1.0),9,6)," (+0.0 -> ) \n"))
    #write (FID,concat("F_adj_range2  -1.000000 ",str(F_adj_range2( +0.0,0.0,1.0,-1.0,1.0),9,6)," (+0.0 -> ) \n"))
    #write (FID,concat("f_range2range +1.000000 ",str(f_range2range(+1.0,0.0,1.0,-1.0,1.0),9,6)," (+1.0 -> ) \n"))
    #write (FID,concat("F_adj_range2  +1.000000 ",str(F_adj_range2( +1.0,0.0,1.0,-1.0,1.0),9,6)," (+1.0 -> ) \n"))
    #write (FID,concat("f_range2range -2.000000 ",str(f_range2range(-0.5,0.0,1.0,-1.0,1.0),9,6)," (-0.5 -> ) \n"))
    #write (FID,concat("F_adj_range2  -2.000000 ",str(F_adj_range2( -0.5,0.0,1.0,-1.0,1.0),9,6)," (-0.5 -> ) \n"))
    #write (FID,concat("f_range2range +2.000000 ",str(f_range2range(+1.5,0.0,1.0,-1.0,1.0),9,6)," (+1.5 -> ) \n"))
    #write (FID,concat("F_adj_range2  +2.000000 ",str(F_adj_range2( +1.5,0.0,1.0,-1.0,1.0),9,6)," (+1.5 -> ) \n"))
    #write (FID,"\n")
    // F_clamp   val --> -1.0,1.0
    #write (FID,"---\n")
    #write (FID,concat("M_clamp  -1.000000 ",str(M_clamp(+1.0,-1.0,1.0),9,6)," (+1.0 -> ) Create F_clamp\n"))
    #write (FID,concat("F_clamp  -1.000000 ",str(F_clamp(+1.0,-1.0,1.0),9,6)," (+1.0 -> ) val --> -1,+1\n"))
    #write (FID,concat("F_clamp  +0.000000 ",str(F_clamp(+0.0,-1.0,1.0),9,6)," (+0.0 -> ) \n"))
    #write (FID,concat("F_clamp  -1.000000 ",str(F_clamp(-1.0,-1.0,1.0),9,6)," (-1.0 -> ) \n"))
    #write (FID,"-\n- showing f_wrap's two modes\n")
    #write (FID,concat("f_wrap  -1.000000 ",str(f_wrap(+1.0,-1.0,1.0,0),9,6)," (+1.0 -> ) mode 0\n"))
    #write (FID,concat("f_wrap  +0.000000 ",str(f_wrap(+0.0,-1.0,1.0,0),9,6)," (+0.0 -> ) \n"))
    #write (FID,concat("f_wrap  -1.000000 ",str(f_wrap(-1.0,-1.0,1.0,0),9,6)," (-1.0 -> ) \n"))
    #write (FID,concat("f_wrap  +1.000000 ",str(f_wrap(+1.0,-1.0,1.0,1),9,6)," (+1.0 -> ) mode 1\n"))
    #write (FID,concat("f_wrap  +0.000000 ",str(f_wrap(+0.0,-1.0,1.0,1),9,6)," (+0.0 -> ) \n"))
    #write (FID,concat("f_wrap  -1.000000 ",str(f_wrap(-1.0,-1.0,1.0,1),9,6)," (-1.0 -> ) \n"))
    #write (FID,"-\n")
    #write (FID,concat("F_clamp  +0.999999 ",str(F_clamp(+1.0-1e-6,-1.0,1.0),9,6)," (+1.0-1e-6 -> ) \n"))
    #write (FID,concat("F_clamp  -0.999999 ",str(F_clamp(+1.0+1e-6,-1.0,1.0),9,6)," (+1.0+1e-6 -> ) \n"))
    #write (FID,concat("F_clamp  +0.999999 ",str(F_clamp(-1.0-1e-6,-1.0,1.0),9,6)," (-1.0-1e-6 -> ) \n"))
    #write (FID,concat("F_clamp  -0.999999 ",str(F_clamp(-1.0+1e-6,-1.0,1.0),9,6)," (-1.0+1e-6 -> ) \n"))
    #write (FID,"-\n")
    #write (FID,concat("F_clamp  -0.500000 ",str(F_clamp(-0.5,-1.0,1.0),9,6)," (-0.5 -> ) \n"))
    #write (FID,concat("F_clamp  +0.500000 ",str(F_clamp(+0.5,-1.0,1.0),9,6)," (+0.5 -> ) \n"))
    #write (FID,concat("F_clamp  +0.500000 ",str(F_clamp(-1.5,-1.0,1.0),9,6)," (-1.5 -> ) \n"))
    #write (FID,concat("F_clamp  -0.500000 ",str(F_clamp(+1.5,-1.0,1.0),9,6)," (+1.5 -> ) \n"))
    #write (FID,concat("F_clamp  +0.000000 ",str(F_clamp(-2.0,-1.0,1.0),9,6)," (-2.0 -> ) \n"))
    #write (FID,concat("F_clamp  +0.000000 ",str(F_clamp(+2.0,-1.0,1.0),9,6)," (+2.0 -> ) \n"))
    #write (FID,"\n")
    // M_clip() --> F_clip()
    #write (FID,"---\n")
    #write (FID,concat("M_clip  0.500000 ",str(M_clip(0.5,0,1),9,-1)," compiling F_clip\n"))
    #write (FID,concat("F_clip  0.500000 ",str(global.F_clip(0.5,0,1),9,-1)," 0.5 to 0-1\n"))
    #write (FID,concat("F_clip  0.000000 ",str(global.F_clip(0.0,0,1),9,-1)," 0.0 to 0-1\n"))
    #write (FID,concat("F_clip  1.000000 ",str(global.F_clip(1.0,0,1),9,-1)," 1.0 to 0-1\n"))
    #write (FID,concat("F_clip  0.000000 ",str(global.F_clip(-0.1,0,1),9,-1)," -0.1 to 0-1\n"))
    #write (FID,concat("F_clip  1.000000 ",str(global.F_clip(+1.1,0,1),9,-1)," +1.1 to 0-1\n"))
    #write (FID,"\n")
    // M_sgn() --> F_sgn()
    #write (FID,"---\n")
    #write (FID,concat("M_sgn  1.000000 ",str(M_sgn(0.5),9,-1)," compiling F_sgn\n"))
    #write (FID,concat("F_sgn  1.000000 ",str(global.F_sgn(0.5),9,-1)," 0.5\n"))
    #write (FID,concat("F_sgn  0.000000 ",str(global.F_sgn(0.0),9,-1)," 0.0\n"))
    #write (FID,concat("F_sgn  1.000000 ",str(global.F_sgn(1.0),9,-1)," 1.0\n"))
    #write (FID,concat("F_sgn -1.000000 ",str(global.F_sgn(-0.1),9,-1)," -0.1\n"))
    #write (FID,concat("F_sgn  1.000000 ",str(global.F_sgn(+0.1),9,-1)," +0.1\n"))
    #write (FID,"\n")
    // M_even() --> F_even()
    #write (FID,"---\n")
    #write (FID,concat("M_even  0.000000 ",str(M_even(0.5),9,-1)," compiling F_even\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(0.1),9,-1)," 0.1\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(0.5),9,-1)," 0.5\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(0.9),9,-1)," 0.9\n"))
    #write (FID,concat("F_even  1.000000 ",str(global.F_even(0.0),9,-1)," 0.0\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(1.0),9,-1)," 1.0\n"))
    #write (FID,concat("F_even  1.000000 ",str(global.F_even(2.0),9,-1)," 2.0\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(3.0),9,-1)," 3.0\n"))
    #write (FID,concat("F_even  1.000000 ",str(global.F_even(4.0),9,-1)," 4.0\n"))
    #write (FID,concat("F_even  1.000000 ",str(global.F_even(-4.0),9,-1)," -4.0\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(-3.0),9,-1)," -3.0\n"))
    #write (FID,concat("F_even  1.000000 ",str(global.F_even(-2.0),9,-1)," -2.0\n"))
    #write (FID,concat("F_even  0.000000 ",str(global.F_even(-1.0),9,-1)," -1.0\n"))
    #write (FID,"\n")
    // M_odd() --> F_odd()
    #write (FID,"---\n")
    #write (FID,concat("M_odd  1.000000 ",str(M_odd(0.5),9,-1)," compiling F_odd\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(0.1),9,-1)," 0.1\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(0.5),9,-1)," 0.5\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(0.9),9,-1)," 0.9\n"))
    #write (FID,concat("F_odd  0.000000 ",str(global.F_odd(0.0),9,-1)," 0.0\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(1.0),9,-1)," 1.0\n"))
    #write (FID,concat("F_odd  0.000000 ",str(global.F_odd(2.0),9,-1)," 2.0\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(3.0),9,-1)," 3.0\n"))
    #write (FID,concat("F_odd  0.000000 ",str(global.F_odd(4.0),9,-1)," 4.0\n"))
    #write (FID,concat("F_odd  0.000000 ",str(global.F_odd(-4.0),9,-1)," -4.0\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(-3.0),9,-1)," -3.0\n"))
    #write (FID,concat("F_odd  0.000000 ",str(global.F_odd(-2.0),9,-1)," -2.0\n"))
    #write (FID,concat("F_odd  1.000000 ",str(global.F_odd(-1.0),9,-1)," -1.0\n"))
    #write (FID,"\n")
    // M_sind() --> F_sind()
    #write (FID,"---\n")
    #write (FID,concat("M_sind +0.544639 ",str(M_sind(33),9,-1)," compiling F_sind +33\n"))
    #write (FID,concat("F_sind +0.544639 ",str(global.F_sind(+33),9,-1)," +33\n"))
    #write (FID,concat("F_sind -0.544639 ",str(global.F_sind(-33),9,-1)," -33\n"))
    #write (FID,concat("F_sind +0.544639 ",str(global.F_sind(+393),9,-1)," +393\n"))
    #write (FID,concat("F_sind -0.544639 ",str(global.F_sind(-393),9,-1)," -393\n"))
    #write (FID,concat("F_sind +0.000000 ",str(global.F_sind(+0),9,-1)," +0\n"))
    #write (FID,concat("F_sind +1.000000 ",str(global.F_sind(+90),9,-1)," +90\n"))
    #write (FID,concat("F_sind -1.000000 ",str(global.F_sind(-90),9,-1)," -90\n"))
    #write (FID,"-\n")
    // M_asind() --> F_asind()
    #write (FID,concat("M_asind +32.999998 ",str(M_asind(+0.544639),9,-1)," compiling F_asind +0.544639\n"))
    #write (FID,concat("F_asind +32.999998 ",str(global.F_asind(+0.544639),9,-1)," +0.544639\n"))
    #write (FID,concat("F_asind -32.999998 ",str(global.F_asind(-0.544639),9,-1)," -0.544639\n"))
    #write (FID,concat("F_asind +32.999998 ",str(global.F_asind(+1.544639),9,-1)," +1.544639\n"))
    #write (FID,concat("F_asind -32.999998 ",str(global.F_asind(-1.544639),9,-1)," -1.544639\n"))
    #write (FID,concat("F_asind +0.000000  ",str(global.F_asind(+0.0),9,-1)," +0.0\n"))
    #write (FID,concat("F_asind +90.000000 ",str(global.F_asind(+1.0),9,-1)," +1.0\n"))
    #write (FID,concat("F_asind -90.000000 ",str(global.F_asind(-1.0),9,-1)," -1.0\n"))
    #write (FID,"\n")
    // M_cosd() --> F_cosd()
    #write (FID,"---\n")
    #write (FID,concat("M_cosd +0.838671 ",str(M_cosd(33),9,-1)," compiling F_cosd +33\n"))
    #write (FID,concat("F_cosd +0.838671 ",str(global.F_cosd(+33),9,-1)," +33\n"))
    #write (FID,concat("F_cosd +0.838671 ",str(global.F_cosd(-33),9,-1)," -33\n"))
    #write (FID,concat("F_cosd +0.838671 ",str(global.F_cosd(+393),9,-1)," +393\n"))
    #write (FID,concat("F_cosd +0.838671 ",str(global.F_cosd(-393),9,-1)," -393\n"))
    #write (FID,concat("F_cosd +1.000000 ",str(global.F_cosd(+0),9,-1)," +0\n"))
    #write (FID,concat("F_cosd +0.000000 ",str(global.F_cosd(+90),9,-1)," +90\n"))
    #write (FID,concat("F_cosd +0.000000 ",str(global.F_cosd(-90),9,-1)," -90\n"))
    #write (FID,concat("F_cosd -1.000000 ",str(global.F_cosd(+180),9,-1)," +180\n"))
    #write (FID,concat("F_cosd -1.000000 ",str(global.F_cosd(-180),9,-1)," -180\n"))
    #write (FID,"-\n")
    // M_acosd() --> F_acosd()
    #write (FID,concat("M_acosd +32.999955  ",str(M_acosd(+0.838671),9,-1)," compiling F_acosd +0.838671\n"))
    #write (FID,concat("F_acosd +32.999955  ",str(global.F_acosd(+0.838671),9,-1)," +0.838671\n"))
    #write (FID,concat("F_acosd +32.999955  ",str(global.F_acosd(+0.838671),9,-1)," +0.838671\n"))
    #write (FID,concat("F_acosd +32.999955  ",str(global.F_acosd(+1.838671),9,-1)," +1.838671\n"))
    #write (FID,concat("F_acosd +32.999955  ",str(global.F_acosd(+1.838671),9,-1)," +1.838671\n"))
    #write (FID,concat("F_acosd +90.000000  ",str(global.F_acosd(+0.0),9,-1)," +0.0\n"))
    #write (FID,concat("F_acosd +0.000000   ",str(global.F_acosd(+1.0),9,-1)," +1.0\n"))
    #write (FID,concat("F_acosd +180.000000 ",str(global.F_acosd(-1.0),9,-1)," -1.0\n"))
    #write (FID,"\n")
    // M_tand() --> F_tand()
    #write (FID,"---\n")
    #write (FID,concat("M_tand +0.649408 ",str(M_tand(33),9,-1)," compiling F_tand +33\n"))
    #write (FID,concat("F_tand +0.649408 ",str(global.F_tand(+33),9,-1)," +33\n"))
    #write (FID,concat("F_tand -0.649408 ",str(global.F_tand(-33),9,-1)," -33\n"))
    #write (FID,concat("F_tand +0.649408 ",str(global.F_tand(+393),9,-1)," +393\n"))
    #write (FID,concat("F_tand -0.649408 ",str(global.F_tand(-393),9,-1)," -393\n"))
    #write (FID,concat("F_tand +0.000000 ",str(global.F_tand(+0),9,-1)," +0\n"))
    #write (FID,concat("F_tand +nan ",str(global.F_tand(+90),19,16)," +90 inf ? \n"))
    #write (FID,concat("F_tand -nan ",str(global.F_tand(-90),19,16)," -90 inf ? \n"))
    #write (FID,concat("F_tand -0.000000 ",str(global.F_tand(+180),9,-1)," +180\n"))
    #write (FID,concat("F_tand +0.000000 ",str(global.F_tand(-180),9,-1)," -180\n"))
    #write (FID,"-\n")
    // M_atand() --> F_atand()
    #write (FID,concat("M_atand +33.000016 ",str(M_atand(+0.649408),9,-1)," compiling F_atand +0.649408\n"))
    #write (FID,concat("F_atand +33.000016 ",str(global.F_atand(+0.649408),9,-1)," +0.649408\n"))
    #write (FID,concat("F_atand -33.000016 ",str(global.F_atand(-0.649408),9,-1)," -0.649408\n"))
    #write (FID,concat("F_atand +58.772483 ",str(global.F_atand(+1.649408),9,-1)," +1.649408\n"))
    #write (FID,concat("F_atand -58.772483 ",str(global.F_atand(-1.649408),9,-1)," -1.649408\n"))
    #write (FID,concat("F_atand +88.970526 ",str(global.F_atand(+55.649408),9,-1)," +55.649408\n"))
    #write (FID,concat("F_atand -88.970526 ",str(global.F_atand(-55.649408),9,-1)," -55.649408\n"))
    #write (FID,"\n")
    // M_atan2d() --> F_atan2d()
    #write (FID,concat("M_atan2d  +45.000000 ",str(M_atan2d(+1,+1),9,-1)," compiling F_atan2d +1 +1\n"))
    #write (FID,concat("F_atan2d  +45.000000 ",str(global.F_atan2d(+1,+1),9,-1)," +1 +1 \n"))
    #write (FID,concat("F_atan2d  -45.000000 ",str(global.F_atan2d(-1,+1),9,-1)," -1 +1 \n"))
    #write (FID,concat("F_atan2d +135.000000 ",str(global.F_atan2d(+1,-1),9,-1)," +1 -1 \n"))
    #write (FID,concat("F_atan2d -135.000000 ",str(global.F_atan2d(-1,-1),9,-1)," -1 -1 \n"))
    #write (FID,"\n")
    // M_since_wave() --> F_sine_wave().  Running too inbuilt f_sine_wave at mode 2 for reference.
    #write (FID,"---\n")
    #write (FID,concat("M_sine_wave -0.809017 ",str(M_sine_wave(-0.3,1,1),9,-1)," compiling F_sine_wave -0.3\n"))
    #write (FID,concat("F_sine_wave -0.809017 ",str(F_sine_wave(-0.3,1,1),9,-1)," -0.3\n"))
    #write (FID,concat("F_sine_wave +1.000000 ",str(F_sine_wave(-1.5,1,1),9,-1)," -1.5\n"))
    #write (FID,concat("F_sine_wave  0.000000 ",str(F_sine_wave(-1.0,1,1),9,-1)," -1.0\n"))
    #write (FID,concat("F_sine_wave -1.000000 ",str(F_sine_wave(-0.5,1,1),9,-1)," -0.5\n"))
    #write (FID,concat("F_sine_wave  0.000000 ",str(F_sine_wave(+0.0,1,1),9,-1)," +0.0\n"))
    #write (FID,concat("F_sine_wave +1.000000 ",str(F_sine_wave(+0.5,1,1),9,-1)," +0.5\n"))
    #write (FID,concat("F_sine_wave  0.000000 ",str(F_sine_wave(+1.0,1,1),9,-1)," +1.0\n"))
    #write (FID,concat("F_sine_wave -1.000000 ",str(F_sine_wave(+1.5,1,1),9,-1)," +1.5\n"))
    #write (FID,"-\n")
    #write (FID,concat("f_sine_wave -0.809017 ",str(f_sine_wave(-0.30,2),9,-1)," at -0.30 mode 2\n"))
    #write (FID,concat("f_sine_wave +1.000000 ",str(f_sine_wave(-1.50,2),9,-1)," at -1.50 mode 2\n"))
    #write (FID,concat("f_sine_wave +1.000000 ",str(f_sine_wave(-1.50,2),9,-1)," at -1.50 mode 2\n"))
    #write (FID,concat("f_sine_wave  0.000000 ",str(f_sine_wave(-1.00,2),9,-1)," at -1.00 mode 2\n"))
    #write (FID,concat("f_sine_wave -1.000000 ",str(f_sine_wave(-0.50,2),9,-1)," at -0.50 mode 2\n"))
    #write (FID,concat("f_sine_wave  0.000000 ",str(f_sine_wave(+0.00,2),9,-1)," at +0.00 mode 2\n"))
    #write (FID,concat("f_sine_wave +1.000000 ",str(f_sine_wave(+0.50,2),9,-1)," at +0.50 mode 2\n"))
    #write (FID,concat("f_sine_wave  0.000000 ",str(f_sine_wave(+1.00,2),9,-1)," at +1.00 mode 2\n"))
    #write (FID,concat("f_sine_wave -1.000000 ",str(f_sine_wave(+1.50,2),9,-1)," at +1.50 mode 2\n"))
    #write (FID,"\n")
    // M_scallop_wave() --> F_scallop_wave().   Running too inbuilt f_scallop_wave at mode 2 for reference.
    #write (FID,"---\n")
    #write (FID,concat("M_scallop_wave +0.809017 ",str(M_scallop_wave(-0.3,1,1),9,-1)," compiling F_scallop_wave -0.3\n"))
    #write (FID,concat("F_scallop_wave +0.809017 ",str(F_scallop_wave(-0.3,1,1),9,-1)," -0.3\n"))
    #write (FID,concat("F_scallop_wave +1.000000 ",str(F_scallop_wave(-1.5,1,1),9,-1)," -1.5\n"))
    #write (FID,concat("F_scallop_wave  0.000000 ",str(F_scallop_wave(-1.0,1,1),9,-1)," -1.0\n"))
    #write (FID,concat("F_scallop_wave +1.000000 ",str(F_scallop_wave(-0.5,1,1),9,-1)," -0.5\n"))
    #write (FID,concat("F_scallop_wave  0.000000 ",str(F_scallop_wave(+0.0,1,1),9,-1)," +0.0\n"))
    #write (FID,concat("F_scallop_wave +1.000000 ",str(F_scallop_wave(+0.5,1,1),9,-1)," +0.5\n"))
    #write (FID,concat("F_scallop_wave  0.000000 ",str(F_scallop_wave(+1.0,1,1),9,-1)," +1.0\n"))
    #write (FID,concat("F_scallop_wave +1.000000 ",str(F_scallop_wave(+1.5,1,1),9,-1)," +1.5\n"))
    #write (FID,"-\n")
    #write (FID,concat("f_scallop_wave +0.809017 ",str(f_scallop_wave(-0.30,0),9,-1)," at -0.30 mode 2\n"))
    #write (FID,concat("f_scallop_wave +1.000000 ",str(f_scallop_wave(-1.50,0),9,-1)," at -1.50 mode 2\n"))
    #write (FID,concat("f_scallop_wave  0.000000 ",str(f_scallop_wave(-1.00,0),9,-1)," at -1.00 mode 2\n"))
    #write (FID,concat("f_scallop_wave +1.000000 ",str(f_scallop_wave(-0.50,0),9,-1)," at -0.50 mode 2\n"))
    #write (FID,concat("f_scallop_wave  0.000000 ",str(f_scallop_wave(+0.00,0),9,-1)," at +0.00 mode 2\n"))
    #write (FID,concat("f_scallop_wave +1.000000 ",str(f_scallop_wave(+0.50,0),9,-1)," at +0.50 mode 2\n"))
    #write (FID,concat("f_scallop_wave  0.000000 ",str(f_scallop_wave(+1.00,0),9,-1)," at +1.00 mode 2\n"))
    #write (FID,concat("f_scallop_wave +1.000000 ",str(f_scallop_wave(+1.50,0),9,-1)," at +1.50 mode 2\n"))
    #write (FID,"\n")
    // M_evaluate_pigment() --> FnToEvalPigmentAtLoc()
    #write (FID,"---\n")
    #declare Pigment00 = pigment { color rgbft <1.0,1.1,1.2,1.3,1.4> }
    #declare Vec00_A   = M_evaluate_pigment("FnPg00",Pigment00,<0,0,0>,false);
    // Showing redefinition of Pigment00 has NO effect. Already compiled.
    #declare Pigment00 = pigment { color rgbft <1.0,1.1,1.2,1.3,1.4>*0.5 }
    #declare Vec00_B   = M_evaluate_pigment("FnPg00",Pigment00,<1,1,1>,false);
    #declare Vec00_C   = FnPg00(2,2,2); // Best performance
    #declare Pigment01 = pigment { color rgbft <0.0,0.1,0.2,0.3,0.4> }
    #declare Vec01_A   = M_evaluate_pigment("FnPg01",Pigment01,<0,0,0>,false);
    // Showing redefinition of Pigment01 has effect. We ask for re-compile.
    #declare Pigment01 = pigment { color rgbft <0.0,0.1,0.2,0.3,0.4>+0.1 }
    #declare Vec01_B   = M_evaluate_pigment("FnPg01",Pigment01,<1,1,1>,true);
    #declare Vec01_C   = FnPg01(2,2,2); // Best performance
    #write (FID,concat("M_evaluate_pigment  1.000000 ",str(Vec00_A.x,9,6)," Vec00_A.x\n"))
    #write (FID,concat("M_evaluate_pigment  1.100000 ",str(Vec00_A.y,9,6)," Vec00_A.y\n"))
    #write (FID,concat("M_evaluate_pigment  1.200000 ",str(Vec00_A.z,9,6)," Vec00_A.z\n"))
    #write (FID,concat("M_evaluate_pigment  1.300000 ",str(Vec00_A.t,9,6)," Vec00_A.t\n"))
    #write (FID,concat("M_evaluate_pigment  1.400000 ",str(Vec00_A.transmit,9,6)," Vec00_A.transmit\n"))
    #write (FID,"-\n")
    #write (FID,concat("M_evaluate_pigment  1.000000 ",str(Vec00_B.x,9,6)," Vec00_B.x\n"))
    #write (FID,concat("M_evaluate_pigment  1.100000 ",str(Vec00_B.y,9,6)," Vec00_B.y\n"))
    #write (FID,concat("M_evaluate_pigment  1.200000 ",str(Vec00_B.z,9,6)," Vec00_B.z\n"))
    #write (FID,concat("M_evaluate_pigment  1.300000 ",str(Vec00_B.t,9,6)," Vec00_B.t\n"))
    #write (FID,concat("M_evaluate_pigment  1.400000 ",str(Vec00_B.transmit,9,6)," Vec00_B.transmit\n"))
    #write (FID,"-\n")
    #write (FID,concat("M_evaluate_pigment  1.000000 ",str(Vec00_C.x,9,6)," Vec00_C.x\n"))
    #write (FID,concat("M_evaluate_pigment  1.100000 ",str(Vec00_C.y,9,6)," Vec00_C.y\n"))
    #write (FID,concat("M_evaluate_pigment  1.200000 ",str(Vec00_C.z,9,6)," Vec00_C.z\n"))
    #write (FID,concat("M_evaluate_pigment  1.300000 ",str(Vec00_C.t,9,6)," Vec00_C.t\n"))
    #write (FID,concat("M_evaluate_pigment  1.400000 ",str(Vec00_C.transmit,9,6)," Vec00_C.transmit\n"))
    #write (FID,"--\n")
    #write (FID,concat("M_evaluate_pigment  0.000000 ",str(Vec01_A.x,9,6)," Vec01_A.x\n"))
    #write (FID,concat("M_evaluate_pigment  0.100000 ",str(Vec01_A.y,9,6)," Vec01_A.y\n"))
    #write (FID,concat("M_evaluate_pigment  0.200000 ",str(Vec01_A.z,9,6)," Vec01_A.z\n"))
    #write (FID,concat("M_evaluate_pigment  0.300000 ",str(Vec01_A.t,9,6)," Vec01_A.t\n"))
    #write (FID,concat("M_evaluate_pigment  0.400000 ",str(Vec01_A.transmit,9,6)," Vec01_A.transmit\n"))
    #write (FID,"-\n")
    #write (FID,concat("M_evaluate_pigment  0.100000 ",str(Vec01_B.x,9,6)," Vec01_B.x\n"))
    #write (FID,concat("M_evaluate_pigment  0.200000 ",str(Vec01_B.y,9,6)," Vec01_B.y\n"))
    #write (FID,concat("M_evaluate_pigment  0.300000 ",str(Vec01_B.z,9,6)," Vec01_B.z\n"))
    #write (FID,concat("M_evaluate_pigment  0.400000 ",str(Vec01_B.t,9,6)," Vec01_B.t\n"))
    #write (FID,concat("M_evaluate_pigment  0.500000 ",str(Vec01_B.transmit,9,6)," Vec01_B.transmit\n"))
    #write (FID,"-\n")
    #write (FID,concat("M_evaluate_pigment  0.100000 ",str(Vec01_C.x,9,6)," Vec01_C.x\n"))
    #write (FID,concat("M_evaluate_pigment  0.200000 ",str(Vec01_C.y,9,6)," Vec01_C.y\n"))
    #write (FID,concat("M_evaluate_pigment  0.300000 ",str(Vec01_C.z,9,6)," Vec01_C.z\n"))
    #write (FID,concat("M_evaluate_pigment  0.400000 ",str(Vec01_C.t,9,6)," Vec01_C.t\n"))
    #write (FID,concat("M_evaluate_pigment  0.500000 ",str(Vec01_C.transmit,9,6)," Vec01_C.transmit\n"))
    #write (FID,"-\n")
    #declare Pigment02 = pigment { color rgbft <-0.0,-0.1,-0.2,-0.3,-0.4> }
    #declare Vec02_A   = M_evaluate_pigment("",Pigment02,<0,0,0>,true);
    #declare Pigment02 = pigment { color rgbft <-0.0,-0.1,-0.2,-0.3,-0.4>-0.1 }
    #declare Vec02_B   = M_evaluate_pigment("",Pigment02,<1,1,1>,true);
    #declare Vec02_C   = global[""](2,2,2); // Null string function... Used for old style,
                                            // compile-always, macro call behavior.
    #write (FID,concat("M_evaluate_pigment -0.000000 ",str(Vec02_A.x,9,6)," Vec01_A.x\n"))
    #write (FID,concat("M_evaluate_pigment -0.100000 ",str(Vec02_A.y,9,6)," Vec01_A.y\n"))
    #write (FID,concat("M_evaluate_pigment -0.200000 ",str(Vec02_A.z,9,6)," Vec01_A.z\n"))
    #write (FID,concat("M_evaluate_pigment -0.300000 ",str(Vec02_A.t,9,6)," Vec01_A.t\n"))
    #write (FID,concat("M_evaluate_pigment -0.400000 ",str(Vec02_A.transmit,9,6)," Vec01_A.transmit\n"))
    #write (FID,"-\n")
    #write (FID,concat("M_evaluate_pigment -0.100000 ",str(Vec02_B.x,9,6)," Vec01_B.x\n"))
    #write (FID,concat("M_evaluate_pigment -0.200000 ",str(Vec02_B.y,9,6)," Vec01_B.y\n"))
    #write (FID,concat("M_evaluate_pigment -0.300000 ",str(Vec02_B.z,9,6)," Vec01_B.z\n"))
    #write (FID,concat("M_evaluate_pigment -0.400000 ",str(Vec02_B.t,9,6)," Vec01_B.t\n"))
    #write (FID,concat("M_evaluate_pigment -0.500000 ",str(Vec02_B.transmit,9,6)," Vec01_B.transmit\n"))
    #write (FID,"-\n")
    #write (FID,concat("M_evaluate_pigment -0.100000 ",str(Vec02_C.x,9,6)," Vec01_C.x\n"))
    #write (FID,concat("M_evaluate_pigment -0.200000 ",str(Vec02_C.y,9,6)," Vec01_C.y\n"))
    #write (FID,concat("M_evaluate_pigment -0.300000 ",str(Vec02_C.z,9,6)," Vec01_C.z\n"))
    #write (FID,concat("M_evaluate_pigment -0.400000 ",str(Vec02_C.t,9,6)," Vec01_C.t\n"))
    #write (FID,concat("M_evaluate_pigment -0.500000 ",str(Vec02_C.transmit,9,6)," Vec01_C.transmit\n"))
    #write (FID,"\n")

    #write (FID,"\n")
    #fclose FID

    #error "\n\n=== Self testing. Stop before render. ===\n"

#end // End self testing

#version Munctions_Inc_Temp;
#end //munctions.inc


